
#ifndef __KIBA_OBJECT_H__
#define __KIBA_OBJECT_H__

#include <gtk/gtk.h>
#include <glib.h>

#include <kiba-window.h>
#include <kiba-types.h>

G_BEGIN_DECLS

#define KIBA_TYPE_OBJECT           (kiba_object_get_type ())
#define KIBA_OBJECT(obj)           (G_TYPE_CHECK_INSTANCE_CAST ((obj), KIBA_TYPE_OBJECT, KibaObject))
#define KIBA_OBJECT_CLASS(obj)     (G_TYPE_CHECK_CLASS_CAST ((obj), KIBA_TYPE_OBJECT,  KibaObjectClass))
#define KIBA_IS_OBJECT(obj)        (G_TYPE_CHECK_INSTANCE_TYPE ((obj), KIBA_TYPE_OBJECT))
#define KIBA_IS_OBJECT_CLASS(obj)  (G_TYPE_CHECK_CLASS_TYPE ((obj), KIBA_TYPE_OBJECT))
#define KIBA_OBJECT_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), KIBA_TYPE_OBJECT, KibaObjectClass))

typedef struct _KibaObject KibaObject;
typedef struct _KibaObjectClass KibaObjectClass;

/**
 * KibaObjectCustomScaleFunc:
 * @Param1: The KibaObject Instance.
 * @Param2: The width scale factor return.
 * @Param3: The height scale factor return.
 * @Param4: User Data.
 *
 * Method used for custom scaling.
 */
typedef void (*KibaObjectCustomScaleFunc) (KibaObject*, float*, float*, gpointer);

/**
 * KibaObjectCustomMaxScaleFunc:
 * @Param1: The KibaObject Instance.
 * @Param2: The maximal width scale factor return.
 * @Param3: The maximal height scale factor return.
 * @Param4: User Data.
 *
 * Gets maximal scale factors for a scaling method.
 */
typedef void (*KibaObjectCustomMaxScaleFunc) (KibaObject*, float*, float*, gpointer);

/**
 * KibaObjectCustomOffsetFunc:
 * @Param1: The KibaObject Instance.
 * @Param2: The width offset return.
 * @Param3: The height offset return.
 * @Param4: User Data.
 *
 * Method used for custom offset on the dock.
 */
typedef void (*KibaObjectCustomOffsetFunc) (KibaObject*, float*, float*, gpointer);

/**
 * KibaObjectCustomMaxOffsetFunc:
 * @Param1: The KibaObject Instance.
 * @Param2: The maximal width offset return.
 * @Param3: The maximal height offset return.
 * @Param4: User Data.
 *
 * Method used for custom offset on the dock.
 */
typedef void (*KibaObjectCustomMaxOffsetFunc) (KibaObject*, float*, float*, gpointer);

typedef enum {
	KIBA_OBJECT_ON_DOCK = 0,
	KIBA_OBJECT_LEFT_OF_DOCK,
	KIBA_OBJECT_RIGHT_OF_DOCK
}KibaObjectPosition;

/**
 * KibaObjectCustomScaleData:
 *
 * Type for custom scale animations.
 */
typedef struct {
	KibaObjectCustomScaleFunc scale_func;
	KibaObjectCustomMaxScaleFunc max_func;
	gboolean scale_widget;
	gpointer data;
}KibaObjectCustomScaleData;

/**
 * KibaObjectCustomOffsetData:
 *
 * Type for custom surface offset animations.
 */
typedef struct {
	KibaObjectCustomScaleFunc place_func;
	KibaObjectCustomMaxScaleFunc max_func;
	gpointer data;
}KibaObjectCustomOffsetData;

/**
 * KibaObjectState:
 *
 * Object flags.
 */
typedef enum {
	KIBA_OBJECT_DRAGGING = 1,
	KIBA_OBJECT_PRELIGHTED,
	KIBA_OBJECT_URGENT,
	KIBA_OBJECT_ACTIVE,
	KIBA_OBJECT_WORKING
}KibaObjectState;

struct _KibaObject {
	GtkBin bin;
	
	KibaObjectPosition position;
	gint state; /**<< state flags **/
	
	KibaAllocation alloc;
	GtkRequisition size;
	float scale_x;        /**< widget width scale factor */
	float scale_y;        /**< widget height scale factor */
	float destroy_factor; /**< Scale factor for the destroy animation */
	
	GList *custom_scale_funcs; /**< list of all KibaObjectCustomScaleFunc's */
	GList *custom_offset_data; /**< list of all KibaObjectCustomOffsetData's */
};

struct _KibaObjectClass {
	GtkBinClass parent_class;
	
	/* signals */
	void (* urgent_toggled) (KibaObject  *object,
                             gboolean urgent);
	void (* active_toggled) (KibaObject  *object,
                             gboolean active);
	void (* working_toggled) (KibaObject  *object,
                              gboolean working);
	void (*dragging_toggled) (KibaObject  *object,
                              gboolean dragging);
};

void kiba_object_set_icon_geometry (KibaObject *object,
                                    KibaWindow *window);

int kiba_object_launch (KibaObject *object,
                        const char *command);

gboolean kiba_object_get_state (KibaObject *object,
                                KibaObjectState state);

void kiba_object_set_state (KibaObject *object,
                            KibaObjectState state);

void kiba_object_unset_state (KibaObject *object,
                              KibaObjectState state);

gboolean kiba_object_is_on_dock (KibaObject *object);

void kiba_object_set_requisition_factor (KibaObject *object,
                                         float      scale_x,
                                         float      scale_y);

void kiba_object_add_custom_offset_function (KibaObject *object,
                                             KibaObjectCustomOffsetFunc place_func,
                                             KibaObjectCustomMaxScaleFunc max_func,
                                             gpointer user_data);

void kiba_object_remove_custom_offset_function_by_data (KibaObject *object,
                                                        gpointer user_data);

gboolean kiba_object_update_scale_factor (KibaObject *object);

void kiba_object_add_custom_scale_function (KibaObject *object,
                                            KibaObjectCustomScaleFunc scale_func,
                                            KibaObjectCustomMaxScaleFunc max_func,
                                            gboolean scale_widget,
                                            gpointer user_data);

void kiba_object_remove_custom_scale_function_by_data (KibaObject *object,
                                                       gpointer user_data);

void kiba_object_get_scale_factor (KibaObject *object,
                                   float *scale_x_ret,
                                   float *scale_y_ret);

void kiba_object_set_tooltip_text(KibaObject *object,
                                  const char *text);

void kiba_object_set_title (KibaObject *object,
                            const char *text);

const char* kiba_object_get_title (KibaObject *object);

void kiba_object_show_title (GtkWidget *widget);

void kiba_object_hide_title (GtkWidget *widget);

float kiba_object_get_destroy_factor (KibaObject *object);

gboolean kiba_object_get_destroying (KibaObject *object);

void kiba_object_destroy (KibaObject *object);

G_END_DECLS

#endif
