/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <gtk/gtk.h>
#include <string.h>
#include <libintl.h>

#include "kiba-settings.h"
#include "kiba-prefs.h"
#include "kiba-gradient-chooser-dialog.h"

#define _(string) gettext (string)

#define KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_GRADIENT_CHOOSER_DIALOG, KibaGradientChooserDialogPrivate))

G_DEFINE_TYPE (KibaGradientChooserDialog, kiba_gradient_chooser_dialog, GTK_TYPE_WINDOW)

typedef struct {
	GtkWidget *combobox;
	GtkWidget *colorChooser;
	
	GList *colors;
} KibaGradientChooserDialogPrivate;

enum {
	SIG_CHANGED = 0,
	SIG_GRADIENT_CHOOSER_DIALOG_LAST
};

static guint kiba_gradient_chooser_dialog_signals[SIG_GRADIENT_CHOOSER_DIALOG_LAST] = { 0 };

/** Private Methods **/

static void __update_color_chooser (KibaGradientChooserDialog *dialog) {
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(dialog);
	GdkColor col;
	gchar *text = gtk_combo_box_get_active_text(GTK_COMBO_BOX(priv->combobox));
	unsigned int r,g,b, alpha;
	
	sscanf(text, "#%02X%02X%02X%02X", &r, &g, &b, &alpha);
	col.red = (float) r*256;
	col.green = (float) g*256;
	col.blue = (float) b*256;
	
	gtk_color_selection_set_current_color(GTK_COLOR_SELECTION(priv->colorChooser), &col);
	gtk_color_selection_set_current_alpha(GTK_COLOR_SELECTION(priv->colorChooser), alpha*256);
}

static void __combobox_changed_cb (GtkComboBox *combobox,
                                   gpointer data) {
	KibaGradientChooserDialog *dialog = data;
	__update_color_chooser(dialog);
}

static void __color_chooser_changed_cb(GtkColorSelection *selection,
                                       gpointer data) {
	KibaGradientChooserDialog *dialog = data;
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(dialog);
	GList *active;
	GdkColor col;
	gchar *text;
	gint position;
	unsigned int alpha;
	gchar *currentText;
	gtk_color_selection_get_current_color (selection, &col);
	
	alpha = gtk_color_selection_get_current_alpha(GTK_COLOR_SELECTION(priv->colorChooser));
	text = g_strdup_printf("#%02X%02X%02X%02X", col.red/256, col.green/256, col.blue/256, alpha/256);
	
	currentText = gtk_combo_box_get_active_text(GTK_COMBO_BOX(priv->combobox));
	if (!currentText || !text || strcmp(currentText, text) == 0)
		return;
	
	position = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->combobox));
	
	g_signal_handlers_disconnect_by_func(G_OBJECT(priv->combobox),
		G_CALLBACK( __combobox_changed_cb), dialog);
	gtk_combo_box_remove_text (GTK_COMBO_BOX(priv->combobox), position);
	gtk_combo_box_insert_text (GTK_COMBO_BOX(priv->combobox), position, text);
	gtk_combo_box_set_active(GTK_COMBO_BOX(priv->combobox), position);
	g_signal_connect(G_OBJECT(priv->combobox), "changed",
		G_CALLBACK(__combobox_changed_cb), dialog);
	
	active = g_list_nth(priv->colors, position);
	if (active->data)
		g_free (active->data);
	active->data = text;
	
	g_signal_emit (dialog,
		kiba_gradient_chooser_dialog_signals[SIG_CHANGED],
		0);
}

static void __add_button_clicked_cb (GtkButton *button,
                                     gpointer data) {
	KibaGradientChooserDialog *dialog = data;
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(dialog);
	kiba_gradient_chooser_dialog_add_color(dialog, "#000000FF");
	gtk_combo_box_set_active(GTK_COMBO_BOX(priv->combobox), g_list_length(priv->colors)-1);
	
	g_signal_emit (dialog,
		kiba_gradient_chooser_dialog_signals[SIG_CHANGED],
		0);
}

static void __remove_button_clicked_cb (GtkButton *button,
                                        gpointer data) {
	KibaGradientChooserDialog *dialog = data;
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(dialog);
	gint position;
	
	if (g_list_length(priv->colors) <= 1)
		return;
	
	position = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->combobox));
	
	if (position == 0)
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->combobox), 1);
	else
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->combobox), position-1);
	gtk_combo_box_remove_text (GTK_COMBO_BOX(priv->combobox), position);
	
	priv->colors = g_list_remove(priv->colors, g_list_nth_data(priv->colors, position));
	
	g_signal_emit (dialog,
		kiba_gradient_chooser_dialog_signals[SIG_CHANGED],
		0);
}

static void __destroy_cb (GtkWidget *button,
                          gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

/** public Methods **/

void kiba_gradient_chooser_dialog_add_color(KibaGradientChooserDialog *chooser,
                                     const char *color) {
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(chooser);
	priv->colors = g_list_append(priv->colors, g_strdup(color));
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->combobox), color);
	if (g_list_length(priv->colors) == 1) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->combobox), 0);
		__update_color_chooser(chooser);
	}
}

GList* kiba_gradient_chooser_dialog_get_colors(KibaGradientChooserDialog *chooser) {
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(chooser);
	return priv->colors;
}

/**
Class Functions
**/

static void kiba_gradient_chooser_dialog_class_init (KibaGradientChooserDialogClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaGradientChooserDialogPrivate));
	
	kiba_gradient_chooser_dialog_signals[SIG_CHANGED] =
		g_signal_new ("changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaGradientChooserDialogClass, changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
}

static void kiba_gradient_chooser_dialog_init (KibaGradientChooserDialog *dialog) {
	KibaGradientChooserDialogPrivate *priv = KIBA_GRADIENT_CHOOSER_DIALOG_GET_PRIVATE(dialog);
	GtkWidget *vbox, *hbox, *bhbox, *button, *label;
	GdkPixbuf *kiba_icon;
	
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
	
	kiba_icon = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gtk_window_set_icon (GTK_WINDOW (dialog), kiba_icon);
	g_object_unref (kiba_icon);
	gtk_window_set_title(GTK_WINDOW(dialog), _("Gradient Chooser"));
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
	
	vbox = gtk_vbox_new(FALSE, 10);
	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(dialog), vbox);
	
	hbox = gtk_hbox_new(FALSE, 5);
	gtk_widget_show(hbox);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	
	priv->combobox = gtk_combo_box_new_text();
	g_signal_connect(G_OBJECT(priv->combobox), "changed",
		G_CALLBACK(__combobox_changed_cb), dialog);
	gtk_widget_show(priv->combobox);
	gtk_box_pack_start(GTK_BOX(hbox), priv->combobox, TRUE, TRUE, 0);
	
	bhbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(bhbox);
	gtk_box_pack_start(GTK_BOX(hbox), bhbox, FALSE, FALSE, 0);
	
	button = gtk_button_new_from_stock(GTK_STOCK_ADD);
	g_signal_connect(G_OBJECT(button), "clicked",
		G_CALLBACK(__add_button_clicked_cb), dialog);
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(bhbox), button, TRUE, FALSE, 0);
	
	button = gtk_button_new_from_stock(GTK_STOCK_REMOVE);
	g_signal_connect(G_OBJECT(button), "clicked",
		G_CALLBACK(__remove_button_clicked_cb), dialog);
	gtk_widget_show(button);
	gtk_box_pack_start(GTK_BOX(bhbox), button, TRUE, FALSE, 0);
	
	priv->colorChooser = gtk_color_selection_new();
	gtk_color_selection_set_has_opacity_control(GTK_COLOR_SELECTION(priv->colorChooser), TRUE);
	g_signal_connect(G_OBJECT(priv->colorChooser), "color-changed",
		G_CALLBACK(__color_chooser_changed_cb), dialog);
	gtk_widget_show(priv->colorChooser);
	gtk_box_pack_start(GTK_BOX(vbox), priv->colorChooser, TRUE, TRUE, 0);
	
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(hbox);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	
	button = gtk_button_new_from_stock(GTK_STOCK_QUIT);
	g_signal_connect(G_OBJECT(button), "clicked",
		G_CALLBACK(__destroy_cb), dialog);
	gtk_widget_show(button);
	gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
}

/**
 * kiba_gradient_chooser_new:
 *
 * Creates and returns a KibaGradientChooserDialog instance.
 *
 * Returns: The KibaGradientChooserDialog instance created.
 */
KibaGradientChooserDialog* kiba_gradient_chooser_dialog_new (void) {
	return g_object_new (KIBA_TYPE_GRADIENT_CHOOSER_DIALOG, NULL);
}
