/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <gtk/gtk.h>

#include "kiba-prefs.h"
#include "kiba-gradient-chooser.h"
#include "kiba-gradient-chooser-dialog.h"
#include "kiba-cairo-utils.h"

#define KIBA_GRADIENT_CHOOSER_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_GRADIENT_CHOOSER, KibaGradientChooserPrivate))

G_DEFINE_TYPE (KibaGradientChooser, kiba_gradient_chooser, GTK_TYPE_HBOX)

typedef struct {
	GtkWidget *button;
	GtkWidget *drawingArea;
	KibaGradientChooserDialog *dialog;
	
	GList *colors;
} KibaGradientChooserPrivate;

enum {
	SIG_CHANGED = 0,
	SIG_KIBA_GRADIENT_CHOOSER_LAST
};

static guint kiba_gradient_chooser_signals[SIG_KIBA_GRADIENT_CHOOSER_LAST] = { 0 };

/** Private Methods **/

static gboolean __drawing_area_expose_cb (GtkWidget *widget,
                                          GdkEventExpose *expose,
                                          gpointer data) {
	KibaGradientChooser *chooser = data;
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern;
	gint i;
	
	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	
	KibaRGBA rgba[g_list_length(priv->colors)];
	for (i = 0; i < g_list_length(priv->colors); i++)
		rgba[i] = kiba_prefs_rgba_from_string((char*) g_list_nth_data(priv->colors, i));
	
	cairo_paint_shape (cr,
		KIBA_ROUNDED, KIBA_ROUNDED,
		KIBA_ROUNDED, KIBA_ROUNDED,
		1.0f, 1.0f,
		widget->allocation.width - 2.0f,
		widget->allocation.height - 2.0f, 5.0f);
	cairo_stroke_preserve(cr);
	pattern = cairo_get_gradient_pattern (0, 0,
			widget->allocation.width,
			widget->allocation.height,
			rgba, G_N_ELEMENTS(rgba));
	cairo_set_source (cr, pattern);
	cairo_fill (cr);
	cairo_pattern_destroy (pattern);
	
	cairo_destroy (cr);
	
	return TRUE;
}

static void __dialog_changed_cb (KibaGradientChooserDialog *dialog,
                                 gpointer data) {
	KibaGradientChooser *chooser = data;
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	GList *l;
	
	for (l = priv->colors; l != NULL; l = priv->colors)
		priv->colors = g_list_remove(priv->colors, l->data);
	for (l = kiba_gradient_chooser_dialog_get_colors(dialog); l != NULL; l = l->next)
		priv->colors = g_list_append(priv->colors, l->data);
	
	gtk_widget_queue_draw(priv->drawingArea);
	
	g_signal_emit (chooser,
		kiba_gradient_chooser_signals[SIG_CHANGED],
		0);
}

static void __dialog_destroy_cb (GtkWidget *widget,
                                 gpointer data) {
	KibaGradientChooser *chooser = data;
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	priv->dialog = NULL;
}

static gboolean __button_clicked_cb (GtkWidget *widget,
                                     gpointer data) {
	KibaGradientChooser *chooser = data;
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	
	if (GTK_IS_WINDOW(priv->dialog))
		gtk_window_present(GTK_WINDOW(priv->dialog));
	else {
		GList *l;
		
		priv->dialog = kiba_gradient_chooser_dialog_new();
		for (l = priv->colors; l != NULL; l = l->next)
			kiba_gradient_chooser_dialog_add_color(priv->dialog, (char*) l->data);
		gtk_window_present(GTK_WINDOW(priv->dialog));
		g_signal_connect(G_OBJECT(priv->dialog), "changed",
			G_CALLBACK(__dialog_changed_cb), chooser);
		g_signal_connect(G_OBJECT(priv->dialog), "destroy",
			G_CALLBACK(__dialog_destroy_cb), chooser);
	}
	
	return TRUE;
}

/** public Methods **/

void kiba_gradient_chooser_add_color(KibaGradientChooser *chooser,
                                     const char          *color) {
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	priv->colors = g_list_prepend(priv->colors, g_strdup(color));
	gtk_widget_queue_draw(priv->drawingArea);
}

void kiba_gradient_chooser_clear(KibaGradientChooser *chooser) {
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	GList *l;
	
	if(priv->dialog) {
		gtk_widget_destroy(GTK_WIDGET(priv->dialog));
		priv->dialog = NULL;
	}
	
	for (l = priv->colors; l != NULL; l = priv->colors)
		priv->colors = g_list_remove(priv->colors, l->data);
	gtk_widget_queue_draw(priv->drawingArea);
}

GList* kiba_gradient_chooser_get_colors(KibaGradientChooser *chooser) {
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	return priv->colors;
}

/**
Class Functions
**/

static void kiba_gradient_chooser_class_init (KibaGradientChooserClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaGradientChooserPrivate));
	
	kiba_gradient_chooser_signals[SIG_CHANGED] =
		g_signal_new ("changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaGradientChooserClass, changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
}

static void kiba_gradient_chooser_init (KibaGradientChooser *chooser) {
	KibaGradientChooserPrivate *priv = KIBA_GRADIENT_CHOOSER_GET_PRIVATE(chooser);
	GtkWidget *image;
	
	priv->button = gtk_button_new();
	image = gtk_image_new_from_stock(GTK_STOCK_EDIT, GTK_ICON_SIZE_DIALOG);
	gtk_button_set_image(GTK_BUTTON(priv->button), image);
	gtk_box_pack_start(GTK_BOX(chooser), priv->button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(priv->button), "clicked",
		G_CALLBACK(__button_clicked_cb), chooser);
	gtk_widget_show(priv->button);
	
	priv->drawingArea = gtk_drawing_area_new();
	gtk_box_pack_start(GTK_BOX(chooser), priv->drawingArea, TRUE, TRUE, 0);
	g_signal_connect(G_OBJECT(priv->drawingArea), "expose-event",
		G_CALLBACK(__drawing_area_expose_cb), chooser);
	gtk_widget_show(priv->drawingArea);
	
	priv->dialog = NULL;
}

/**
 * kiba_gradient_chooser_new:
 *
 * Creates and returns a KibaGradientChooser instance.
 *
 * Returns: The KibaGradientChooser instance created.
 */
GtkWidget* kiba_gradient_chooser_new (void) {
	return g_object_new (KIBA_TYPE_GRADIENT_CHOOSER, NULL);
}
