/*
 *  Copyright (C) 2007 Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Torbjörn Svangård  <trptorbjorn@gmail.com>
 *              Daniel    Bessler    <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <locale.h>

#include <libxml/parser.h>

#include "kiba-settings.h"
#include "settings-dbus.h"
#include "kiba-prefs.h"
#include "prefs-private.h"
#include "kiba-prefs-schema.h"
#include "widgets.h"

static GtkTooltips  *tooltips = NULL;

static PrefMember kiba_settings_options[] = {
	{ "kiba_settings_geometry", "650x400-1-1", KIBA_STRING }
};

static gboolean type_is_valid (const char *type) {
	const char *kiba_pref_types[] = {
		"char", "int", "double", "bool", "list"
	};
	
	int i;
	
	for (i = 0; i < G_N_ELEMENTS (kiba_pref_types); i++) {
		if (strcmp (kiba_pref_types[i], type) == 0)
			return TRUE;
	}
	
	fprintf (stderr, "Invalid Type '%s'\n"
			 "Please use one of the following types: ",
			 type);
	for (i = 0; i < G_N_ELEMENTS (kiba_pref_types); i++) {
		fprintf (stderr, "%s", kiba_pref_types[i]);
		if (i + 1 < G_N_ELEMENTS (kiba_pref_types))
			fprintf (stderr, ", ");
		else
			fprintf (stderr, "\n");
	}
	
	return FALSE;
}

GtkTooltips* kiba_settings_get_tooltips (void) {
	if (!tooltips)
		tooltips = gtk_tooltips_new();
	
	return tooltips;
}

int main (int argc, char *argv[]) {
	KibaPrefs *prefs;
	KibaSettings *settings;
	KibaPrefsFile *file;
	KibaPrefsFile *generalFile;
	gchar *configFile = NULL;
	gchar *userPath, *configPath;
	GDir *configDir;
	GList *l;
	gint i;
	
	gtk_set_locale();
	gtk_disable_setlocale();
	
	setlocale (LC_NUMERIC, "C");  /* use decimal point instead of comma */
	setlocale (LC_CTYPE, "");
	setlocale (LC_MESSAGES, "");
	
#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, DATADIR"/locale");
	textdomain (PACKAGE);
	bind_textdomain_codeset (PACKAGE, "UTF-8");	
#endif
	
	gtk_init(&argc, &argv);
	
	i = 0;
	while (i < argc) {
		if (strcmp (argv[i], "--help") == 0 ||
		    strcmp (argv[i], "-h") == 0) {
			fprintf (stderr,
				 "Usage: kiba-settings [--get-'$type' '$name']"
						  "[--set-'$type' '$name' '$value']"
						  "[--config-file config_file]"
						  "...\n");
			kiba_prefs_fini();
			return 0;
		}
		else if (g_strrstr (argv[i], "--config-file") != NULL) {
			configFile = g_strdup (argv[i+1]);
			if (!g_file_test (configFile, G_FILE_TEST_EXISTS)) {
				g_free (configFile);
				configFile = NULL;
			}
			break;
		}
		i++;
	}
	
	prefs = kiba_prefs_get();
	
	/* create kiba home directory */
	userPath = g_strdup_printf ("%s/.kiba-dock", g_get_home_dir ());
	if (!g_file_test (userPath, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (userPath, 7664) != 0) {
			fprintf (stderr, "Failed to create user directory at '%s'!\n", userPath);
			exit(0);
		}
	}
	
	/* create config directory */
	configPath = g_strdup_printf ("%s/config", userPath);
	g_free(userPath);
	if (!g_file_test (configPath, G_FILE_TEST_IS_DIR)) {
		if (g_file_test (configPath, G_FILE_TEST_EXISTS))
			g_remove(configPath);
		if (g_mkdir (configPath, 7664) != 0) {
			fprintf (stderr, "Failed to create config directory at '%s'!\n", configPath);
			g_free(configPath);
			exit(0);
		}
	}
	
	/* adding general config file */
	char *generalConfigPath = g_strdup_printf ("%s/general.xml", configPath);
	generalFile = kiba_prefs_add_file_with_type(generalConfigPath, KIBA_PREFS_FILE_GENERAL);
	kiba_prefs_update_config_file(prefs, generalFile);
	g_free(generalConfigPath);
	
	/* create dock config directory */
	char *dockConfigPath = g_strdup_printf ("%s/docks", configPath);
	g_free(configPath);
	if (!g_file_test (dockConfigPath, G_FILE_TEST_IS_DIR)) {
		if (g_file_test (dockConfigPath, G_FILE_TEST_EXISTS))
			g_remove(dockConfigPath);
		if (g_mkdir (dockConfigPath, 7664) != 0) {
			fprintf (stderr, "Failed to create dock config directory at '%s'!\n", dockConfigPath);
			g_free(dockConfigPath);
			exit(EXIT_FAILURE);
		}
	}
	
	/* adding global config file */
	char *globalConfigPath = g_strdup_printf ("%s/" KIBA_PREFS_GLOBAL_FILE, dockConfigPath);
	kiba_prefs_add_file_with_type(globalConfigPath, KIBA_PREFS_FILE_GLOBAL);
	g_free(globalConfigPath);
	
	/* initiall preferences */
	if (configFile && !g_str_has_prefix(configFile, dockConfigPath)) {
		fprintf (stderr, "Invalid config directory argument '%s'!\n", configFile);
		g_free(configFile);
		configFile = NULL;
	}
	if (!configFile)
		configFile = g_strdup_printf ("%s/default_dock.xml", dockConfigPath);
	file = kiba_prefs_add_config_file(configFile);
	kiba_prefs_update_config_file(prefs, file);
	g_free(configFile);
	
	i = 0;
	while (i < argc) {
		/* get Value from settings file */
		if (g_strrstr (argv[i], "--get-")) {
			char *value;
			
			g_free(dockConfigPath);
			
			if (!type_is_valid (argv[i]+6)) {
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+1 >= argc) {
				fprintf (stderr, "No Setting Name defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			value = kiba_prefs_get_option_from_file (file, argv[i+1], argv[i]+6);
			if (!value)
				fprintf (stderr, "'%s' not found\n", argv[i+1]);
			else
				fprintf (stderr, "%s\n", value);
			
			kiba_prefs_fini();
			exit (0);
		}
		/* write Value to settings */
		else if (g_strrstr (argv[i], "--set-")) {
			g_free(dockConfigPath);
			
			if (!type_is_valid (argv[i]+6)) {
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+1 >= argc) {
				fprintf (stderr, "No Setting Name defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+2 >= argc) {
				fprintf (stderr, "No Setting Value defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			
			if (!kiba_prefs_write_to_config_file (file,
					argv[i+1],
					argv[i]+6,
					argv[i+2]))
				fprintf (stderr, "Failed to set '%s' to '%s'\n",
						 argv[i+1], argv[i+2]);
			
			kiba_prefs_fini();
			exit (0);
		}
		i++;
	}
	
	/* initiall libxml */
	LIBXML_TEST_VERSION
	
	settings = g_new0(KibaSettings, 1);
	
	kiba_settings_dbus_init(settings);
	
	configDir = g_dir_open (dockConfigPath, 0, NULL);
	if (!configDir) {
		g_free(dockConfigPath);
		exit(EXIT_FAILURE);
	}
	
	while (TRUE) {
		const char *file = g_dir_read_name(configDir);
		if (!file) break;
		if (!g_str_has_suffix(file, ".xml") || strcmp(file, KIBA_PREFS_GLOBAL_FILE) == 0) continue;
		char *configPath = g_strdup_printf("%s/%s", dockConfigPath, file);
		KibaPrefsFile *prefFile = kiba_prefs_get_file (configPath);
		if (!prefFile)
			kiba_prefs_add_config_file(configPath);
		g_free(configPath);
	}
	g_dir_close(configDir);
	g_free(dockConfigPath);
	
	/* Load kiba schemas, so that we can acces the data for creating the widgets */
	settings->general_schema = kiba_prefs_schema_get(KIBA_SCHEMA_GENERAL);
	if (g_list_length(settings->general_schema->sections) == 0) {
		g_free(settings);
		fprintf (stderr, "No valid .schema file(s) found\n");
		return EXIT_FAILURE;
	}
	
	settings->schema = kiba_prefs_schema_get(KIBA_SCHEMA_DOCK);
	if (g_list_length(settings->schema->sections) == 0) {
		g_free(settings);
		kiba_prefs_schema_fini(settings->general_schema);
		fprintf (stderr, "No valid .schema file(s) found\n");
		return EXIT_FAILURE;
	}
	
	settings->sectionLabel = NULL;
	settings->sectionDescription = NULL;
	settings->sectionImage = NULL;
	settings->notebook = NULL;
	settings->activeSection = NULL;
	settings->generalFile = generalFile;
	settings->dockFile = file;
	settings->globalFile = FALSE;
	
	kiba_prefs_schema_set_options_from_config(settings->general_schema, generalFile);
	
	/* create the widgets
	   here is a small schema of the Layout:
		VBox {  //<< Main Container
			Frame { //<< Top Container
				HBox {
					Frame { HBox { VButtonBox }},
					VBox {
						Label, //<< Active Section/Description
						Notebook { HBox { SettingWidgets }}}}},
			HBox { Button }} //<< Quit Button
	*/
	GtkWidget *window = kiba_settings_create_window (settings);
	
	kiba_prefs_add_options (generalFile, kiba_settings_options, G_N_ELEMENTS(kiba_settings_options));
	char *geometryString = kiba_prefs_get_string(generalFile, "kiba_settings_geometry", NULL);
	if (geometryString)
		gtk_window_parse_geometry (GTK_WINDOW(window), geometryString);
	
	gtk_widget_show_all (window);
	
	gtk_main();
	
	/* save window geometry */
	gint x,y,width,height;
	gtk_window_get_position(GTK_WINDOW(window), &x, &y);
	gtk_window_get_size(GTK_WINDOW(window), &width, &height);
	geometryString = g_strdup_printf ("%ix%i%s%i%s%i",
		width, height, x < 0 ? "-" : "+", x, y < 0 ? "-" : "+", y);
	kiba_prefs_write_to_config_file(generalFile, "kiba_settings_geometry", "char", geometryString);
	g_free(geometryString);
	
	kiba_prefs_schema_fini (settings->schema);
	kiba_prefs_schema_fini (settings->general_schema);
	g_free(settings);
	kiba_prefs_fini();
	
	return EXIT_SUCCESS;
}

