/*
 *  Copyright (C) 2007 Daniel Bessler  <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 */

#include <glib.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include "kiba-debug.h"
#include "settings-dbus.h"
#include "settings-notebook.h"

#define KIBA_SETTINGS_DBUS_PATH  "/org/kiba/dock/KibaSettings"
#define KIBA_SETTINGS_DBUS_NAMESPACE  "org.kiba.dock.KibaSettings"

/*** Signal handler ***/

/* TODO: catch option changes from the kiba-dock and apply them to kiba-settings.
         currently this function is also called if a option changes from kiba-settings,
         dbus_connection_remove_filter/dbus_connection_add_filter didnt worked, seems like
         dbus_g_proxy_call is processed with a delay or something.
*/
static DBusHandlerResult __signal_filter (DBusConnection *dbus,
						      DBusMessage    *message,
						      void           *user_data) {
	KibaSettings *settings = user_data;
	DBusError error;
	gchar *configPath;
	gchar *option_id;
	
	dbus_error_init (&error);
	
	if (dbus_message_is_signal (message, KIBA_SETTINGS_DBUS_NAMESPACE, "IntOptionChanged")) {
		gint value;
		
		dbus_message_get_args (message, &error,
					DBUS_TYPE_STRING, &configPath,
					DBUS_TYPE_STRING, &option_id,
					DBUS_TYPE_INT32, &value,
					DBUS_TYPE_INVALID);
		dbus_error_free (&error);
		
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	else if (dbus_message_is_signal (message, KIBA_SETTINGS_DBUS_NAMESPACE, "DoubleOptionChanged")) {
		gdouble value;
		
		dbus_message_get_args (message, &error,
					DBUS_TYPE_STRING, &configPath,
					DBUS_TYPE_STRING, &option_id,
					DBUS_TYPE_DOUBLE, &value,
					DBUS_TYPE_INVALID);
		dbus_error_free (&error);
		
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	else if (dbus_message_is_signal (message, KIBA_SETTINGS_DBUS_NAMESPACE, "BoolOptionChanged")) {
		gboolean value;
		
		dbus_message_get_args (message, &error,
					DBUS_TYPE_STRING, &configPath,
					DBUS_TYPE_STRING, &option_id,
					DBUS_TYPE_BOOLEAN, &value,
					DBUS_TYPE_INVALID);
		dbus_error_free (&error);
		
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	else if (dbus_message_is_signal (message, KIBA_SETTINGS_DBUS_NAMESPACE, "CharOptionChanged")) {
		gchar *value;
		
		dbus_message_get_args (message, &error,
					DBUS_TYPE_STRING, &configPath,
					DBUS_TYPE_STRING, &option_id,
					DBUS_TYPE_STRING, &value,
					DBUS_TYPE_INVALID);
		dbus_error_free (&error);
		
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	else if (dbus_message_is_signal (message, KIBA_SETTINGS_DBUS_NAMESPACE, "ListOptionChanged")) {
		gchar *value;
		
		dbus_message_get_args (message, &error,
					DBUS_TYPE_STRING, &configPath,
					DBUS_TYPE_STRING, &option_id,
					DBUS_TYPE_STRING, &value,
					DBUS_TYPE_INVALID);
		dbus_error_free (&error);
		
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/*** dbus method calls ***/

void kiba_settings_dbus_set_int (const gchar *configPath,
                                 const gchar *option_id,
                                 gint value) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "SetIntOption", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_STRING, option_id,
			G_TYPE_INT, value,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_set_bool (const gchar *configPath,
                                  const gchar *option_id,
                                  gboolean value) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "SetBoolOption", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_STRING, option_id,
			G_TYPE_BOOLEAN, value,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_set_float (const gchar *configPath,
                                   const gchar *option_id,
                                   gfloat value) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "SetDoubleOption", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_STRING, option_id,
			G_TYPE_DOUBLE, value,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_set_char (const gchar *configPath,
                                  const gchar *option_id,
                                  const gchar *value) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "SetCharOption", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_STRING, option_id,
			G_TYPE_STRING, value,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_set_list (const gchar *configPath,
                                  const gchar *option_id,
                                  const gchar *value) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "SetListOption", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_STRING, option_id,
			G_TYPE_STRING, value,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_add_dock (const gchar *configPath) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "AddDockConfigFile", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_INVALID);
}

void kiba_settings_dbus_remove_dock (const gchar *configPath) {
	DBusGConnection *connection;
	DBusConnection *bus;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	
	if (connection == NULL)
		return;
	
	bus = dbus_g_connection_get_connection (connection);
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	g_clear_error (&error);
	
	dbus_g_proxy_call(proxy, "RemoveDockConfigFile", &error,
			G_TYPE_STRING, configPath,
			G_TYPE_INVALID);
}

/*** Init ***/

static void listenKibaSettingsSignals (DBusGConnection *connection, KibaSettings *settings) {
	DBusConnection *bus;
	DBusError error;
	
	bus = dbus_g_connection_get_connection (connection);
	
	dbus_error_init (&error);
	dbus_bus_add_match (bus, "type='signal',interface='org.kiba.dock.KibaSettings'", &error);
	dbus_connection_add_filter (bus, __signal_filter, settings, NULL);
}

void kiba_settings_dbus_init (KibaSettings *settings) {
	DBusGConnection *connection;
	DBusGProxy *proxy;
	GError *error = NULL;
	
	connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
	if (connection == NULL)
		return;
	
	proxy = dbus_g_proxy_new_for_name(
			connection,
			KIBA_SETTINGS_DBUS_NAMESPACE,
			KIBA_SETTINGS_DBUS_PATH,
			KIBA_SETTINGS_DBUS_NAMESPACE);
	
	listenKibaSettingsSignals(connection, settings);
}
