/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libintl.h>

#include "kiba-prefs-schema.h"
#include "kiba-prefs.h"
#include "kiba-dialog.h"
#include "prefs-private.h"

#include "settings-notebook.h"
#include "kiba-settings.h"
#include "kiba-gradient-chooser.h"
#include "settings-dbus.h"
#include "widgets.h"

#define _(string) gettext (string)

#define KIBA_DATA_DIR DATADIR "/kiba-dock"

#define EDITABLE_RANGE "editable"
#define TEXT_RANGE "text"
#define ICON_RANGE "pngfile"
#define RGBA_RANGE "rgba"
#define RGB_RANGE "rgb"
#define ENCRYPTED_RANGE "encrypt"

#define SYSINFO_THEME_DIR  KIBA_DATA_DIR "/icons/sysinfo_themes"
#define CAIRO_CLOCK_THEME_DIR "/usr/share/cairo-clock/themes/"
#define CLOCK_DIR "/clock"
#define SYSINFO_DIR "/sysinfo"

struct _KibaSettingsNotebook {
	GtkNotebook parent;
};

struct _KibaSettingsNotebookClass {
	GtkNotebookClass class;
};

G_DEFINE_TYPE (KibaSettingsNotebook, kiba_settings_notebook, GTK_TYPE_NOTEBOOK)

enum {
	TITLE_COLUMN,
	N_COLUMNS
};

static KibaSettings* kiba_settings = NULL;

/*** private methods ***/

static char *fixName(char *nameToFix) {
	char *ret;
	
	ret = g_strdup(g_ascii_strdown(nameToFix, strlen(nameToFix)));
	g_strdelimit(ret, "_", ' ');
	
	return ret;
}

typedef struct {
	GtkWidget *widget;
	SchemaMember *member;
}ResetClosure;

static void combobox_set_value (GtkWidget  *combobox,
				  const char *default_value) {
	GtkTreeModel *model;
	GtkTreeIter   iter;
	gboolean      valid;
	
	model = gtk_combo_box_get_model (GTK_COMBO_BOX (combobox));
	valid = gtk_tree_model_get_iter_first (model, &iter);
	
	while (valid) {
		gchar *str_data;
	
		gtk_tree_model_get (model, &iter, 0, &str_data, -1);
		if (strcmp (default_value, str_data) == 0) {
			gtk_combo_box_set_active_iter (
				GTK_COMBO_BOX (combobox), &iter);
			break;
		}
		g_free (str_data);
		
		valid = gtk_tree_model_iter_next (model, &iter);
	}
}

static void __involve_dbus(const char *id,
                           const char *value,
                           const char *type,
                           const char *configPath) {
	if (strcmp (type, "int") == 0)
		kiba_settings_dbus_set_int (configPath, id, atoi(value));
	else if (strcmp (type, "double") == 0)
		kiba_settings_dbus_set_float (configPath, id, g_ascii_strtod(value, NULL));
	else if (strcmp (type, "char") == 0)
		kiba_settings_dbus_set_char (configPath, id, value);
	else if (strcmp (type, "bool") == 0)
		kiba_settings_dbus_set_bool (configPath, id,
			(gboolean) (strcmp(value, "TRUE") == 0));
	else if (strcmp (type, "list") == 0)
		kiba_settings_dbus_set_list (configPath, id, value);
}

static void reset_to_default (GtkButton *button,
			      gpointer   user_data) {
	ResetClosure *closure = user_data;
	KibaPrefsFile *file;
	KibaPrefsFile *globalFile;
	
	globalFile = kiba_prefs_get_global_file(kiba_prefs_get());
	if (kiba_settings->globalFile)
		file = globalFile;
	else
		file = (closure->member->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	const char *path = kiba_prefs_get_file_path(file);
	
	if (file != globalFile && kiba_prefs_has_global_option(file, closure->member->widget_name)) {
		kiba_settings->globalFile = TRUE;
		reset_to_default(button, user_data);
		kiba_settings->globalFile = FALSE;
		gtk_widget_set_sensitive(closure->widget, TRUE);
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(), GTK_WIDGET(button), _("Reset to default") , NULL);
		return;
	}
	
	const char *setting_name = gtk_widget_get_name (closure->widget);
	g_return_if_fail (setting_name != NULL);
	
	if (!kiba_settings->globalFile) {
		if (GTK_IS_CHECK_BUTTON (closure->widget))
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (closure->widget),
				strcmp (closure->member->default_value, "TRUE") == 0);
		else if (GTK_IS_COLOR_BUTTON (closure->widget)) {
			GdkColor col;
			unsigned int r,g,b, a;
			
			sscanf (closure->member->default_value,
					"#%02X%02X%02X%02X",
					&r, &g, &b, &a);
			col.red    = (float) r * 256;
			col.green  = (float) g * 256;
			col.blue   = (float) b * 256;
			gtk_color_button_set_color(GTK_COLOR_BUTTON (closure->widget), &col);
		
			if (gtk_color_button_get_use_alpha (GTK_COLOR_BUTTON (closure->widget)))
				gtk_color_button_set_alpha  (GTK_COLOR_BUTTON (closure->widget), a * 256);
		}
		else if (GTK_IS_SCALE (closure->widget))
			gtk_range_set_value (GTK_RANGE (closure->widget),
				g_ascii_strtod (closure->member->default_value, NULL));
		else if (GTK_IS_SPIN_BUTTON (closure->widget))
			gtk_spin_button_set_value (GTK_SPIN_BUTTON (closure->widget),
				g_ascii_strtod (closure->member->default_value, NULL));
		else if (GTK_IS_ENTRY (closure->widget))
			gtk_entry_set_text (GTK_ENTRY (closure->widget),
				closure->member->default_value);
		else if (GTK_IS_FILE_CHOOSER (closure->widget))
			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (closure->widget),
							closure->member->default_value);
		else if (GTK_IS_COMBO_BOX (closure->widget)) {
			combobox_set_value (closure->widget,
				closure->member->default_value);
		}
		else if (KIBA_IS_GRADIENT_CHOOSER (closure->widget)) {
			GSList *items, *l;
			kiba_gradient_chooser_clear(KIBA_GRADIENT_CHOOSER(closure->widget));
			items = kiba_prefs_get_list_from_string (closure->member->default_value);
			for (l = items; l != NULL; l = l->next)
				kiba_gradient_chooser_add_color(KIBA_GRADIENT_CHOOSER(closure->widget), (char*) l->data);
	// 		g_signal_connect(G_OBJECT(widget), "changed",
	// 			G_CALLBACK(__gradient_chooser_changed), option);
	//TODO
		}
	//TODO list widgets
		else if (GTK_IS_HBOX (closure->widget)) {
			GList     *childs;
			GtkWidget *combobox;
			
			childs    = gtk_container_get_children (GTK_CONTAINER (closure->widget));
			combobox  = childs->data;
			combobox_set_value (combobox, closure->member->default_value);
			g_list_free (childs);
		}
		
		kiba_prefs_write_to_config_file (file,
				closure->member->widget_name,
				closure->member->type,
				closure->member->default_value);
		
		__involve_dbus(closure->member->widget_name,
				closure->member->default_value,
				closure->member->type, path);
	} else {
		kiba_prefs_remove_from_config_file (file,
				closure->member->widget_name,
				closure->member->type);
		
		kiba_settings_dbus_set_char (path, closure->member->widget_name, "REMOVE");
	}
	status_bar_push_message("Config File Saved");
}

static void set_spinbutton_value_from_range (GtkRange *range,
					     gpointer  user_data) {
	GtkWidget *spinbutton = user_data;
	
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),
		gtk_range_get_value (range));
}

static void set_range_value_from_spinbutton (GtkSpinButton *spinbutton,
					     gpointer       user_data) {
	GtkWidget *range = user_data;
	
	gtk_range_set_value (GTK_RANGE (range),
		gtk_spin_button_get_value (spinbutton));
}

static void checkbutton_cb (GtkButton *check_button,
		   	    gpointer   user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	const char *name;
	const char *value;
	
	name = gtk_widget_get_name (GTK_WIDGET (check_button));
	value = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(check_button)) ? "TRUE" : "FALSE";
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "bool", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_bool (path, name, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(check_button)));
	
	status_bar_push_message("Config File Saved");
}

static void __gradient_chooser_changed (KibaGradientChooser *chooser,
                                        gpointer data) {
	SchemaMember *settings = data;
	KibaPrefsFile *file = (settings->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	GList *colors = kiba_gradient_chooser_get_colors(chooser);
	char *text;
	
	const char* name = gtk_widget_get_name (GTK_WIDGET (chooser));
	
	if (settings->value) g_free(settings->value);
	settings->value = kiba_prefs_list_to_string(colors);
	kiba_prefs_write_to_config_file (file, name, "list", settings->value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_list (path, name, settings->value);
	
	status_bar_push_message("Config File Saved");
}

static void double_range_cb (GtkRange *range,
			     gpointer  user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	char *value;
	
	const char* name = gtk_widget_get_name (GTK_WIDGET (range));
	
	/* save two decimals */
	value = g_strdup_printf("%0.2f", gtk_range_get_value (range));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "double", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_float (path, name, gtk_range_get_value (range));
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void rgb_cb (GtkColorButton *widget,
		    gpointer        user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	GdkColor col;
	char *value;
	
	gtk_color_button_get_color(widget, &col);
	value = g_strdup_printf("#%02X%02X%02X",
				col.red / 256,
				col.green / 256,
				col.blue / 256);
	const char* name = gtk_widget_get_name(GTK_WIDGET(widget));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void rgba_cb (GtkColorButton *widget,
		     gpointer        user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	GdkColor col;
	int   alpha;
	char *value;
	
	gtk_color_button_get_color(widget, &col);
	alpha = gtk_color_button_get_alpha(widget);
	value = g_strdup_printf("#%02X%02X%02X%02X",
				col.red / 256,
				col.green / 256,
				col.blue / 256,
				alpha / 256);
	const char* name = gtk_widget_get_name(GTK_WIDGET(widget));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void combobox_cb (GtkComboBox *combobox,
			 gpointer     user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	const char* name = gtk_widget_get_name (GTK_WIDGET(combobox));
	const char* value = gtk_combo_box_get_active_text (combobox);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
}

static void font_cb (GtkFontButton *font_button,
		     gpointer       user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	const char* name  = gtk_widget_get_name (GTK_WIDGET(font_button));
	const char* value = gtk_font_button_get_font_name (font_button);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
}

static void entry_cb (GtkEntry *entry,
		      gpointer  user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	
	const char *name  = gtk_widget_get_name(GTK_WIDGET(entry));
	const char* value = gtk_entry_get_text (entry);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
}

static void filechooser_cb (GtkFileChooser *chooser,
			    gpointer        user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	const char *name  = gtk_widget_get_name (GTK_WIDGET(chooser));
	const char* value = gtk_file_chooser_get_filename (chooser);
	if (!value || strlen (value) < 2 || strcmp (value, "(null)") == 0 ||
	    g_str_has_suffix (value, ".desktop") || strcmp(value, data->value) == 0)
		return;
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (file, name, "char", value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_char (path,  name, value);
	status_bar_push_message("Config File Saved");
}

static void spinbutton_cb (GtkSpinButton *button,
			   gpointer       user_data) {
	SchemaMember *data = user_data;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	float value;
	char *char_value;
	
	const char *name  = gtk_widget_get_name (GTK_WIDGET (button));
	
	value = gtk_spin_button_get_value (button);
	char_value = g_strdup_printf ("%g", value);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(char_value);
	kiba_prefs_write_to_config_file (file, name, "int", char_value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_int (path, name, value);
	status_bar_push_message("Config File Saved");
	
	g_free (char_value);
}

void update_preview_cb (GtkFileChooser *file_chooser,
			gpointer        data) {
	GtkWidget *preview;
	GdkPixbuf *pixbuf;
	
	preview = GTK_WIDGET(data);
	const char* filename = gtk_file_chooser_get_preview_filename(file_chooser);
	
	if (filename != NULL) {
		pixbuf = gdk_pixbuf_new_from_file_at_size(filename, 128, 128, NULL);
		
		gtk_image_set_from_pixbuf (GTK_IMAGE(preview), pixbuf);
		
		gtk_file_chooser_set_preview_widget_active(file_chooser, pixbuf != NULL);
		if (pixbuf)
			gdk_pixbuf_unref(pixbuf);
	}
}

static void add_dialog_destroy_cb (GtkButton *check_button,
		   		   gpointer   user_data) {
	GtkWidget *dialog = user_data;
	
	gtk_widget_destroy (dialog);
}

static void add_dialog_accept_cb (GtkButton *button,
		   		  gpointer   user_data) {
	GtkTreeView *view = user_data;
	SchemaMember *settings;
	GtkTreeStore *store;
	GtkTreeIter iter;
	GtkWidget *entry;
	const char *text;
	char *buf;
	
	entry = g_object_get_data (G_OBJECT(button), "entry-widget");
	settings = g_object_get_data (G_OBJECT(view), "Pref-Member");
	text = gtk_entry_get_text (GTK_ENTRY(entry));
	
	KibaPrefsFile *file = (settings->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	
	if (!text || strlen(text) < 3)
		return;
	
	buf = g_strdup (settings->value);
	buf[strlen(buf)-1] = '\0';
	if (strcmp (buf, "[]") == 0)
		settings->value = g_strdup_printf ("[%s]", text);
	else
		settings->value = g_strdup_printf ("%s,%s]", buf, text);
	kiba_prefs_write_to_config_file (file, settings->widget_name, "list", settings->value);
	const char *path = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_list (path,  settings->widget_name, settings->value);
	g_free (buf);
	
	store = GTK_TREE_STORE(gtk_tree_view_get_model(view));
	gtk_tree_store_append (store, &iter, NULL);  /* Acquire an iterator */
	
	gtk_tree_store_set (store, &iter, TITLE_COLUMN, text, -1);
}

static void cb_add_item_to_list (GtkWidget  *button,
				 gpointer    data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	GtkWidget *dialog, *entry, *button_box, *box, *dialog_button;
	
	dialog = gtk_dialog_new();
	
	box = gtk_vbox_new(TRUE, 5);
	button_box = gtk_hbox_new(FALSE, 5);
	
	entry = gtk_entry_new ();
	gtk_widget_show (entry);
	gtk_box_pack_start (GTK_BOX (box), entry, TRUE, TRUE, 0);
	gtk_widget_show (box);
	
	dialog_button = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_object_set_data (G_OBJECT(dialog_button), "entry-widget", entry);
	gtk_box_pack_start (GTK_BOX (button_box), dialog_button, FALSE, FALSE, 0);
	gtk_widget_show (dialog_button);
	g_signal_connect (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_accept_cb), view);
	g_signal_connect_after (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_destroy_cb), dialog);
	dialog_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	gtk_box_pack_start (GTK_BOX (button_box), dialog_button, FALSE, FALSE, 0);
	gtk_widget_show (dialog_button);
	g_signal_connect (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_destroy_cb), dialog);

	gtk_box_pack_start (GTK_BOX (box), button_box, TRUE, TRUE, 0);
	gtk_widget_show (button_box);

	gtk_dialog_add_action_widget (GTK_DIALOG(dialog), box, 13);
	gtk_widget_show (dialog);
}

static void cb_remove_item_from_list_real (GtkTreeModel  *model,
					   GtkTreePath   *path,
					   GtkTreeIter   *iter,
					   gpointer       data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	SchemaMember *settings;
	GSList *list, *l;
	gchar *text, *string = NULL;
	gint count;
	
	settings = g_object_get_data (G_OBJECT(view), "Pref-Member");
	list = kiba_prefs_get_list_from_string(settings->value);
	
	KibaPrefsFile *file = (settings->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	
	gtk_tree_model_get (model, iter, TITLE_COLUMN, &text, -1);
	
	count = 0;
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf = NULL;
		
		if (strcmp (item, text) == 0)
			continue;
		if (string) {
			buf = g_strdup (string);
			g_free (string);
			string = NULL;
		}
		
		if (!count)
			string = g_strdup_printf ("[%s,", item);
		else
			string = g_strdup_printf ("%s%s,", buf, item);
		
		count++;
	}
	if (string)
		string[strlen(string)-1] = ']';
	else
		string = g_strdup ("[]");
	settings->value = string;
	kiba_prefs_write_to_config_file (file, settings->widget_name, "list", string);
	const char *configPath = kiba_prefs_get_file_path(file);
	kiba_settings_dbus_set_list (configPath,  settings->widget_name, string);
	status_bar_push_message("Config File Saved");
	
	gtk_tree_model_row_deleted (model, path);
}

static void cb_remove_item_from_list (GtkWidget *button,
				      gpointer   data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	GtkTreeSelection  *selection;
	
	selection = gtk_tree_view_get_selection (view);
	gtk_tree_selection_selected_foreach (selection,
		cb_remove_item_from_list_real, view);
}

static GtkWidget* get_hbox_from_widget (GtkWidget *widget) {
	GtkWidget *box;
	
	box = gtk_hbox_new (FALSE, 0);
	
	gtk_box_pack_start (GTK_BOX (box),
			    widget,
			    FALSE, FALSE, 0);
	
	return box;
}

static GtkWidget* get_entry_widget (const char* id,
			     const char* value,
			     const char* description) {
	GtkWidget *entry;
	
// 	widget = gtk_event_box_new ();
	entry = gtk_entry_new ();
	
	if (value)
		gtk_entry_set_text (GTK_ENTRY (entry), value);
	
	gtk_widget_set_name(entry, id);
	
	/* fix for tooltips not showing up,
	   remove the eventbox when http://bugzilla.gnome.org/show_bug.cgi?id=303717 is closed */
// 	gtk_container_add (GTK_CONTAINER (widget), entry);
// 	if (description)
// 		gtk_tooltips_set_tip (kiba_settings_get_tooltips(),
// 			widget, description, NULL);
	
// 	return widget;
	return entry;
}

static GtkWidget *create_double_hscale_widget (SchemaMember *data) {
	GtkWidget *widget;
	GtkObject *adjustment;
	char *low = NULL,*high = NULL;
	
	/* data.the[i].range+1  if the first char is - (negative) */
	high = g_strstr_len(data->range+1, strlen(data->range), "-") + 1;
	low = g_strndup(data->range, strlen(data->range) - strlen(high) - 1);
	adjustment = gtk_adjustment_new(g_ascii_strtod(data->value, NULL),
					g_ascii_strtod(low, NULL),
					g_ascii_strtod(high, NULL),
					0.1, 0.0, 0.0);
	g_free(low);
	
	widget = gtk_hscale_new(GTK_ADJUSTMENT(adjustment));
	gtk_scale_set_draw_value (GTK_SCALE(widget), TRUE);
	gtk_scale_set_digits (GTK_SCALE(widget), 2);
	gtk_widget_set_name(widget, data->widget_name);
	
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
 	g_signal_connect (G_OBJECT (widget), "value-changed",
		G_CALLBACK (double_range_cb), data);
	
	return widget;
}

static GtkWidget *create_label_widget (const char *text) {
	GtkWidget *widget;
	
	widget = gtk_label_new(text);
	gtk_misc_set_alignment (GTK_MISC(widget), 0.0, 0.5);
	
	gtk_widget_set_name(widget, text);
	
	return widget;
}

static GtkWidget *create_colour_widget (SchemaMember *data) {
	GtkWidget *widget;
	GdkColor col;
	unsigned int r,g,b;
	
	widget = gtk_color_button_new();
	
	sscanf(data->value, "#%02X%02X%02X", &r, &g, &b);
	col.red    = (float) r * 256;
	col.green  = (float) g * 256;
	col.blue   = (float) b * 256;
	gtk_color_button_set_color((GtkColorButton *)widget, &col);
	
 	g_signal_connect (G_OBJECT (widget), "color-set",
		G_CALLBACK (rgb_cb), data);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	return widget;
}

static GtkWidget *create_rgba_widget (SchemaMember *data) {
	GtkWidget *widget;
	GdkColor col;
	unsigned int r,g,b, alpha;
	
	widget = gtk_color_button_new();
	
	sscanf(data->value, "#%02X%02X%02X%02X", &r, &g, &b, &alpha);
	col.red = (float) r * 256;
	col.green = (float) g * 256;
	col.blue = (float) b * 256;
	gtk_color_button_set_color(GTK_COLOR_BUTTON (widget), &col);
	
	gtk_color_button_set_use_alpha (GTK_COLOR_BUTTON (widget), TRUE);
	gtk_color_button_set_alpha  (GTK_COLOR_BUTTON (widget), alpha * 256);
	
 	g_signal_connect (G_OBJECT (widget), "color-set",
		G_CALLBACK (rgba_cb), data);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	return widget;
}

static void get_clock_themes (GtkWidget *widget, const char *theme_directory) {
	GDir    *dir;
	GError  *error = NULL;
	char    *buf;
	
	dir = g_dir_open (theme_directory, 0, &error);
	if (error != NULL) {
		g_message ("Failed to open Theme Directory %s\notification: %s\n",
					theme_directory, error->message);
		g_error_free (error);
		return;
	}
	
	while ((buf = (char *) g_dir_read_name (dir)) != NULL)
		gtk_combo_box_append_text (GTK_COMBO_BOX(widget), buf);
}

static GtkWidget *create_combobox_widget (SchemaMember *data) {
	GtkWidget *combobox;
	
	combobox = gtk_combo_box_new_text();
	/* append the string from the configfile first to the combobox*/
	if (strcmp(data->value, ICON_RANGE) != 0 && !g_str_has_suffix (data->value, ".png") &&
	   !g_str_has_suffix (data->value, ".svg"))
		gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), data->value);
	
	/* if a range exist */
	if (data->range) {
		guint f;
		gchar **strings = g_strsplit(data->range, ",", -1);
	
		/* append all strings to the combobox... but not the one that is set in config file */
		for (f = 0; f <  g_strv_length(strings); f++)
			if (strcmp(data->value, strings[f]) != 0 &&
			    strcmp(strings[f], ICON_RANGE) != 0 &&
			   !g_str_has_suffix (strings[f], ".png") &&
			   !g_str_has_suffix (strings[f], ".svg"))
				gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), strings[f]);
	
		g_strfreev(strings);
	}
	else if (strcmp(data->widget_name, "clock_theme") == 0) {
		char *theme_directory = NULL;
	
		theme_directory = g_strdup_printf ("%s/.kiba-dock" CLOCK_DIR, g_get_home_dir ());
		if (g_file_test (theme_directory, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, theme_directory);
		g_free (theme_directory);
		
		if (g_file_test (CAIRO_CLOCK_THEME_DIR, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, CAIRO_CLOCK_THEME_DIR);
	}
	else if (strcmp(data->widget_name, "sysinfo_theme") == 0) {
		char *theme_directory = NULL;
	
		theme_directory = g_strdup_printf ("%s/.kiba-dock" SYSINFO_DIR "/themes", g_get_home_dir ());
		if (g_file_test (theme_directory, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, theme_directory);
		g_free (theme_directory);
		
		if (g_file_test (SYSINFO_THEME_DIR, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, SYSINFO_THEME_DIR);
	}
	
	gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0);

	g_signal_connect (G_OBJECT (combobox), "changed",
		G_CALLBACK (combobox_cb), data);
	
	gtk_widget_set_name(combobox, data->widget_name);
	
	return combobox;
}

static GtkWidget *create_checkbox_widget (SchemaMember *data) {
	GtkWidget *widget;
	
	widget = gtk_check_button_new();
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),
				(g_str_has_prefix(data->value, "TRUE")) ? TRUE : FALSE);
 	g_signal_connect (G_OBJECT (widget), "clicked",
		G_CALLBACK (checkbutton_cb), data);
	
	return widget;
}

static GtkWidget *create_filechooser_widget(SchemaMember *data) {
	GtkWidget *widget, *filechooser;
	GtkWidget *preview;
	char      *name;
	
	widget = gtk_event_box_new ();
	
	name = fixName(data->widget_name);
	filechooser = gtk_file_chooser_button_new(name, GTK_FILE_CHOOSER_ACTION_OPEN);
	g_free(name);
	
	gtk_widget_set_name (filechooser, data->widget_name);
	if (g_strrstr (data->value, ".png") ||  g_strrstr (data->value, ".svg"))
		gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (filechooser), data->value);
	else if (g_str_has_suffix (data->widget_name, "_background") &&
		 g_file_test ("/usr/share/nautilus/patterns", G_FILE_TEST_IS_DIR)) {
		gtk_file_chooser_set_current_folder (
			GTK_FILE_CHOOSER(filechooser),
			"/usr/share/nautilus/patterns");
		gtk_file_chooser_add_shortcut_folder (
			GTK_FILE_CHOOSER(filechooser),
			"/usr/share/nautilus/patterns", NULL);
	}
	g_signal_connect (G_OBJECT (filechooser), "selection-changed",
		G_CALLBACK (filechooser_cb), data);
	
	preview = gtk_image_new ();
	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER(filechooser), preview);
	g_signal_connect (GTK_FILE_CHOOSER(filechooser), "update-preview",
			G_CALLBACK(update_preview_cb), preview);
	
	/* fix for tooltips not showing up,
	   remove the eventbox when http://bugzilla.gnome.org/show_bug.cgi?id=303717 is closed */
	gtk_container_add (GTK_CONTAINER (widget), filechooser);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
		widget, data->description, NULL);
	
	return widget;
}

static GtkWidget *create_font_widget (SchemaMember *data) {
	GtkWidget *widget;
	KibaPrefsFile *file = (data->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
	char *name, *font_description;
	
	name = fixName(data->widget_name);
	widget = gtk_font_button_new();
	g_free(name);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	font_description = kiba_prefs_get_option_from_file (file, data->widget_name, "char");
	gtk_font_button_set_font_name (GTK_FONT_BUTTON (widget), font_description);
	
 	g_signal_connect (G_OBJECT (widget), "font-set",
		G_CALLBACK (font_cb), data);
	
	return widget;
}

static GtkWidget *get_float_spinbutton_widget (const char* key,
					       const char* value,
					       const char *range) {
	GtkWidget *spinbutton;
	float min = 0.0, max = 100.0, current;
	
	current = g_ascii_strtod (value, NULL);
	
	if (range && strlen (range) > 2) {
		char *min_string  = NULL,
		     *max_string = NULL;
		
		max_string = g_strstr_len (range+1,
				strlen(range), "-") + 1;
		min_string = g_strndup (range,
				strlen(range) - strlen(max_string) - 1);
		
		min = g_ascii_strtod (min_string, NULL);
		max = g_ascii_strtod (max_string, NULL);
		
		g_free(min_string);
	}
	
	spinbutton = gtk_spin_button_new_with_range (min, max, max/100);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton), current);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 2);
	gtk_entry_set_width_chars (GTK_ENTRY (spinbutton), 7);
	gtk_widget_set_name (spinbutton, key);
	
	return spinbutton;
}

static GtkWidget *create_list_widget (SchemaMember *data) {
	GtkTreeStore *store;
	GtkWidget *tree;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GSList *items, *l;
	
	items = kiba_prefs_get_list_from_string (data->value);
	g_return_val_if_fail (items != NULL, NULL);
	
	/* Create a model. */
	store = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING);
	
	/* fill the model with data */
	for (l = items; l != NULL; l = l->next) {
		char *item = l->data;
		GtkTreeIter iter;
		
		gtk_tree_store_append (store, &iter, NULL);  /* Acquire an iterator */
		
		gtk_tree_store_set (store, &iter, TITLE_COLUMN, item, -1);
	}
	
	/* Create a view */
	tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
	// gtk_tree_view_set_enable_search (GTK_TREE_VIEW (tree), TRUE);
	// gtk_tree_view_set_enable_tree_lines (GTK_TREE_VIEW (tree), TRUE);
	// gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (tree), FALSE);
	g_object_unref (G_OBJECT (store));
	
	/* Create a cell render */
	renderer = gtk_cell_renderer_text_new ();
	
	/* Create a column, associating the "text" attribute of the
	* cell_renderer to the first column of the model */
	column = gtk_tree_view_column_new_with_attributes (data->description, renderer,
							"text", TITLE_COLUMN,
							NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
	
	return tree;
}


static GtkWidget *get_spinbutton_widget (SchemaMember *data) {
	GtkWidget *spinbutton;
	gint min = 0, max = 100, current;
	
	current = atoi (data->value);
	
	if (data->range && strlen (data->range) > 2) {
		char *min_string  = NULL,
		     *max_string = NULL;
		
		max_string = g_strstr_len (data->range+1,
				strlen(data->range), "-") + 1;
		min_string = g_strndup (data->range,
				strlen(data->range) - strlen(max_string) - 1);
		
		min = atoi (min_string);
		max = atoi (max_string);
		
		g_free(min_string);
	}
	
	spinbutton = gtk_spin_button_new_with_range (min, max, 1);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton), current);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 0);
	gtk_widget_set_name (spinbutton, data->widget_name);
 	g_signal_connect (G_OBJECT (spinbutton), "value-changed",
		G_CALLBACK (spinbutton_cb), data);
	
	return spinbutton;
}

static gboolean __subsection_exist (const gchar *section,
                                    GtkNotebook *notebook) {
	gint i;
	
	/* check all pages in notebook if section name exist */
	for (i = 0; i < gtk_notebook_get_n_pages(notebook); i++) {
		if (strcmp (section, gtk_notebook_get_tab_label_text(notebook,
		    gtk_notebook_get_nth_page(notebook, i))) == 0)
			return TRUE;
	}
	
	return FALSE;
}

static void __clear_notebook (GtkNotebook *notebook) {
	gint i, num_pages;
	
	num_pages = gtk_notebook_get_n_pages (notebook);
	
	for (i = 0; i < num_pages; i++)
		gtk_notebook_remove_page (notebook, -1);
}

static void __free_closure (GtkWidget *widget,
                            gpointer data) {
	ResetClosure *closure = data;
	g_free(closure);
}

static GtkWidget *fill_subsection(SchemaSubSection *subsection) {
	GtkWidget *container;
	GtkWidget *widget;
	GtkWidget *scrolledwindow;
	int i=-1, c=0;
	GList *l;
	
	scrolledwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
	/* insert a table inside the scrollable window */
	container = gtk_table_new(g_list_length(subsection->options), 3, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE(container), 5);
	gtk_table_set_row_spacings (GTK_TABLE(container), 5);
	
	for(l = subsection->options; l != NULL; l = l->next) {
		SchemaMember *option = l->data;
		i++;
		if (strcmp(option->type, "bool") == 0) {
			
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			
			widget = create_checkbox_widget(option);
			gtk_table_attach (GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "int") == 0) {
			/* skip position widgets for now */
			if (g_str_has_suffix (option->widget_name, "_position"))
				continue;
	
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
	
			widget = get_spinbutton_widget(option);
			gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
				widget, option->description, NULL);
			gtk_table_attach (GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "double") == 0 &&
			 strcmp(option->range, TEXT_RANGE) != 0 &&
			 strcmp(option->range, "") != 0 ) {
			GtkWidget *box, *spinbutton;
			
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			
			box = gtk_hbox_new (FALSE, 0);
			widget = create_double_hscale_widget(option);
			gtk_box_pack_start (GTK_BOX (box),
					widget,
					TRUE, TRUE, 0);
// 			spinbutton = get_float_spinbutton_widget (option->title,
// 					option->value,
// 					option->range);
// 			gtk_box_pack_start (GTK_BOX (box),
// 					spinbutton,
// 					FALSE, FALSE, 0);
			
// 			g_signal_connect (G_OBJECT(widget), "value-changed",
// 				G_CALLBACK (set_spinbutton_value_from_range), spinbutton);
// 			g_signal_connect (G_OBJECT(spinbutton), "value-changed",
// 				G_CALLBACK (set_range_value_from_spinbutton), widget);
			
			gtk_table_attach (GTK_TABLE(container), box,
				2, 3, c, 1+c, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "list") == 0) {
			if (option->range &&
			    strcmp (option->range, RGBA_RANGE) == 0) {
				GSList *items, *l;
				widget = kiba_gradient_chooser_new();
				gtk_widget_set_name(widget, option->widget_name);
				items = kiba_prefs_get_list_from_string (option->value);
				for (l = items; l != NULL; l = l->next)
					kiba_gradient_chooser_add_color(KIBA_GRADIENT_CHOOSER(widget), (char*) l->data);
				g_signal_connect(G_OBJECT(widget), "changed",
					G_CALLBACK(__gradient_chooser_changed), option);
				gtk_table_attach(GTK_TABLE(container), widget,
					2, 3, c, 1+c, GTK_FILL,GTK_FILL, 0, 0);
				GtkWidget *label_widget = create_label_widget(option->title);
				gtk_table_attach (GTK_TABLE(container), label_widget,
					1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			}
			else {
				GtkWidget *button;
				GtkWidget *vbox;
				GtkWidget *image;
				GtkWidget *list_box;
				
				list_box = gtk_hbox_new (FALSE, 0);
				widget = create_list_widget (option);
				g_object_set_data  (G_OBJECT(widget), "Pref-Member", option);
				gtk_box_pack_start (GTK_BOX(list_box), widget, TRUE, TRUE, 0);
				
				vbox = gtk_vbox_new (TRUE, 0);
				button = gtk_button_new ();
				image = gtk_image_new_from_stock (GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON);
				gtk_button_set_image (GTK_BUTTON (button), image);
				g_signal_connect (G_OBJECT(button), "clicked",
					G_CALLBACK (cb_add_item_to_list), widget);
				gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, 0);
				
				button = gtk_button_new ();
				image = gtk_image_new_from_stock (GTK_STOCK_REMOVE, GTK_ICON_SIZE_BUTTON);
				gtk_button_set_image (GTK_BUTTON (button), image);
				g_signal_connect (G_OBJECT(button), "clicked",
					G_CALLBACK (cb_remove_item_from_list), widget);
				gtk_box_pack_end (GTK_BOX(vbox), button, FALSE, FALSE, 0);
	
				gtk_box_pack_start (GTK_BOX(list_box), vbox, FALSE, FALSE, 0);
	
				gtk_table_attach(GTK_TABLE(container), list_box,
					1, 3 , c, 1+c,
					GTK_FILL | GTK_EXPAND, GTK_FILL,
					0, 0);
			}
		}
		else if (strcmp(option->type, "char") == 0) {
			if (option->range && strcmp (option->range, RGBA_RANGE) == 0)
				widget = create_rgba_widget(option);
			else if (option->range && strcmp (option->range, RGB_RANGE) == 0)
				widget = create_colour_widget(option);
			/* combobox + filechooser */
			else if (option->range &&
				 g_strrstr (option->range, ICON_RANGE)) {
				GtkWidget *file_chooser;
				int        fnord = 1;
				
				widget = gtk_hbox_new (FALSE, 5);
				
				if (strcmp (option->range, ICON_RANGE) != 0) {
					GtkWidget *combobox;
					
					combobox = create_combobox_widget(option);
					gtk_box_pack_start (GTK_BOX (widget),
							combobox,
							FALSE, FALSE, 0);
					fnord++;
				}
				
				file_chooser = create_filechooser_widget(option);
				gtk_box_pack_start (GTK_BOX (widget),
						file_chooser,
						FALSE, FALSE, 0);
			}
			else if (option->range && g_strrstr (option->range, TEXT_RANGE))
				widget = create_font_widget(option);
			else if (option->range && strcmp (option->range, EDITABLE_RANGE) == 0) {
				widget = get_entry_widget (option->widget_name,
						option->value,
						option->description);
				g_signal_connect (G_OBJECT (widget), "activate",
					G_CALLBACK (entry_cb), option);
			}
			else
				widget = create_combobox_widget (option);
			
			GtkWidget *label_widget;
			
			gtk_table_attach(GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL,GTK_FILL, 0, 0);
			label_widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), label_widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else continue;
		
		GtkWidget *image, *button, *hbox;
		
		KibaPrefsFile *file = (option->schema == kiba_settings->general_schema ? kiba_settings->generalFile : kiba_settings->dockFile);
		KibaPrefsFile *globalFile = kiba_prefs_get_global_file(kiba_prefs_get());
		const char *tip;
		if (file != globalFile && kiba_prefs_has_global_option(file, option->widget_name)) {
			gtk_widget_set_sensitive(widget,FALSE);
			tip =  _("Remove global option, using the individual dock option again.");
		} else if (file == globalFile) {
			tip =  _("Remove global option, using the individual dock option again.");
		} else {
			tip =  _("Reset to default");
		}
		
		hbox = gtk_vbox_new(FALSE, 0);
		button = gtk_button_new ();
		image = gtk_image_new_from_stock (GTK_STOCK_CLEAR,
						GTK_ICON_SIZE_BUTTON);
		gtk_button_set_image (GTK_BUTTON (button), image);
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
			button, tip , NULL);
		gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
		gtk_widget_show(hbox);
		gtk_table_attach (GTK_TABLE(container), hbox,
			0, 1 , c, 1+c, 0, GTK_FILL, 0, 0);
		
		ResetClosure *closure = g_new0 (ResetClosure, 1);
		closure->widget = widget;
		closure->member = option;
		g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK (reset_to_default), closure);
		g_signal_connect (G_OBJECT (button), "destroy",
			G_CALLBACK (__free_closure), closure);
		
		c++;
	}
	
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolledwindow), container);
	
	return scrolledwindow;
}

/*** public methods ***/

void kiba_settings_notebook_set_section (KibaSettingsNotebook *notebook,
                                         SchemaSection *section) {
	GtkNotebook *gnotebook = GTK_NOTEBOOK (notebook);
	GList *l;
	
	__clear_notebook (gnotebook);
	
	for (l = section->subsections; l != NULL; l = l->next) {
		SchemaSubSection *subsection = l->data;
		if (!__subsection_exist(subsection->subsection_name, gnotebook)) {
			gtk_notebook_append_page(gnotebook,
				fill_subsection(subsection),
				gtk_label_new(subsection->subsection_name));
		}
	}
	
	gtk_widget_show_all(GTK_WIDGET(notebook));
}

/*** Class init ***/

static void kiba_settings_notebook_class_init (KibaSettingsNotebookClass *class) {}

static void kiba_settings_notebook_init (KibaSettingsNotebook *notebook) {}

GtkWidget* kiba_settings_notebook_new (KibaSettings *settings) {
	kiba_settings = settings;
	return g_object_new (KIBA_TYPE_SETTINGS_NOTEBOOK, NULL);
}
