/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>

#include "kiba-utils.h"

#include "settings-section-box.h"

#define _(string) gettext (string)

typedef struct {
	KibaSettingsSectionBox *box;
	SchemaSection *schema_section;
	GtkWidget *button;
	GtkWidget *label;
	GtkWidget *image;
}KibaSettingSectionButton;

struct _KibaSettingsSectionBox {
	GtkVBox parent;
	
	GList *sections;
	KibaSettingSectionButton *selected_section;
};

struct _KibaSettingsSectionBoxClass {
	GtkVBoxClass class;
	
	/* signals */
	void (* selected_section_changed) (KibaSettingsSectionBox *box);
};

G_DEFINE_TYPE (KibaSettingsSectionBox, kiba_settings_section_box, GTK_TYPE_VBOX)

/* signal ids */
enum {
	SIG_SELECTED_SECTION_CHANGED = 0,
	SIG_KIBA_SETTINGS_SECTION_BOX_LAST
};

static guint settings_section_box_signals[SIG_KIBA_SETTINGS_SECTION_BOX_LAST] = { 0 };

/*** private methods ***/

static GdkPixbuf* __get_pixbuf_from_file (const char *file) {
	GdkPixbuf *icon = NULL;
	GtkIconInfo *info = NULL;
	
	if (file) {
		icon = gdk_pixbuf_new_from_file_at_size (file, 48, -1, NULL);
		if (icon)
			return icon;
		
		info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), file, 48, 0);
		if (info != NULL) {
			icon= gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (info), 48, -1, NULL);
			gtk_icon_info_free (info);
			info = NULL;
		}
		if (icon) return icon;
	}
	
	info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), GTK_STOCK_MISSING_IMAGE, 48, 0);
	if (info != NULL) {
		icon= gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (info), 48, -1, NULL);
		gtk_icon_info_free (info);
	}
	
	return icon;
}

static GdkPixbuf* __get_pixbuf_from_icon_list (GSList *icon_list) {
	GSList *l;
	GdkPixbuf *ret = NULL;
	
	for (l = icon_list; l != NULL; l = l->next) {
		char *icon_file = l->data;
		
		ret = __get_pixbuf_from_file (icon_file);
		if (ret) break;
	}
	
	return ret;
}

static gboolean __section_exist (const gchar *section,
                                 GtkWidget *button_box) {
	GList *buttons, *l;
	
	buttons = gtk_container_get_children (GTK_CONTAINER (button_box));
	
	for (l = buttons; l != NULL; l = l->next) {
		GtkWidget *box = l->data;
		GList     *box_childs, *li;
		
		box_childs = gtk_container_get_children (GTK_CONTAINER (box));
		
		for (li = box_childs; li != NULL; li = li->next) {
			GtkWidget *child = li->data;
			
			if (!GTK_IS_LABEL (child))
				continue;
			
			const gchar* label_text = gtk_label_get_text (GTK_LABEL (child));
			
			if (strcmp (g_ascii_strdown (label_text, strlen(label_text)),
				        g_ascii_strdown (section, strlen(section))) == 0)
				return TRUE;
			
			break;
		}
	}
	
	return FALSE;
}

static void __unselect_section (KibaSettingsSectionBox *box,
                                KibaSettingSectionButton *section) {
	GtkLabel *label = GTK_LABEL(section->label);
	const char *text = gtk_label_get_text(label);
	
	gtk_label_set_text (label, text);
	gtk_widget_set_state (section->image, GTK_STATE_NORMAL);
}

static void __select_section (GtkWidget *button,
                              gpointer data) {
	KibaSettingSectionButton *section = data;
	KibaSettingsSectionBox *box = section->box;
	GtkLabel *label = GTK_LABEL(section->label);
	const char *text = gtk_label_get_text(label);
	char *markup;
	
	if (box->selected_section == section)
		return;
	
	if (box->selected_section != NULL) {
		__unselect_section(box, box->selected_section);
	}
	
	gtk_widget_set_state (section->image, GTK_STATE_SELECTED);
	
	markup = g_markup_printf_escaped ("<span style=\"italic\"><b>%s</b></span>", text);
	gtk_label_set_markup (label, markup);
	g_free(markup);
	
	box->selected_section = section;
	
	g_signal_emit (box, settings_section_box_signals[SIG_SELECTED_SECTION_CHANGED], 0);
}

static void __section_button_destroyed (GtkWidget *button,
                                        gpointer data) {
	KibaSettingSectionButton *section = data;
	KibaSettingsSectionBox *box = section->box;
	
	if (box->selected_section == section)
		box->selected_section = NULL;
	
	box->sections = g_list_remove(box->sections, section);
	
	g_free(section);
}

static GtkWidget* __get_section_button_box (KibaSettingsSectionBox *box,
                                            const char *title,
                                            gboolean with_scrolled_window) {
	GtkWidget *frame;
	GtkWidget *button_box;
	GtkWidget *scrolled_window;
	
	frame = gtk_frame_new (title);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_set_border_width (GTK_CONTAINER (frame), 5);
	button_box = gtk_vbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (button_box), GTK_BUTTONBOX_START);
	if (with_scrolled_window) {
		scrolled_window = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
		gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled_window), button_box);
		gtk_container_add(GTK_CONTAINER(frame), scrolled_window);
	} else {
		gtk_container_add(GTK_CONTAINER(frame), button_box);
	}
	gtk_box_pack_start (GTK_BOX (box),
		frame,
		with_scrolled_window, with_scrolled_window, 0);
	gtk_widget_show_all(frame);
	
	return button_box;
}

static void __fill_section_box (KibaSettingsSectionBox *box,
                                GtkWidget *section_box,
                                GList *sections) {
	GList *l;
	SchemaSection *section;
	KibaSchemaType last_type;
	
	if (sections) {
		section = sections->data;
		last_type = section->type;
	}
	
	for (l = sections; l != NULL; l = l->next) {
		section = l->data;
		
		if (!__section_exist(section->section_name, section_box)) {
			KibaSettingSectionButton *section_button;
			GtkWidget *box_hbox, *button, *button_label, *button_image;
			GdkPixbuf *icon = NULL;
			
			if (last_type != section->type) {
				box_hbox = gtk_vbox_new (FALSE, 0);
				
				button = gtk_hseparator_new ();
				gtk_box_pack_start (GTK_BOX (box_hbox),
						button,
						TRUE, TRUE, 0);
				
				gtk_widget_show_all(box_hbox);
				gtk_container_add (GTK_CONTAINER (section_box), box_hbox);
			}
			last_type = section->type;
			
			button = gtk_button_new();
			gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
			gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
			
			box_hbox = gtk_hbox_new (FALSE, 0);
			gtk_container_add (GTK_CONTAINER (button), box_hbox);
			icon = __get_pixbuf_from_icon_list (section->icon_list);
			button_image = gtk_image_new_from_pixbuf (icon);
			gtk_box_pack_start (GTK_BOX (box_hbox),
					button_image,
					FALSE, FALSE, 0);
			
			button_label = gtk_label_new(section->section_name);
			gtk_box_pack_start (GTK_BOX (box_hbox),
					button_label,
					FALSE, FALSE, 5);
			
			gtk_container_add (GTK_CONTAINER (section_box), button);
			
			section_button = g_new0(KibaSettingSectionButton ,1);
			section_button->box = box;
			section_button->button = button;
			section_button->image = button_image;
			section_button->label = button_label;
			section_button->schema_section = section;
			
			box->sections = g_list_append(box->sections, section_button);
			
			g_signal_connect(G_OBJECT(button), "clicked",
				G_CALLBACK(__select_section), section_button);
			g_signal_connect(G_OBJECT(button), "destroy",
				G_CALLBACK(__section_button_destroyed), section_button);
		}
	}
}

static void __add_general_sections (KibaSettingsSectionBox *box,
                                    KibaSchema *general_schema) {
	GtkWidget *section_box;
	
	section_box = __get_section_button_box (box, _("General configuration"), FALSE);
	__fill_section_box(box, section_box, general_schema->sections);
}

static gint __compare_sections (gconstpointer apointer,
                                gconstpointer bpointer) {
	const SchemaSection *asection = apointer;
	const SchemaSection *bsection = bpointer;
	
	if (asection->type != bsection->type)
		return asection->type - bsection->type;
	
	return kiba_utils_sort_string_list(asection->section_name, bsection->section_name);
}

static void __add_dock_sections (KibaSettingsSectionBox *box,
                                 KibaSchema *dock_schema) {
	GtkWidget *section_box;
	GList *sorted_sections;
	
	section_box = __get_section_button_box (box, _("Dock configuration"), TRUE);
	sorted_sections = g_list_copy(dock_schema->sections);
	sorted_sections = g_list_sort(sorted_sections, __compare_sections);
	__fill_section_box(box, section_box, sorted_sections);
	g_list_free(sorted_sections);
}

/*** public methods ***/

SchemaSection* kiba_settings_section_box_get_selected (KibaSettingsSectionBox *box) {
	return box->selected_section->schema_section;
}

GtkWidget* kiba_settings_section_box_get_selected_image (KibaSettingsSectionBox *box) {
	return box->selected_section->image;
}

void kiba_settings_section_box_set_sections (KibaSettingsSectionBox *box,
                                             KibaSchema *general_schema,
                                             KibaSchema *dock_schema) {
	GList *l;
	
	for (l = box->sections; l != NULL; l = box->sections) {
		KibaSettingSectionButton *section = l->data;
		
		gtk_widget_destroy(section->button);
	}
	
	__add_general_sections(KIBA_SETTINGS_SECTION_BOX(box), general_schema);
	__add_dock_sections(KIBA_SETTINGS_SECTION_BOX(box), dock_schema);
	
	if (box->sections)
		__select_section(NULL, box->sections->data);
}

/*** Class init ***/

static void kiba_settings_section_box_class_init (KibaSettingsSectionBoxClass *class) {
	settings_section_box_signals[SIG_SELECTED_SECTION_CHANGED] =
		g_signal_new ("selected-section-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaSettingsSectionBoxClass, selected_section_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0,
			G_TYPE_NONE);
}

static void kiba_settings_section_box_init (KibaSettingsSectionBox *box) {
	box->sections = NULL;
	box->selected_section = NULL;
}

GtkWidget* kiba_settings_section_box_new (void) {
	return g_object_new (KIBA_TYPE_SETTINGS_SECTION_BOX, NULL);
}
