/*
 *  Copyright (C) 2007 Torbjörn Svangård  <trptorbjorn@gmail.com>
 *  Copyright (C) 2007 Daniel Bessler  <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Torbjörn Svangård  <trptorbjorn@gmail.com>
 *              Daniel    Bessler    <danielb@kiba-dock.org>
 */

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>

#include "kiba-settings.h"
#include "kiba-dialog.h"
#include "widgets.h"

#define _(string) gettext (string)

#define STATUS_PUSH(bar, txt)  gtk_statusbar_push (bar, gtk_statusbar_get_context_id(bar, "info"), txt)

#include "settings-section-box.h"
#include "settings-notebook.h"

#define WINDOW_LABEL _("Kiba-Dock Settings Editor")

static GtkWidget *status_bar = NULL;
static gint status_pop_id = 0;

void status_bar_push_message (const char *message) {
	STATUS_PUSH (GTK_STATUSBAR(status_bar), message);
}

static gint status_pop (gpointer data) {
	GtkStatusbar *bar = data;
	
	gtk_statusbar_pop (bar, gtk_statusbar_get_context_id(bar, "info"));
	
	status_pop_id = 0;
	
	return FALSE;
}

static void status_pushed (GtkStatusbar *statusbar,
			   guint         context_id,
			   gchar        *text,
			   gpointer      user_data) {
	if (status_pop_id) {
		gtk_statusbar_pop (statusbar, gtk_statusbar_get_context_id(statusbar, "info"));
		g_source_remove (status_pop_id);
	}
	
	status_pop_id = g_timeout_add (2*1000, status_pop, statusbar);
}

static void activate_section (KibaSettings *settings,
			      GtkWidget *selected_image,
			      SchemaSection  *section) {
	GdkPixbuf *pixbuf;
	gchar *markup;
	GList *l;
	
	settings->activeSection = section;
	
	/* update the image */
	pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(selected_image));
	gtk_image_set_from_pixbuf(GTK_IMAGE(settings->sectionImage), pixbuf);
	
	/* set title text */
	markup = g_markup_printf_escaped ("<big><b>%s</b></big>", section->section_name);
	gtk_label_set_markup (GTK_LABEL (settings->sectionLabel), markup);
	g_free (markup);
	
	/* set description text */
	gtk_label_set_text (GTK_LABEL (settings->sectionDescription), section->description);
	
	/* update the notebook */
	kiba_settings_notebook_set_section(KIBA_SETTINGS_NOTEBOOK(settings->notebook), section);
	gtk_widget_grab_focus (settings->notebook);
}

static void __activate_config_file (KibaSettings *settings,
                                    KibaPrefsFile *file) {
	GdkPixbuf *icon;
	char *title;
	
	settings->dockFile = file;
	
	const char *path = kiba_prefs_get_file_path(file);
	
	title = g_strdup_printf ("%s - %s", _("Kiba-Dock Settings Editor"), g_strrstr(path, "/")+1);
	gtk_window_set_title(GTK_WINDOW(settings->window), title);
	g_free(title);
	
	kiba_prefs_schema_set_options_from_config(settings->schema, file);
	
	/* reload currently active section */
	activate_section (settings, settings->sectionImage, settings->activeSection);
	
	gtk_widget_grab_focus (settings->notebook);
	gtk_widget_show_all (settings->notebook);
}

static void __choose_config_file_cb (GtkButton *button, gpointer data) {
	KibaSettings *settings = data;
	GtkWidget *configDialog = NULL;
	char *configDir;
	
	configDialog = gtk_file_chooser_dialog_new (_("Please choose a dock configuration file"),
							NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
							GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
							GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
	configDir = g_strdup_printf ("%s/.kiba-dock/config/docks", g_get_home_dir ());
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (configDialog), configDir);
	
	if (gtk_dialog_run (GTK_DIALOG (configDialog)) == GTK_RESPONSE_ACCEPT) {
		char *filename;
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (configDialog));
		if (!g_str_has_prefix(filename, configDir))
			fprintf (stderr, "Invalid config file '%s'!\n", filename);
		else {
			KibaPrefsFile *file;
			file = kiba_prefs_get_file(filename);
			if (!file) {
				file = kiba_prefs_add_config_file(filename);
				if (!g_str_has_suffix(filename, KIBA_PREFS_GLOBAL_FILE))
					kiba_prefs_update_config_file(kiba_prefs_get (), file);
			}
			if (!g_str_has_suffix(filename, KIBA_PREFS_GLOBAL_FILE)) {
				settings->globalFile = FALSE;
			} else {
				settings->globalFile = TRUE;
			}
			__activate_config_file(settings, file);
		}
		g_free (filename);
	}
	g_free(configDir);
	
	gtk_widget_destroy (configDialog);
}

static GtkWidget *dock_add_dialog = NULL;
static GtkWidget *dock_add_entry = NULL;

static void __destroy_dock_add_dialog_cb (GtkWidget *widget, gpointer data) {
	dock_add_dialog = NULL;
}

static void __destroy_dock_add_dialog (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog = data;
	gtk_widget_destroy(dialog);
}

static void __dock_create_config_file_cb (GtkWidget *widget, gpointer data) {
	KibaSettings *settings = data;
	KibaPrefsFile *file;
	const char *dockID;
	char *configPath;
	
	dockID = gtk_entry_get_text(GTK_ENTRY(dock_add_entry));
	if (!dockID || strlen(dockID) == 0) return;
	
	configPath = g_strdup_printf ("%s/.kiba-dock/config/docks/%s.xml", g_get_home_dir (), dockID);
	file = kiba_prefs_get_file(configPath);
	if (!file) {
		file = kiba_prefs_add_config_file(configPath);
		kiba_prefs_update_config_file(kiba_prefs_get(), file);
		__activate_config_file(settings, file);
		kiba_settings_dbus_add_dock(configPath);
		gtk_widget_destroy(dock_add_dialog);
	} else {
		kiba_dialog_error(_("Failed to create dock!"), _("This ID is allready used in the preferences database."));
	}
	g_free(configPath);
}

static void __remove_dock_cb (GtkButton *button, gpointer data) {
	KibaSettings *settings = data;
	GtkWidget *dialog;
	const char *path = kiba_prefs_get_file_path(settings->dockFile);
	gboolean removingAllowed = FALSE;
	char *dockConfigPath = g_strdup_printf ("%s/.kiba-dock/config/docks", g_get_home_dir ());
	gint response;
	
	GDir *configDir = g_dir_open (dockConfigPath, 0, NULL);
	if (!configDir) {
		g_free(dockConfigPath);
		return;
	}
	g_free(dockConfigPath);
	
	while (TRUE) {
		const char *file = g_dir_read_name(configDir);
		if (!file) break;
		if (!g_str_has_suffix(file, ".xml") || strcmp(file, KIBA_PREFS_GLOBAL_FILE) == 0) continue;
		if (!g_str_has_suffix(path, file)) {
			removingAllowed = TRUE;
			break;
		}
	}
	g_dir_close(configDir);
	
	if (!removingAllowed) {
		kiba_dialog_error(_("Failed to remove dock!"), _("There must be at least one dock."));
		return;
	}
	
	char *title = g_strdup_printf (_("Removing dock '%s'"), path);
	dialog = kiba_dialog_confirm(title, _("Are you sure you want to remove this dock?"));
	response = gtk_dialog_run (GTK_DIALOG (dialog));
	switch (response) {
		case GTK_RESPONSE_YES: {
			const char *path = kiba_prefs_get_file_path(settings->dockFile);
			kiba_settings_dbus_remove_dock(path);
			g_remove(path);
			gtk_widget_destroy(GTK_WIDGET(dialog));
			dialog = NULL;
			__choose_config_file_cb(NULL, data);
			break;
		}
	}
	if (dialog)
		gtk_widget_destroy(GTK_WIDGET(dialog));
}

static void __create_dock_cb (GtkButton *button, gpointer data) {
	KibaSettings *settings = data;
	GtkWidget *dialog;
	GtkWidget *image;
	GdkPixbuf *pixbuf;
	GtkWidget *entry;
	GtkWidget *label;
	GtkWidget *hbox, *vbox, *main_vbox;
	GtkWidget *but;
	
	if (dock_add_dialog) {
		gtk_window_present(GTK_WINDOW(dock_add_dialog));
		return;
	}
	
	dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	dock_add_dialog = dialog;
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
	gtk_window_set_title(GTK_WINDOW(dialog), _("Create a new dock"));
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
	gtk_window_set_type_hint(GTK_WINDOW(dialog), GDK_WINDOW_TYPE_HINT_DIALOG);
	label = gtk_label_new(_("Please choose a ID for the new dock:"));
	pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	image = gtk_image_new_from_pixbuf(pixbuf);
	gtk_window_set_icon (GTK_WINDOW (dialog), pixbuf);
	g_object_unref (pixbuf);
	entry = gtk_entry_new();
	dock_add_entry = entry;
	
	main_vbox = gtk_vbox_new(FALSE, 5);
	vbox = gtk_vbox_new(FALSE, 5);
	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start (GTK_BOX (hbox),
			image,
			TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX (vbox),
			label,
			TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (vbox),
			entry,
			TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (hbox),
			vbox,
			TRUE, TRUE, 5);
	gtk_box_pack_start (GTK_BOX (main_vbox),
			hbox,
			TRUE, TRUE, 5);
	
	hbox = gtk_hbox_new(FALSE, 5);
	but = gtk_button_new_from_stock (GTK_STOCK_QUIT);
	g_signal_connect(G_OBJECT(but), "released",
		G_CALLBACK(__destroy_dock_add_dialog), dialog);
	gtk_box_pack_end (GTK_BOX (hbox),
			but,
			FALSE, FALSE, 5);
	but = gtk_button_new_from_stock (GTK_STOCK_OK);
	g_signal_connect(G_OBJECT(but), "released",
		G_CALLBACK(__dock_create_config_file_cb), settings);
	gtk_box_pack_end (GTK_BOX (hbox),
			but,
			FALSE, FALSE, 5);
	gtk_box_pack_start (GTK_BOX (main_vbox),
			hbox,
			TRUE, TRUE, 5);
	
	gtk_container_add(GTK_CONTAINER(dialog), main_vbox);
	
	gtk_widget_show_all (dialog);
	
	g_signal_connect(G_OBJECT(dialog), "destroy",
		G_CALLBACK(__destroy_dock_add_dialog_cb), NULL);
}

static void __section_selected (KibaSettingsSectionBox *box,
                                gpointer data) {
	KibaSettings *settings = data;
	
	activate_section (settings,
		kiba_settings_section_box_get_selected_image(box),
		kiba_settings_section_box_get_selected(box));
}

GtkWidget* kiba_settings_create_window (KibaSettings *settings) {
	GtkWidget *window,
		  *main_container,
		  *top_container,
		  *setting_box,
		  *section_vbox,
		  *bottom_container,
		  *quit_button,
		  *title_hbox,
		  *title_vbox,
		  *main_frame;
	GtkWidget *real_section_box;
	GdkPixbuf *kiba_icon;
	
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), WINDOW_LABEL);
	g_signal_connect (G_OBJECT (window), "delete_event",
		G_CALLBACK(gtk_main_quit), NULL);
	settings->window = window;
	
	kiba_icon = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gtk_window_set_icon (GTK_WINDOW (window), kiba_icon);
	g_object_unref (kiba_icon);
	
	main_container = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (main_container), 5);
	gtk_container_add(GTK_CONTAINER(window), main_container);
	
	main_frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (main_frame), GTK_SHADOW_ETCHED_IN);
	
	top_container = gtk_hbox_new (FALSE, 0);
	
	/* create Section container Widgets */
	section_vbox = gtk_vbox_new (FALSE, 5);
		real_section_box = kiba_settings_section_box_new();
		gtk_box_pack_start (GTK_BOX (section_vbox),
					real_section_box,
					TRUE, TRUE, 0);
		gtk_widget_show_all(real_section_box);
	
		quit_button = gtk_button_new_with_label (_("Change dock"));
		GtkWidget *image = gtk_image_new_from_stock(GTK_STOCK_PREFERENCES, GTK_ICON_SIZE_BUTTON);
		gtk_button_set_image(GTK_BUTTON(quit_button), image);
		char *tip = g_strdup_printf ("%s\n%s",
			_("Choose a dock configuration file you want to edit in kiba-settings."),
			_("Please select 'global.xml' to edit settings for all docks."));
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(), quit_button, tip, NULL);
		g_free(tip);
		g_signal_connect (G_OBJECT (quit_button), "released",
				G_CALLBACK(__choose_config_file_cb), settings);
		gtk_box_pack_start (GTK_BOX (section_vbox),
					quit_button,
					FALSE, FALSE, 3);
	
		quit_button = gtk_button_new_with_label (_("Remove dock"));
		image = gtk_image_new_from_stock(GTK_STOCK_REMOVE, GTK_ICON_SIZE_BUTTON);
		gtk_button_set_image(GTK_BUTTON(quit_button), image);
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(), quit_button, _("Remove the currently active dock configuration file."), NULL);
		g_signal_connect (G_OBJECT (quit_button), "released",
				G_CALLBACK(__remove_dock_cb), settings);
		gtk_box_pack_start (GTK_BOX (section_vbox),
					quit_button,
					FALSE, FALSE, 0);
		
		quit_button = gtk_button_new_with_label (_("New dock"));
		image = gtk_image_new_from_stock(GTK_STOCK_NEW, GTK_ICON_SIZE_BUTTON);
		gtk_button_set_image(GTK_BUTTON(quit_button), image);
		tip = g_strdup_printf ("%s", _("Create a new dock configuration file."));
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(), quit_button, tip, NULL);
		g_free(tip);
		g_signal_connect (G_OBJECT (quit_button), "released",
				G_CALLBACK(__create_dock_cb), settings);
		gtk_box_pack_start (GTK_BOX (section_vbox),
					quit_button,
					FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (top_container),
			    section_vbox,
			    FALSE, FALSE, 0);
	
	/* create Settings container widgets */
	setting_box = gtk_vbox_new (FALSE, 5);
	gtk_container_set_border_width (GTK_CONTAINER (setting_box), 5);
	title_hbox = gtk_hbox_new (FALSE, 5);

	settings->sectionImage = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE, GTK_ICON_SIZE_DIALOG);
	gtk_box_pack_start (GTK_BOX (title_hbox),
			    settings->sectionImage,
			    FALSE, FALSE, 0);
	
	title_vbox = gtk_vbox_new (FALSE, 0);
	settings->sectionLabel = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (settings->sectionLabel), 0.0, 0.5);
	gtk_box_pack_start (GTK_BOX (title_vbox),
			    settings->sectionLabel,
			    FALSE, FALSE, 0);
	settings->sectionDescription = gtk_label_new (NULL);
// 	gtk_misc_set_alignment (GTK_MISC (settings->sectionLabel), 0.0, 0.5);
	gtk_label_set_line_wrap (GTK_LABEL(settings->sectionDescription), TRUE);
	gtk_box_pack_start (GTK_BOX (title_vbox),
			    settings->sectionDescription,
			    FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (title_hbox),
			    title_vbox,
			    FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (setting_box),
			    title_hbox,
			    FALSE, FALSE, 0);
	settings->notebook = kiba_settings_notebook_new(settings);
	gtk_notebook_set_tab_pos (GTK_NOTEBOOK(settings->notebook), GTK_POS_TOP);
	gtk_notebook_set_scrollable (GTK_NOTEBOOK(settings->notebook), TRUE);
	gtk_box_pack_start (GTK_BOX (setting_box),
			    settings->notebook,
			    TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (top_container),
			    setting_box,
			    TRUE, TRUE, 0);
	
	gtk_container_add (GTK_CONTAINER(main_frame), top_container);
	
	/* create Bottom Container (Quit Button) */
	bottom_container = gtk_hbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (bottom_container), 5);
	
	quit_button = gtk_button_new_from_stock (GTK_STOCK_QUIT);
	g_signal_connect (G_OBJECT (quit_button), "released", gtk_main_quit, NULL);
	gtk_box_pack_end (GTK_BOX (bottom_container),
			    quit_button,
			    FALSE, FALSE, 3);
	
	status_bar = gtk_statusbar_new();
	gtk_container_set_border_width (GTK_CONTAINER(status_bar), 0);
	gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR(status_bar), FALSE);
	g_signal_connect_after (G_OBJECT (status_bar),
		"text-pushed", G_CALLBACK(status_pushed), NULL);
	gtk_box_pack_start (GTK_BOX (bottom_container),
			    status_bar,
			    TRUE, TRUE, 0);
	
	/* grab focus on the notebook, since we dont wanna have focus on the buttonbox */
	gtk_widget_grab_focus (settings->notebook);
	
	/* fill main container */
	gtk_box_pack_start (GTK_BOX (main_container),
			    main_frame,
			    TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (main_container),
			    bottom_container,
			    FALSE, FALSE, 0);
	
	g_signal_connect(G_OBJECT(real_section_box), "selected-section-changed",
		G_CALLBACK(__section_selected), settings);
	kiba_settings_section_box_set_sections (KIBA_SETTINGS_SECTION_BOX(real_section_box),
				settings->general_schema,
				settings->schema);
	
	__activate_config_file(settings, settings->dockFile);
	
	return window;
}
