/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-app-chooser
 * @short_description: Application chooser dialog.
 * @see_also:
 * @stability: Unstable
 *
 * Application chooser dialog.
 */

#include <libintl.h>
#include <string.h>

#include "kiba-app-chooser.h"
#include "kiba-utils.h"

#define _(string)  gettext (string)

G_DEFINE_TYPE (KibaAppChooser, kiba_app_chooser, G_TYPE_OBJECT)

enum {
	SIG_APP_ACTIVATE = 0,
	SIG_KIBA_APP_CHOOSER_LAST
};

static guint kiba_app_chooser_signals[SIG_KIBA_APP_CHOOSER_LAST] = { 0 };

static char *section_data[] = {
	"All", "System", "Settings", "Office", "Graphic", "Audio/Video",
	"Network", "Games", "Development", "Tools"
};

static KibaAppChooserSection kiba_app_chooser_section_from_string (const char *section) {
	int i;
	
	for (i = 0; i < G_N_ELEMENTS(section_data); i++) {
		if (strcmp (section_data[i], section) == 0)
			return i;
	}
	
	return KIBA_APP_SECTION_ALL;
}

static void kiba_app_chooser_destroy_cb (GtkWidget *widget,
				         gpointer   data) {
	KibaAppChooser *chooser = data;
	KibaAppChooserApp *app;
	GList *l;
	
	for (l = chooser->apps; l != NULL; l = chooser->apps) {
		app = l->data;
		
		if (GTK_IS_WIDGET(app->box))
			gtk_widget_destroy (app->box);
		g_free (app->uri);
		chooser->apps = g_list_remove (chooser->apps, app);
		g_free (app);
	}
	g_list_free (chooser->apps);
	g_object_unref (chooser);
}

static gboolean app_button_press_callback (GtkWidget      *widget,
					   GdkEventButton *event,
					   gpointer        user_data) {
	KibaAppChooserApp *app = user_data;
	
	kiba_icon_view_select_child (KIBA_ICON_VIEW(app->chooser->view_win->icon_view), app->box);
	gtk_widget_queue_draw (app->chooser->view_win->icon_view);
	
	if (event->type != GDK_2BUTTON_PRESS) {
		return FALSE;
	}
	
	g_signal_emit (app->chooser,
		kiba_app_chooser_signals[SIG_APP_ACTIVATE],
		0, app->uri);
	
	gtk_widget_destroy (app->chooser->view_win->win);
	
	return TRUE;
}

static void kiba_app_chooser_combobox_changed_cb (GtkComboBox *combobox,
						  gpointer     user_data) {
	KibaAppChooser *chooser = user_data;
	KibaAppChooserApp *app;
	KibaAppChooserSection selected_section;
	KibaIconView *icon_view;
	GList *l;
	
	icon_view = KIBA_ICON_VIEW(chooser->view_win->icon_view);
	
	selected_section = kiba_app_chooser_section_from_string (
				gtk_combo_box_get_active_text(combobox));
	
	for (l = chooser->apps; l != NULL; l = l->next) {
		app = l->data;
		
		if (selected_section == KIBA_APP_SECTION_ALL ||  app->section == selected_section) {
			if (!app->added) {
				kiba_icon_view_win_add (chooser->view_win, app->box);
				gtk_widget_show (app->box);
				app->added = TRUE;
			}
		}
		else if (app->added) {
			app->box = gtk_widget_ref (app->box);
			app->image->surface = cairo_surface_reference (app->image->surface);
			kiba_icon_view_win_remove (chooser->view_win, app->box);
			gtk_widget_hide (app->box);
			app->added = FALSE;
		}
	}
	
	chooser->active_section = selected_section;
}

static void kiba_icon_view_child_selected_cb (KibaIconView *icon_view,
					      GtkWidget    *child,
					      gpointer      user_data) {
	KibaAppChooser *chooser = user_data;
	KibaAppChooserApp *app = NULL;
	GList *l;
	char *markup;
	
	for (l = chooser->apps; l != NULL; l = l->next) {
		KibaAppChooserApp *check_app = l->data;
		
		if (check_app->box == child) {
			app = check_app;
			break;
		}
	}
	
	if (!app)
		return;
	
	kiba_icon_view_select_child (KIBA_ICON_VIEW(chooser->view_win->icon_view), app->box);
	markup = g_markup_printf_escaped ("<span size=\"large\" weight=\"bold\">%s</span>", app->name);
	gtk_label_set_markup (GTK_LABEL(chooser->name_label), markup);
	g_free (markup);
	gtk_label_set_text (GTK_LABEL(chooser->description_label), app->description);
	gtk_widget_queue_draw (chooser->view_win->icon_view);
}

static gboolean kiba_icon_view_expose_cb (GtkWidget      *widget,
				          GdkEventExpose *expose,
					  gpointer        data) {
	KibaAppChooser *chooser = data;
	KibaInfoWin *win;
	cairo_t *cr = NULL;
	KibaRGBA borderColor;
	
	if (!GTK_WIDGET_REALIZED (chooser->view_win->win))
		return FALSE;
	
	win = KIBA_INFO_WIN(chooser->view_win->win);
	
	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	borderColor = kiba_get_info_win_border_color();
	
	cairo_set_source_rgba (cr,
		borderColor.r,
		borderColor.g,
		borderColor.b,
		borderColor.a);
	cairo_set_line_width (cr, 3);
	
	cairo_move_to (cr,
		0,
		chooser->name_label->allocation.y - 5);
	cairo_rel_line_to (cr,
		chooser->view_win->win->allocation.width,
		0);
	cairo_stroke (cr);
	
	cairo_move_to (cr,
		0,
		chooser->filter->allocation.y + 5 +
		chooser->filter->allocation.height);
	cairo_rel_line_to (cr,
		chooser->view_win->win->allocation.width,
		0);
	cairo_stroke (cr);
	
	cairo_destroy (cr);
	
	return FALSE;
}

static void kiba_app_drag_data_get_cb (GtkWidget        *widget,
				       GdkDragContext   *drag_context,
				       GtkSelectionData *data,
				       guint             info,
				       guint             time,
				       gpointer          user_data) {
	KibaAppChooserApp *app = user_data;
	
	gtk_selection_data_set(data,
		GDK_SELECTION_TYPE_STRING,
		8,	/* 8 bits per character. */
		(guchar *)app->uri,
		strlen(app->uri));
}

static void add_apps_from_directory (KibaAppChooser *chooser,
			             const char     *app_path) {
	static const char desktop_entry_group[] = "Desktop Entry";
	static const char desktop_file_suffix[] = ".desktop";
	static const struct { const char *name; KibaAppChooserSection section; } categories_data[] = {
		{ "AudioVideo", KIBA_APP_SECTION_AUDIO_VIDEO },
		{ "WebBrowser", KIBA_APP_SECTION_NETWORK },
		{ "Game", KIBA_APP_SECTION_GAMES },
		{ "Graphics", KIBA_APP_SECTION_GRAPHICS },
		{ "Office", KIBA_APP_SECTION_OFFICE },
		{ "Network", KIBA_APP_SECTION_NETWORK },
		{ "Development", KIBA_APP_SECTION_DEVELOPMENT },
		{ "Utility", KIBA_APP_SECTION_TOOLS },
		{ "Filemanager", KIBA_APP_SECTION_TOOLS },
		{ "Accessibility", KIBA_APP_SECTION_TOOLS },
		{ "Settings", KIBA_APP_SECTION_SETTINGS },
		{ "X-XFCE", KIBA_APP_SECTION_SYSTEM },
		{ "KDE", KIBA_APP_SECTION_SYSTEM },
		{ "System", KIBA_APP_SECTION_SYSTEM },
		{ "GNOME", KIBA_APP_SECTION_SYSTEM },
	};
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaAppChooserApp *app;
	GtkWidget *label;
	GDir *app_dir;
	char *filename;
	char *markup;
	GError *error = NULL;
	gint i;
	
	app_dir = g_dir_open (app_path, 0, &error);
	g_clear_error (&error);
	error = NULL;
	
	if (!app_dir)
		return;
	
	while ((filename = (char*) g_dir_read_name (app_dir)) != NULL) {
		GtkWidget *vbox;
		GKeyFile *file;
		GdkPixbuf *pixbuf = NULL;
		char *icon;
		char *path;
		char *name;
		char *categories;
		char *description;
		
		path = g_strdup_printf ("%s/%s", app_path, filename);
		
		if (g_file_test (path, (G_FILE_TEST_IS_DIR)) == TRUE) {
			add_apps_from_directory (chooser, path);
			g_free (path);
			continue;
		}
		if (!g_str_has_suffix (filename, desktop_file_suffix)) {
			g_free (path);
			continue;
		}
		
		file = g_key_file_new ();
		if (!g_key_file_load_from_file (file, path, 0, &error)) {
			g_key_file_free (file);
			g_clear_error (&error);
			g_free (path);
			continue;
		}
		
		icon = g_key_file_get_value (file, desktop_entry_group, "Icon", &error);
		g_clear_error (&error);
		error = NULL;
		if (!icon) {
			g_free (path);
			continue;
		}
		
		categories = g_key_file_get_value (file, desktop_entry_group, "Categories", &error);
		g_clear_error (&error);
		error = NULL;
		
		name = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Name",
				NULL,
				&error);
		g_clear_error (&error);
		error = NULL;
		if (!name) {
			name = g_key_file_get_value (file, desktop_entry_group, "Name", &error);
			g_clear_error (&error);
			error = NULL;
		}
		
		description = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Comment",
				NULL,
				&error);
		g_clear_error (&error);
		error = NULL;
		if (!description) {
			description = g_key_file_get_value (file, desktop_entry_group, "Comment", &error);
			g_clear_error (&error);
			error = NULL;
		}
		g_key_file_free (file);
		
		kiba_load_icon (icon, 48, &pixbuf);
		if (!pixbuf)
			continue;
		
		app = g_new0 (KibaAppChooserApp, 1);
		
		vbox = gtk_vbox_new(FALSE, 0);
		/* enabling drag and drop */
		gtk_drag_source_set (vbox,
			GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_COPY);
		
		app->image = KIBA_IMAGE(kiba_image_new());
		gtk_widget_set_usize (GTK_WIDGET(app->image), 48, 48);
		kiba_image_set_pixbuf (app->image, pixbuf, 48, 48);
		g_object_unref(pixbuf);
		g_free (icon);
		gtk_widget_show (GTK_WIDGET(app->image));
		gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET(app->image), FALSE, FALSE, 10);
		
		app->section = -1;
		if (categories) {
			for (i = 0; i < G_N_ELEMENTS(categories_data); i++) {
				const char *section = categories_data[i].name;
				
				if (g_strrstr (categories, section)) {
					app->section = categories_data[i].section;
					break;
				}
			}
		}
		if (app->section == -1) {
			app->section = KIBA_APP_SECTION_TOOLS;
		}
		
		chooser->apps = g_list_append (chooser->apps, app);
		
		label = gtk_label_new(name);
		gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
		gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
		gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
		markup = g_markup_printf_escaped ("<span size=\"small\">%s</span>",
				GTK_LABEL (label)->text);
		gtk_label_set_markup (GTK_LABEL (label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
		gtk_widget_show(label);
		
		app->box = vbox;
		app->uri = path;
		app->added = FALSE;
		app->chooser = chooser;
		app->name = name;
		app->description = description;
		
		g_signal_connect (G_OBJECT (vbox), 
				  "button_press_event",
				  G_CALLBACK (app_button_press_callback),
				  app);
		g_signal_connect (GTK_OBJECT(vbox),
				  "drag-data-get",
				  G_CALLBACK(kiba_app_drag_data_get_cb),
				  app);
	}
	g_dir_close (app_dir);
}

static void filter_activate_cb (GtkEntry *entry,
				gpointer  data) {
	KibaAppChooser *chooser = data;
	KibaAppChooserApp *app;
	GList *l;
	const char *entry_text;
	char *filter_text;
	KibaIconView *icon_view;
	gint activeIcons = 0;
	
	icon_view = KIBA_ICON_VIEW(chooser->view_win->icon_view);
	
	entry_text = gtk_entry_get_text(entry);
	filter_text = g_strdup(entry_text);
	kiba_strdown (&filter_text);
	
	for (l = chooser->apps; l != NULL; l = l->next) {
		app = l->data;
		char *name = NULL;
		char *desc = NULL;
		
		if (chooser->active_section != KIBA_APP_SECTION_ALL &&
		    app->section != chooser->active_section)
			continue;
		
		if (app->name) {
			name = g_strdup(app->name);
			kiba_strdown (&name);
		}
		
		if (app->description) {
			desc = g_strdup(app->description);
			kiba_strdown (&desc);
		}
		
		if (app->description) {
			desc = g_strdup(app->description);
			kiba_strdown (&desc);
		}
		
		if (!app->added) {
			if ((name && g_strrstr (name, filter_text)) || (desc && g_strrstr (desc, filter_text))) {
				gtk_widget_show (app->box);
				kiba_icon_view_win_add (chooser->view_win, app->box);
				app->added = TRUE;
				activeIcons++;
			}
		}
		else if ((!name || !g_strrstr (name, filter_text)) && (!desc || !g_strrstr (desc, filter_text))) {
			app->box = gtk_widget_ref (app->box);
			app->image->surface = cairo_surface_reference (app->image->surface);
			kiba_icon_view_win_remove (chooser->view_win, app->box);
			gtk_widget_hide (app->box);
			app->added = FALSE;
		}
		
		if (name) {
			g_free (name);
		}
		
		if (desc) {
			g_free (desc);
		}
	}
	
	if (activeIcons == 0) {
		gtk_label_set_text (GTK_LABEL(chooser->description_label), NULL);
		gtk_label_set_text (GTK_LABEL(chooser->name_label), NULL);
	}
	else {
		for (l = chooser->apps; l != NULL; l = l->next) {
			app = l->data;
			
			if (kiba_icon_view_get_selected_child(icon_view) == app->box)
				break;
// 			if (icon_view->active_children->data == app->box)
// 				break;
		}
		
		char *markup;
		markup = g_markup_printf_escaped (
				"<span size=\"large\" weight=\"bold\">%s</span>",
				app->name);
		gtk_label_set_markup (GTK_LABEL(chooser->name_label), markup);
		g_free (markup);
		gtk_label_set_text (GTK_LABEL(chooser->description_label), app->description);
	}
	
	g_free (filter_text);
}

/**
Class Functions
**/

static void kiba_app_chooser_class_init (KibaAppChooserClass *class) {
	kiba_app_chooser_signals[SIG_APP_ACTIVATE] =
		g_signal_new ("app-activate",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaAppChooserClass, app_activate),
			NULL, NULL,
			g_cclosure_marshal_VOID__STRING,
			G_TYPE_NONE, 1,
			G_TYPE_STRING);
}

static void kiba_app_chooser_init (KibaAppChooser *chooser) {
	chooser->apps = NULL;
	chooser->view_win = NULL;
}

/**
 * kiba_app_chooser_new:
 * @app_path: The Path to .desktop Application file(s).
 *
 * Creates and returns a KibaAppChooser Instance.
 *
 * Returns: The KibaAppChooser Instance created.
 */
KibaAppChooser* kiba_app_chooser_new (const char *app_path) {
	KibaAppChooser *chooser;
	KibaAppChooserApp *app;
	GtkWidget *box;
	GtkWidget *section_label;
	GtkWidget *filter_label;
	GList *l;
	char *markup;
	gint i;
	
	g_return_val_if_fail (app_path != NULL, NULL);
	
	chooser = g_object_new (KIBA_TYPE_APP_CHOOSER, NULL);
	
	add_apps_from_directory (chooser, app_path);
	
	chooser->view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT, 6, 4, TRUE);
	gtk_window_set_role (GTK_WINDOW(chooser->view_win->win), "kiba-app-chooser");
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(chooser->view_win->win), TRUE);
	gtk_window_set_position (GTK_WINDOW(chooser->view_win->win), GTK_WIN_POS_CENTER);
	
	/* create a tree model containing all sections */
	box = gtk_hbox_new (FALSE, 5);
		section_label = gtk_label_new(NULL);
		gtk_misc_set_alignment (GTK_MISC(section_label), 0.1, 0.5);
		markup = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", _("Section"));
		gtk_label_set_markup (GTK_LABEL (section_label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (box),
			section_label, FALSE, FALSE, 5);
		gtk_widget_show(section_label);
		
		chooser->combobox = gtk_combo_box_new_text ();
		for (i = 0; i < KIBA_APP_SECTION_LAST; i++) {
			const char *section = section_data[i];
			
			gtk_combo_box_append_text (GTK_COMBO_BOX(chooser->combobox), section);
		}
		gtk_combo_box_set_active(GTK_COMBO_BOX(chooser->combobox), 0);
		gtk_widget_set_usize (chooser->combobox, 50, -1);
		gtk_box_pack_start (GTK_BOX (box),
			chooser->combobox, TRUE, TRUE, 5);
		gtk_widget_show (chooser->combobox);
		gtk_box_pack_start (GTK_BOX (chooser->view_win->top_box),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	box = gtk_hbox_new (FALSE, 5);
		filter_label = gtk_label_new(NULL);
		gtk_misc_set_alignment (GTK_MISC(filter_label), 0.1, 0.5);
		markup = g_markup_printf_escaped ("<span weight=\"bold\">%s</span>", _("Filter"));
		gtk_label_set_markup (GTK_LABEL (filter_label), markup);
		g_free (markup);
		gtk_box_pack_start (GTK_BOX (box),
			filter_label, FALSE, FALSE, 5);
		gtk_widget_show(filter_label);
		
		chooser->filter = gtk_entry_new();
		gtk_box_pack_start (GTK_BOX (box),
			chooser->filter, TRUE, TRUE, 5);
		gtk_widget_show(chooser->filter);
		g_signal_connect (G_OBJECT(chooser->filter), "activate",
			G_CALLBACK(filter_activate_cb), chooser);
		
		gtk_box_pack_start (GTK_BOX (chooser->view_win->top_box),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	GtkRequisition req;
	if (filter_label->allocation.width > section_label->allocation.width) {
		gtk_widget_size_request (filter_label, &req);
		gtk_widget_set_usize (section_label, req.width, -1);
	}
	else {
		gtk_widget_size_request (section_label, &req);
		gtk_widget_set_usize (filter_label, req.width, -1);
	}
	
	g_signal_connect (G_OBJECT(chooser->combobox), "changed",
		G_CALLBACK(kiba_app_chooser_combobox_changed_cb), chooser);
	
	for (l = chooser->apps; l != NULL; l = l->next) {
		app = l->data;
		
		kiba_icon_view_win_add (chooser->view_win, app->box);
		app->added = TRUE;
	}
	
	app = chooser->apps->data;
	
	chooser->name_label = gtk_label_new(app->name);
	gtk_label_set_line_wrap(GTK_LABEL(chooser->name_label), TRUE);
	gtk_label_set_justify(GTK_LABEL(chooser->name_label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(chooser->name_label), PANGO_ELLIPSIZE_END);
	markup = g_markup_printf_escaped ("<span size=\"large\" weight=\"bold\">%s</span>",
			GTK_LABEL (chooser->name_label)->text);
	gtk_label_set_markup (GTK_LABEL (chooser->name_label), markup);
	g_free (markup);
	gtk_box_pack_start (GTK_BOX (chooser->view_win->bottom_box),
		chooser->name_label, FALSE, FALSE, 5);
	gtk_widget_show(chooser->name_label);
	
	chooser->description_label = gtk_label_new(app->description);
	gtk_label_set_line_wrap(GTK_LABEL(chooser->description_label), TRUE);
	gtk_label_set_justify(GTK_LABEL(chooser->description_label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(chooser->description_label), PANGO_ELLIPSIZE_END);
	gtk_box_pack_start (GTK_BOX (chooser->view_win->bottom_box),
		chooser->description_label, FALSE, FALSE, 10);
	gtk_widget_show(chooser->description_label);
	
	g_signal_connect (G_OBJECT(chooser->view_win->win), "destroy",
		G_CALLBACK(kiba_app_chooser_destroy_cb), chooser);
	g_signal_connect (GTK_OBJECT(chooser->view_win->icon_view), "child-selected",
		G_CALLBACK(kiba_icon_view_child_selected_cb), chooser);
	g_signal_connect_after(G_OBJECT (chooser->view_win->win), "expose-event",
		G_CALLBACK(kiba_icon_view_expose_cb), chooser);
	
	gtk_widget_grab_focus (chooser->filter);
	
	chooser->active_section = KIBA_APP_SECTION_ALL;
	
	return chooser;
}
