#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdk.h>
#include <pango/pango.h>

#include "kiba-cairo-utils.h"
#include "kiba-types.h"

#define NUM_CHANNELS 4

/**
 * SECTION:kiba-cairo-utils
 * @short_description: Utility methods for cairo.
 * @see_also:
 * @stability: Unstable
 *
 * Utility methods for cairo.
 */

/**
Ofscreen Pixmap Rendering
**/
GdkPixmap* kiba_cairo_utils_create_pixmap (int w, int h) {
	GdkPixmap *pixmap;
	GdkVisual *visual;
	GdkColormap *colormap;
	
	visual = gdk_visual_get_best_with_depth (32); /**< TODO: 32 is only valif for composited windows,
                                                       not composited windows need 24 here */
	if (!visual)
		return NULL;
	
	pixmap = gdk_pixmap_new (NULL, w, h, 32);
	if (!pixmap)
		return NULL;
	
	colormap = gdk_colormap_new (visual, FALSE);
	if (!colormap) {
		gdk_pixmap_unref (pixmap);
		return NULL;
	}
	
	gdk_drawable_set_colormap (GDK_DRAWABLE (pixmap), colormap);
	gdk_colormap_unref (colormap);
	
	return pixmap;
}

/**
 * cairo_create_pixmap_context:
 * @width: The drawing context width.
 * @height: The drawing context height.
 * @pixmap: The GdkPixmap return.
 * @gc: The GdkGC return.
 *
 * Creates and returns a context on a pixmap.
 *
 * Returns: The drawing context created.
 */
cairo_t* cairo_create_pixmap_context (int         width,
				      int         height,
				      GdkPixmap **pixmap,
				      GdkGC     **gc) {
	*pixmap = kiba_cairo_utils_create_pixmap (width, height);
	if (*pixmap == NULL)
		return NULL;
	
	*gc = gdk_gc_new (*pixmap);
	if (*gc == NULL)
		return NULL;
	
	return gdk_cairo_create (*pixmap);
}

/**
Puc Surface
**/

/**
 * cairo_get_image_surface_from_puc_data:
 * @puc_data: The puc data to create the surface from.
 * @width: The surface width.
 * @height: The surface height.
 *
 * Creates and returns a image surface from puc data.
 *
 * Returns: The surface created.
 */
cairo_surface_t* cairo_get_image_surface_from_puc_data (unsigned char **puc_data,
							int             width,
							int             height) {
	cairo_surface_t *image_surface;
	
	/* Create a cairo surface for our allocated image. */
	image_surface = cairo_image_surface_create_for_data (*puc_data, CAIRO_FORMAT_ARGB32,
				width, height, NUM_CHANNELS * width);
	if (cairo_surface_status (image_surface) != CAIRO_STATUS_SUCCESS) {
		fprintf (stderr, "Couldn't create cairo puc surface\n");
		return NULL;
	}
	
	return image_surface;
}

/**
 * cairo_get_context_from_puc_data:
 * @puc_data: The puc data to create the drawing context from.
 * @width: The drawing context width.
 * @height: The drawing context height.
 *
 * Creates and returns a context from puc data.
 *
 * Returns: The drawing context created.
 */
cairo_t* cairo_get_context_from_puc_data (unsigned char **puc_data,
					  int             width,
					  int             height) {
	cairo_t *draw_context = NULL;
	cairo_surface_t *image_surface = NULL;
	
	image_surface = cairo_get_image_surface_from_puc_data (puc_data, width, height);
	if (!image_surface)
		return NULL;
	
	/* Create a cairo drawing context */
	draw_context = cairo_create (image_surface);
	if (cairo_status (draw_context) != CAIRO_STATUS_SUCCESS) {
		fprintf (stderr, "Couldn't create cairo context for puc Surface\n");
		cairo_surface_destroy (image_surface);
		return NULL;
	}
	
	cairo_surface_destroy (image_surface);
	
	return draw_context;
}

/**
 * cairo_create_puc_data:
 * @width: The puc data width.
 * @height: The puc data height.
 *
 * Creates puc data with dimensions defined by the given width and height.
 *
 * Returns: The puc data created.
 */
unsigned char* cairo_create_puc_data (int width,
				      int height) {
	unsigned char* puc_data;
	
	/* allocate new space for the pixel data */
	puc_data = calloc (NUM_CHANNELS * width * height, sizeof (unsigned char));
	if (!puc_data)
		fprintf (stderr, "Couldn't allocate Pixel Data\n");
	
	return puc_data;
}

/**
utils
**/

/**
 * cairo_paint_with_prelight:
 * @cr: The drawing context.
 * @alpha: The alpha for the prelight.
 * @width: ...
 * @height: ...
 *
 * Renders the current context with prelight.
 */
void cairo_paint_with_prelight (cairo_t *cr,
				float   alpha,
				int      width,
				int      height) {
	cairo_pattern_t *prelight_pattern;
	
	prelight_pattern = cairo_pattern_create_linear (0, 0, width, height);
	cairo_pattern_add_color_stop_rgba (prelight_pattern,
		0.5 , 1, 1, 1, alpha);
	cairo_mask (cr, prelight_pattern);
	cairo_pattern_destroy (prelight_pattern);
}

/**
 * cairo_rotate_centered:
 * @cr: The drawing context.
 * @angle: The angle to rotate.
 * @x: The x surface offset.
 * @y: The y surface offset.
 *
 * Rotates the context centered.
 */
void cairo_rotate_centered (cairo_t *cr,
			    float   angle,
			    int      x,
			    int      y) {
	cairo_translate (cr, x, y);
	cairo_rotate (cr, M_PI * angle);
	cairo_translate (cr, -x, -y);
}

/**
 * cairo_update_pango_layout:
 * @text: The layout text.
 * @description: The layout font description.
 * @cr: The drawing context.
 * @layout: The PangoLayout return.
 *
 * Creates/Updates a PangoLayout.
 */
void cairo_update_pango_layout (const char   *text,
			        const char   *description,
			        cairo_t      *cr,
			        PangoLayout **layout) {
	cairo_font_options_t *font_options;
	PangoFontDescription *font_description;
	
	if (text == NULL)
		return;
	
	if (*layout != NULL) {
		g_object_unref (*layout);
		*layout = NULL;
	}
	
	*layout = pango_cairo_create_layout (cr);
	font_options = cairo_font_options_create ();
	font_description = pango_font_description_new ();
	
	cairo_font_options_set_hint_metrics (font_options, CAIRO_HINT_METRICS_ON);
	cairo_font_options_set_hint_style (font_options, CAIRO_HINT_STYLE_NONE);
	cairo_font_options_set_antialias (font_options, CAIRO_SUBPIXEL_ORDER_DEFAULT);
	cairo_font_options_set_subpixel_order (font_options, CAIRO_SUBPIXEL_ORDER_RGB);
	pango_cairo_context_set_font_options (pango_layout_get_context (*layout), font_options);
	cairo_font_options_destroy (font_options);
	
	font_description = pango_font_description_from_string (description);
	pango_layout_set_font_description (*layout, font_description);
	pango_font_description_free (font_description);
	
	if (!g_utf8_validate (text, -1, NULL)) {
		char *utf8;
		
		utf8 = g_locale_to_utf8 (text, -1, NULL, NULL, NULL);
		if (utf8) {
			pango_layout_set_text (*layout, utf8, -1);
			g_free (utf8);
		}
	}
	else {
		pango_layout_set_text (*layout, text, -1);
	}
}

/**
 * cairo_set_source_rgba_from_string:
 * @cr: The Cairo Context.
 * @alpha_factor: Alpha Factor.
 * @rgba: The RGBA String (#FFFFFFFF for 100% opacity and white color).
 *
 * Sets cairo context's RGBA from string.
 */
void cairo_set_source_rgba_from_string (cairo_t    *cr,
					float      alpha_factor,
					const char *rgba) {
	int r,g,b,a;
	
	sscanf (rgba, "#%02X%02X%02X%02X", &r, &g, &b, &a);
	
	cairo_set_source_rgba (cr, (float) r/256,
		(float) g/256,
		(float) b/256,
		(float) alpha_factor*a/256);
}

/**
 * cairo_get_png_pattern:
 * @file: The PNG file.
 * @alpha: The alpha for the pattern.
 *
 * Creates a png file pattern.
 *
 * Returns: The cairo_pattern_t created.
 */
cairo_pattern_t* cairo_get_png_pattern (const char *file,
					float      alpha) {
	cairo_pattern_t *pattern;
	cairo_surface_t *image;
	cairo_t *cr;
	
	image = cairo_image_surface_create_from_png (file);
	if (image == NULL) {
		fprintf (stderr, "Failed to load file '%s' to image Surface\n", file);
		return NULL;
	}
	
	if (alpha < 1.0 && alpha > 0.0) {
		cr = cairo_create (image);
		if (cr) {
			cairo_set_operator (cr, CAIRO_OPERATOR_XOR);
			cairo_paint_with_alpha (cr, 1.0 - alpha);
			cairo_destroy (cr);
		}
	}
	
	pattern = cairo_pattern_create_for_surface (image);
	cairo_surface_destroy (image);
	
#if CAIRO_VERSION_MAJOR < 1 || (CAIRO_VERSION_MAJOR == 1 && CAIRO_VERSION_MINOR < 4)
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
#else
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REFLECT);
#endif
	
	return pattern;
}

/**
 * cairo_get_gradient_pattern:
 * @x0: The topleft x coordinate.
 * @y0: The topleft y coordinate.
 * @x1: The bottomright x coordinate.
 * @y1: The bottomright y coordinate.
 * @rgba: The gradient array.
 * @elemets: Number of Colors in @rgba.
 *
 * Creates a gradient pattern.
 *
 * Returns: The cairo_pattern_t created.
 */
cairo_pattern_t* cairo_get_gradient_pattern (int x0,
                                             int y0,
                                             int x1,
                                             int y1,
                                             KibaRGBA  *rgba,
                                             int elements) {
	cairo_pattern_t *pattern;
	gint i;
	float offset = 0.0;
	float step;
	
	if (elements > 1)
		step = 1.0f/((float) (elements-1));
	else
		step = 0.0f;
	
	pattern = cairo_pattern_create_linear (x0, y0, x1, y1);
	
	for (i = 0; i < elements; i++) {
		cairo_pattern_add_color_stop_rgba (pattern, offset,
			rgba[i].r, rgba[i].g, rgba[i].b, rgba[i].a);
		offset += step;
	}
	
	return pattern;
}

/**
 * cairo_get_radial_pattern:
 * @x0: The topleft x coordinate.
 * @y0: The topleft y coordinate.
 * @x1: The bottomright x coordinate.
 * @y1: The bottomright y coordinate.
 * @rgba_1: The inner radial color.
 * @rgba_2: The outer radial color.
 * @step_size: The space between lines.
 *
 * Creates a radial pattern.
 *
 * Returns: The cairo_pattern_t created.
 */
cairo_pattern_t* cairo_get_radial_pattern (int        x0,
					   int        y0,
					   int        x1,
					   int        y1,
					   KibaRGBA   rgba_1,
					   KibaRGBA   rgba_2,
					   float     step_size) {
	cairo_pattern_t *pattern;
	cairo_matrix_t matrix;
	gint steps, i;
	gint px0, px1, py0, py1;
	float step_offset;
	
	px0 = MIN (x0,x1);
	px1 = MAX (x0,x1);
	py0 = MIN (y0,y1);
	py1 = MAX (y0,y1);
	
	pattern = cairo_pattern_create_radial (x0 + x1/2, y0 + y1/2,
			MIN(px1-px0,py1-py0) * 0.4,
			x0 + x1/2, y0 + y1/2,
			MIN(px1-px0,py1-py0) * 0.8);
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
	
	cairo_matrix_init_translate (&matrix, 1, 1);
	cairo_pattern_set_matrix (pattern, &matrix);
	
	step_offset = step_size/75.0;
	steps = (int) (1.0/(4.0*step_offset));
	step_offset = 1.0/(4*steps);
	
	for (i = 0; i < steps; i++) {
		float offset;
		
		offset = i*4*step_offset;
		
		cairo_pattern_add_color_stop_rgba (pattern, offset,
			rgba_1.r, rgba_1.g, rgba_1.b, rgba_1.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + step_offset,
			rgba_2.r, rgba_2.g, rgba_2.b, rgba_2.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + 2*step_offset,
			rgba_2.r, rgba_2.g, rgba_2.b, rgba_2.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + 3*step_offset,
			rgba_1.r, rgba_1.g, rgba_1.b, rgba_1.a);
	}
	
	return pattern;
}

/**
 * cairo_get_linear_pattern:
 * @x0: The topleft x coordinate.
 * @y0: The topleft y coordinate.
 * @x1: The bottomright x coordinate.
 * @y1: The bottomright y coordinate.
 * @rgba_1: The inner line color.
 * @rgba_2: The outer line color.
 * @affection: The affection of the rendered lines.
 *
 * Creates a linear pattern.
 *
 * Returns: The cairo_pattern_t created.
 */
cairo_pattern_t* cairo_get_linear_pattern (int        x0,
					   int        y0,
					   int        x1,
					   int        y1,
					   KibaRGBA   rgba_1,
					   KibaRGBA   rgba_2,
					   float     affection,
					   float     step_size) {
	cairo_pattern_t *pattern;
	cairo_matrix_t matrix;
	gint steps, i;
	gint px0, px1, py0, py1;
	float step_offset;
	
	px0 = MIN (x0,x1);
	px1 = MAX (x0,x1);
	py0 = MIN (y0,y1);
	py1 = MAX (y0,y1);
	
	pattern = cairo_pattern_create_linear (0, 0,
		MIN(px1-px0,py1-py0) * sin ((affection / 180.0) * M_PI),
		MIN(px1-px0,py1-py0) * sin ((affection / 180.0) * M_PI - M_PI/2));
	cairo_pattern_set_extend (pattern, CAIRO_EXTEND_REPEAT);
	
	cairo_matrix_init_translate (&matrix, 1, 1);
	cairo_pattern_set_matrix (pattern, &matrix);
	
	step_offset = step_size/75.0;
	steps = (int) (1.0/(4.0*step_offset));
	step_offset = 1.0/(4*steps);
	
	for (i = 0; i < steps; i++) {
		float offset;
		
		offset = i*4*step_offset;
		
		cairo_pattern_add_color_stop_rgba (pattern, offset,
			rgba_1.r, rgba_1.g, rgba_1.b, rgba_1.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + step_offset,
			rgba_2.r, rgba_2.g, rgba_2.b, rgba_2.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + 2*step_offset,
			rgba_2.r, rgba_2.g, rgba_2.b, rgba_2.a);
		cairo_pattern_add_color_stop_rgba (pattern, offset + 3*step_offset,
			rgba_1.r, rgba_1.g, rgba_1.b, rgba_1.a);
	}
	
	return pattern;
}

/**
 * cairo_paint_shape:
 * @cr: The drawing context.
 * @topleft_corner: The topleft corner shape.
 * @topright_corner: The topright corner shape.
 * @bottomleft_corner: The bottomleft corner shape.
 * @bottomright_corner: The bottomright corner shape.
 * @x0: The topleft x coordinate.
 * @y0: The topleft y coordinate.
 * @x1: The bottomright x coordinate.
 * @y1: The bottomright y coordinate.
 * @radius: The corner radius.
 *
 * Renders shape to the cairo_t.
 */
void cairo_paint_shape (cairo_t               *cr,
			KibaCornerShape        topleft_corner,
			KibaCornerShape        topright_corner,
			KibaCornerShape        bottomleft_corner,
			KibaCornerShape        bottomright_corner,
			float                 x0,
			float                 y0,
			float                 x1,
			float                 y1,
			float                 radius) {
	cairo_move_to (cr, x0, y0 + radius);
	
	switch (topleft_corner) {
		case KIBA_ROUNDED:
			cairo_arc (cr, x0 + radius, y0 + radius, radius, M_PI, 3 * M_PI / 2);
			break;
		case KIBA_MITERED:
			cairo_line_to (cr, x0 + radius, y0);
			break;
		case KIBA_NOTCHED:
			cairo_arc_negative (cr, x0, y0, radius, M_PI / 2, 2 * M_PI);
			break;
		default:
			cairo_line_to (cr, x0, y0);
			cairo_line_to (cr, x0 + radius, y0);
			break;
	}
	
	cairo_line_to (cr, x1 - radius, y0);
	
	switch (topright_corner) {
		case KIBA_ROUNDED:
			cairo_arc (cr, x1 - radius, y0 + radius, radius, 3 * M_PI / 2, 2 * M_PI);
			break;
		case KIBA_MITERED:
			cairo_line_to (cr, x1, y0 + radius);
			break;
		case KIBA_NOTCHED:
			cairo_arc_negative (cr, x1, y0, radius, M_PI, M_PI/2);
			break;
		default:
			cairo_line_to (cr, x1, y0);
			cairo_line_to (cr, x1, y0 + radius);
			break;
	}
	
	cairo_line_to (cr, x1, y1 - radius);
	
	switch (bottomright_corner) {
		case KIBA_ROUNDED:
			cairo_arc (cr, x1 - radius, y1 - radius, radius, 0, M_PI / 2);
			break;
		case KIBA_MITERED:
			cairo_line_to (cr, x1 - radius, y1);
			break;
		case KIBA_NOTCHED:
			cairo_arc_negative (cr, x1, y1, radius, 3 * M_PI / 2, M_PI);
			break;
		default:
			cairo_line_to (cr, x1, y1);
			cairo_line_to (cr, x1 - radius, y1);
			break;
	}
	
	cairo_line_to (cr, x0 + radius, y1);
	
	switch (bottomleft_corner) {
		case KIBA_ROUNDED:
			cairo_arc (cr, x0 + radius, y1 - radius, radius, M_PI / 2, M_PI);
			break;
		case KIBA_MITERED:
			cairo_line_to (cr, x0, y1 - radius);
			break;
		case KIBA_NOTCHED:
			cairo_arc_negative (cr, x0, y1, radius, 2 * M_PI, 3 * M_PI / 2);
			break;
		default:
			cairo_line_to (cr, x0, y1);
			cairo_line_to (cr, x0, y1 - radius);
			break;
	}
	
	cairo_line_to (cr, x0, y0 + radius);
	
	cairo_close_path (cr);
}

/**
 * cairo_update_surface_at_size:
 * @cr: The drawing context.
 * @surface: The surface to update.
 * @width: The surface width.
 * @height: The surface height.
 * @draw_function: The method to draw on the surface.
 * @data: The User data.
 *
 * Updates surface with the given size.
 */
void cairo_update_surface_at_size (cairo_t          *cr,
				   cairo_surface_t **surface,
				   gint              width,
				   gint              height,
				   CairoSurfaceFunc  draw_function,
				   void             *data) {
	cairo_surface_t *pNewSurface;
	cairo_t *pDrawContext;
	
	pNewSurface = cairo_surface_create_similar (cairo_get_target (cr),
				CAIRO_CONTENT_COLOR_ALPHA,
				width, height);
	
	if (cairo_surface_status (pNewSurface) != CAIRO_STATUS_SUCCESS)
		return;
	
	pDrawContext = cairo_create (pNewSurface);
	if (cairo_status (pDrawContext) != CAIRO_STATUS_SUCCESS)
		return;
	
	cairo_set_tolerance (pDrawContext, 0.1);
	cairo_set_source_rgba (pDrawContext, 0, 0, 0, 0);
	cairo_set_operator (pDrawContext, CAIRO_OPERATOR_SOURCE);
	cairo_paint (pDrawContext);
	
	cairo_set_operator (pDrawContext, CAIRO_OPERATOR_OVER);
	draw_function (pDrawContext, pNewSurface, width, height, data);
	
	cairo_destroy (pDrawContext);
	
	if (!pNewSurface)
		return;
	
	if (*surface)
		cairo_surface_destroy (*surface);
	*surface = pNewSurface;
}

/**
 * cairo_update_surface:
 * @cr: The drawing context.
 * @surface_data: The surface to update.
 * @draw_function: The method to draw on the surface.
 * @data: The User data.
 *
 * Updates surface (calling cairo_update_surface_at_size).
 */
void cairo_update_surface (cairo_t          *cr,
			   CairoSurfaceData *surface_data,
			   CairoSurfaceFunc  draw_function,
			   void             *data) {
	cairo_update_surface_at_size (cr,
		&surface_data->surface,
		surface_data->width, surface_data->height,
		draw_function, data);
}

/* FIXME: transparent parts are black */
/**
 * cairo_get_pixbuf_from_surface:
 * @surface: The surface to create the GdkPixbuf from.
 * @width: The GdkPixbuf width.
 * @height: The GdkPixbuf height.
 *
 * Creates a GdkPixbuf from cairo_surface_t.
 *
 * Returns: The GdkPixbuf created.
 */
GdkPixbuf* cairo_get_pixbuf_from_surface (cairo_surface_t *surface,
					  gint             width,
					  gint             height) {
	GdkPixmap *pixmap = NULL;
	GdkPixbuf *pixbuf = NULL;
	cairo_t *cr;
	
	g_return_val_if_fail (surface != NULL, NULL);
	
	pixmap = gdk_pixmap_new (NULL, width, height, 24);
	
	cr = gdk_cairo_create (GDK_DRAWABLE(pixmap));
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS) {
		g_free (pixbuf);
		return NULL;
	}
	
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_set_source_surface (cr, surface, 0, 0);
	cairo_paint (cr);
	cairo_destroy (cr);
	
	pixbuf = gdk_pixbuf_get_from_drawable (NULL, GDK_DRAWABLE(pixmap), NULL,
			0, 0, 0, 0, width, height);
	
	gdk_drawable_unref (GDK_DRAWABLE(pixmap));
	
	return pixbuf;
}
