/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-object
 * @short_description: Base class for objects on #KibaDock.
 * @see_also: #KibaWin,#KibaDock
 * @stability: Unstable
 *
 * #KibaObject implementation with container functionality.
 */

#include "../config.h"

#include "kiba.h"
#include "kiba-dock.h"
#include "kiba-container-object.h"

#define MIN_SIZE 28

G_DEFINE_TYPE (KibaContainerObject, kiba_container_object, KIBA_TYPE_OBJECT)

/*** Class Implementations ***/

static void kiba_container_object_screen_changed (GtkWidget *widget,
                                                  GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgb_colormap (screen);
	visual   = gdk_screen_get_rgb_visual (screen);
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

/**
 * gtk_widget_size_allocat implementation.
 **/
static void kiba_container_object_size_allocate (GtkWidget *widget,
                                                 GtkAllocation *allocation) {
	KibaContainerObject *object = KIBA_CONTAINER_OBJECT(widget);
	GtkWidget *child;
	const KibaLayoutData *d;
	GList *l;
	
	widget->allocation = *allocation;
	child = gtk_bin_get_child (GTK_BIN(widget));
	if (child) {
		GtkAllocation child_allocation;
		GtkRequisition requisition;
		
		gtk_widget_size_request (child, &requisition);
		child_allocation.width = requisition.width;
		child_allocation.height = requisition.height;
		child_allocation.x = 0;
		child_allocation.y = 0;
		gtk_widget_size_allocate (child, &child_allocation);
		
		if (GDK_IS_WINDOW (widget->window))
			gdk_window_move_resize (widget->window,
				widget->allocation.x + (allocation->width-requisition.width)/2,
				widget->allocation.y + (allocation->height-requisition.height)/2,
				child_allocation.width,
				child_allocation.height);
	} else if (GDK_IS_WINDOW (widget->window))
		gdk_window_move_resize (widget->window,
			widget->allocation.x,
			widget->allocation.y,
			widget->allocation.width,
			widget->allocation.height);
}

/**
 * gtk_widget_size_request implementation.
 **/
static void kiba_container_object_size_request (GtkWidget *widget,
                                                GtkRequisition *requisition) {
	KibaObject *object = KIBA_OBJECT(widget);
	GtkWidget *child = gtk_bin_get_child (GTK_BIN(widget));
	GtkWidget *dock = GTK_WIDGET(object)->parent;
	
	/* using the child allocation, if there is a child */
	if (child) {
		float scale_x = 1.0f;
		float scale_y = 1.0f;
		GList *l;
		
		scale_x = object->scale_x;
		scale_y = object->scale_y;
		
		for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
			KibaObjectCustomScaleData *data = l->data;
			
			/* skip scale method that just scale the surface */
			if (!data->scale_widget) continue;
			
			float custom_factor_x = 1.0f;
			float custom_factor_y = 1.0f;
			
			data->scale_func (object,
				&custom_factor_x,
				&custom_factor_y,
				data->data);
			
			scale_x *= custom_factor_x;
			scale_y *= custom_factor_y;
		}
		
		gtk_widget_size_request (child, requisition);
		object->size = (GtkRequisition) {requisition->width, requisition->height};
		requisition->width = (int) (requisition->width*scale_x);
		requisition->height = (int) (requisition->height*scale_y);
	} else {
		(* GTK_WIDGET_CLASS (kiba_container_object_parent_class)->size_request) (widget, requisition);
	}
}

/**
 * gtk_widget_unmap implementation.
 **/
static void kiba_container_object_unmap (GtkWidget *widget) {
	GtkBin *bin;
	
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_hide (widget->window);
	
	if (bin->child && GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_unmap (bin->child);
}

/**
 * gtk_widget_map implementation.
 **/
static void kiba_container_object_map (GtkWidget *widget) {
	GtkBin *bin;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_show (widget->window);
	
	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void kiba_container_object_unrealize (GtkWidget *widget) {
	(* GTK_WIDGET_CLASS (kiba_container_object_parent_class)->unrealize) (widget);
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_container_object_realize (GtkWidget *widget) {
	KibaContainerObject *object = KIBA_CONTAINER_OBJECT(widget);
	Kiba *kiba = kiba_get();
	GdkWindowAttr attributes;
	gint attributes_mask;
	gint border_width;
	GtkRequisition requisition;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	border_width = GTK_CONTAINER (widget)->border_width;
	
	gtk_widget_size_request (widget, &requisition);
	
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window); /* increase the reference count of the window */
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.visual = gtk_widget_get_visual(widget);
	attributes.colormap = gtk_widget_get_colormap(widget);
	attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	if (attributes.visual != NULL && attributes.colormap != NULL)
		attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;
	
	widget->window = gdk_window_new (gtk_widget_get_parent_window(widget), &attributes, attributes_mask);
	gdk_window_set_user_data (widget->window, widget); /* catching the window events on the widget */
	gdk_window_register_dnd (widget->window); /* allowing dnd an the window */
	gdk_window_set_type_hint (widget->window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	gdk_window_set_background (widget->window, &widget->style->base[GTK_WIDGET_STATE (widget)]);
}

/*** Class init ***/

static void kiba_container_object_class_init (KibaContainerObjectClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	/* GtkWidget method implementations */
	widget_class->realize = kiba_container_object_realize;
	widget_class->unrealize = kiba_container_object_unrealize;
	widget_class->map = kiba_container_object_map;
	widget_class->unmap = kiba_container_object_unmap;
	widget_class->size_allocate = kiba_container_object_size_allocate;
	widget_class->size_request = kiba_container_object_size_request;
	widget_class->screen_changed = kiba_container_object_screen_changed;
}

static void kiba_container_object_init (KibaContainerObject *object) {
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);
	
	kiba_container_object_screen_changed(GTK_WIDGET(object), NULL);
}

/**
 * kiba_container_object_new:
 * @group: The group on the dock.
 *
 * Creates and returns a KibaContainerObject container.
 *
 * Returns: The KibaContainerObject instance created.
 */
GtkWidget* kiba_container_object_new (const char *group) {
	GtkWidget *object = g_object_new (KIBA_TYPE_CONTAINER_OBJECT,
                                 "group", group, NULL);
	return object;
}
