/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-debug
 * @short_description: Debugging methods.
 * @see_also:
 * @stability: Unstable
 *
 * Debugging methods.
 */

#include "../config.h"
#include "kiba-debug.h"
#include <stdio.h>

typedef struct {
	gint        level;
	const char  *message;
} KibaVerboseLevelData;

static const char *level_data[] = {
	"Misc Message", "Info Message", "Warning", "Error", "FATAL"
};

static gboolean DEBUG_ENABLE = 0;

static void kiba_debug_vargs(int         level,
			     const char *file,
			     signed int  line,
			     const char *format,
			     va_list     args) {
	char  *arg_s = NULL;
	char  *debugMessage;
	
	arg_s = g_strdup_vprintf(format, args);
	debugMessage = g_strdup_printf ("%s (%s @ line %d):\n"
	                 "\t%s", level_data[level], file, line, arg_s);
	g_free(arg_s);

	switch(level) {
		case KIBA_DEBUG_MISC:
		case KIBA_DEBUG_INFO:
		case KIBA_DEBUG_WARNING:
			g_printf("%s\n", debugMessage);
			break;
		case KIBA_DEBUG_ERROR:
			g_warning("%s", debugMessage);
			g_fprintf(stderr, "\tFor a core dump, run kiba-dock with --g-fatal-warnings.\n");
			break;
		case KIBA_DEBUG_FATAL:
			g_error("%s", debugMessage);
	}

	g_free(debugMessage);	
}

/**
 * kiba_debug_message:
 * @level: level of Debugging.
 * @file: __FILE__ value.
 * @line: __LINE__ value.
 * @format: ...
 * @...: ...
 *
 * Prints a Debug Message with KIBA_DEBUG_ERROR level.
 */
void kiba_debug_message (int         level,
			 const char *file,
			 signed int  line,
			 const char *format,
			 ...) {
#ifdef ENABLE_DEBUG_MODE
	va_list args;
	
	if (!DEBUG_ENABLE && level < KIBA_DEBUG_ERROR)
		return;
	g_return_if_fail(format != NULL);
	
	va_start(args, format);
	kiba_debug_vargs(level, file, line, format, args);
	va_end(args);
#endif
}

/**
 * kiba_debug_is_enabled:
 *
 * Returns wether Debugging is enabled.
 *
 * Returns: TRUE if debugging is enabled, otherwise FALSE is returned.
 */
gboolean kiba_debug_is_enabled() {
	return DEBUG_ENABLE;
}

/**
 * kiba_debug_init:
 *
 * Initiall/Enable Debugging.
 */
void kiba_debug_init (void) {
	DEBUG_ENABLE = TRUE;
}
