/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-desktop-editor
 * @short_description: Editor for .desktop files.
 * @see_also:#KibaDesktopIcon
 * @stability: Unstable
 *
 * Editor for .desktop files.
 */

#include <libintl.h>
#include <string.h>

#include "kiba-desktop-editor.h"
#include "kiba-key-file.h"
#include "kiba-utils.h"
#include "kiba-image.h"

#define _(string)  gettext (string)
#define CHOOSER_COLUMNS 8

G_DEFINE_TYPE (KibaDesktopEditor, kiba_desktop_editor, G_TYPE_OBJECT)

enum {
	SIG_KIBA_DESKTOP_FILE_SAVED = 0,
	SIG_KIBA_DESKTOP_EDITOR_LAST
};

static guint kiba_desktop_editor_signals[SIG_KIBA_DESKTOP_EDITOR_LAST] = { 0 };

static void kiba_desktop_editor_destroy_cb (GtkWidget *widget,
				            gpointer   data) {
	KibaDesktopEditor *editor = data;
	KibaDesktopEditorFile *desktop_file;
	GList *l;
	
	for (l = editor->desktop_files; l != NULL; l = editor->desktop_files) {
		desktop_file = l->data;
		
		editor->desktop_files = g_list_remove (editor->desktop_files, desktop_file);
		g_free (desktop_file->name);
		g_free (desktop_file->description);
		g_free (desktop_file->exec);
		g_free (desktop_file->group);
		g_free (desktop_file->path);
		g_free (desktop_file->icon);
		g_free (desktop_file);
	}
	g_list_free (editor->desktop_files);
	g_object_unref (editor);
}

static void kiba_desktop_editor_update_preview (KibaDesktopEditor     *editor,
						KibaDesktopEditorFile *file,
						gboolean               update_chooser) {
	GdkPixbuf *pixbuf = NULL;
	
	if (!file->icon)
		return;
	
	kiba_load_icon (file->icon, 48, &pixbuf);
	if (!pixbuf) {
		gtk_image_set_from_stock (GTK_IMAGE(editor->image),
			GTK_STOCK_MISSING_IMAGE, GTK_ICON_SIZE_DIALOG);
	}
	else {
		gtk_image_set_from_pixbuf (GTK_IMAGE(editor->image),
				pixbuf);
		g_object_unref (G_OBJECT(pixbuf));
	}
	
	if (!update_chooser)
		return;
	
	if (g_strrstr (file->icon, "/")) {
		gtk_file_chooser_set_filename (
			GTK_FILE_CHOOSER(editor->icon_chooser),
			file->icon);
	}
	else {
		GtkIconInfo *icon_info;
		char *suffix;
		
		icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
				file->icon, 48, 0);
		if (icon_info) {
			gtk_file_chooser_set_filename (
				GTK_FILE_CHOOSER(editor->icon_chooser),
				gtk_icon_info_get_filename (icon_info));
			gtk_icon_info_free (icon_info);
			return;
		}
		
		char *icon_name = g_strdup(file->icon);
		
		suffix = strstr (icon_name, ".png");
		if (suffix == NULL)
			suffix = strstr (icon_name, ".svg");
		if (suffix == NULL)
			return;
		*suffix = '\0';
		icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
					icon_name, 48, 0);
		g_free (icon_name);
		if (icon_info != NULL) {
			gtk_file_chooser_set_filename (
				GTK_FILE_CHOOSER(editor->icon_chooser),
				gtk_icon_info_get_filename (icon_info));
			gtk_icon_info_free (icon_info);
			return;
		}
	}
}

/**
 * kiba_desktop_editor_activate_file:
 * @editor: The KibaDesktopEditorFile instance.
 * @file: The KibaDesktopEditorFile to activate.
 *
 * Shows a KibaDesktopEditorFile in the editor.
 */
void kiba_desktop_editor_activate_file (KibaDesktopEditor     *editor,
				        KibaDesktopEditorFile *file) {
	g_return_if_fail (KIBA_IS_DESKTOP_EDITOR(editor));
	g_return_if_fail (file != NULL);
	
	editor->active_file = file;
	
	gtk_entry_set_text (GTK_ENTRY(editor->name_entry),
		file->name);
	gtk_entry_set_text (GTK_ENTRY(editor->description_entry),
		file->description);
	gtk_entry_set_text (GTK_ENTRY(editor->exec_entry),
		file->exec);
	gtk_entry_set_text (GTK_ENTRY(editor->group_entry),
		file->group);
	
	if (file) {
		kiba_desktop_editor_update_preview (editor, file, TRUE);
	}
	else {
		gtk_image_set_from_stock (GTK_IMAGE(editor->image),
			GTK_STOCK_MISSING_IMAGE, GTK_ICON_SIZE_DIALOG);
	}
}

static void kiba_desktop_editor_combobox_changed_cb (GtkComboBox *combobox,
						     gpointer     user_data) {
	KibaDesktopEditor *editor = user_data;
	KibaDesktopEditorFile *file = NULL;
	const char *selected_file;
	char *selected_path;
	GList *l;
	
	selected_file = gtk_combo_box_get_active_text(combobox);
	selected_path = g_strdup_printf ("%s/%s", editor->path, selected_file);
	
	for (l = editor->desktop_files; l != NULL; l = l->next) {
		KibaDesktopEditorFile *check_file = l->data;
		
		if (strcmp (check_file->path, selected_path) == 0) {
			file = check_file;
			break;
		}
	}
	g_free (selected_path);
	
	if (!file)
		return;
	
	kiba_desktop_editor_activate_file (editor, file);
}

static void save_button_press_cb (GtkWidget      *widget,
				  GdkEventButton *event,
				  gpointer        data) {
	KibaDesktopEditor *editor = data;
	KibaDesktopEditorFile *file = editor->active_file;
	const char *name;
	const char *description;
	const char *exec;
	const char *group;
	
	name = gtk_entry_get_text(GTK_ENTRY(editor->name_entry));
	if (name && (!file->name || strcmp (name, file->name) != 0)) {
		kiba_key_file_write_locale_desktop_entry (file->path,
			"Name", name);
		if (file->name)
			g_free (file->name);
		file->name = g_strdup (name);
	}
	description = gtk_entry_get_text(GTK_ENTRY(editor->description_entry));
	if (description && (!file->description || strcmp (description, file->description) != 0)) {
		kiba_key_file_write_locale_desktop_entry (file->path,
			"Description", description);
		if (file->description)
			g_free (file->description);
		file->description = g_strdup (description);
	}
	exec = gtk_entry_get_text(GTK_ENTRY(editor->exec_entry));
	if (exec && (!file->exec || strcmp (exec, file->exec) != 0)) {
		kiba_key_file_write_desktop_entry (file->path,
			"Exec", exec);
		if (file->exec)
			g_free (file->exec);
		file->exec = g_strdup (exec);
	}
	group = gtk_entry_get_text(GTK_ENTRY(editor->group_entry));
	if (group && (!file->group || strcmp (group, file->group) != 0)) {
		kiba_key_file_write_desktop_entry (file->path,
			"LauncherGroup", group);
		if (file->group)
			g_free (file->group);
		file->group = g_strdup (group);
	}
	if (file->icon_changed) {
		file->icon_changed = FALSE;
		kiba_key_file_write_desktop_entry (file->path,
			"Icon", file->icon);
	}
	
	g_signal_emit (editor,
		kiba_desktop_editor_signals[SIG_KIBA_DESKTOP_FILE_SAVED],
		0, file->path);
}

static void quit_button_press_cb (GtkWidget      *widget,
				  GdkEventButton *event,
				  gpointer        data) {
	GtkWidget *win = data;
	
	gtk_widget_destroy (win);
}

static void icon_chooser_changed_cb (GtkFileChooser *chooser,
				     gpointer        user_data) {
	KibaDesktopEditor *editor = user_data;
	GtkFileChooserDialog *dialog;
	const char *icon = gtk_file_chooser_get_filename (chooser);
	KibaDesktopEditorFile *file = editor->active_file;
	
	if (!icon || strcmp (icon, "(null)") == 0)
		return;
	
	if (file->icon) {
		if (strcmp (icon, file->icon) == 0)
			return;
		g_free (file->icon);
	}
	file->icon = g_strdup (icon);
	
	kiba_desktop_editor_update_preview (editor, file, FALSE);
	file->icon_changed = TRUE;
}

void update_preview_cb (GtkFileChooser *file_chooser,
			gpointer        data) {
	GtkWidget *preview;
	GdkPixbuf *pixbuf;
	
	preview = GTK_WIDGET(data);
	const char* filename = gtk_file_chooser_get_preview_filename(file_chooser);
	
	if (filename != NULL) {
		pixbuf = gdk_pixbuf_new_from_file_at_size(filename, 128, 128, NULL);
		
		gtk_image_set_from_pixbuf (GTK_IMAGE(preview), pixbuf);
		
		gtk_file_chooser_set_preview_widget_active(file_chooser, pixbuf != NULL);
		if (pixbuf)
			gdk_pixbuf_unref(pixbuf);
	}
}

static void add_file_from_path (KibaDesktopEditor *editor,
				const char        *path) {
	static const char desktop_entry_group[] = "Desktop Entry";
	static const char desktop_file_suffix[] = ".desktop";
	
	KibaDesktopEditorFile *desktop_file;
	GKeyFile *file;
	GError *error = NULL;
	
	if (g_file_test (path, (G_FILE_TEST_IS_DIR)) == TRUE) {
		return;
	}
	
	if (!g_str_has_suffix (path, desktop_file_suffix)) {
		return;
	}
	
	if (g_file_test (path, (G_FILE_TEST_EXISTS)) != TRUE) {
		FILE *new_file;
		
		new_file = fopen (path, "w");
		fclose (new_file);
	}
	
	desktop_file = g_new0 (KibaDesktopEditorFile, 1);
	desktop_file->path = g_strdup(path);
	
	file = g_key_file_new ();
	if (g_key_file_load_from_file (file, path, 0, &error)) {
		desktop_file->name = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Name",
				NULL, &error);
		g_clear_error (&error);
		error = NULL;
		
		desktop_file->description = g_key_file_get_locale_string (file,
				desktop_entry_group,
				"Comment",
				NULL, &error);
		g_clear_error (&error);
		error = NULL;
		
		desktop_file->exec = g_key_file_get_value (file,
				desktop_entry_group,
				"Exec", &error);
		g_clear_error (&error);
		error = NULL;
		
		desktop_file->group = g_key_file_get_value (file,
				desktop_entry_group,
				"LauncherGroup", &error);
		g_clear_error (&error);
		error = NULL;
		
		desktop_file->icon = g_key_file_get_value (file,
				desktop_entry_group,
				"Icon", &error);
		g_clear_error (&error);
		error = NULL;
		
		g_key_file_free (file);
	}
	else {
		g_clear_error (&error);
		
		desktop_file->name = NULL;
		desktop_file->description = NULL;
		desktop_file->exec = NULL;
		desktop_file->group = NULL;
		desktop_file->icon = NULL;
	}
	
	editor->desktop_files = g_list_append (editor->desktop_files, desktop_file);
}

static void add_files_from_directory (KibaDesktopEditor *editor,
			              const char        *desktop_path) {
	GtkWidget *label;
	GDir *app_dir;
	char *filename;
	char *markup;
	GError *error = NULL;
	gint i;
	
	app_dir = g_dir_open (desktop_path, 0, &error);
	g_clear_error (&error);
	error = NULL;
	
	if (!app_dir)
		return;
	
	while ((filename = (char*) g_dir_read_name (app_dir)) != NULL) {
		char *path;
		
		path = g_strdup_printf ("%s/%s", desktop_path, filename);
		add_file_from_path (editor, path);
		g_free (path);
	}
	g_dir_close (app_dir);
}

/**
Class Functions
**/

static void kiba_desktop_editor_class_init (KibaDesktopEditorClass *class) {
	kiba_desktop_editor_signals[SIG_KIBA_DESKTOP_FILE_SAVED] =
		g_signal_new ("file-saved",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDesktopEditorClass, file_saved),
			NULL, NULL,
			g_cclosure_marshal_VOID__STRING,
			G_TYPE_NONE, 1,
			G_TYPE_STRING);
}

static void kiba_desktop_editor_init (KibaDesktopEditor *editor) {
	editor->desktop_files = NULL;
	editor->win = NULL;
}

/**
 * kiba_desktop_editor_new:
 * @app_path: The .desktop file(s) path, can be a directory or a single file
 *
 * Creates and returns a KibaDesktopEditor Instance.
 *
 * Returns: The KibaDesktopEditor instance created.
 */
KibaDesktopEditor* kiba_desktop_editor_new (const char *app_path) {
	KibaDesktopEditor *editor;
	KibaDesktopEditorFile *desktop_file;
	GtkWidget *box, *mainbox;
	GtkWidget *description_box;
	GtkWidget *label, *button;
	GList *l;
	char *markup;
	gint i;
	gint size;
	
	editor = g_object_new (KIBA_TYPE_DESKTOP_EDITOR, NULL);
	editor->path = g_strdup (app_path);
	
	editor->win = kiba_info_win_new();
	gtk_window_set_skip_taskbar_hint(GTK_WINDOW(editor->win), TRUE);
	gtk_window_set_role (GTK_WINDOW(editor->win), "kiba-desktop-chooser");
	gtk_window_set_position (GTK_WINDOW(editor->win), GTK_WIN_POS_CENTER);
	
	mainbox = gtk_vbox_new (FALSE, 0);
	
	if (g_file_test(app_path, G_FILE_TEST_IS_DIR)) {
		add_files_from_directory (editor, app_path);
		/* create a tree model containing all desktop files */
		box = gtk_hbox_new (FALSE, 5);
			label = gtk_label_new(NULL);
			markup = g_markup_printf_escaped (
					"<span weight=\"bold\">%s</span>",
					_("Desktop file"));
			gtk_label_set_markup (GTK_LABEL (label), markup);
			g_free (markup);
			gtk_box_pack_start (GTK_BOX (box),
				label, FALSE, FALSE, 5);
			gtk_widget_show(label);
			
			editor->combobox = gtk_combo_box_new_text ();
			for (l = editor->desktop_files; l != NULL; l = l->next) {
				desktop_file = l->data;
				
				gtk_combo_box_append_text (GTK_COMBO_BOX(editor->combobox),
					g_strrstr (desktop_file->path, "/")+1);
			}
			gtk_combo_box_set_active(GTK_COMBO_BOX(editor->combobox), 0);
			gtk_widget_set_usize (editor->combobox, 50, -1);
			gtk_box_pack_start (GTK_BOX (box),
				editor->combobox, TRUE, TRUE, 5);
			gtk_widget_show (editor->combobox);
			gtk_box_pack_start (GTK_BOX (mainbox),
				box, FALSE, FALSE, 5);
			
			g_signal_connect (G_OBJECT(editor->combobox), "changed",
				G_CALLBACK(kiba_desktop_editor_combobox_changed_cb), editor);
		gtk_widget_show (box);
	}
	else {
		add_file_from_path (editor, editor->path);
	}
	
	editor->image = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE, GTK_ICON_SIZE_DIALOG);
	gtk_box_pack_start (GTK_BOX (mainbox),
		editor->image, FALSE, FALSE, 5);
	gtk_widget_show (editor->image);
	
	size = 0;
	description_box = gtk_hbox_new (FALSE, 5);
		GtkRequisition req;
		
		label = gtk_label_new(_("Description"));
		gtk_box_pack_start (GTK_BOX (description_box),
			label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		
		gtk_widget_size_request (label, &req);
		size = req.width;
		
		editor->description_entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX (description_box),
			editor->description_entry, TRUE, TRUE, 5);
		gtk_widget_show(editor->description_entry);
// 	gtk_widget_show (box);
	
	box = gtk_hbox_new (FALSE, 5);
		label = gtk_label_new(_("Name"));
		gtk_widget_set_usize (label, size, -1);
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_box_pack_start (GTK_BOX (box),
			label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		
		editor->name_entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX (box),
			editor->name_entry, TRUE, TRUE, 5);
		gtk_widget_show(editor->name_entry);
		
		gtk_box_pack_start (GTK_BOX (mainbox),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	box = gtk_hbox_new (FALSE, 5);
		label = gtk_label_new(_("Exec"));
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_widget_set_usize (label, size, -1);
		gtk_box_pack_start (GTK_BOX (box),
			label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		
		editor->exec_entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX (box),
			editor->exec_entry, TRUE, TRUE, 5);
		gtk_widget_show(editor->exec_entry);
		
		gtk_box_pack_start (GTK_BOX (mainbox),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	box = gtk_hbox_new (FALSE, 5);
		label = gtk_label_new(_("Group"));
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_widget_set_usize (label, size, -1);
		gtk_box_pack_start (GTK_BOX (box),
			label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		
		editor->group_entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX (box),
			editor->group_entry, TRUE, TRUE, 5);
		gtk_widget_show(editor->group_entry);
		
		gtk_box_pack_start (GTK_BOX (mainbox),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	gtk_box_pack_start (GTK_BOX (mainbox),
		description_box, FALSE, FALSE, 5);
	gtk_widget_show (description_box);
	
	box = gtk_hbox_new (FALSE, 5);
		GtkWidget *preview;
		
		label = gtk_label_new(_("Icon"));
		gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
		gtk_widget_set_usize (label, size, -1);
		gtk_box_pack_start (GTK_BOX (box),
			label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		
		editor->icon_chooser = gtk_file_chooser_button_new (_("Choose an icon"),
						GTK_FILE_CHOOSER_ACTION_OPEN);
		gtk_box_pack_start (GTK_BOX (box),
			editor->icon_chooser, TRUE, TRUE, 5);
		gtk_widget_show(editor->icon_chooser);
		g_signal_connect (G_OBJECT(editor->icon_chooser), "selection-changed",
			G_CALLBACK(icon_chooser_changed_cb), editor);
		
		preview = gtk_image_new ();
		gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER(editor->icon_chooser), preview);
		g_signal_connect (GTK_FILE_CHOOSER(editor->icon_chooser), "update-preview",
				G_CALLBACK(update_preview_cb), preview);
		
		gtk_box_pack_start (GTK_BOX (mainbox),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	box = gtk_hbox_new (TRUE, 5);
		button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
		gtk_box_pack_start (GTK_BOX (box),
			button, TRUE, TRUE, 5);
		gtk_widget_show(button);
		g_signal_connect (G_OBJECT(button), "button-press-event",
			G_CALLBACK(quit_button_press_cb), editor->win);
		
		button = gtk_button_new_from_stock(GTK_STOCK_SAVE);
		gtk_box_pack_start (GTK_BOX (box),
			button, TRUE, TRUE, 5);
		gtk_widget_show(button);
		g_signal_connect (G_OBJECT(button), "button-press-event",
			G_CALLBACK(save_button_press_cb), editor);
		
		gtk_box_pack_start (GTK_BOX (mainbox),
			box, FALSE, FALSE, 5);
	gtk_widget_show (box);
	
	gtk_container_add (GTK_CONTAINER(editor->win), mainbox);
	gtk_widget_show (mainbox);
	
	desktop_file = editor->desktop_files->data;
	kiba_desktop_editor_activate_file (editor, desktop_file);
	
	g_signal_connect (G_OBJECT(editor->win), "destroy",
		G_CALLBACK(kiba_desktop_editor_destroy_cb), editor);
	
	return editor;
}
