/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-desktop-icon
 * @short_description: Class representing a .desktop file.
 * @see_also:#KibaDesktopEditor
 * @stability: Unstable
 *
 * Class representing a .desktop file.
 */

#include <string.h>
#include "kiba.h"
#include "kiba-win.h"
#include "kiba-debug.h"
#include "kiba-key-file.h"
#include "kiba-object.h"
#include "kiba-inline-pixbufs.h"
#include "kiba-svg-macro.h"
#include "kiba-desktop-icon.h"
#ifdef USE_SVG
  #include <librsvg/rsvg.h>
#endif

typedef enum {
	KIBA_DESKTOP_TYPE_ENTRY = 0,
	KIBA_DESKTOP_NAME_ENTRY,
	KIBA_DESKTOP_COMMENT_ENTRY,
	KIBA_DESKTOP_EXEC_ENTRY,
	KIBA_DESKTOP_URL_ENTRY,
	KIBA_DESKTOP_PATH_ENTRY,
	KIBA_DESKTOP_ICON_ENTRY,
	KIBA_DESKTOP_GROUP_ENTRY,
	KIBA_DESKTOP_CLASS_ENTRY
}KibaDesktopKey;

enum {
	ACTIVATE_EVENT = 0,
	DEACTIVATE_EVENT,
	FILE_SET_EVENT,
	OBJECT_SET_EVENT,
	SIG_DESKTOP_ICON_LAST
};

static guint desktop_icon_signals[SIG_DESKTOP_ICON_LAST] = { 0 };

G_DEFINE_TYPE (KibaDesktopIcon, kiba_desktop_icon, G_TYPE_OBJECT)

/**
Helper Functions
**/

static void kiba_desktop_icon_draw_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaDesktopIcon *icon = data;
	gint size;
	GdkPixbuf *pixbuf = NULL;
#ifdef USE_SVG
	RsvgHandle *handle = NULL;
#endif
	
	size = MAX(width,height);
	
#ifdef USE_SVG
	if (kiba_load_svg (icon->entry.icon, &size, &handle)) {
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
		return;
	}
#endif
	if (kiba_load_icon (icon->entry.icon, size, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}
	
	pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
	g_object_unref (pixbuf);
}

void kiba_desktop_icon_update_surface (KibaDesktopIcon *icon) {
	KibaWin *default_win;
	KibaObject *object;
	KibaDrawableObject *drawable_object;
	cairo_t *cr;
	
	object =  KIBA_OBJECT(icon->object);
	drawable_object = KIBA_DRAWABLE_OBJECT(object);
	
	icon->surface->width  = KIBA_OBJECT(icon->object)->size.width;
	icon->surface->height = KIBA_OBJECT(icon->object)->size.height;
	
	default_win = kiba_get_win(kiba_get());
	cr = kiba_win_cairo_create (default_win);
	cairo_update_surface (cr, icon->surface,
			kiba_desktop_icon_draw_surface, icon);
	cairo_destroy (cr);
	
	if (kiba_desktop_icon_get_state(icon) == KIBA_DESKTOP_STATE_ACTIVE) {
		drawable_object->surface_data->surface = cairo_surface_reference (icon->surface->surface);
		kiba_drawable_object_queue_draw (drawable_object);
	}
}

static KibaDesktopIconType kiba_desktop_icon_get_type_from_string (const char *string) {
	static const char *icon_types [] = {
		"Application", "Link", "Directory", "Reminder"
	};
	
	gint i;
	
	g_return_val_if_fail (string != NULL, KIBA_DESKTOP_TYPE_INVALID);
	
	for (i = 0; i < G_N_ELEMENTS(icon_types); i++) {
		if (strcmp (string, icon_types[i]) == 0)
			return i;
	}
	
	return KIBA_DESKTOP_TYPE_INVALID;
}

static gboolean kiba_desktop_icon_load_file (KibaDesktopIcon *icon,
					     const char      *uri,
					     GError         **error) {
	static const char file_uri_prefix[] = "file://";
	static const char desktop_entry_group[] = "Desktop Entry";
	static const char desktop_file_suffix[] = ".desktop";
	static const struct { const char *string; KibaDesktopKey key; size_t offset; } desktop_data[] = {
		{ G_KEY_FILE_DESKTOP_KEY_TYPE, KIBA_DESKTOP_TYPE_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.type) },
		{ G_KEY_FILE_DESKTOP_KEY_NAME, KIBA_DESKTOP_NAME_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.name) },
		{ G_KEY_FILE_DESKTOP_KEY_COMMENT, KIBA_DESKTOP_COMMENT_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.comment) },
		{ G_KEY_FILE_DESKTOP_KEY_EXEC, KIBA_DESKTOP_EXEC_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_URL, KIBA_DESKTOP_URL_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_PATH, KIBA_DESKTOP_PATH_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.exec) },
		{ G_KEY_FILE_DESKTOP_KEY_ICON, KIBA_DESKTOP_ICON_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.icon) },
		{ "LauncherGroup", KIBA_DESKTOP_GROUP_ENTRY, G_STRUCT_OFFSET (KibaDesktopIcon, entry.group) }
	};
	static const char *desktop_keys[G_N_ELEMENTS (desktop_data)];
	
	GKeyFile *key_file;
	char *filename = NULL;
	char *s;
	int i;
	
	for (i = 0; i < G_N_ELEMENTS (desktop_keys); i++) {
		desktop_keys[i] = desktop_data[i].string;
	}
	
	if (!g_file_test(uri, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_SYMLINK)) {
		DEBUG_ERROR ("'%s' is not a valid File\n", uri);
		return FALSE;
	}
	if (!g_str_has_suffix(uri, desktop_file_suffix)) {
		DEBUG_ERROR ("'%s' is not a .desktop File\n", uri);
		return FALSE;
	}
	
	if (g_file_test(uri, G_FILE_TEST_IS_SYMLINK)) {
		filename = g_file_read_link (uri, error);
		if (filename == NULL)
			return FALSE;
	}
	else if (g_str_has_prefix (uri, file_uri_prefix)) {
		filename = g_strdup(uri + strlen (file_uri_prefix));
	}
	else {
		filename = g_strdup(uri);
	}
	
	if (!kiba_key_file_add_missing_desktop_entries (filename,
					desktop_entry_group,
					desktop_keys,
					G_N_ELEMENTS (desktop_keys))) {
		g_free (filename);
		return FALSE;
	}
	
	key_file = g_key_file_new ();
	if (!g_key_file_load_from_file (key_file, filename, G_KEY_FILE_KEEP_COMMENTS, error)) {
		g_free (filename);
		return FALSE;
	}
	
	for (i = 0; i < G_N_ELEMENTS (desktop_data); i++) {
		/* FIXME: if there are unescaped percentage characters some long numbers are used instead,
		          percentage characters must be escaped like this: %% */
		s = g_key_file_get_string (key_file, desktop_entry_group, desktop_data[i].string, error);
		g_clear_error (error);
		*error = NULL;
		
		if (!s)
			continue;
		
		switch (desktop_data[i].key) {
			case KIBA_DESKTOP_TYPE_ENTRY:
				G_STRUCT_MEMBER (int , icon, desktop_data[i].offset) =
					kiba_desktop_icon_get_type_from_string (s);
				break;
			case KIBA_DESKTOP_NAME_ENTRY:
			case KIBA_DESKTOP_COMMENT_ENTRY: {
				char *translated;
				
				translated = g_key_file_get_locale_string (key_file,
							desktop_entry_group,
							desktop_data[i].string,
							NULL,
							error);
				g_clear_error (error);
				*error = NULL;
				
				if (translated) {
					g_free (s);
					G_STRUCT_MEMBER (char *, icon, desktop_data[i].offset) = translated;
				}
				else
					G_STRUCT_MEMBER (char *, icon, desktop_data[i].offset) = s;
				break;
			}
			case KIBA_DESKTOP_EXEC_ENTRY:
			case KIBA_DESKTOP_ICON_ENTRY:
			case KIBA_DESKTOP_GROUP_ENTRY:
				G_STRUCT_MEMBER (char *, icon, desktop_data[i].offset) = s;
				break;
			case KIBA_DESKTOP_URL_ENTRY:
			case KIBA_DESKTOP_PATH_ENTRY: {
				char *command;
				
				if (G_STRUCT_MEMBER (char *, icon, desktop_data[i].offset))
					break;
				
				command = g_strdup_printf ("nautilus --no-desktop %s", s);
				g_free (s);
				G_STRUCT_MEMBER (char *, icon, desktop_data[i].offset) = command;
				break;
			}
		}
	}
	g_key_file_free (key_file);
	
	if (icon->entry.uri)
		g_free (icon->entry.uri);
	icon->entry.uri = filename;
	
	return TRUE;
}

/**
Properties
**/

void kiba_desktop_icon_set_state (KibaDesktopIcon      *icon,
				  KibaDesktopIconState  state) {
	g_return_if_fail (KIBA_IS_DESKTOP_ICON(icon));
	
	if (state == icon->state)
		return;
	
	icon->state = state;
	
	if (state == KIBA_DESKTOP_STATE_ACTIVE) {
		if (icon->object) {
			if (icon->entry.uri) {
				kiba_desktop_icon_update_surface (icon);
				kiba_object_set_title (KIBA_OBJECT(icon->object), icon->entry.name);
			}
		}
		
		g_signal_emit (icon, desktop_icon_signals[ACTIVATE_EVENT], 0);
	}
	else
		g_signal_emit (icon, desktop_icon_signals[DEACTIVATE_EVENT], 0);
}

KibaDesktopIconState kiba_desktop_icon_get_state (KibaDesktopIcon *icon) {
	g_return_val_if_fail (KIBA_IS_DESKTOP_ICON(icon), KIBA_DESKTOP_STATE_INACTIVE);
	
	return icon->state;
}

gboolean kiba_desktop_icon_set_file (KibaDesktopIcon *icon,
				     const char      *file) {
	GError *error = NULL;
	
	g_return_if_fail (KIBA_IS_DESKTOP_ICON(icon));
	g_return_if_fail (file != NULL);
	
	if (icon->entry.name)
		g_free (icon->entry.name);
	if (icon->entry.comment)
		g_free (icon->entry.comment);
	if (icon->entry.exec)
		g_free (icon->entry.exec);
	if (icon->entry.icon)
		g_free (icon->entry.icon);
	if (icon->entry.group) {
		g_free (icon->entry.group);
		icon->entry.group = NULL;
	}
	
	if (!kiba_desktop_icon_load_file(icon, file, &error)) {
		DEBUG_ERROR ("Failed to parse File '%s'\n", file);
		if (error != NULL) {
			DEBUG_ERROR ("%s\n", error->message);
			g_free (error);
		}
		return FALSE;
	}
	
	if (!icon->entry.group)
		icon->entry.group = g_strdup("Default");
	
	if (icon->object) {
		kiba_desktop_icon_update_surface (icon);
		kiba_object_set_title (KIBA_OBJECT(icon->object), icon->entry.name);
	}
	
	g_signal_emit (icon, desktop_icon_signals[FILE_SET_EVENT], 0, file);
	
	return TRUE;
}

char* kiba_desktop_icon_get_file (KibaDesktopIcon *icon) {
	g_return_val_if_fail (KIBA_IS_DESKTOP_ICON(icon), NULL);
	
	return icon->entry.uri;
}

void kiba_desktop_icon_set_icon (KibaDesktopIcon *icon,
				 const char      *file) {
	g_return_if_fail (KIBA_IS_DESKTOP_ICON(icon));
	g_return_if_fail (file != NULL);
	
	if (icon->entry.icon) {
		if (strcmp (icon->entry.icon, file) == 0)
			return;
		g_free (icon->entry.icon);
	}
	icon->entry.icon = g_strdup (file);
	
	kiba_desktop_icon_update_surface (icon);
}

void kiba_desktop_icon_set_object (KibaDesktopIcon *icon,
				   GtkWidget       *widget) {
	GError *error = NULL;
	
	g_return_if_fail (KIBA_IS_DESKTOP_ICON(icon));
	g_return_if_fail (GTK_IS_WIDGET(widget));
	
	icon->object = widget;
	
	icon->surface->width  = KIBA_OBJECT(icon->object)->size.width;
	icon->surface->height = KIBA_OBJECT(icon->object)->size.height;
	
	if (icon->entry.uri) {
		if (kiba_desktop_icon_get_state(icon) == KIBA_DESKTOP_STATE_ACTIVE)
			kiba_object_set_title (KIBA_OBJECT(icon->object), icon->entry.name);
		if (icon->surface->width != 1 && icon->surface->height != 1)
			kiba_desktop_icon_update_surface (icon);
	}
	
	g_signal_emit (icon, desktop_icon_signals[OBJECT_SET_EVENT], 0, widget);
}

GtkWidget* kiba_desktop_icon_get_object (KibaDesktopIcon *icon) {
	g_return_val_if_fail (KIBA_IS_DESKTOP_ICON(icon), NULL);
	
	return icon->object;
}

/**
Class Functions
**/

static void kiba_desktop_icon_class_init (KibaDesktopIconClass *class) {
	desktop_icon_signals[ACTIVATE_EVENT] =
		g_signal_new ("activate-event",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaDesktopIconClass, activate),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	
	desktop_icon_signals[DEACTIVATE_EVENT] =
		g_signal_new ("deactivate-event",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaDesktopIconClass, deactivate),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	
	desktop_icon_signals[FILE_SET_EVENT] =
		g_signal_new ("file-set",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaDesktopIconClass, file_set),
				NULL, NULL,
				g_cclosure_marshal_VOID__STRING,
				G_TYPE_NONE, 1,
				G_TYPE_CHAR);
	
	desktop_icon_signals[OBJECT_SET_EVENT] =
		g_signal_new ("object-set",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaDesktopIconClass, object_set),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				GTK_TYPE_WIDGET);
}

static void kiba_desktop_icon_init (KibaDesktopIcon *icon) {
	icon->state = KIBA_DESKTOP_STATE_INACTIVE;
	
	icon->object = NULL;
	
	icon->entry.type    = KIBA_DESKTOP_TYPE_INVALID;
	icon->entry.uri     = NULL;
	icon->entry.name    = NULL;
	icon->entry.comment = NULL;
	icon->entry.exec    = NULL;
	icon->entry.icon    = NULL;
	icon->entry.group   = NULL;
	
	icon->surface = g_new0 (CairoSurfaceData, 1);
	icon->surface->surface = NULL;
	icon->surface->width   = -1;
	icon->surface->height  = -1;
}

KibaDesktopIcon* kiba_desktop_icon_new (void) {
	KibaDesktopIcon *icon;
	
	icon = g_object_new (KIBA_TYPE_DESKTOP_ICON, NULL);
	
	return icon;
}

void kiba_desktop_icon_destroy (KibaDesktopIcon *icon) {
	GList *l;
	
	if (icon->entry.uri)
		g_free (icon->entry.uri);
	if (icon->entry.name)
		g_free (icon->entry.name);
	if (icon->entry.comment)
		g_free (icon->entry.comment);
	if (icon->entry.exec)
		g_free (icon->entry.exec);
	if (icon->entry.icon)
		g_free (icon->entry.icon);
	if (icon->entry.group)
		g_free (icon->entry.group);
	
	cairo_surface_destroy (icon->surface->surface);
	g_free (icon->surface);
	
	g_object_unref (icon);
}
