/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-dialog
 * @short_description: Dialog routines.
 * @see_also:#KibaDesktopEditor
 * @stability: Unstable
 *
 * Dialog routines.
 */

#include <gtk/gtk.h>
#include <libintl.h>

#include "../config.h"
#include "kiba-dialog.h"
#include "kiba-debug.h"

#define _(string)  gettext (string)

/**
About Dialog
**/

void kiba_dialog_about (void) {
	static GtkWidget *about = NULL;
	char *authors [] = {
		"Daniel Bessler",
		"Kristian Høgsberg",
		"Christian Bodenstein",
		"Torbjörn Svangård",
		NULL
	};
	char *documenters[] = {
		NULL
	  };
	int   i;

	if (about) {
		gtk_window_present (GTK_WINDOW (about));
		return;
	}
	
	for (i = 0; authors [i]; i++)
		authors [i] = _(authors [i]);
	
	about = gtk_about_dialog_new ();
	g_object_set (about,
		      "name",  "Kiba-Dock",
		      "version", VERSION,
		      "copyright", "Copyright (c) 2006-2008 Daniel Bessler.",
		      "comments", _("This program is responsible for launching other "
				    "applications and embedding Plugins within itself."),
		      "authors", authors,
		      "documenters", documenters,
		      "translator-credits", _("translator-credits"),
		      "logo-icon-name", "kiba-dock",
		      NULL);
	
	gtk_window_set_wmclass (GTK_WINDOW (about), "about_dialog", "Panel");
	gtk_window_set_screen (GTK_WINDOW (about),
			       gdk_screen_get_default ());
	g_signal_connect (about, "destroy",
			  G_CALLBACK (gtk_widget_destroyed),
			  &about);
	
	g_signal_connect (about, "response",
			  G_CALLBACK (gtk_widget_destroy),
			  NULL);

	gtk_widget_show (about);
}

/**
Tips on Startup Dialog
**/

enum {
	TIPS_RESPONSE_NEXT = 72,
	TIPS_RESPONSE_PREV = 73
};

void kiba_dialog_tips () {
	const char *tips[] = {
		"TIP1",
		"TIP2"
	};
	
	GtkWidget *dialog;
	int response;
	int tip_number = 0;
	
	dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL,
					GTK_MESSAGE_INFO,
					GTK_BUTTONS_NONE, _("<b>Tip of the Day...</b>\n"));
	
	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
			_("Previous"), TIPS_RESPONSE_PREV,
			_("Next"), TIPS_RESPONSE_NEXT,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
	
	gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog), tips[tip_number]);
	
	gtk_widget_show (dialog);
	
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);
	
	while ((response = gtk_dialog_run (GTK_DIALOG (dialog)) > GTK_RESPONSE_CANCEL)) {
		switch (response) {
			case TIPS_RESPONSE_PREV:
				if (tip_number > 0)
					tip_number--;
				else
					tip_number=G_N_ELEMENTS (tips) - 1;
				gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog),
					tips[tip_number]);
				break;
			case TIPS_RESPONSE_NEXT:
				if (tip_number < G_N_ELEMENTS (tips) - 1)
					tip_number++;
				else
					tip_number=0;
				gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog),
					tips[tip_number]);
				break;
		}
	}
	
	gtk_object_destroy (GTK_OBJECT (dialog));
}

/**
 * kiba_dialog_confirm:
 * @title: The Window title.
 * @format: printf format.
 * @...: ...
 *
 * Display a confirmation dialog.
 */
GtkWidget* kiba_dialog_confirm (const char *title,
			        const char *format,
			        ...) {
	GtkWidget *dialog;
	va_list args;
	char *arg_s = NULL;
	
	dialog = gtk_message_dialog_new (NULL, 0,
					GTK_MESSAGE_QUESTION,
					GTK_BUTTONS_NONE, title);
	
	va_start(args, format);
		arg_s = g_strdup_vprintf(format, args);
	va_end(args);
	
	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
					arg_s);
	
	gtk_widget_realize (dialog);
	
	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_YES, GTK_RESPONSE_YES,
				NULL);
	
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);
	gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);
	gtk_window_set_title (GTK_WINDOW (dialog), "");
	
	gtk_widget_show (dialog);
	
	return dialog;
}

/**
 * kiba_dialog_error:
 * @title: The Window title.
 * @format: printf format.
 * @...: ...
 *
 * Display a error dialog.
 */
void kiba_dialog_error (const char *title,
			const char *format,
			...) {
	GtkWidget *dialog;
	va_list args;
	char *arg_s = NULL;
	
	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_CLOSE, title);
	
	va_start(args, format);
		arg_s = g_strdup_vprintf(format, args);
	va_end(args);
	
	DEBUG_ERROR(arg_s);
	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), arg_s);
	
	gtk_widget_show (dialog);
	g_signal_connect (dialog, "response", G_CALLBACK (gtk_widget_destroy), NULL);
}
