/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

//FIXME: KibaObject looses prelight sometimes.

#include "kiba-win.h"
#include "kiba-dock.h"
#include "dnd-window.h"

static GtkWidget *dnd_window = NULL;

/*** signal handler ***/

static gboolean __expose_event_cb (GtkWidget *widget,
                                   GdkEventExpose *expose,
                                   gpointer data) {
	KibaObject *object = KIBA_OBJECT(data);
	GtkWidget *dnd_widget = GTK_WIDGET(object);
	GtkWidget *dock = dnd_widget->parent;
	cairo_surface_t *surface = NULL;
	float factor_x = 1.2f, factor_y = 1.2f;
	gint window_max_width, window_max_height;
	const KibaLayoutData *d;
	GtkWidget *child;
	cairo_t *cr;
	GList *l;
	
	d = &kiba_layout_data[KIBA_DOCK(dnd_widget->parent)->d.position];
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
			
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	window_max_width  = factor_x*dnd_widget->allocation.width;
	window_max_height = factor_y*dnd_widget->allocation.height;
	
	cr = gdk_cairo_create (expose->window);
	if (!cr) return TRUE;
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		cairo_set_source_rgba (cr, 0, 0, 0, 0);
		cairo_paint (cr);
		
		if (KIBA_IS_DRAWABLE_OBJECT(object)) {
			KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
			if (drawable_object->surface_data && drawable_object->surface_data->surface) {
				cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
				cairo_translate (cr,
					window_max_width/2 - dnd_widget->allocation.width/2,
					window_max_height/2 - dnd_widget->allocation.height/2);
				if (kiba_drawable_object_scale (drawable_object, cr)) {
					if (GET_FLAG (KIBA_DOCK(dock)->flags, KIBA_DOCK_DRAW_SHADOW)) {
						cairo_save (cr);
							kiba_drawable_object_paint_shadow (drawable_object, cr);
						cairo_restore (cr);
					}
					cairo_save(cr);
						if (!drawable_object->paint)
							kiba_drawable_object_paint (drawable_object, cr, NULL);
						else
							drawable_object->paint (drawable_object, cr, drawable_object->paint_data);
					cairo_restore (cr);
				}
			}
		} else {
			//TODO get pixmap from widget or so
		}
	cairo_destroy (cr);
	
	child = gtk_bin_get_child(GTK_BIN(widget));
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return TRUE;
}

static void __destroy_cb (GtkWidget *widget,
                          gpointer data) {
	dnd_window = NULL;
}

/*** public methods ***/

GtkWidget* kiba_dnd_window_new (KibaObject *object) {
	GdkScreen *screen = gdk_screen_get_default ();
	gint height, width;
	float factor_x = 1.2f, factor_y = 1.2f;
	GList *l;
	
	g_return_val_if_fail (KIBA_IS_OBJECT(object), dnd_window);
	g_return_val_if_fail (dnd_window == NULL, dnd_window);
	
	dnd_window = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_window_set_decorated (GTK_WINDOW(dnd_window), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(dnd_window), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(dnd_window), TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(dnd_window), GDK_WINDOW_TYPE_HINT_DND);
	
	gtk_widget_set_colormap (dnd_window, gdk_screen_get_rgba_colormap (screen));
	
	width = GTK_WIDGET(object)->allocation.width;
	height = GTK_WIDGET(object)->allocation.height;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->max_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
			
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	width *= factor_x;
	height *= factor_y;
	
	gtk_widget_set_usize (dnd_window,
		width,
		height);
	
	g_signal_connect (G_OBJECT(dnd_window), "expose-event",
		G_CALLBACK(__expose_event_cb), object);
	g_signal_connect (G_OBJECT(dnd_window), "destroy",
		G_CALLBACK(__destroy_cb), object);
	
	gtk_window_present (GTK_WINDOW(dnd_window));
	
	return dnd_window;
}
