/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <gtk/gtk.h>
#include <libintl.h>

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-separator.h"
#include "kiba-dialog.h"
#include "kiba-utils.h"
#include "kiba-menu-items.h"
#include "kiba-prefs.h"

#include "prefs-dbus.h"

#define _(string)  gettext (string)

static void __quit_cb (GtkWidget *widget,
                       gpointer data) {
	Kiba *kiba = kiba_get();
	kiba_quit(kiba);
}

static void __about_cb (GtkWidget *widget,
                        gpointer data) {
	kiba_dialog_about();
}

static void __settings_editor_cb (GtkWidget *widget,
                                  gpointer data) {
	KibaDock *dock = data;
	KibaPrefs *prefs;
	KibaPrefsFile *configFile;
	char *command;
	
	prefs = kiba_prefs_get();
	
	configFile = kiba_get_dock_config_file(kiba_get(), dock);
	command = g_strdup_printf ("kiba-settings --config-file %s\n", kiba_prefs_get_file_path(configFile));
	kiba_launch_programm(command);
	g_free (command);
}

static void __toggle_auto_hide_cb (GtkWidget *widget,
                                   gpointer data) {
	KibaDock *dock = data;
	const char *id = "auto_hide_enable";
	const char *type = "bool";
	const char *value;
	KibaPrefsFile *configFile;
	
	if (GET_FLAG(dock->flags, KIBA_DOCK_AUTOHIDE))
		value = "FALSE";
	else
		value = "TRUE";
	
	configFile = kiba_get_dock_config_file(kiba_get(), dock);
	kiba_prefs_write_to_config_file (configFile, id, type, value);
	kiba_prefs_set_option (configFile, id, type, value);
	prefs_dbus_set_bool (kiba_prefs_get_file_path(configFile), id, (strcmp(value, "TRUE") == 0));
}

static void __add_separator_cb (GtkWidget *widget,
                                gpointer data) {
	KibaDock *dock = data;
	GtkWidget *separator;
	GList *l;
	gint i = 1;
	
	for (l = gtk_container_get_children(GTK_CONTAINER(dock)); l != NULL; l = l->next) {
		if (KIBA_IS_SEPARATOR(l->data))
			i++;
	}
	
	char *group = g_strdup_printf("separator%i", i);
	separator = kiba_separator_new (group);
	gtk_container_add(GTK_CONTAINER(dock), separator);
	g_free(group);
	gtk_widget_show (separator);
}

static void __populate_dock_cb (GtkWidget *widget,
                                gpointer user_data) {
	KibaDockPopulateData *data = user_data;
	
	data->func(data->dock, data->user_data);
}

gboolean kiba_dock_show_menu (GtkWidget      *widget,
			      GdkEventButton *event) {
	KibaDock *dock  = KIBA_DOCK(widget);
	GtkMenu *menu;
	GtkWidget *menu_item, *image;
	GtkWidget *populate_menu;
	GdkPixbuf *pixbuf = NULL;
	GList *l;
	
	if (event->button != 3) {
		return FALSE;
	}
	
	menu = GTK_MENU (gtk_menu_new ());
	
	menu_item = gtk_image_menu_item_new_with_label (_("Kiba Settings"));
	gtk_tooltips_set_tip (kiba_get_tooltips(), menu_item, _("Kiba setting editor"), NULL);
	image = gtk_image_new_from_icon_name ("kiba-dock", GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate", G_CALLBACK (__settings_editor_cb), dock);
	gtk_widget_show_all (GTK_WIDGET (menu));
	
	menu_item = gtk_check_menu_item_new_with_label (_("Autohide"));
	gtk_tooltips_set_tip (kiba_get_tooltips(), menu_item, _("Toggle Autohide"), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	if (GET_FLAG(dock->flags, KIBA_DOCK_AUTOHIDE))
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menu_item), TRUE);
	else
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (menu_item), FALSE);
	g_signal_connect (G_OBJECT (menu_item), "activate", G_CALLBACK (__toggle_auto_hide_cb), dock);
	
	kiba_menu_add_separator (GTK_WIDGET (menu));
	
	populate_menu = gtk_menu_new ();
	
	menu_item = gtk_image_menu_item_new_with_label (_("Populate Dock"));
	image = gtk_image_new_from_icon_name (GTK_STOCK_ADD, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), GTK_WIDGET(populate_menu));
	gtk_widget_show (menu_item);
	
	for (l = dock->populate_data; l != NULL; l = l->next) {
		KibaDockPopulateData *data = l->data;
		
		menu_item = gtk_image_menu_item_new_with_label (data->title);
		kiba_load_icon (data->icon, 24, &pixbuf);
		image = gtk_image_new_from_pixbuf (pixbuf);
		g_object_unref (pixbuf);
		pixbuf = NULL;
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (populate_menu), menu_item);
		g_signal_connect (G_OBJECT (menu_item), "activate",
				G_CALLBACK (__populate_dock_cb), data);
	}
	
	menu_item = gtk_image_menu_item_new_with_label (_("Add Separator"));
	image = gtk_image_new_from_icon_name (GTK_STOCK_ADD, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	gtk_tooltips_set_tip (kiba_get_tooltips(), menu_item,
		_("Add a new Separator Object to the Dock"), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (populate_menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate",
			G_CALLBACK (__add_separator_cb), dock);
	
	kiba_menu_add_separator (GTK_WIDGET (menu));
	
	menu_item = gtk_image_menu_item_new_with_label (_("About"));
	gtk_tooltips_set_tip (kiba_get_tooltips(), menu_item, _("About Kiba-Dock"), NULL);
	image = gtk_image_new_from_icon_name (GTK_STOCK_ABOUT, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate", G_CALLBACK (__about_cb), dock);
	gtk_widget_show_all (GTK_WIDGET (menu));
	
	menu_item = gtk_image_menu_item_new_with_label (_("Quit"));
	gtk_tooltips_set_tip (kiba_get_tooltips(), menu_item, _("Quit Kiba-Dock"), NULL);
	image = gtk_image_new_from_icon_name (GTK_STOCK_QUIT, GTK_ICON_SIZE_MENU);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menu_item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
	g_signal_connect (G_OBJECT (menu_item), "activate", G_CALLBACK (__quit_cb), dock);
	gtk_widget_show_all (GTK_WIDGET (menu));
	
	gtk_menu_popup (menu, NULL, NULL, NULL, NULL, event->button, event->time);
	
	return TRUE;
}
