/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-dock
 * @short_description: Container for #KibaObject's.
 * @see_also: #KibaWin,#KibaObject
 * @stability: Unstable
 *
 * Container for #KibaObject's.
 */

#include "../config.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-drawable-object.h"
#include "kiba-fixed-object.h"
#include "kiba-separator.h"
#include "kiba-info-win.h"
#include "kiba-image.h"
#include "kiba-screen.h"
#include "kiba-x11.h"
#include "kiba-svg-macro.h"
#include "dnd-window.h"
#ifdef USE_SVG
  #include <librsvg/rsvg.h>
#endif
#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#define KIBA_DOCK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_DOCK, KibaDockPrivate))

#define MIN_OBJECT_SPACE   28
#define MIN_DOCK_HEIGHT    30

#define MAX_DRAG_DISTANCE_FOR_EXEC      15
#define MAX_OBJ_DISTANCE_FOR_SWITCHING  10

#define SCROLL_ARROW_SIZE   10
#define SCROLL_STEP         2

#define OBJECT_FADE_IN_STEP    0.075f
#define OBJECT_FADE_OUT_STEP   0.05f

/* rendering the objects as black rectangles */
#define OBJECT_SOLID_RECTANGLES FALSE

G_DEFINE_TYPE (KibaDock, kiba_dock, GTK_TYPE_BOX)

typedef struct {
	char *group;
	KibaGroupSortFunc sort_func;
	gpointer user_data;
}KibaDockGroupData;

typedef struct {
	GdkWindow *event_window; /**<< pointer to the event catcher window **/
	
	gint scroll_counter; /**<< counter for scroll acceleration **/
	gint objects_right_of_dock; /**<< number of objects right of the dock **/
	gint objects_left_of_dock; /**<< number of objects left of the dock **/
	
	GdkRectangle mouse_anchor;
	gint move_object_id;
	
	GtkAllocation drag_allocation;
	GtkWidget *dragging_child; /**<< currently dragged child or NULL **/
	GtkWidget *dragging_window; /**<< pointer to the temporary dnd window **/
	gint press_count; /**<< number of button presses on the dragged object **/
	guint drag_button;
	
	GList *group_data; /**<< list of all added KibaDockGroupData instances **/
	
	GList *drawableObjects; /**<< list of all added KibaDrawableObject instances **/
	
	float fade_alpha; /**<< dock alpha **/
} KibaDockPrivate;


enum {
	SIG_POSITION_CHANGED = 0,
	SIG_ORIENTATION_CHANGED,
	SIG_AUTOHIDE_TOGGLED,
	SIG_OBJECT_SIZE_CHANGED,
	SIG_HEIGHT_CHANGED,
	SIG_DRAG_STARTED,
	SIG_KIBA_DOCK_LAST
};

static guint kiba_dock_signals[SIG_KIBA_DOCK_LAST] = { 0 };

/** Private Methods **/

/*** Helper Methods ***/

/**
 * Returns the allocation of the input catcher window.
 **/
static GtkAllocation __get_dock_allocation (GtkWidget *widget) {
	GtkAllocation event_allocation;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[KIBA_DOCK(widget)->d.position];
	
	event_allocation = widget->allocation;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.height = KIBA_DOCK(widget)->d.height;
	else
		event_allocation.width = KIBA_DOCK(widget)->d.height;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.y += (d->valignment/2)*
			(widget->allocation.height - KIBA_DOCK(widget)->d.height);
	else
		event_allocation.x += (d->halignment/2)*
			(widget->allocation.width - KIBA_DOCK(widget)->d.height);
	
	return event_allocation;
}

static GtkAllocation __get_event_allocation (GtkWidget *widget) {
	GtkAllocation event_allocation;
	const KibaLayoutData *d;
	GList *l;
	gint maxSize = 0;
	
	d = &kiba_layout_data[KIBA_DOCK(widget)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		for (l = gtk_container_get_children(GTK_CONTAINER(widget)); l != NULL; l = l->next) {
			GtkWidget *object = l->data;
			if (object->allocation.height > maxSize)
				maxSize = object->allocation.height;
		}
	} else {
		for (l = gtk_container_get_children(GTK_CONTAINER(widget)); l != NULL; l = l->next) {
			GtkWidget *object = l->data;
			if (object->allocation.width > maxSize)
				maxSize = object->allocation.width;
		}
	}
	maxSize += KIBA_DOCK(widget)->d.object_margin;
	
	event_allocation = widget->allocation;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.height = maxSize;
	else
		event_allocation.width = maxSize;
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		event_allocation.y += (d->valignment/2)*
			(widget->allocation.height - event_allocation.height);
	else
		event_allocation.x += (d->halignment/2)*
			(widget->allocation.width - event_allocation.width);
	
	return event_allocation;
}

static gboolean __group_is_added (KibaDock   *dock,
					  const char *group) {
	GList *l;
	
	g_return_val_if_fail(group != NULL, TRUE);
	
	for (l = dock->groups; l != NULL; l = l->next) {
		char *data = l->data;
		
		if (strcmp (data, group) == 0)
			return TRUE;
	}
	
	return FALSE;
}

static KibaDockGroupData* __get_group_data (KibaDock   *dock,
						    const char *group) {
	KibaDockPrivate *priv;
	GList *l;
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	for (l = priv->group_data; l != NULL; l = l->next) {
		KibaDockGroupData *data = l->data;
		
		if (strcmp (data->group, group) == 0)
			return data;
	}
	
	return NULL;
}

static GList* __add_objects_from_group (KibaDock *dock,
						GList    *sorted,
						GList    *children,
						char     *group) {
	KibaDockGroupData *group_data;
	GList *l, *sorted_group = NULL;
	
	if (!group)
		return NULL;
	
	for (l = children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		gchar *data;
		
		g_object_get (G_OBJECT(child->widget), "group", &data, NULL);
		
		if (data) {
			if (strcmp (data, group) == 0)
				sorted_group = g_list_append (sorted_group, child);
			g_free(data);
		}
	}
	
	/* using custom sort function, if available */
	group_data = __get_group_data(dock, group);
	if (group_data != NULL) {
		GList *ret;
		
		ret = group_data->sort_func(dock, group, sorted_group, group_data->user_data);
		if (g_list_length(ret) == g_list_length(sorted_group))
			sorted_group = ret;
	}
	
	for (l = sorted_group; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		
		sorted = g_list_append (sorted, child);
	}
	
	return sorted;
}

static gint __scroll_counter_increase (gpointer data) {
	KibaDockPrivate *priv = data;
	priv->scroll_counter++;
	return FALSE;
}

static gint __scroll_counter_decrease (gpointer data) {
	KibaDockPrivate *priv = data;
	priv->scroll_counter--;
	return FALSE;
}

#ifdef USE_LIBSN
static void __sn_launch_started (KibaSn  *sn,
                                 GObject *object,
                                 gpointer data) {
	kiba_object_set_state(KIBA_OBJECT(object), KIBA_OBJECT_WORKING);
}

static void __sn_launch_finished (KibaSn  *sn,
                                 gboolean success,
                                 GObject *object,
                                 gpointer data) {
	kiba_object_unset_state(KIBA_OBJECT(object), KIBA_OBJECT_WORKING);
}
#endif

/** default method to place the objects on the dock **/
static void __place_objects (KibaDock *dock,
                             gpointer data) {
	KibaDockPrivate *priv;
	GtkWidget *widget = GTK_WIDGET(dock);
	GList *l;
	GtkAllocation child_allocation;
	const KibaLayoutData *d;
	gint border_width;
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	d = &kiba_layout_data[KIBA_DOCK(widget)->d.position];
	border_width = gtk_container_get_border_width(GTK_CONTAINER(widget));
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		child_allocation.x = dock->d.cap_size + border_width + widget->allocation.x;
		child_allocation.y = widget->allocation.y;
	}
	else {
		child_allocation.x = widget->allocation.x;
		child_allocation.y = dock->d.cap_size + border_width + widget->allocation.y;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		child_allocation.x -= dock->scroll_offset;
	else
		child_allocation.y -= dock->scroll_offset;
	
	for (l = GTK_BOX(widget)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		GtkAllocation object_allocation;
		GtkRequisition child_requisition;
		KibaObject *object = KIBA_OBJECT(child->widget);
		float destroy_factor;
		
		gtk_widget_size_request (child->widget, &child_requisition);
		
		destroy_factor = kiba_object_get_destroy_factor (object);
		object_allocation.width  = child_requisition.width*destroy_factor;
		object_allocation.height = child_requisition.height*destroy_factor;
		object_allocation.x = child_allocation.x;
		object_allocation.y = child_allocation.y;
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			child_allocation.x += object_allocation.width + dock->d.object_space*destroy_factor;
		else
			child_allocation.y += object_allocation.height + dock->d.object_space*destroy_factor;
		
		if (!KIBA_IS_FIXED_OBJECT(object)) {
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				object_allocation.y += (d->valignment/2)*
					(widget->allocation.height - object_allocation.height) -
					(d->valignment-1)*dock->d.object_margin;
			else
				object_allocation.x += (d->halignment/2)*
					(widget->allocation.width - object_allocation.width) -
					(d->halignment-1)*dock->d.object_margin;
		}
		else {
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				object_allocation.y += (d->valignment/2)*
					(widget->allocation.height - dock->d.height + border_width) +
					(dock->d.height - child_requisition.height)/2 - border_width;
			else
				object_allocation.x += (d->halignment/2)*
					(widget->allocation.width - dock->d.height + border_width) +
					(dock->d.height - child_requisition.width)/2 - border_width;
		}
		
		if (destroy_factor != 1.0f) {
			object_allocation.x -= (object_allocation.width*(1.0f-destroy_factor))/2;
			object_allocation.y -= (object_allocation.height*(1.0f-destroy_factor))/2;
		}
		
		object->alloc.x = (float) object_allocation.x;
		object->alloc.y = (float) object_allocation.y;
		object->alloc.width = (float) object_allocation.width;
		object->alloc.height = (float) object_allocation.height;
		
		gtk_widget_size_allocate (child->widget, &object_allocation);
	}
}

/*** Rendering ***/

static void __queue_draw (GtkWidget *widget) {
	KibaDock *dock = KIBA_DOCK(widget);
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GdkRectangle area;
	const KibaLayoutData *d;
	GList *l;
	
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	d = &kiba_layout_data[dock->d.position];
	
//TODO shouldnt we use event area here??
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		area.x = widget->allocation.x;
		area.y = (d->valignment/2)*(widget->allocation.height - dock->d.height) + widget->allocation.y;
		area.width = widget->allocation.width;
		area.height = dock->d.height + 1;
	}
	else {
		area.x = (d->halignment/2)*(widget->allocation.width - dock->d.height) + widget->allocation.x;
		area.y = widget->allocation.y;
		area.width = dock->d.height + 1;
		area.height = widget->allocation.height;
	}
	
	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaDrawableObject *object = l->data;
		kiba_drawable_object_queue_draw(object);
	}
	
	/* putting damaged area in KibaWin's expose queue */
	kiba_win_invalidate_area (widget->parent, area);
}

static void __draw_surface (cairo_t          *cr,
				    cairo_surface_t  *surface,
				    gint              width,
				    gint              height,
				    void             *data) {
	KibaDock *dock = data;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];
	
	if (dock->shape.draw_func == NULL)
		return;
	
	cairo_save (cr);
		/* calling custom shape draw method */
		dock->shape.draw_func (dock, cr, dock->shape.draw_data);
	cairo_restore (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
	
	/* drawing curves */
	if (GET_FLAG (dock->flags, KIBA_DOCK_DRAW_CURVES)) {
		cairo_set_source_rgba (cr,
			dock->style.curve_color.r,
			dock->style.curve_color.g,
			dock->style.curve_color.b,
			dock->style.curve_color.a);
		cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			cairo_move_to (cr, 0, height/2);
			
			if (d->anchor.y == 0) {
				cairo_line_to (cr, 0, 0);
				cairo_line_to (cr, width, 0);
				cairo_line_to (cr, width, height/2);
				cairo_curve_to (cr,
					width*(4.0/5.0), height/16,
					width*(3.0/5.0), height/2,
					0, height/2);
			}
			else {
				cairo_line_to (cr, 0, height);
				cairo_line_to (cr, width, height);
				cairo_line_to (cr, width, height/2);
				cairo_curve_to (cr,
					width*(4.0/5.0), height/4,
					width*(3.0/5.0), 1.3*height,
					0, height/2);
			}
		}
		else {
			cairo_move_to (cr, width/2, 0);
			
			if (d->anchor.x == 0) {
				cairo_line_to (cr, 0, 0);
				cairo_line_to (cr, 0, height);
				cairo_line_to (cr, width/2, height);
				cairo_curve_to (cr,
					width/2, height*(4.0/5.0),
					width/12, height*(3.0/5.0),
					width/2, 0);
			}
			else {
				cairo_line_to (cr, width, 0);
				cairo_line_to (cr, width, height);
				cairo_line_to (cr, width/2, height);
				cairo_curve_to (cr,
					width/4, height*(4.0/5.0),
					width, height*(3.0/5.0),
					width/2, 0);
			}
		}
		cairo_fill (cr);
	}
}

static void __draw_gear_surface (cairo_t *cr,
				    cairo_surface_t  *surface,
				    gint              width,
				    gint              height,
				    void             *data) {
	KibaDock *dock = KIBA_DOCK(data);
	if (dock->continuousProgress.icon == NULL)
		return;
#ifdef USE_SVG
	gint size = dock->continuousProgress.size;
	RsvgHandle *handle = NULL;
	if (kiba_load_svg (dock->continuousProgress.icon, &size, &handle)) {
		rsvg_handle_render_cairo_sub (handle, cr, "#layer1");
		rsvg_handle_free (handle);
	}
#endif
}

static void __draw_gear_surface_overlay (cairo_t *cr,
				    cairo_surface_t  *surface,
				    gint              width,
				    gint              height,
				    void             *data) {
	KibaDock *dock = KIBA_DOCK(data);
	if (dock->continuousProgress.icon == NULL)
		return;
#ifdef USE_SVG
	gint size = dock->continuousProgress.size;
	RsvgHandle *handle = NULL;
	if (kiba_load_svg (dock->continuousProgress.icon, &size, &handle)) {
		rsvg_handle_render_cairo_sub (handle, cr, "#layer2");
		rsvg_handle_free (handle);
	}
#endif
}

static void __update_gear_surface (KibaDock *dock) {
	GtkWidget *widget = GTK_WIDGET(dock);
	KibaWin *win;
	cairo_t *cr;
	
	g_return_if_fail (KIBA_IS_WIN(widget->parent));
	
	win = KIBA_WIN(widget->parent);
	
	cr = kiba_win_cairo_create (win);
	if (!cr)
		return;
	cairo_update_surface_at_size (cr,
		&dock->gearSurface,
		dock->continuousProgress.size, dock->continuousProgress.size,
		__draw_gear_surface,
		dock);
	cairo_update_surface_at_size (cr,
		&dock->gearSurfaceOverlay,
		dock->continuousProgress.size, dock->continuousProgress.size,
		__draw_gear_surface_overlay,
		dock);
	cairo_destroy (cr);
}

/*** KibaObject Stuff ***/

static void __sort_objects (KibaDock *dock) {
	GList *l;
	GList *sorted_objects = NULL;
	KibaDockPrivate *priv;
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	for (l = dock->groups; l != NULL; l = l->next) {
		sorted_objects = __add_objects_from_group (dock,
					sorted_objects,
					GTK_BOX(dock)->children,
					(char*) l->data);
	}
	
	GTK_BOX(dock)->children = sorted_objects;
}

static gboolean __object_is_group_drag_dest (GtkWidget *dock,
						     GtkWidget *drag_dest) {
	GList *l;
	KibaDockPrivate *priv;
	KibaObject *dragging_object;
	const KibaLayoutData *d;
	gint drag_distance;
	gchar *dest_group;
	gchar *dragging_group;
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	dragging_object = KIBA_OBJECT(priv->dragging_child);
	
	g_object_get (G_OBJECT(drag_dest), "group", &dest_group, NULL);
	g_object_get (G_OBJECT(dragging_object), "group", &dragging_group, NULL);
	
	/* skip objects with the same group */
	if (strcmp(dragging_group, dest_group) == 0) {
		g_free(dragging_group);
		g_free(dest_group);
		return FALSE;
	}
	g_free(dragging_group);
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		drag_distance = priv->dragging_child->allocation.x - drag_dest->allocation.x;
	else
		drag_distance = priv->dragging_child->allocation.y - drag_dest->allocation.y;
	drag_distance = MAX(drag_distance, -drag_distance);
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		gint distance;
		gchar *check_group;
		
		g_object_get (G_OBJECT(child->widget), "group", &check_group, NULL);
		
		/* skip objects with the same group */
		if (strcmp (dest_group, check_group) != 0) {
			g_free(check_group);
			continue;
		}
		g_free(check_group);
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			distance = priv->dragging_child->allocation.x - child->widget->allocation.x;
		else
			distance = priv->dragging_child->allocation.y - child->widget->allocation.y;
		distance = MAX(distance, -distance);
		
		if (distance > drag_distance) {
			g_free (dest_group);
			return FALSE;
		}
	}
	g_free (dest_group);
	
	return TRUE;
}

static void __check_switch_objects (GtkWidget *dock) {
	GList *l;
	GtkWidget *win = dock->parent;
	KibaDockPrivate *priv;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object;
		GtkAllocation draggingAllocation;
		gint dx, dy;
		
		if (child->widget == priv->dragging_child)
			continue;
		
		object = KIBA_OBJECT(child->widget);
		
		if (!kiba_object_is_on_dock(object))
			continue;
		
		if (priv->dragging_window) {
			gint x,y;
			gtk_window_get_position(GTK_WINDOW(priv->dragging_window), &x, &y);
			draggingAllocation.x = x - win->allocation.x;
			draggingAllocation.y = y - win->allocation.y;
			draggingAllocation.width = priv->dragging_child->allocation.width;
			draggingAllocation.height = priv->dragging_child->allocation.height;
		}
		else
			draggingAllocation = priv->dragging_child->allocation;
		
		dx = (draggingAllocation.x + draggingAllocation.width/2) -
		     (child->widget->allocation.x + child->widget->allocation.width/2);
		dy = (draggingAllocation.y + draggingAllocation.height/2) -
		     (child->widget->allocation.y + child->widget->allocation.height/2);
		
		/* ignoring the unrelevant orientation for fixed objects */
		if (KIBA_IS_FIXED_OBJECT(priv->dragging_child)) {
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				dy = dx;
			else
				dx = dy;
		}
		
		if (MAX(dx, -dx) > MAX_OBJ_DISTANCE_FOR_SWITCHING || MAX(dy, -dy) > MAX_OBJ_DISTANCE_FOR_SWITCHING)
			continue;
		
		/* switching groups */
		if (__object_is_group_drag_dest(dock, child->widget)) {
			GList *li;
			gpointer *list_data = NULL;
			gpointer *switch_data;
			gchar *dragging_group;
			
			g_object_get (G_OBJECT(priv->dragging_child), "group", &dragging_group, NULL);
			
			for (li = KIBA_DOCK(dock)->groups; li != NULL; li = li->next) {
				char *group = li->data;
				GList *list;
				gchar *object_group;
				
				if (strlen(group) < 4)
					continue;
				
				g_object_get (G_OBJECT(object), "group", &object_group, NULL);
				
				if (strcmp (group, object_group) == 0) {
					list_data = li->data;
					
					for (list = li->next; list != NULL; list = list->next) {
						char *check_group = list->data;
						
						if (strcmp (check_group, dragging_group) == 0) {
							break;
						}
					}
					
					list_data = list->data;
					
					for (list = list->prev; list != NULL; list = list->prev) {
						char *check_group = list->data;
						
						list->next->data = list->data;
						
						if (strcmp (check_group, object_group) == 0) {
							list->data = list_data;
							break;
						}
					}
					
					g_free(object_group);
					break;
				}
				else if (strcmp (group, dragging_group) == 0) {
					list_data = li->data;
					
					for (list = li->next; list != NULL; list = list->next) {
						char *check_group = list->data;
						
						list->prev->data = list->data;
						
						if (strcmp (check_group, object_group) == 0) {
							list->data = list_data;
							break;
						}
					}
					g_free(object_group);
					break;
				}
			}
			g_free(dragging_group);
		}
		__sort_objects(KIBA_DOCK(dock));
		KIBA_DOCK(dock)->allocate_children (KIBA_DOCK(dock), KIBA_DOCK(dock)->allocate_children_data);
		gtk_window_move(GTK_WINDOW(priv->dragging_window),
			priv->dragging_child->allocation.x, priv->dragging_child->allocation.y);
		
		break;
	}
}

static gint __move_object (gpointer data) {
	GtkWidget *dock = data;
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	GtkWidget *win = dock->parent;
	KibaObject *object;
	GtkAllocation child_allocation;
	gint x, y;
	gint dx, dy, drag_distance;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	g_return_val_if_fail (priv->dragging_child != NULL, FALSE);
	
	object = KIBA_OBJECT(priv->dragging_child);
	
	if (kiba_object_get_destroying(object)) {
		if (GTK_IS_WINDOW(priv->dragging_window)) {
			gtk_widget_destroy (priv->dragging_window);
			priv->dragging_window = NULL;
			priv->dragging_child = NULL;
			return FALSE;
		}
		
		priv->dragging_child = NULL;
		
		return FALSE;
	}
	
	child_allocation = priv->dragging_child->allocation;
	
	gdk_window_get_pointer (GTK_WIDGET(dock)->window, &x, &y, NULL);
	if (KIBA_IS_FIXED_OBJECT(object)) {
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			x -= priv->mouse_anchor.x;
		else
			y -= priv->mouse_anchor.y;
	}
	else {
		x -= priv->mouse_anchor.x;
		y -= priv->mouse_anchor.y;
	}
	
	__check_switch_objects (dock);
	
// 	if (!KIBA_IS_FIXED_OBJECT(object)) {
		if (!GTK_IS_WINDOW(priv->dragging_window)) {
			dx = priv->dragging_child->allocation.x - x;
			dy = priv->dragging_child->allocation.y - y;
			drag_distance = sqrt (dx * dx + dy * dy);
			
			if (drag_distance > MAX_DRAG_DISTANCE_FOR_EXEC) {
				priv->dragging_window = kiba_dnd_window_new(object);
				kiba_object_set_state(object, KIBA_OBJECT_DRAGGING);
			}
		}
		if (GTK_IS_WINDOW(priv->dragging_window)) {
			GtkWidget *window = data;
			GList *l;
			float factor_x = 1.2f, factor_y = 1.2f;
			float offsetX = 0.0f, offsetY = 0.0f;
			
			for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
				KibaObjectCustomScaleData *data = l->data;
				
				if (data->scale_widget)
					continue;
				
				float custom_factor_x = 1.0f;
				float custom_factor_y = 1.0f;
				
				data->scale_func (object,
					&custom_factor_x,
					&custom_factor_y,
					data->data);
					
				factor_x *= custom_factor_x;
				factor_y *= custom_factor_y;
			}
			
			for (l = object->custom_offset_data; l != NULL; l = l->next) {
				float customX = 0.0f, customY = 0.0f;
				KibaObjectCustomOffsetData *offset_data = l->data;
				
				offset_data->place_func (object, &customX, &customY, offset_data->data);
				offsetX += customX;
				offsetY += customY;
			}
			
			x -= (factor_x*priv->dragging_child->allocation.width)/2 - priv->dragging_child->allocation.width/2 + offsetX;
			y -= (factor_y*priv->dragging_child->allocation.height)/2 - priv->dragging_child->allocation.height/2 + offsetY;
			
			x += win->allocation.x;
			y += win->allocation.y;
			
			gtk_window_move (GTK_WINDOW(priv->dragging_window), x, y);
			
			/* force the widget to emit the size-allocate signal */
			GtkAllocation alloc;
			alloc = priv->dragging_window->allocation;
			gtk_widget_queue_resize_no_redraw(priv->dragging_window);
			gtk_widget_size_allocate (priv->dragging_window, &alloc);
			
			if (KIBA_IS_DRAWABLE_OBJECT(priv->dragging_child))
				kiba_drawable_object_queue_draw(KIBA_DRAWABLE_OBJECT(priv->dragging_child));
		}
// 	}
	
	return TRUE;
}

static void __check_objects_on_dock (GtkWidget *widget) {
	GList *l;
	KibaDockPrivate *priv;
	const KibaLayoutData *d;
	gint border_width;
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	d = &kiba_layout_data[KIBA_DOCK(widget)->d.position];
	border_width = gtk_container_get_border_width (GTK_CONTAINER(widget));
	
	for (l = GTK_BOX(widget)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObjectPosition position;
		KibaObject *object;
		gint object_position = 0;
		
		object = KIBA_OBJECT(child->widget);
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			object_position = child->widget->allocation.x + child->widget->allocation.width;
		else
			object_position = child->widget->allocation.y + child->widget->allocation.height;
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			if (object_position > widget->allocation.x + widget->allocation.width -
					KIBA_DOCK(widget)->d.cap_size - border_width)
				position = KIBA_OBJECT_RIGHT_OF_DOCK;
			else if ((object_position - child->widget->allocation.width) <
			         (widget->allocation.x + KIBA_DOCK(widget)->d.cap_size + border_width))
				position = KIBA_OBJECT_LEFT_OF_DOCK;
			else
				position = KIBA_OBJECT_ON_DOCK;
		}
		else {
			if (object_position > widget->allocation.y +
			    widget->allocation.height - KIBA_DOCK(widget)->d.cap_size - border_width)
				position = KIBA_OBJECT_RIGHT_OF_DOCK;
			else if (object_position - child->widget->allocation.height <
			         widget->allocation.y + KIBA_DOCK(widget)->d.cap_size + border_width)
				position = KIBA_OBJECT_LEFT_OF_DOCK;
			else
				position = KIBA_OBJECT_ON_DOCK;
		}
		
		if (object->position == position)
			continue;
		
		if (object->position == KIBA_OBJECT_RIGHT_OF_DOCK) {
			priv->objects_right_of_dock--;
			if (!priv->objects_right_of_dock)
				__queue_draw (widget);
		}
		else if (object->position == KIBA_OBJECT_LEFT_OF_DOCK) {
			priv->objects_left_of_dock--;
			if (!priv->objects_left_of_dock)
				__queue_draw (widget);
		}
		if (position == KIBA_OBJECT_RIGHT_OF_DOCK) {
			if (!priv->objects_right_of_dock)
				__queue_draw (widget);
			priv->objects_right_of_dock++;
		}
		else if (position == KIBA_OBJECT_LEFT_OF_DOCK) {
			if (!priv->objects_left_of_dock)
				__queue_draw (widget);
			priv->objects_left_of_dock++;
		}
		object->position = position;
		
		if (position == KIBA_OBJECT_ON_DOCK) {
			gtk_widget_show(child->widget);
		}
		else {
			gtk_widget_hide(child->widget);
		}
	}
}

/*** KibaObject signal handler ***/

static gboolean __object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	GtkWidget *dock = data;
	KibaObject *object;
	KibaDockPrivate *priv;
	gint x, y;
	gint fX, fY;
	
	priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	if (!kiba_object_is_on_dock(KIBA_OBJECT(widget)))
		return TRUE;
	
	if (event->button == 3) {
		if (GTK_IS_WINDOW(priv->dragging_window)) {
			if (event->type == GDK_BUTTON_PRESS)
				priv->press_count++;
			return TRUE;
		}
		return FALSE;
	}
	
	object = KIBA_OBJECT(widget);
	priv->dragging_child = widget;
	
	kiba_object_hide_title (widget);
	
	gdk_window_get_pointer (GTK_WIDGET(dock)->window, &x, &y, NULL);
	
	fX = (int) object->alloc.x;
	fY = (int) object->alloc.y;
	
	priv->mouse_anchor.x = x - fX;
	priv->mouse_anchor.y = y - fY;
	
	if (event->type == GDK_BUTTON_PRESS)
		priv->press_count++;
	if (!priv->drag_button)
		priv->drag_button = event->button;
	
	if (event->button != 1)
		return TRUE;
	
	if (!priv->move_object_id)
		priv->move_object_id = g_timeout_add (20, __move_object, dock);
	
	return TRUE;
}

static gboolean __object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaDock *dock = KIBA_DOCK(data);
	KibaObject *object = KIBA_OBJECT(widget);
	KibaDockPrivate *priv;
	gint x, y, dx, dy, drag_distance;
	const KibaLayoutData *d;
	
// 	if (event->button != 1)
// 		return;
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	d = &kiba_layout_data[dock->d.position];
	
	priv->press_count--;
	if (priv->press_count > 0 && event->button != priv->drag_button)
		return TRUE;
	priv->press_count = 0;
	priv->drag_button = 0;
	
	if (!kiba_object_is_on_dock(object))
		return TRUE;
	
	gdk_window_get_pointer (GTK_WIDGET(dock)->window, &x, &y, NULL);
	if (KIBA_IS_FIXED_OBJECT(object)) {
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			x -= priv->mouse_anchor.x;
		else
			y -= priv->mouse_anchor.y;
	}
	else {
		x -= priv->mouse_anchor.x;
		y -= priv->mouse_anchor.y;
	}
	
	dx = widget->allocation.x - x;
	dy = widget->allocation.y - y;
	if (KIBA_IS_FIXED_OBJECT(priv->dragging_child)) {
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			dy = dx;
		else
			dx = dy;
	}
	drag_distance = sqrt (dx * dx + dy * dy);
	
	if (!KIBA_IS_FIXED_OBJECT(object) && drag_distance < MAX_DRAG_DISTANCE_FOR_EXEC)
		kiba_object_show_title (widget);
	
	if (priv->move_object_id) {
		g_source_remove (priv->move_object_id);
		priv->move_object_id = 0;
	}
	
	if (GET_FLAG (KIBA_OBJECT(widget)->state, KIBA_OBJECT_DRAGGING))
		kiba_object_unset_state(KIBA_OBJECT(widget), KIBA_OBJECT_DRAGGING);
	
	dock->allocate_children (dock, dock->allocate_children_data);
	
	if (GTK_IS_WINDOW(priv->dragging_window)) {
		gtk_widget_destroy (priv->dragging_window);
		priv->dragging_window = NULL;
		priv->dragging_child = NULL;
		return TRUE;
	}
	
	priv->dragging_child = NULL;
	
	return drag_distance > MAX_DRAG_DISTANCE_FOR_EXEC;
}

static gboolean __object_enter_notify_cb (GtkWidget        *widget,
					     GdkEventCrossing *event,
					     gpointer          data) {
	KibaDock *dock = KIBA_DOCK(data);
	KibaObject *object = KIBA_OBJECT(widget);
	
	if (GET_FLAG(object->state, KIBA_OBJECT_DRAGGING))
		return FALSE;
	if (!kiba_object_is_on_dock(object))
		return FALSE;
	
	gtk_container_set_focus_child (GTK_CONTAINER(dock), widget);
	
	if (KIBA_IS_SEPARATOR(widget))
		return FALSE;
	
	kiba_object_set_state (object, KIBA_OBJECT_PRELIGHTED);
	kiba_object_show_title (widget);

	return FALSE;
}

static gboolean __object_leave_notify_cb (GtkWidget        *widget,
					     GdkEventCrossing *event,
					     gpointer          data) {
	KibaObject *object = KIBA_OBJECT(widget);
	
	if (!kiba_object_is_on_dock(object))
		return FALSE;
	
	kiba_object_hide_title (widget);
	
	if (!GET_FLAG(object->state, KIBA_OBJECT_DRAGGING))
		kiba_object_unset_state (object, KIBA_OBJECT_PRELIGHTED);
	
	return FALSE;
}

static gboolean __object_scroll_notify_cb (GtkWidget      *widget,
					      GdkEventScroll *event) {
	/* dont let the dock get the signal */
	return !KIBA_IS_FIXED_OBJECT(widget);
}

/*** core methods, you can always trust that the pointer is valid here ***/

void kiba_dock_set_shape (KibaDock      *dock,
			  KibaDockShape  shape) {
	dock->shape = shape;
	
	if (!GTK_WIDGET_REALIZED(dock))
		return;
	
	kiba_dock_update_surface (dock);
}

void kiba_dock_set_max_size (KibaDock *dock,
			     gint      max_size) {
	gint diff;
	
	if (max_size < 40) {
		const KibaLayoutData *layout;
		
		layout = &kiba_layout_data[dock->d.position];
		if (layout->orientation == GTK_ORIENTATION_VERTICAL)
			max_size = KIBA_SCREEN_DEFAULT->geometry.height;
		else
			max_size = KIBA_SCREEN_DEFAULT->geometry.width;
	}
	diff = (max_size - dock->d.max_size)/2;
	dock->d.max_size = max_size;
	
	if (diff > 0) {
		dock->scroll_offset -= diff;
		dock->scroll_offset = MAX(dock->scroll_offset, 0);
	}
	
	if (!GTK_WIDGET_REALIZED(dock))
		return;
	
	gtk_widget_queue_resize(GTK_WIDGET(dock));
}

void kiba_dock_set_layout (KibaDock       *dock,
			   KibaDockLayout  d) {
	gboolean height_changed;
	
	if (d.height != dock->d.height)
		height_changed = TRUE;
	
	dock->d = d;
	
	if (height_changed)
		g_signal_emit (dock, kiba_dock_signals[SIG_HEIGHT_CHANGED], 0, dock->d.height);
	
	/* apply new layout to the dock */
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

void kiba_dock_set_style (KibaDock       *dock,
			  KibaDockStyle   style) {
	dock->style = style;
	
	if (!GTK_WIDGET_REALIZED(dock))
		return;
	
	kiba_dock_update_surface (dock);
}

void kiba_dock_set_flag (KibaDock *dock,
			 gint      flag) {
	if (GET_FLAG(dock->flags, flag))
		return;
	
	SET_FLAG(dock->flags, flag);
	
	switch (flag) {
		case KIBA_DOCK_AUTOHIDE:
			g_signal_emit (dock, kiba_dock_signals[SIG_AUTOHIDE_TOGGLED],
				0, TRUE);
			break;
		case KIBA_DOCK_DRAW_CURVES:
			kiba_dock_update_surface (dock);
			break;
		case KIBA_DOCK_DRAWABLE:
		case KIBA_DOCK_DRAW_SEPARATOR:
		case KIBA_DOCK_DRAW_SHADOW:
		case KIBA_DOCK_DRAW_REFLECTIONS:
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_SET_STRUT:
			kiba_dock_update_strut (GTK_WIDGET(dock));
			break;
	}
}

void kiba_dock_unset_flag (KibaDock *dock,
			   gint      flag) {
	if (!GET_FLAG(dock->flags, flag))
		return;
	
	DEL_FLAG(dock->flags, flag);
	
	switch (flag) {
		case KIBA_DOCK_AUTOHIDE:
			g_signal_emit (dock, kiba_dock_signals[SIG_AUTOHIDE_TOGGLED],
				0, FALSE);
			break;
		case KIBA_DOCK_DRAW_CURVES:
			kiba_dock_update_surface (dock);
			break;
		case KIBA_DOCK_DRAWABLE:
		case KIBA_DOCK_DRAW_SEPARATOR:
		case KIBA_DOCK_DRAW_SHADOW:
		case KIBA_DOCK_DRAW_REFLECTIONS:
			__queue_draw (GTK_WIDGET(dock));
			break;
		case KIBA_DOCK_SET_STRUT:
			kiba_dock_update_strut (GTK_WIDGET(dock));
			break;
	}
}

void kiba_dock_set_alpha (KibaDock *dock,
			  float    alpha) {
	KibaDockPrivate *priv;
	
	priv = KIBA_DOCK_GET_PRIVATE (KIBA_DOCK(dock));
	
	priv->fade_alpha = alpha;
	__queue_draw (GTK_WIDGET(dock));
}

void kiba_dock_set_object_size (KibaDock *dock,
                                gint size) {
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	GList *l;
	
	dock->d.object_size = size;
	
	if (!GTK_WIDGET_REALIZED(dock))
		return;
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		GtkWidget *widget = child->widget;
		GtkAllocation allocation;
		GtkRequisition requisition;
		
		allocation = widget->allocation;
		
		gtk_widget_size_request (widget, &requisition);
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
		gtk_widget_queue_resize_no_redraw (widget);
	}
	
	gtk_widget_queue_resize (GTK_WIDGET(dock));
	
	g_signal_emit (dock,
		kiba_dock_signals[SIG_OBJECT_SIZE_CHANGED],
		0, dock->d.object_size);
}

void kiba_dock_set_group_list (KibaDock *dock,
			       GSList   *group_list) {
	GSList *l;
	
	if (group_list == NULL)
		return;
	
	if (dock->groups) {
		GList *li;
		
		for (li = dock->groups; li != NULL; li = dock->groups) {
			dock->groups = g_list_remove(dock->groups, li->data);
		}
	}
	for (l = group_list; l != NULL; l = l->next) {
		dock->groups = g_list_append(dock->groups, l->data);
	}
	__sort_objects(dock);
}

GList* kiba_dock_get_group_list (KibaDock *dock) {
	return dock->groups;
}

cairo_pattern_t* kiba_dock_get_pattern (KibaDock *dock) {
	cairo_pattern_t *ret;
	gint border_width, x0, x1, y0, y1;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));
	
	x0 = border_width/2;
	y0 = border_width/2;
	x1 = dock->surface_data.width  - border_width/2;
	y1 = dock->surface_data.height - border_width/2;
	
	switch (dock->style.background.style) {
		case KIBA_GRADIENT:
			ret = cairo_get_gradient_pattern (x0, y0, x1, y1,
				dock->style.gradient, dock->style.numColors);
			break;
		default:
			ret = cairo_get_png_pattern (dock->style.background.file, 1.0);
			break;
	}
	
	return ret;
}

GtkWidget* kiba_dock_get_dragging_child(KibaDock *dock) {
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	return priv->dragging_child;
}

GtkWidget* kiba_dock_get_dragging_window(KibaDock *dock){
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	return priv->dragging_window;
}

/*** Public Methods ***/

/*** Render Methods ***/

void kiba_dock_paint_object (KibaDock *dock,
                             KibaDrawableObject *drawable_object,
                             cairo_t *cr) {
	KibaDockPrivate *priv;
	KibaObject *object;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	g_return_if_fail (cr != NULL);
	
	if (!drawable_object->surface_data->surface)
		return;
	
	object = KIBA_OBJECT(drawable_object);
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	if (kiba_drawable_object_scale (drawable_object, cr)) {
		GList *custom_funcs;
		
		kiba_drawable_object_rotate (drawable_object, cr);
		
		for (custom_funcs = drawable_object->custom_paint_funcs_before;
		     custom_funcs != NULL; custom_funcs = custom_funcs->next) {
			KibaDrawableObjectCustomPaintData *paint_data = custom_funcs->data;
			
			cairo_save(cr);
				paint_data->func (drawable_object, cr, paint_data->data);
			cairo_restore (cr);
		}
		
		if (priv->dragging_child != GTK_WIDGET(drawable_object) || !GTK_IS_WINDOW(priv->dragging_window)) {
			if (GET_FLAG (dock->flags, KIBA_DOCK_DRAW_SHADOW)) {
				cairo_save (cr);
					kiba_drawable_object_paint_shadow (drawable_object, cr);
				cairo_restore (cr);
			}
			cairo_save(cr);
#if OBJECT_SOLID_RECTANGLES == TRUE
				cairo_rectangle (cr, 0, 0,
					object->alloc.width, object->alloc.height);
				cairo_set_source_rgba (cr, 0, 0, 0, 1);
				cairo_fill (cr);
#else
				if (!drawable_object->paint)
					kiba_drawable_object_paint (drawable_object, cr, NULL);
				else
					drawable_object->paint (drawable_object, cr, drawable_object->paint_data);
#endif
			cairo_restore (cr);
		}
		
		for (custom_funcs = drawable_object->custom_paint_funcs_after;
		     custom_funcs != NULL; custom_funcs = custom_funcs->next) {
			KibaDrawableObjectCustomPaintData *paint_data = custom_funcs->data;
			
			cairo_save(cr);
				paint_data->func (drawable_object, cr, paint_data->data);
			cairo_restore(cr);
		}
	}
}

/**
 * kiba_dock_expose:
 * @dock: The KibaDock instance.
 * @cr: The drawing context.
 *
 * Recursively renders the dock and its children.
 */
void kiba_dock_expose (KibaDock *dock,
		       cairo_t  *cr) {
	GList *l;
	GtkWidget *win;
	const KibaLayoutData *d;
	KibaDockPrivate *priv;
	gint border_width;
	GtkWidget *focus_child;
	
	g_return_if_fail (KIBA_IS_DOCK(dock));
	
	win = GTK_WIDGET(dock)->parent;
	border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	d = &kiba_layout_data[dock->d.position];
	
	/* rendering the focussed child first */
	focus_child = GTK_CONTAINER(dock)->focus_child;
	if (KIBA_IS_DRAWABLE_OBJECT(focus_child)) {
		priv->drawableObjects = g_list_remove (priv->drawableObjects, focus_child);
		priv->drawableObjects = g_list_append (priv->drawableObjects, focus_child);
	}
	
	if (GET_FLAG (dock->flags, KIBA_DOCK_DRAWABLE)) {
		float alpha;
		GtkAllocation dockAllocation;
		
		dockAllocation = __get_dock_allocation(GTK_WIDGET(dock));
		
		alpha = priv->fade_alpha*dock->style.background_alpha;
		
		cairo_save (cr);
		
		cairo_save (cr);
		cairo_translate (cr, dockAllocation.x, dockAllocation.y);
		cairo_scale (cr, (float) dockAllocation.width/dock->surface_data.width,
				 (float) dockAllocation.height/dock->surface_data.height);
		
		/* drawing the Dock */
		if (alpha) {
			cairo_save (cr);
				cairo_set_source_surface (cr, dock->surface_data.surface, 0, 0);
				cairo_paint_with_alpha (cr, priv->fade_alpha*dock->style.background_alpha);
			cairo_restore (cr);
		}
		cairo_restore (cr);
		
		cairo_save(cr);
		
		cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
		
		cairo_save(cr);
		if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_SEPARATOR) &&
		    dock->shape.separator_draw_func != NULL)
			dock->shape.separator_draw_func (dock, cr, dock->shape.draw_data);
		cairo_restore (cr);
		
		if (dock->shape.clip_func)
			dock->shape.clip_func (dock, cr, dock->shape.draw_data);
		
		/* Reflection ontop of the dock */
		if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_REFLECTIONS)) {
			for (l = priv->drawableObjects; l != NULL; l = l->next) {
				KibaDrawableObject *drawableObject = l->data;
				if (!GTK_WIDGET_MAPPED(drawableObject))
					continue;
				cairo_save(cr);
					kiba_drawable_object_paint_reflection (drawableObject, cr);
				cairo_restore (cr);
			}
		}
		
		if (!dock->shape.triangle_ontop) {
			cairo_save (cr);
				kiba_triangle_draw (KIBA_WIN(win)->triangle, d, cr);
			cairo_restore (cr);
		}
		
		cairo_restore (cr);
	}
	
	/* drawing the objects */
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	for (l = priv->drawableObjects; l != NULL; l = l->next) {
		KibaDrawableObject *drawable_object = l->data;
		GtkWidget *widget = GTK_WIDGET(drawable_object);
		KibaObject *object = KIBA_OBJECT(drawable_object);
		GList *funcs;
		float offsetX = 0.0f, offsetY = 0.0f;
		
		if (!GTK_WIDGET_MAPPED(widget))
			continue;
		
		if (priv->dragging_child == widget && GTK_IS_WINDOW(priv->dragging_window)) {
			gdk_window_invalidate_rect (priv->dragging_window->window, NULL, TRUE);
		}
		
		for (funcs = object->custom_offset_data; funcs != NULL; funcs = funcs->next) {
			float customX = 0.0f, customY = 0.0f;
			KibaObjectCustomOffsetData *data = funcs->data;
			
			data->place_func (object, &customX, &customY, data->data);
			offsetX += customX;
			offsetY += customY;
		}
		
		cairo_save(cr);
			cairo_translate (cr,
				object->alloc.x - (d->halignment-1)*offsetX,
				object->alloc.y - (d->valignment-1)*offsetY);
				kiba_dock_paint_object (dock, drawable_object, cr);
		cairo_restore (cr);
		
		/* the object has some work in progress */
		if (dock->continuousProgress.enable) {
			cairo_save(cr);
				kiba_drawable_object_paint_gear(drawable_object, cr);
			cairo_restore (cr);
		}
	}
	
	if (dock->shape.triangle_ontop) {
		cairo_save (cr);
			kiba_triangle_draw (KIBA_WIN(win)->triangle, d, cr);
		cairo_restore (cr);
	}
}

/**
 * kiba_dock_update_surface:
 * @dock: The KibaDock instance.
 *
 * Updates/creates the surface buffer used by @dock.
 */
void kiba_dock_update_surface (KibaDock *dock) {
	GtkWidget *widget = GTK_WIDGET(dock);
	KibaWin *win;
	cairo_t *cr;
	const KibaLayoutData *d;
	
	g_return_if_fail (KIBA_IS_WIN(widget->parent));
	
	win = KIBA_WIN(widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		dock->surface_data.height = dock->d.height;
		dock->surface_data.width = widget->allocation.width;
	}
	else {
		dock->surface_data.width = dock->d.height;
		dock->surface_data.height = widget->allocation.height;
	}
	
	cr = kiba_win_cairo_create (win);
	if (!cr)
		return;
	cairo_update_surface (cr,
		&dock->surface_data,
		__draw_surface,
		dock);
	cairo_destroy (cr);
	__queue_draw (widget);
}

/*** Properties ***/

gboolean kiba_dock_object_drag_active (KibaDock *dock) {
	KibaDockPrivate *priv;
	
	g_return_val_if_fail (KIBA_IS_DOCK(dock), FALSE);
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	return (priv->dragging_window != NULL);
}

/**
 * kiba_dock_set_position:
 * @dock: The KibaDock instance.
 * @position: The position to set.
 *
 * Sets dock position.
 */
void kiba_dock_set_position (KibaDock     *dock,
			     KibaPosition  position) {
	const KibaLayoutData *old_d;
	const KibaLayoutData *new_d;
	KibaDockPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (position >= 0);
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	if (position == dock->d.position)
		return;
	
	old_d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	new_d = &kiba_layout_data[position];
	
	dock->d.position = position;
	dock->d.offset = 0;
	if (dock->d.max_size < 40) {
		if (new_d->orientation == GTK_ORIENTATION_VERTICAL)
			dock->d.max_size = KIBA_SCREEN_DEFAULT->geometry.height;
		else
			dock->d.max_size = KIBA_SCREEN_DEFAULT->geometry.width;
	}
	
	if (old_d->orientation != new_d->orientation)
		g_signal_emit (dock,
			kiba_dock_signals[SIG_ORIENTATION_CHANGED],
			0, new_d->orientation);
	g_signal_emit (dock,
		kiba_dock_signals[SIG_POSITION_CHANGED],
		0, position);
	
	/* queue resize on fixed objects */
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		
		if (KIBA_IS_FIXED_OBJECT(child->widget))
			gtk_widget_queue_resize (child->widget);
	}
	
	if (GTK_WIDGET_REALIZED(dock))
		kiba_dock_update_surface (dock);
}

/**
 * kiba_dock_get_position:
 * @dock: The KibaDock instance.
 *
 * Returns dock position.
 *
 * Returns: The dock position.
 */
KibaPosition kiba_dock_get_position (KibaDock *dock) {
	g_return_val_if_fail (KIBA_IS_DOCK(dock), KIBA_POSITION_BOTTOM_EDGE);
	
	return dock->d.position;
}

/**
 * kiba_dock_set_continuous_progress_icon:
 * @dock: The KibaDock instance.
 * @icon: The icon file to use for the continuous progress (currently only SVG files are supported).
 *
 * Set the size of the continuous progress icon.
 */
void kiba_dock_set_continuous_progress_icon(KibaDock *dock,
                                            const char *icon) {
	if (dock->continuousProgress.icon != NULL)
		g_free(dock->continuousProgress.icon);
	dock->continuousProgress.icon = g_strdup(icon);
	__update_gear_surface(dock);
}

/**
 * kiba_dock_set_continuous_progress_size:
 * @dock: The KibaDock instance.
 * @size: The size to use for the continuous progress icon.
 *
 * Set the size of the continuous progress icon.
 */
void kiba_dock_set_continuous_progress_size(KibaDock *dock,
                                            gint size) {
	dock->continuousProgress.size = size;
	__update_gear_surface(dock);
}

// TODO set partial strut
// void kiba_set_strut_partial (Window Xid,
// 		int left, int right, int top, int bottom,
// 		int left_start_y, int left_end_y, int right_start_y, int right_end_y,
// 		int top_start_x, int top_end_x, int bottom_start_x, int bottom_end_x) {
// 	g_return_if_fail (Xid > 0);
// 
// 	gulong iGeometryStrut[12] = {
// 		left, right, top, bottom,
// 		left_start_y, left_end_y, right_start_y, right_end_y,
// 		top_start_x, top_end_x, bottom_start_x, bottom_end_x};
// 	
// 	XChangeProperty (gdk_display,
// 		Xid,
// 		XInternAtom (gdk_display, "_NET_WM_STRUT", False),
// 		XA_CARDINAL, 32, PropModeReplace,
// 		(guchar *) iGeometryStrut, 12);
// }

/**
 * kiba_dock_update_strut:
 * @widget: The KibaDock instance.
 *
 * Sets/updates _NET_WM_STRUT property on the dock window.
 * This will prevent other windows to overlap with the dock in maximized state.
 */
void kiba_dock_update_strut (GtkWidget *widget) {
	KibaStrut strut;
	KibaDock *dock;
	const KibaLayoutData *d;
	
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	dock = KIBA_DOCK(widget);
	d = &kiba_layout_data[dock->d.position];
	
	if (!GET_FLAG(dock->flags, KIBA_DOCK_SET_STRUT) || dock->autohide.state == DOCK_HIDDEN ||
	     GET_FLAG(KIBA_WIN(widget->parent)->probs, KIBA_WIN_BELOW)) {
		strut.left = 0;
		strut.right = 0;
		strut.top = 0;
		strut.bottom = 0;
	}
	else if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		strut.left = 0;
		strut.right = 0;
		strut.top = dock->d.height * ((2-d->valignment)/2);
		strut.bottom = dock->d.height * (d->valignment/2);
	}
	else {
		strut.left = dock->d.height * ((2-d->halignment)/2);
		strut.right = dock->d.height * (d->halignment/2);
		strut.top = 0;
		strut.bottom = 0;
	}
	
	kiba_set_wm_strut (gdk_x11_drawable_get_xid (widget->window), &strut);
}

/**
 * kiba_dock_remove_custom_children_position_func_by_data:
 * @dock: The KibaDock instance.
 * @data: The user data associated to the method.
 *
 * Removes method to calculate the object positions on the dock.
 *
 * Returns: TRUE if the method was removed succesfully, otherwise FALSE is returned.
 */
gboolean kiba_dock_remove_custom_children_position_func_by_data (KibaDock *dock,
								 gpointer  data) {
	if (dock->allocate_children_data != data)
		return FALSE;
	
	dock->allocate_children = __place_objects;
	dock->allocate_children_data = NULL;
	
	gtk_widget_queue_resize(GTK_WIDGET(dock));
	
	return TRUE;
}

/**
 * kiba_dock_set_custom_children_position_func:
 * @dock: The KibaDock instance.
 * @func: The method to place the children.
 * @data: The user data passed to @func.
 *
 * Sets method to calculate the object positions on the dock.
 * All objects on the dock must be placed in @func.
 *
 * Returns: TRUE if the method was removed succesfully, otherwise FALSE is returned.
 */
gboolean kiba_dock_set_custom_children_position_func (KibaDock                 *dock,
						      KibaDockChildPositionFunc func,
						      gpointer                  data) {
	if (dock->allocate_children != __place_objects)
		return FALSE;
	
	dock->allocate_children = func;
	dock->allocate_children_data = data;
	
	gtk_widget_queue_resize(GTK_WIDGET(dock));
	
	return TRUE;
}

/**
 * kiba_dock_add_populate_func:
 * @dock: The KibaDock instance.
 * @func: The method to populate the dock.
 * @icon: The menu item icon path or name.
 * @title: The menu item title.
 * @user_data: The user data.
 *
 * Adds method to populate the dock with objects.
 */
void kiba_dock_add_populate_func (KibaDock             *dock,
				  KibaDockPopulateFunc  func,
				  const char           *icon,
				  const char           *title,
				  gpointer              user_data) {
	KibaDockPopulateData *data;
	GList *l;
	
	g_return_if_fail (dock != NULL);
	g_return_if_fail (func != NULL);
	g_return_if_fail (icon != NULL);
	g_return_if_fail (title != NULL);
	
	for (l = dock->populate_data; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->func == func) {
			data->user_data = data;
			return;
		}
	}
	
	data = g_new0 (KibaDockPopulateData, 1);
	data->dock = dock;
	data->func = func;
	data->icon = g_strdup(icon);
	data->title = g_strdup(title);
	data->user_data = user_data;
	
	dock->populate_data = g_list_append (dock->populate_data, data);
}

/**
 * kiba_dock_remove_populate_func:
 * @dock: The KibaDock instance.
 * @func: The method to populate the dock.
 *
 * Removes method to populate the dock with objects.
 */
void kiba_dock_remove_populate_func (KibaDock             *dock,
				     KibaDockPopulateFunc  func) {
	GList *l;
	
	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (func != NULL);
	
	for (l = dock->populate_data; l != NULL; l = l->next) {
		KibaDockPopulateData *data = l->data;
		
		if (data->func == func) {
			dock->populate_data = g_list_remove (dock->populate_data, data);
			g_free (data->icon);
			g_free (data->title);
			g_free (data);
			return;
		}
	}
}

/**
 * kiba_dock_set_group_sort_func:
 * @dock: The KibaDock instance.
 * @group: The group id.
 * @func: The method to sort the group.
 * @data: The user data.
 *
 * Sets method to sort objects associated to @group.
 */
void kiba_dock_set_group_sort_func (KibaDock         *dock,
				    const char       *group,
				    KibaGroupSortFunc func,
				    gpointer          data) {
	KibaDockPrivate *priv;
	KibaDockGroupData *new_data;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (group != NULL);
	g_return_if_fail (func != NULL);
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	for (l = priv->group_data; l != NULL; l = l->next) {
		KibaDockGroupData *data = l->data;
		
		if (strcmp (data->group, group) == 0) {
			data->sort_func = func;
			data->user_data = data;
			return;
		}
	}
	
	new_data = g_new0 (KibaDockGroupData, 1);
	new_data->group = g_strdup (group);
	new_data->sort_func = func;
	new_data->user_data = data;
	
	priv->group_data = g_list_append (priv->group_data, new_data);
}

/**
 * kiba_dock_remove_group_sort_func:
 * @dock: The KibaDock instance.
 * @group: The group id.
 *
 * Removes currently used method to sort objects associated to @group.
 */
void kiba_dock_remove_group_sort_func (KibaDock    *dock,
				       const char  *group) {
	KibaDockPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DOCK(dock));
	g_return_if_fail (group != NULL);
	
	priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	for (l = priv->group_data; l != NULL; l = l->next) {
		KibaDockGroupData *data = l->data;
		
		if (strcmp (data->group, group) == 0) {
			priv->group_data = g_list_remove (priv->group_data, data);
			g_free (data->group);
			g_free (data);
			data = NULL;
			return;
		}
	}
}

/*** Class Implementations ***/

/**
 * gtk_container_add implementation.
 **/
static void kiba_dock_add (GtkContainer *container,
			   GtkWidget    *widget) {
	KibaDock *dock;
	KibaObject *object;
	KibaDockPrivate *priv;
	gchar *group;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	dock = KIBA_DOCK(container);
	priv = KIBA_DOCK_GET_PRIVATE (container);
	object = KIBA_OBJECT(widget);
	
	if (KIBA_IS_DRAWABLE_OBJECT(object))
		priv->drawableObjects = g_list_append(priv->drawableObjects, object);
	
	/* connecting object signal handler */
	g_signal_connect (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(__object_button_press_cb), container);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(__object_button_release_cb), container);
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(__object_enter_notify_cb), container);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(__object_leave_notify_cb), container);
	g_signal_connect_after (G_OBJECT(widget), "scroll-event",
		G_CALLBACK(__object_scroll_notify_cb), container);
	
	kiba_object_set_requisition_factor (object, 1.0, 1.0);
	
	g_object_get (G_OBJECT(object), "group", &group, NULL);
	if (!__group_is_added (dock, group))
		dock->groups = g_list_append (dock->groups, group);
	else g_free(group);
	
	(* GTK_CONTAINER_CLASS (kiba_dock_parent_class)->add) (container, widget);
	
	/* sorting object groups */
	__sort_objects(dock);
}

/**
 * gtk_container_remove implementation.
 **/
static void kiba_dock_remove (GtkContainer *container,
			      GtkWidget    *widget) {
	KibaDock *dock = KIBA_DOCK(container);
	KibaDockPrivate *priv;
	KibaObject *object;
	GtkBox *box;
	GtkBoxChild *child;
	GList *children;
	
	object = KIBA_OBJECT(widget);
	priv = KIBA_DOCK_GET_PRIVATE (container);
	
	if (g_list_find(priv->drawableObjects, object))
		priv->drawableObjects = g_list_remove(priv->drawableObjects, object);
	
	if (dock->scroll_offset > 0) {
		dock->scroll_offset -= widget->allocation.width;
		dock->scroll_offset = MAX(KIBA_DOCK(dock)->scroll_offset, 0);
	}
	
	if (object->position == KIBA_OBJECT_RIGHT_OF_DOCK)
		priv->objects_right_of_dock--;
	else if (object->position == KIBA_OBJECT_LEFT_OF_DOCK)
		priv->objects_left_of_dock--;
	
	(* GTK_CONTAINER_CLASS (kiba_dock_parent_class)->remove) (container, widget);
	
	dock->allocate_children (dock, dock->allocate_children_data);
	__check_objects_on_dock (GTK_WIDGET(dock));
}

/**
 * gtk_widget_scroll_event implementation.
 **/
static gboolean kiba_dock_scroll_event (GtkWidget      *widget,
					GdkEventScroll *event) {
	KibaDockPrivate *priv;
	KibaDock *dock = KIBA_DOCK(widget);
	gint objects_left, objects_right;
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	objects_left = priv->objects_left_of_dock;
	objects_right = priv->objects_right_of_dock;
	
	/* moving objects to the left */
	if (event->direction == GDK_SCROLL_UP) {
		if (priv->objects_right_of_dock <= 0)
			return TRUE;
		priv->scroll_counter++;
		priv->scroll_counter = MIN(priv->scroll_counter, 15);
		g_timeout_add (100, __scroll_counter_decrease, priv);
	}
	/* moving objects to the right */
	else if (event->direction == GDK_SCROLL_DOWN) {
		if (priv->objects_left_of_dock <= 0)
			return TRUE;
		priv->scroll_counter--;
		priv->scroll_counter = MAX(priv->scroll_counter, -15);
		g_timeout_add (100, __scroll_counter_increase, priv);
	}
	KIBA_DOCK(widget)->scroll_offset += priv->scroll_counter*SCROLL_STEP;
	
	dock->allocate_children (dock, dock->allocate_children_data);
	__check_objects_on_dock (widget);
	
	return TRUE;
}

/**
 * gtk_widget_size_request implementation
 **/
static void kiba_dock_size_request (GtkWidget      *widget,
				    GtkRequisition *requisition) {
	KibaDock *dock;
	const KibaLayoutData *d;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DOCK(widget));
	
	dock = KIBA_DOCK(widget);
	
	requisition->width = 0;
	requisition->height = 0;
	
	d = &kiba_layout_data[dock->d.position];
	
	struct {
		gint *w, *h;
		gint alignment;
	} magic[2] = {
		{ &requisition->width,  &requisition->height, d->valignment},
		{ &requisition->height, &requisition->width,  d->halignment }
	};
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object = KIBA_OBJECT(child->widget);
		GtkRequisition child_requisition;
		GList *li;
		float destroy_factor;
		float max_factor_x = 1.0f;
		float max_factor_y = 1.0f;
		
		destroy_factor = kiba_object_get_destroy_factor (object);
		gtk_widget_size_request (child->widget, &child_requisition);
		
		child_requisition.width*=destroy_factor;
		child_requisition.height*=destroy_factor;
		
		/* making sure the widget has enough space for scale operations */
		for (li = object->custom_scale_funcs; li != NULL; li = li->next) {
			KibaObjectCustomScaleData *data = li->data;
			float custom_factor_x = 1.0f;
			float custom_factor_y = 1.0f;
			
			data->max_func (object,
				&custom_factor_x,
				&custom_factor_y,
				data->data);
			
			max_factor_x = MAX(max_factor_x, custom_factor_x);
			max_factor_y = MAX(max_factor_y, custom_factor_y);
		}
		gint object_size;
		if (KIBA_IS_SEPARATOR(object))
			object_size = dock->d.height;
		else
			object_size = dock->d.object_size;
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			child_requisition.width = MAX(child_requisition.width, object_size*object->scale_x*max_factor_x);
		}
		else {
			child_requisition.height = MAX(child_requisition.height, object_size*object->scale_y*max_factor_y);
		}
		
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			if (child_requisition.width > requisition->width)
				requisition->width = child_requisition.width;
			requisition->height += child_requisition.height + dock->d.object_space*destroy_factor;
		}
		else {
			if (child_requisition.height > requisition->height)
				requisition->height = child_requisition.height;
			requisition->width += child_requisition.width + dock->d.object_space*destroy_factor;
		}
	}
	if (d->orientation == GTK_ORIENTATION_VERTICAL)
		requisition->height -= dock->d.object_space;
	else
		requisition->width -= dock->d.object_space;
	
	if (*magic[d->orientation].w < MIN_OBJECT_SPACE)
		*magic[d->orientation].w = MIN_OBJECT_SPACE;
	if (*magic[d->orientation].h < MIN_OBJECT_SPACE)
		*magic[d->orientation].h = MIN_DOCK_HEIGHT;
	
	*magic[d->orientation].w += 2*dock->d.cap_size +
		2*gtk_container_get_border_width (GTK_CONTAINER(dock));
	*magic[d->orientation].h += dock->d.object_margin +
		gtk_container_get_border_width (GTK_CONTAINER(dock));
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (requisition->height > dock->d.max_size)
			requisition->height = dock->d.max_size;
		if (requisition->width > KIBA_SCREEN_DEFAULT->geometry.width)
			requisition->width = KIBA_SCREEN_DEFAULT->geometry.width;
	}
	else {
		if (requisition->width > dock->d.max_size)
			requisition->width = dock->d.max_size;
		if (requisition->height > KIBA_SCREEN_DEFAULT->geometry.height)
			requisition->height = KIBA_SCREEN_DEFAULT->geometry.height;
	}
}

/**
 * gtk_widget_size_allocate implementation.
 **/
static void kiba_dock_size_allocate (GtkWidget     *widget,
				     GtkAllocation *allocation) {
	KibaDockPrivate *priv;
	KibaDock *dock;
	const KibaLayoutData *d;
	gint dx, dy;
	GtkAllocation event_allocation;
	
	dock = KIBA_DOCK(widget);
	d = &kiba_layout_data[dock->d.position];
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	
	if (widget->allocation.width != allocation->width ||
	    widget->allocation.height != allocation->height ||
	    widget->allocation.y != allocation->y ||
	    widget->allocation.x != allocation->x) {
		__queue_draw (GTK_WIDGET(dock));
		widget->allocation = *allocation;
		__queue_draw (GTK_WIDGET(dock));
	}
	
	dock->allocate_children (dock, dock->allocate_children_data);
	__check_objects_on_dock (widget);
	
	if (!GTK_WIDGET_REALIZED (widget))
		return;
	
	event_allocation = __get_event_allocation (widget);
	gdk_window_move_resize (priv->event_window,
				event_allocation.x,
				event_allocation.y,
				event_allocation.width,
				event_allocation.height);
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		dy = dock->surface_data.height - dock->d.height;
		dx = dock->surface_data.width - GTK_WIDGET(dock)->allocation.width;
	}
	else {
		dx = dock->surface_data.width - dock->d.height;
		dy = dock->surface_data.height - GTK_WIDGET(dock)->allocation.height;
	}
	/* avoiding to update the dock surface to often,
	   that may would make some widget scale operastions slow */
	if (MAX(dx, -dx) > 100 || MAX(dy, -dy) > 100)
		kiba_dock_update_surface (dock);
	
	/* updating the _NET_WM_STRUT property */
	kiba_dock_update_strut (widget);
}

/**
 * gtk_widget_unmap implementation.
 **/
static void kiba_dock_unmap (GtkWidget *widget) {
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (widget);
	
	gdk_window_hide (priv->event_window);
	
	/* calling the GtkWidget implementation */
	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->unmap) (widget);
}

/**
 * gtk_widget_map_implemantation.
 **/
static void kiba_dock_map (GtkWidget *widget) {
	KibaDockPrivate *priv;
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	
	/* calling the GtkWidget implementation */
	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->map) (widget);
	
	gdk_window_show (priv->event_window);
}

/**
 * gtk_widget_unrealize implementaion.
 **/
static void kiba_dock_unrealize (GtkWidget *widget) {
	KibaDockPrivate *priv;
	GList *l;
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	
#ifdef USE_LIBSN
	KibaSn *sn = kiba_sn_get();
	g_signal_handlers_disconnect_by_func (G_OBJECT(sn),
		G_CALLBACK(__sn_launch_finished), widget);
	g_signal_handlers_disconnect_by_func (G_OBJECT(sn),
		G_CALLBACK(__sn_launch_started), widget);
#endif
	
	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	if (KIBA_DOCK(widget)->surface_data.surface)
		cairo_surface_destroy (KIBA_DOCK(widget)->surface_data.surface);
	
	for (l = KIBA_DOCK(widget)->groups; l != NULL; l = KIBA_DOCK(widget)->groups)
		KIBA_DOCK(widget)->groups = g_list_remove (KIBA_DOCK(widget)->groups, l->data);
	g_list_free (KIBA_DOCK(widget)->groups);
	
	(* GTK_WIDGET_CLASS (kiba_dock_parent_class)->unrealize) (widget);
	
	if (priv->move_object_id)
		g_source_remove (priv->move_object_id);
	
	for (l = priv->group_data; l != NULL; l = priv->group_data) {
		KibaDockGroupData *data = l->data;
		
		priv->group_data = g_list_remove (priv->group_data, data);
		g_free (data->group);
		g_free (data);
	}
	g_list_free (priv->group_data);
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_dock_realize (GtkWidget *widget) {
	GdkWindowAttr attributes;
	gint attributes_mask;
	KibaDockPrivate *priv;
	
	/* make sure the dock is properly size allocated */
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		gtk_widget_size_request (widget, &requisition);
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	priv = KIBA_DOCK_GET_PRIVATE (widget);
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window); /* higher the reference count of the window */
	
	/* creating a event catcher window */
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	
	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	gdk_window_set_user_data (priv->event_window, widget); /* catching the window events on the widget */
	gdk_window_register_dnd (priv->event_window); /* allowing dnd an the window */
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	/* creating the dock surface */
	kiba_dock_update_surface (KIBA_DOCK(widget));
	__update_gear_surface(KIBA_DOCK(widget));
	
#ifdef USE_LIBSN
	KibaSn *sn = kiba_sn_get();
	g_signal_connect(G_OBJECT(sn), "launch-finished",
		G_CALLBACK(__sn_launch_finished), widget);
	g_signal_connect(G_OBJECT(sn), "launch-started",
		G_CALLBACK(__sn_launch_started), widget);
#endif
}

static void kiba_dock_class_init (KibaDockClass *class) {
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
	GtkWidgetClass    *widget_class    = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaDockPrivate));
	
	/* custom widget implementations */
	widget_class->realize = kiba_dock_realize;
	widget_class->unrealize = kiba_dock_unrealize;
	widget_class->map = kiba_dock_map;
	widget_class->unmap = kiba_dock_unmap;
	widget_class->size_request = kiba_dock_size_request;
	widget_class->size_allocate = kiba_dock_size_allocate;
	widget_class->scroll_event = kiba_dock_scroll_event;
	
	/* custom container implementations */
	container_class->add = kiba_dock_add;
	container_class->remove = kiba_dock_remove;
	
	/**
	 * KibaDock::position-changed:
	 * @dock: the object on which the signal is emitted
	 * @position: the new dock Position.
	 *
	 * The position-changed signal is emitted when a new position has been set
	 * on @dock.
	 */
	kiba_dock_signals[SIG_POSITION_CHANGED] =
		g_signal_new ("position-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, position_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	kiba_dock_signals[SIG_ORIENTATION_CHANGED] =
		g_signal_new ("orientation-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, orientation_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	kiba_dock_signals[SIG_AUTOHIDE_TOGGLED] =
		g_signal_new ("autohide-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, autohide_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	kiba_dock_signals[SIG_OBJECT_SIZE_CHANGED] =
		g_signal_new ("object-size-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, object_size_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	kiba_dock_signals[SIG_HEIGHT_CHANGED] =
		g_signal_new ("height-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDockClass, height_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
}

static void kiba_dock_init (KibaDock *dock) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaDockPrivate *priv = KIBA_DOCK_GET_PRIVATE (dock);
	
	/* using no widget window */
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (dock), GTK_NO_WINDOW);
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (dock), GTK_DOUBLE_BUFFERED);
	
	gtk_widget_set_app_paintable (GTK_WIDGET (dock), TRUE);
	gtk_widget_add_events (GTK_WIDGET (dock),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	/* initialling defaults */
	dock->d = (KibaDockLayout) {
		KIBA_POSITION_BOTTOM_EDGE, 5, 10, 10, 38, 25, 0, 0.0
	};
	dock->d.max_size = KIBA_SCREEN_DEFAULT->geometry.width;
	KibaRGBA *gradient = g_new0(KibaRGBA, 1);
	gradient[0] = (KibaRGBA) {1.0,1.0,1.0,1.0};
	dock->style = (KibaDockStyle) {
		TRUE,
		KIBA_GRADIENT, 0.9f, 45.0f, 5.0f, 5.0f, 2.0f, 0.6f, 0.75f,
		{ 0.6f, 0.6f, 0.6f, 0.7f },
		{ 1.0f, 1.0f, 1.0f, 0.4f }, { 0.0f, 0.0f, 0.0f, 0.3f },
		gradient, 1
	};
	dock->autohide = (KibaAutoHide) {
		DOCK_UNHIDDEN, 25, 3, 2
	};
	dock->groups = NULL;
	dock->surface_data.surface = NULL;
	dock->gearSurface = NULL;
	dock->gearSurfaceOverlay = NULL;
	dock->scroll_offset = 0.0f;
	
	dock->shape.draw_func = NULL;
	dock->shape.separator_draw_func = NULL;
	dock->shape.clip_func = NULL;
	dock->shape.draw_data = NULL;
	dock->shape.triangle_ontop = FALSE;
	dock->shape.topleft = KIBA_ROUNDED;
	dock->shape.topright = KIBA_ROUNDED;
	dock->shape.bottomright = KIBA_ROUNDED;
	dock->shape.bottomleft = KIBA_ROUNDED;
	dock->shape.corner_radius = 25.0f;
	
	dock->continuousProgress.enable = TRUE;
	dock->continuousProgress.icon = NULL;
	dock->continuousProgress.size = 26;
	dock->continuousProgress.step = 0.05f;
	
	dock->allocate_children = __place_objects;
	dock->allocate_children_data = NULL;
	
	gtk_container_set_border_width (GTK_CONTAINER(dock), 5);
	
	SET_FLAG(dock->flags, KIBA_DOCK_DRAW_CURVES);
	
	priv->fade_alpha = 1.0f;
	priv->scroll_counter = 0;
	priv->dragging_child = NULL;
	priv->dragging_window = NULL;
	priv->press_count = 0;
	priv->drag_button = 0;
	
	/* enabling drag and drop */
	gtk_drag_dest_set (GTK_WIDGET (dock),
			GTK_DEST_DEFAULT_ALL,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
	gtk_drag_dest_add_uri_targets (GTK_WIDGET (dock));
}

/**
 * kiba_dock_new:
 *
 * Creates and returns a new KibaDock Container.
 *
 * Returns: The KibaDock Container created.
 */
GtkWidget* kiba_dock_new (void) {
	return g_object_new (KIBA_TYPE_DOCK, NULL);
}
