/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-object
 * @short_description: Base class for objects on #KibaDock.
 * @see_also: #KibaWin,#KibaDock,#KibaObject
 * @stability: Unstable
 *
 * #KibaObject implementation that allows custom drawing.
 */

#include "../config.h"
#include <string.h>
#include <math.h>
#include <gdk/gdkx.h>
#include <cairo-xlib.h>

#include "kiba.h"
#include "kiba-drawable-object.h"
#include "kiba-object.h"
#include "kiba-win.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-x11.h"

#include "dock-private.h"

#define OBJECT_FADE_STEP 0.01f

#define KIBA_DRAWABLE_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_DRAWABLE_OBJECT, KibaDrawableObjectPrivate))

G_DEFINE_TYPE (KibaDrawableObject, kiba_drawable_object, KIBA_TYPE_OBJECT)

static const float deg_to_rad_factor = 6.283f/360.0f;

typedef struct {
	GdkWindow *event_window;
	
	gint fade_id;
	float fade_alpha;
	
	float gearsRotateAngle;
	float gearsAlpha;
	gint rotateGearsID;
	
	GtkAllocation reflection_allocation;
} KibaDrawableObjectPrivate;


enum {
	SIG_SWITCH_EVENT,
	SIG_KIBA_OBJECT_LAST
};

static guint kiba_drawable_object_signals[SIG_KIBA_OBJECT_LAST] = { 0 };

/*** Private Methods ***/

static void __dock_object_size_changed_cb (KibaDock *dock,
                                           gint size,
                                           gpointer data) {
	KibaObject *object = data;
	
	object->size.width = size;
	object->size.height = size;
}

static gint __fade_timeout (gpointer data) {
	KibaDrawableObject *object = data;
	KibaDrawableObjectPrivate *priv;
	
	if (!KIBA_IS_DRAWABLE_OBJECT(object))
		return FALSE;
	
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	kiba_drawable_object_queue_draw (object);
	
	if (object->alpha < priv->fade_alpha) {
		object->alpha += OBJECT_FADE_STEP;
		if (object->alpha > priv->fade_alpha) {
			object->alpha = priv->fade_alpha;
			priv->fade_id = 0;
			return FALSE;
		}
	}
	else {
		object->alpha -= OBJECT_FADE_STEP;
		if (object->alpha < priv->fade_alpha) {
			object->alpha = priv->fade_alpha;
			priv->fade_id = 0;
			return FALSE;
		}
	}
	
	return TRUE;
}

static void __place_object_reflection (KibaDrawableObject *drawable_object,
                                       GtkWidget *dragging_window) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	GtkWidget *widget = GTK_WIDGET(drawable_object);
	GtkWidget *dragging_child;
	KibaDock *dock = KIBA_DOCK(widget->parent);
	const KibaLayoutData *d;
	GtkAllocation dragAllocation;
	
	if (!GET_FLAG(dock->flags, KIBA_DOCK_DRAW_REFLECTIONS))
		return;
	
	d = &kiba_layout_data[dock->d.position];
	
	kiba_drawable_object_queue_draw(drawable_object);
	
	priv->reflection_allocation = widget->allocation;
	dragging_child = kiba_dock_get_dragging_child(dock);
	if (dragging_window && dragging_child == widget) {
		KibaObject *object = KIBA_OBJECT(widget);
		GtkWidget *win = GTK_WIDGET(dock)->parent;
		GList *l;
		gint x,y;
		float factor_x = 1.2f, factor_y = 1.2f;
		
		gtk_window_get_position(GTK_WINDOW(dragging_window), &x, &y);
		dragAllocation.x = x - win->allocation.x;
		dragAllocation.y = y - win->allocation.y;
		
		for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
			KibaObjectCustomScaleData *data = l->data;
			
			if (data->scale_widget)
				continue;
			
			float custom_factor_x = 1.0f;
			float custom_factor_y = 1.0f;
			
			data->scale_func (object,
				&custom_factor_x,
				&custom_factor_y,
				data->data);
				
			factor_x *= custom_factor_x;
			factor_y *= custom_factor_y;
		}
		
		dragAllocation.x += 0.5f*dragging_window->allocation.width*(factor_x*0.5 - 0.5f);
		dragAllocation.y += 0.5f*dragging_window->allocation.height*(factor_y*0.5 - 0.5f);
	}
	else
		dragAllocation = widget->allocation;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		priv->reflection_allocation.x = dragAllocation.x;
		if (d->valignment == 0) {
			priv->reflection_allocation.y = 2*widget->allocation.y - dragAllocation.y - widget->allocation.height - 4;
			if (priv->reflection_allocation.y > widget->allocation.y - widget->allocation.height)
				priv->reflection_allocation.y = dragAllocation.y - widget->allocation.height;
		}
		else {
			priv->reflection_allocation.y = 2*widget->allocation.y - dragAllocation.y + widget->allocation.height + 4;
			if (priv->reflection_allocation.y < widget->allocation.y + widget->allocation.height + 4)
				priv->reflection_allocation.y = dragAllocation.y + widget->allocation.height + 4;
		}
	}
	else {
		if (d->halignment == 0) {
			priv->reflection_allocation.x = 2*widget->allocation.x - dragAllocation.x - widget->allocation.width - 4;
			if (priv->reflection_allocation.x > widget->allocation.x - widget->allocation.width)
				priv->reflection_allocation.x = widget->allocation.x - widget->allocation.width;
		}
		else {
			priv->reflection_allocation.x = 2*widget->allocation.x - dragAllocation.x + widget->allocation.width + 4;
			if (priv->reflection_allocation.x < widget->allocation.x + widget->allocation.width)
				priv->reflection_allocation.x = widget->allocation.x + widget->allocation.width;
		}
	}
	kiba_drawable_object_queue_draw(drawable_object);
}

static GtkAllocation __get_gear_allocation (KibaDrawableObject *object) {
	GtkAllocation ret;
	GtkWidget *widget = GTK_WIDGET(object);
	KibaDock *dock = KIBA_DOCK(widget->parent);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	
	ret.width = dock->continuousProgress.size;
	ret.height = ret.width;
	ret.x = widget->allocation.x;
	ret.y = widget->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		ret.x -= dock->continuousProgress.size/3;
		if (d->valignment == 0)
			ret.y -= dock->continuousProgress.size/3;
		else
			ret.y += widget->allocation.height - 2*dock->continuousProgress.size/3;
	}
	else {
		ret.y -= dock->continuousProgress.size/3;
		if (d->halignment == 0)
			ret.x -= dock->continuousProgress.size/3;
		else
			ret.x += widget->allocation.width - 2*dock->continuousProgress.size/3;
	}
	
	return ret;
}

static void __gears_queue_draw (KibaDrawableObject *object) {
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *dock = widget->parent;
	if (!KIBA_DOCK(dock)->continuousProgress.enable)
		return;
	GdkRectangle area;
	GtkAllocation allocation = __get_gear_allocation(object);
	
	area.width = allocation.width;
	area.height = allocation.height;
	area.x = allocation.x;
	area.y = allocation.y;
	
	kiba_win_invalidate_area (dock->parent, area);
}

static gboolean __rotate_gears (gpointer data) {
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(data);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	
	priv->gearsRotateAngle += dock->continuousProgress.step;
	if (priv->gearsRotateAngle > 2.0f)
		priv->gearsRotateAngle = 0.0f;
	
	__gears_queue_draw(object);
	
	if (kiba_object_get_state(KIBA_OBJECT(object), KIBA_OBJECT_WORKING)) {
		if (priv->gearsAlpha < 1.0f)
			priv->gearsAlpha += 0.05f;
	}
	else {
		if (priv->gearsAlpha > 0.0f)
			priv->gearsAlpha -= 0.05f;
	}
	
	if (priv->gearsAlpha > 0.0f)
		return TRUE;
	else {
		priv->rotateGearsID = 0;
		return FALSE;
	}
}

static void __working_toggled (KibaObject *object,
                               gboolean working,
                               gpointer data) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	if (working && !priv->rotateGearsID)
		priv->rotateGearsID = g_timeout_add(50, __rotate_gears, drawable_object);
}

static void __drag_window_size_allocate (GtkWidget *widget,
                                         GtkAllocation *allocation,
                                         gpointer data) {
	KibaDrawableObject *drawable_object = data;
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(drawable_object)->parent);
	__place_object_reflection(drawable_object, kiba_dock_get_dragging_window(dock));
	kiba_drawable_object_queue_draw(drawable_object);
}

static void __dragging_toggled (KibaObject *object,
                                gboolean dragging,
                                gpointer data) {
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *dragging_window;
	KibaDock *dock = KIBA_DOCK(widget->parent);
	
	dragging_window = kiba_dock_get_dragging_window(dock);
	if (!dragging_window)
		return;
	if (dragging) {
		g_signal_connect(G_OBJECT(dragging_window), "size-allocate",
			G_CALLBACK(__drag_window_size_allocate), drawable_object);
	} else {
		g_signal_handlers_disconnect_by_func (G_OBJECT(dragging_window),
			G_CALLBACK(__drag_window_size_allocate), drawable_object);
		__place_object_reflection(drawable_object, NULL);
		kiba_drawable_object_queue_draw(drawable_object);
	}
}

/*** Public methods ***/

/**
 * kiba_drawable_object_queue_draw:
 * @object: The KibaDrawableObject instance.
 *
 * Invalidates the area on the dock used by the object.
 */
void kiba_drawable_object_queue_draw (KibaDrawableObject *drawable_object) {
	KibaObject *object;
	KibaDrawableObjectPrivate *priv;
	GtkWidget *widget;
	GtkWidget *win;
	GtkWidget *dock;
	GdkWindow *window;
	GdkRectangle area;
	GList *l;
	const KibaLayoutData *d;
	float width, height;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	float shadow_offset_X, shadow_offset_y;
	float offsetX = 0.0f, offsetY = 0.0f;
	float rad;
	float rotateAngle = 0.0f;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	
	widget = GTK_WIDGET(drawable_object);
	object = KIBA_OBJECT(drawable_object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	dock = widget->parent;
	win = dock->parent;
	
	float alpha = drawable_object->alpha * drawable_object->scroll_alpha;
	if (alpha == 0.0f)
		return;
	
	window = gtk_widget_get_parent_window (GTK_WIDGET(object)->parent);
	g_return_if_fail (GDK_IS_WINDOW(window));
	
	width = object->alloc.width;
	height = object->alloc.height;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomOffsetData *data = l->data;
		
		data->place_func (object, &customX, &customY, data->data);
		offsetX += customX;
		offsetY += customY;
	}
	
	for (l = drawable_object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaDrawableObjectCustomRotateData *data = l->data;
		
		data->func (drawable_object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		data->scale_func (object, &customX, &customY, data->data);
		scale_x *= customX;
		scale_y *= customY;
	}
	
	area.x = object->alloc.x - (width *(scale_x-1.0f))/2;
	area.y = object->alloc.y - (height*(scale_y-1.0f))/2;
	
	area.y -= (d->valignment-1)*(offsetY);
	area.x -= (d->halignment-1)*(offsetX);
	
	if (rotateAngle != 0.0f) {
		area.height = MAX(scale_x, scale_y)*
				sqrt (height * height + width  * width);
		area.width = area.height;
		area.y -= (area.height - scale_y*height)/2;
		area.x -= (area.width - scale_x*width)/2;
	}
	else {
		area.height = scale_y*height;
		area.width  = scale_x*width;
	}
	// 	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
	// 		area.height += (d->valignment-1)*2;
	// 	else
	// 		area.width  += (d->halignment-1)*2;
		
	if (GET_FLAG (KIBA_DOCK(dock)->flags, KIBA_DOCK_DRAW_SHADOW)) {
		rad = deg_to_rad_factor*drawable_object->shadow_affection;
		shadow_offset_X = sin(rad)*drawable_object->shadow_offset;
		shadow_offset_y = cos(rad)*drawable_object->shadow_offset;
		
		shadow_offset_X = (shadow_offset_X*area.width)/(((float) object->size.width)*object->scale_x);
		shadow_offset_y = (shadow_offset_y*area.height)/(((float) object->size.height)*object->scale_y);
		
		if (shadow_offset_y < 0.0f) {
			area.height -= (int) shadow_offset_y+1;
			area.y += (int) shadow_offset_y-1;
		}
		else
			area.height += (int) shadow_offset_y+1;
		if (shadow_offset_X < 0.0f) {
			area.width -= (int) shadow_offset_X+1;
			area.x += (int) shadow_offset_X-1;
		}
		else
			area.width += (int) shadow_offset_X+1;
	}
	area.x -= 2;
	area.y -= 2;
	area.width += 4;
	area.height += 4;
	kiba_win_invalidate_area (win, area);
	if (GET_FLAG (KIBA_DOCK(dock)->flags, KIBA_DOCK_DRAW_SHADOW))
		if (shadow_offset_y < 0.0f) {
			area.y -= (int) shadow_offset_y+1;
		if (shadow_offset_X < 0.0f)
			area.x -= (int) shadow_offset_X+1;
	}
		
	if (GET_FLAG (KIBA_DOCK(dock)->flags, KIBA_DOCK_DRAW_REFLECTIONS)) {
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			area.y = priv->reflection_allocation.y + (d->valignment-1)*(offsetY +
				priv->reflection_allocation.height*(scale_y/1.1 - 1.0)) -
				(priv->reflection_allocation.height*(scale_y/1.1 - 1.0))/2;
			area.x += priv->reflection_allocation.x - widget->allocation.x;
		}
		else {
			area.x = priv->reflection_allocation.x + (d->halignment-1)*(offsetX +
				priv->reflection_allocation.width*(scale_x/1.1 - 1.0)) +
			(priv->reflection_allocation.width*(scale_x/1.1 - 1.0))/2;
			area.y += priv->reflection_allocation.y - widget->allocation.y;
		}
		if (GET_FLAG (KIBA_DOCK(dock)->flags, KIBA_DOCK_DRAW_SHADOW)) {
			if (shadow_offset_y > 0.0f)
				area.y -= (int) shadow_offset_y-1;
			if (shadow_offset_X < 0.0f)
				area.x += (int) shadow_offset_X-1;
		}
		if (rotateAngle != 0.0f) {
			area.y -= (area.height - scale_y*height)/2;
			area.x -= (area.width - scale_x*width)/2;
		}
		kiba_win_invalidate_area (win, area);
	}
}

/**
 * kiba_drawable_object_set_alpha:
 * @object: The KibaDrawableObject instance.
 * @alpha: The alpha of KibaDrawableObject.
 *
 * Sets the alpha factor used for rendering the object surface.
 */
void kiba_drawable_object_set_alpha (KibaDrawableObject *object,
			    float      alpha) {
	KibaDrawableObjectPrivate *priv;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	g_return_if_fail (0.0f <= alpha || alpha <= 1.0f);
	
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	if (alpha == priv->fade_alpha)
		return;
	
	priv->fade_alpha = alpha;
	
	if (!priv->fade_id)
		priv->fade_id = g_timeout_add (20, __fade_timeout, object);
}

/**
 * kiba_drawable_object_switch_surface:
 * @object: The KibaDrawableObject instance.
 *
 * Switches the object's surface, sending the "switched" signal.
 */
void kiba_drawable_object_switch_surface (KibaDrawableObject *object) {
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	
	//TODO there should be done more here,
	// the switch plugin shouldnt access the surface struct member.
	g_signal_emit (object,
		kiba_drawable_object_signals[SIG_SWITCH_EVENT], 0);
}

/**
 * kiba_drawable_object_set_paint_function:
 * @object: The KibaDrawableObject instance.
 * @func: The default rendering method.
 * @data: The user data.
 *
 * Sets custom render function.
 */
void kiba_drawable_object_set_paint_function (KibaDrawableObject          *object,
				     KibaDrawableObjectPaintFunc  func,
				     gpointer             data) {
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	
	object->paint = func;
	object->paint_data = data;
}

/**
 * kiba_drawable_object_paint_reflection:
 * @object: The KibaDrawableObject instance.
 * @cr: The drawing context.
 *
 * Renders the object reflection to the context.
 */
void kiba_drawable_object_paint_reflection (KibaDrawableObject *drawable_object,
                                            cairo_t *cr) {
	KibaDrawableObjectPrivate *priv;
	KibaObject *object;
	KibaDock *dock;
	cairo_surface_t *surface;
	const KibaLayoutData *d;
	float alpha;
	gint unscaled_width;
	gint unscaled_height;
	float rotateAngle = 0.0f;
	float buf_alpha;
	float fX, fY;
	float offsetX = 0.0f, offsetY = 0.0f;
	float scale_x = 1.0f, scale_y = 1.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	g_return_if_fail (GTK_WIDGET(drawable_object)->parent != NULL);
	
	if (!drawable_object->surface_data->surface)
		return;
	
	object = KIBA_OBJECT(drawable_object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	d = &kiba_layout_data[dock->d.position];
	
	unscaled_width  = (int) ((object->size.width) * object->scale_x);
	unscaled_height = (int) ((object->size.height) * object->scale_y);
	
	if (drawable_object->switch_surface)
		surface = drawable_object->switch_surface;
	else
		surface = drawable_object->surface_data->surface;
	
	g_return_if_fail (surface != NULL);
	
	alpha = drawable_object->alpha*dock->style.reflection_alpha;
		
	for (l = drawable_object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaDrawableObjectCustomRotateData *data = l->data;
		
		data->func (drawable_object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomOffsetData *data = l->data;
		
		data->place_func (object, &customX, &customY, data->data);
		offsetX += customX;
		offsetY += customY;
	}
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		float customX = 0.0f, customY = 0.0f;
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		data->scale_func (object, &customX, &customY, data->data);
		if (customX > 1.0f) /* fix for bounce scaling, shouldnt offset handle that allready? */
			scale_x *= customX;
		if (customY > 1.0f) /* fix for bounce scaling, shouldnt offset handle that allready? */
			scale_y *= customY;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		cairo_translate (cr,
			priv->reflection_allocation.x,
			priv->reflection_allocation.y + (d->valignment-1)*(offsetY +
			priv->reflection_allocation.height*(scale_y - 1.0)));
		cairo_scale (cr, -1, 1);
		cairo_translate (cr, -GTK_WIDGET(object)->allocation.width, 0);
	}
	else {
		cairo_translate (cr,
			priv->reflection_allocation.x + (d->halignment-1)*(offsetX +
			priv->reflection_allocation.width*(scale_x - 1.0)),
			priv->reflection_allocation.y);
		cairo_scale (cr, 1, -1);
		cairo_translate (cr, 0, -GTK_WIDGET(object)->allocation.height);
	}
	
	if (!kiba_drawable_object_scale (drawable_object, cr))
		return;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		cairo_scale (cr, 1, dock->style.reflection_scale_factor);
		cairo_translate (cr, 0, (-(d->valignment-2)/2)*
			(1.0f-dock->style.reflection_scale_factor)*
			unscaled_height);
	}
	else {
		cairo_scale (cr, dock->style.reflection_scale_factor, 1);
		cairo_translate (cr, (-(d->halignment-2)/2)*
			(1.0f-dock->style.reflection_scale_factor)*
			unscaled_width, 0);
	}
	
	cairo_translate (cr,
		unscaled_width/2,
		unscaled_height/2);
	cairo_rotate (cr, M_PI * (rotateAngle + 1.0));
	cairo_translate (cr,
		-unscaled_width/2,
		-unscaled_height/2);
	
	buf_alpha = drawable_object->alpha;
	drawable_object->alpha = alpha;
	if (GET_FLAG (dock->flags, KIBA_DOCK_DRAW_SHADOW)) {
		cairo_save (cr);
			kiba_drawable_object_paint_shadow (drawable_object, cr);
		cairo_restore (cr);
	}
	cairo_save(cr);
		if (!drawable_object->paint)
			kiba_drawable_object_paint (drawable_object, cr, NULL);
		else
			drawable_object->paint (drawable_object, cr, drawable_object->paint_data);
	cairo_restore (cr);
	drawable_object->alpha = buf_alpha;
}

void kiba_drawable_object_paint_gear (KibaDrawableObject *object,
                                      cairo_t *cr) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	GtkWidget *widget = GTK_WIDGET(object);
	KibaDock *dock = KIBA_DOCK(widget->parent);
	GtkAllocation allocation = __get_gear_allocation(object);
	float alpha;
	
	alpha = priv->gearsAlpha*object->alpha;
	
	cairo_translate (cr, allocation.x, allocation.y);
	
	cairo_save(cr);
		cairo_translate(cr, dock->continuousProgress.size/2, dock->continuousProgress.size/2);
		cairo_rotate(cr, priv->gearsRotateAngle*M_PI);
		cairo_translate(cr, -dock->continuousProgress.size/2, -dock->continuousProgress.size/2);
		
		cairo_set_source_surface (cr, dock->gearSurface, 0, 0);
		cairo_paint_with_alpha(cr, alpha);
	cairo_restore(cr);
	
	cairo_set_source_surface (cr, dock->gearSurfaceOverlay, 0, 0);
	cairo_paint_with_alpha(cr, alpha);
}

/**
 * kiba_drawable_object_paint_shadow:
 * @object: The KibaDrawableObject instance.
 * @cr: The drawing context.
 *
 * Renders the object shadow to the context.
 */
void kiba_drawable_object_paint_shadow (KibaDrawableObject *drawble_object,
                                        cairo_t *cr) {
	KibaDrawableObjectPrivate *priv;
	KibaObject *object;
	cairo_surface_t *surface;
	float offset_X, offset_y;
	float rad;
	float rotateAngle = 0.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawble_object));
	g_return_if_fail (cr != NULL);
	
	object = KIBA_OBJECT(drawble_object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawble_object);
	
	if (drawble_object->switch_surface)
		surface = drawble_object->switch_surface;
	else
		surface = drawble_object->surface_data->surface;
	
	g_return_if_fail (surface != NULL);
		
	for (l = drawble_object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaDrawableObjectCustomRotateData *data = l->data;
		
		data->func (drawble_object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	rad = deg_to_rad_factor*drawble_object->shadow_affection;
	offset_X = sin(rad)*drawble_object->shadow_offset;
	offset_y = cos(rad)*drawble_object->shadow_offset;
	
	/* dont rotate the translation, the light is fixed ;) */
	cairo_rotate_centered (cr,
		(2.0-rotateAngle),
		GTK_WIDGET(drawble_object)->allocation.width /2,
		GTK_WIDGET(drawble_object)->allocation.height/2);
	cairo_translate (cr, offset_X, offset_y);
	cairo_rotate_centered (cr,
		rotateAngle,
		GTK_WIDGET(drawble_object)->allocation.width /2,
		GTK_WIDGET(drawble_object)->allocation.height/2);
	
	cairo_save (cr);
		cairo_set_source_rgba (cr, drawble_object->shadow_color.r,
			drawble_object->shadow_color.g, drawble_object->shadow_color.b,
			drawble_object->alpha*drawble_object->scroll_alpha*drawble_object->shadow_color.a);
		/* using alpha mask from the object surface */
		cairo_mask_surface (cr, surface, 0, 0);
	cairo_restore (cr);
}

/**
 * kiba_drawable_object_remove_custom_rotate_function_by_data:
 * @object: The KibaDrawableObject instance.
 * @user_data: The user data.
 *
 * Removes custom rotate function.
 */
void kiba_drawable_object_remove_custom_rotate_function_by_data (KibaDrawableObject *object,
                                                                 gpointer user_data) {
	KibaDrawableObjectCustomRotateData *data;
	GList *l;
	
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_rotate_funcs = g_list_remove (object->custom_rotate_funcs, l->data);
			
			return;
		}
	}
}

/**
 * kiba_drawable_object_add_custom_rotate_function:
 * @object: The KibaDrawableObject instance.
 * @func: The rotate method.
 * @user_data: The user data.
 *
 * Adds rotate function, for a custom rotate angle.
 */
void kiba_drawable_object_add_custom_rotate_function (KibaDrawableObject *object,
                                                      KibaDrawableObjectCustomRotateFunc  func,
                                                      gpointer user_data) {
	KibaDrawableObjectCustomRotateData *data;
	GList *l;
	
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->func == func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaDrawableObjectCustomRotateData, 1);
	data->func = func;
	data->data = user_data;
	
	object->custom_rotate_funcs = g_list_append (object->custom_rotate_funcs, data);
}

/**
 * kiba_drawable_object_remove_custom_paint_function_by_data:
 * @object: The KibaDrawableObject instance.
 * @user_data: The user data.
 *
 * Removes custom paint function.
 */
void kiba_drawable_object_remove_custom_paint_function_by_data (KibaDrawableObject *object,
                                                                gpointer user_data) {
	KibaDrawableObjectCustomPaintData *data;
	GList *l;
	
	for (l = object->custom_paint_funcs_before; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_paint_funcs_before =
				g_list_remove (object->custom_paint_funcs_before, l->data);
			
			return;
		}
	}
	
	for (l = object->custom_paint_funcs_after; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_paint_funcs_after =
				g_list_remove (object->custom_paint_funcs_after, l->data);
			
			return;
		}
	}
}

/**
 * kiba_drawable_object_add_custom_paint_function:
 * @object: The KibaDrawableObject instance.
 * @func: The paint method.
 * @before: TRUE to call this method before the default rendering method, otherwise FALSE.
 * @user_data: The user data.
 *
 * Adds paint function, for custom rendering ontop/below the object.
 */
void kiba_drawable_object_add_custom_paint_function (KibaDrawableObject *object,
                                                     KibaDrawableObjectPaintFunc func,
                                                     gboolean before,
                                                     gpointer user_data) {
	KibaDrawableObjectCustomPaintData *data;
	GList **data_list;
	GList *l;
	
	if (before)
		data_list = &object->custom_paint_funcs_before;
	else
		data_list = &object->custom_paint_funcs_after;
	
	for (l = *data_list; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->func == func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaDrawableObjectCustomPaintData, 1);
	data->func = func;
	data->data = user_data;
	
	*data_list = g_list_append (*data_list, data);
}

/*** Rendering functions ***/

/**
 * kiba_drawable_object_scale:
 * @object: The KibaDrawableObject instance.
 * @cr: The drawing context.
 *
 * Scales the context by object's scale factor(s).
 *
 * Returns: TRUE if the scale factor is > 0.0, otherwise FALSE is returned.
 */
gboolean kiba_drawable_object_scale (KibaDrawableObject *object,
                                     cairo_t *cr) {
	KibaDrawableObjectPrivate *priv;
	KibaObject *obj;
	float factor_x = 1.0f;
	float factor_y = 1.0f;
	float scale_x;
	float scale_y;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(object));
	g_return_if_fail (cr != NULL);
	
	obj = KIBA_OBJECT(object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	scale_x = 0.0f;
	scale_y = 0.0f;
	kiba_object_get_scale_factor(obj, &scale_x, &scale_y);
	if (scale_x == 0.0f || scale_y == 0.0f)
		return FALSE;
	
	for (l = obj->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (obj,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	cairo_translate (cr,
		((1.0-factor_x)*GTK_WIDGET(object)->allocation.width)/2,
		((1.0-factor_y)*GTK_WIDGET(object)->allocation.height)/2);
	cairo_scale (cr, scale_x, scale_y);
	
	return TRUE;
}

/**
 * kiba_drawable_object_rotate:
 * @object: The KibaDrawableObject instance.
 * @cr: The drawing context.
 *
 * Rotates the context by object's rotate angle.
 */
void kiba_drawable_object_rotate (KibaDrawableObject *drawable_object,
                                  cairo_t *cr) {
	KibaDrawableObjectPrivate *priv;
	KibaObject *object;
	gint unscaled_width;
	gint unscaled_height;
	float rotateAngle = 0.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	g_return_if_fail (cr != NULL);
	
	object = KIBA_OBJECT(drawable_object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
		
	for (l = drawable_object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaDrawableObjectCustomRotateData *data = l->data;
		
		data->func (drawable_object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	unscaled_width  = (int) ((object->size.width) * object->scale_x);
	unscaled_height = (int) ((object->size.height) * object->scale_y);
	
	if (rotateAngle == 0.0f)
		return;
	
	cairo_rotate_centered (cr,
		rotateAngle,
		unscaled_width/2,
		unscaled_height/2);
}

/**
 * kiba_drawable_object_paint_with_prelight:
 * @object: The KibaDrawableObject instance.
 * @surface: The surface to paint with prelight.
 * @cr: The drawing context.
 *
 * Renders the context with prelight.
 */
void kiba_drawable_object_paint_with_prelight (KibaDrawableObject *drawable_object,
                                               cairo_surface_t *surface,
                                               cairo_t *cr) {
	KibaDrawableObjectPrivate *priv;
	KibaObject *object;
	KibaDock *dock;
	float alpha;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT(drawable_object));
	g_return_if_fail (GTK_WIDGET(drawable_object)->parent != NULL);
	g_return_if_fail (cr != NULL);
	
	object = KIBA_OBJECT(drawable_object);
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (drawable_object);
	alpha = drawable_object->alpha * drawable_object->scroll_alpha;
	dock = KIBA_DOCK(GTK_WIDGET(drawable_object)->parent);
	
	cairo_paint_with_alpha (cr, alpha);
	
	if (!GET_FLAG (dock->flags, KIBA_DOCK_DRAW_PRELIGHT))
		return;
	if (!GET_FLAG(object->state, KIBA_OBJECT_PRELIGHTED))
		return;
	
	cairo_set_source_rgba (cr, dock->style.prelight_rgba.r,
		dock->style.prelight_rgba.g,
		dock->style.prelight_rgba.b,
		alpha * dock->style.prelight_rgba.a);
	cairo_mask_surface (cr, surface, 0, 0);
}

/**
 * kiba_drawable_object_paint:
 * @object: The KibaDrawableObject instance.
 * @cr: The drawing context.
 * @data: The user data.
 *
 * Default rendering method.
 */
void kiba_drawable_object_paint (KibaDrawableObject *object,
                                 cairo_t *cr,
                                 gpointer data) {
	cairo_surface_t *surface;
	
	g_return_if_fail (KIBA_IS_DRAWABLE_OBJECT (object));
	
	if (object->switch_surface)
		surface = object->switch_surface;
	else
		surface = object->surface_data->surface;
	g_return_if_fail (surface != NULL);
	
	cairo_set_source_surface (cr, surface, 0, 0);
	kiba_drawable_object_paint_with_prelight (object, surface, cr);
}

/*** Class Implementations ***/

/**
 * gtk_widget_size_allocate implementation.
 **/
static void kiba_drawable_object_size_allocate (GtkWidget *widget,
                                                GtkAllocation *allocation) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	KibaDock *dock = KIBA_DOCK(widget->parent);
	
	if (widget->allocation.width != allocation->width ||
	    widget->allocation.height != allocation->height ||
	    widget->allocation.y != allocation->y ||
	    widget->allocation.x != allocation->x) {
		kiba_drawable_object_queue_draw (object);
		
		widget->allocation = *allocation;
		
		kiba_drawable_object_queue_draw (object);
	}
	
	if (GDK_IS_WINDOW(priv->event_window))
		gdk_window_move_resize (priv->event_window,
				widget->allocation.x,
				widget->allocation.y,
				widget->allocation.width,
				widget->allocation.height);
	
	__place_object_reflection(object, kiba_dock_get_dragging_window(dock));
}

/**
 * gtk_widget_unmap implementation.
 **/
static void kiba_drawable_object_unmap (GtkWidget *widget) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_hide (priv->event_window);
	
	if (bin->child && GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_unmap (bin->child);
}

/**
 * gtk_widget_map implementation.
 **/
static void kiba_drawable_object_map (GtkWidget *widget) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_show (priv->event_window);
	
	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void kiba_drawable_object_unrealize (GtkWidget *widget) {
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	GList *l;
	
	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	if (priv->fade_id)
		g_source_remove (priv->fade_id);
	if (priv->rotateGearsID)
		g_source_remove (priv->rotateGearsID);
	
	if (object->surface_data) {
		if (object->surface_data->surface)
			cairo_surface_destroy (object->surface_data->surface);
		g_free (object->surface_data);
	}
	
	for (l = object->custom_paint_funcs_after; l != NULL; l = object->custom_paint_funcs_after) {
		KibaDrawableObjectCustomPaintData *data = l->data;
		
		g_free (data);
		object->custom_paint_funcs_after = g_list_remove (object->custom_paint_funcs_after, l->data);
	}
	g_list_free (object->custom_paint_funcs_after);
	
	for (l = object->custom_paint_funcs_before; l != NULL; l = object->custom_paint_funcs_before) {
		KibaDrawableObjectCustomPaintData *data = l->data;
		
		g_free (data);
		object->custom_paint_funcs_before =
			g_list_remove (object->custom_paint_funcs_before, l->data);
	}
	g_list_free (object->custom_paint_funcs_before);
	
	(* GTK_WIDGET_CLASS (kiba_drawable_object_parent_class)->unrealize) (widget);
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_drawable_object_realize (GtkWidget *widget) {
	KibaDrawableObjectPrivate *priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (widget);
	KibaDrawableObject *object = KIBA_DRAWABLE_OBJECT(widget);
	Kiba *kiba = kiba_get();
	GdkWindowAttr attributes;
	gint attributes_mask;
	gint border_width;
	GtkRequisition requisition;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	border_width = GTK_CONTAINER (widget)->border_width;
	
	gtk_widget_size_request (widget, &requisition);
	
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window); /* increase the reference count of the window */
	
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	
	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	gdk_window_set_user_data (priv->event_window, widget); /* catching the window events on the widget */
	gdk_window_register_dnd (priv->event_window); /* allowing dnd an the window */
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	object->surface_data = g_new0 (CairoSurfaceData, 1);
	object->surface_data->width  = requisition.width;
	object->surface_data->height = requisition.height;
	object->surface_data->surface = NULL;
	
	priv->fade_id = g_timeout_add (20, __fade_timeout, object);
}

static void kiba_drawable_object_parent_set(GtkWidget *widget,
                                 GtkWidget *old_parent) {
	if (!widget->parent) return;
	KibaDock *dock = KIBA_DOCK(widget->parent);
	g_signal_connect(G_OBJECT(widget->parent), "object_size_changed",
		G_CALLBACK(__dock_object_size_changed_cb), widget);
	KIBA_OBJECT(widget)->size = (GtkRequisition) {dock->d.object_size, dock->d.object_size};
}

/*** Class init ***/

static void kiba_drawable_object_class_init (KibaDrawableObjectClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaDrawableObjectPrivate));
	
	/* GtkWidget method implementations */
	widget_class->realize = kiba_drawable_object_realize;
	widget_class->unrealize = kiba_drawable_object_unrealize;
	widget_class->map = kiba_drawable_object_map;
	widget_class->unmap = kiba_drawable_object_unmap;
	widget_class->size_allocate = kiba_drawable_object_size_allocate;
	widget_class->parent_set = kiba_drawable_object_parent_set;
	
	/* KibaDrawableObject signals */
	kiba_drawable_object_signals[SIG_SWITCH_EVENT] =
		g_signal_new ("switch-event",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaDrawableObjectClass, switch_event),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
}

static void kiba_drawable_object_init (KibaDrawableObject *object) {
	KibaDrawableObjectPrivate *priv;
	
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);
	
	priv = KIBA_DRAWABLE_OBJECT_GET_PRIVATE (object);
	
	object->paint = kiba_drawable_object_paint;
	object->surface_data = NULL;
	object->switch_surface = NULL;
	object->alpha = 0.1f;
	object->scroll_alpha = 1.0f;
	object->custom_paint_funcs_after = NULL;
	object->custom_paint_funcs_before = NULL;
	object->shadow_offset = 5;
	
	priv->gearsAlpha = 0.0f;
	priv->gearsRotateAngle = 0.0f;
	priv->fade_alpha = 1.0f;
	
	priv->reflection_allocation =  (GtkAllocation) { 0, 0, 0, 0 };
	
	object->shadow_color = (KibaRGBA) { 0.0f, 0.0f, 0.0f, 0.6f };
	object->shadow_affection = 45;
	
	g_signal_connect(G_OBJECT(object), "working-toggled",
		G_CALLBACK(__working_toggled), NULL);
	g_signal_connect(G_OBJECT(object), "dragging-toggled",
		G_CALLBACK(__dragging_toggled), NULL);
}

/**
 * kiba_drawable_object_new:
 * @group: The group on the dock.
 *
 * Creates and returns a KibaDrawableObject container.
 *
 * Returns: The KibaDrawableObject instance created.
 */
GtkWidget* kiba_drawable_object_new (const char *group) {
	GtkWidget *object = g_object_new (KIBA_TYPE_DRAWABLE_OBJECT,
                                 "group", group, NULL);
	return object;
}
