
/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-fixed-object
 * @short_description: A fixed #KibaObject.
 * @see_also:#KibaWin,#KibaDock
 * @stability: Unstable
 *
 * A fixed #KibaObject.
 */

#include "kiba-fixed-object.h"
#include "kiba-types.h"
#include "kiba-dock.h"

#define KIBA_FIXED_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_FIXED_OBJECT, KibaFixedObjectPrivate))

G_DEFINE_TYPE (KibaFixedObject, kiba_fixed_object, KIBA_TYPE_OBJECT)

typedef struct {
	GdkWindow *event_window;
} KibaFixedObjectPrivate;

/*** private ***/

static void __dock_height_changed_cb (KibaDock *dock,
                                      gint new_height,
                                      gpointer data) {
	KibaFixedObject *object = data;
	
	KIBA_OBJECT(object)->size.width = new_height;
	kiba_fixed_object_set_size(object, KIBA_OBJECT(object)->size.height);
	gtk_widget_queue_resize(GTK_WIDGET(object));
}

/*** public ***/

/**
 * kiba_fixed_object_set_size:
 * @fixed_object: The KibaFixedObject instance.
 * @size: The size of the object.
 *
 * Sets the size of the fixed object.
 */
void kiba_fixed_object_set_size (KibaFixedObject *fixed_object,
				 gint             size) {
	KibaObject *object;
	float scale;
	
	g_return_if_fail (KIBA_IS_FIXED_OBJECT(fixed_object));
	
	object = KIBA_OBJECT(fixed_object);
	object->size.height = size;
	scale = (float) size/object->size.width;
	kiba_object_set_requisition_factor (object, scale, scale);
}

/*** widget implementations ***/

static void kiba_fixed_object_size_request (GtkWidget *widget,
                                            GtkRequisition *requisition) {
	KibaObject *object;
	GtkWidget *dock;
	const KibaLayoutData *d;
	GList *l;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	
	g_return_if_fail (KIBA_IS_FIXED_OBJECT(widget));
	
	object = KIBA_OBJECT(widget);
	dock = widget->parent;
	
	if (!dock)
		d = &kiba_layout_data[KIBA_POSITION_BOTTOM_EDGE];
	else
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	scale_x = object->scale_x;
	scale_y = object->scale_y;
	requisition->width = object->size.width;
	requisition->height = object->size.width;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (!data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		scale_x *= custom_factor_x;
		scale_y *= custom_factor_y;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		requisition->width  = (int) requisition->width*scale_x;
	else
		requisition->height = (int) requisition->width*scale_y;
}

/**
 * gtk_widget_size_allocate implementation.
 **/
static void kiba_fixed_object_size_allocate (GtkWidget *widget,
                                             GtkAllocation *allocation) {
	KibaFixedObjectPrivate *priv = KIBA_FIXED_OBJECT_GET_PRIVATE (widget);
	KibaFixedObject *object = KIBA_FIXED_OBJECT(widget);
	
	if (GDK_IS_WINDOW(priv->event_window))
		gdk_window_move_resize (priv->event_window,
				widget->allocation.x,
				widget->allocation.y,
				widget->allocation.width,
				widget->allocation.height);
	
	(* GTK_WIDGET_CLASS (kiba_fixed_object_parent_class)->size_allocate) (widget, allocation);
}

static void kiba_fixed_object_parent_set(GtkWidget *widget,
                                         GtkWidget *old_parent) {
	if (!widget->parent) return;
	KibaDock *dock = KIBA_DOCK(widget->parent);
	g_signal_connect(G_OBJECT(dock), "height-changed",
		G_CALLBACK(__dock_height_changed_cb), widget);
	KIBA_OBJECT(widget)->size = (GtkRequisition) {dock->d.height, dock->d.height};
	gtk_widget_queue_resize(widget);
}

/**
 * gtk_widget_unmap implementation.
 **/
static void kiba_fixed_object_unmap (GtkWidget *widget) {
	KibaFixedObjectPrivate *priv = KIBA_FIXED_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_hide (priv->event_window);
	
	if (bin->child && GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_unmap (bin->child);
}

/**
 * gtk_widget_map implementation.
 **/
static void kiba_fixed_object_map (GtkWidget *widget) {
	KibaFixedObjectPrivate *priv = KIBA_FIXED_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_show (priv->event_window);
	
	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void kiba_fixed_object_unrealize (GtkWidget *widget) {
	KibaFixedObject *object = KIBA_FIXED_OBJECT(widget);
	KibaFixedObjectPrivate *priv = KIBA_FIXED_OBJECT_GET_PRIVATE (widget);
	
	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	(* GTK_WIDGET_CLASS (kiba_fixed_object_parent_class)->unrealize) (widget);
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_fixed_object_realize (GtkWidget *widget) {
	KibaFixedObjectPrivate *priv = KIBA_FIXED_OBJECT_GET_PRIVATE (widget);
	KibaFixedObject *object = KIBA_FIXED_OBJECT(widget);
	GdkWindowAttr attributes;
	gint attributes_mask;
	gint border_width;
	GtkRequisition requisition;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	border_width = GTK_CONTAINER (widget)->border_width;
	
	gtk_widget_size_request (widget, &requisition);
	
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window); /* increase the reference count of the window */
	
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	
	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	gdk_window_set_user_data (priv->event_window, widget); /* catching the window events on the widget */
	gdk_window_register_dnd (priv->event_window); /* allowing dnd an the window */
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
}

/*** Init/Fini ***/

static void kiba_fixed_object_class_init (KibaFixedObjectClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaFixedObjectPrivate));
	
	widget_class->size_request = kiba_fixed_object_size_request;
	widget_class->parent_set = kiba_fixed_object_parent_set;
	widget_class->realize = kiba_fixed_object_realize;
	widget_class->unrealize = kiba_fixed_object_unrealize;
	widget_class->map = kiba_fixed_object_map;
	widget_class->unmap = kiba_fixed_object_unmap;
	widget_class->size_allocate = kiba_fixed_object_size_allocate;
}

static void kiba_fixed_object_init (KibaFixedObject *fixed_object) {
	KibaObject *object = KIBA_OBJECT(fixed_object);
	object->size = (GtkRequisition) {0,0};
}

/**
 * kiba_fixed_object_new:
 *
 * Creates and returns a KibaFixedObject Instance.
 *
 * Returns: The KibaFixedObject object created.
 */
GtkWidget* kiba_fixed_object_new (const char *group) {
	GtkWidget *widget =  g_object_new (KIBA_TYPE_FIXED_OBJECT,
                                 "group", group, NULL);
	return widget;	
}
