/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-icon-view-win
 * @short_description: Window containing a #KibaIconView instance.
 * @see_also:#KibaIconView
 * @stability: Unstable
 *
 * Window containing a #KibaIconView instance.
 */

#include <string.h>

#include "kiba-utils.h"
#include "kiba-icon-view-win.h"

#include "marshallers.h"

#define VIEW_WIN_COLUMNS(view_win) kiba_icon_view_get_columns(KIBA_ICON_VIEW(view_win->icon_view))
#define KIBA_ICON_VIEW_MAX_ICONS(view_win)  view_win->rows*VIEW_WIN_COLUMNS(view_win)

G_DEFINE_TYPE (KibaIconViewWin, kiba_icon_view_win, G_TYPE_OBJECT)

static void kiba_icon_view_win_destroy_cb (GtkWidget *widget,
				           gpointer   data) {
	KibaIconViewWin *view_win = data;
	
	g_object_unref (view_win);
}

static void __toggle_up_arrow (KibaIconView *icon_view,
                               gboolean show_arrow,
                               gpointer data) {
	KibaIconViewWin *view_win = data;
	if (show_arrow) {
		gtk_widget_show(view_win->up_arrow);
	} else {
		gtk_widget_hide(view_win->up_arrow);
	}
}

static void __toggle_down_arrow (KibaIconView *icon_view,
                               gboolean show_arrow,
                               gpointer data) {
	KibaIconViewWin *view_win = data;
	if (show_arrow) {
		gtk_widget_show(view_win->down_arrow);
	} else {
		gtk_widget_hide(view_win->down_arrow);
	}
}

static gboolean kiba_icon_view_win_button_release_cb (GtkWidget      *widget,
					              GdkEventButton *event,
					              gpointer        data) {
	KibaIconViewWin *view_win = (KibaIconViewWin*) data;
	
	if (event->button != 2)
		return FALSE;
	
	view_win->win = NULL;
	gtk_widget_destroy (widget);
	
	return TRUE;
}

static gboolean icon_view_arrow_release_cb (GtkWidget      *widget,
					    GdkEventButton *event,
					    gpointer        data) {
	KibaIconViewWin *view_win = (KibaIconViewWin*) data;
	KibaIconView *icon_view = KIBA_ICON_VIEW(view_win->icon_view);
	GtkWidget *child;
	gint i, columns;
	GList *l;
	
	columns = VIEW_WIN_COLUMNS(view_win);
	
	if (widget == view_win->up_arrow->parent) {
		kiba_icon_view_hide_first_row(icon_view);
	}
	else if (widget == view_win->down_arrow->parent) {
		kiba_icon_view_hide_last_row(icon_view);
	}
	
	return TRUE;
}

static void kiba_icon_view_size_allocate_cb (GtkWidget     *widget,
					     GtkAllocation *allocation,
					     gpointer       data) {
	gtk_widget_queue_draw (((KibaIconViewWin*) data)->win);
}

/**
Public Functions
**/

/**
 * kiba_icon_view_win_add:
 * @view_win: The KibaIconViewWin instance.
 * @widget: The widget to add.
 *
 * Adds widget to the grid of view_win.
 */
void kiba_icon_view_win_add (KibaIconViewWin *view_win,
			     GtkWidget       *widget) {
	KibaIconView *icon_view;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW_WIN(view_win));
	g_return_if_fail (GTK_IS_WIDGET(widget));
	
	icon_view = KIBA_ICON_VIEW(view_win->icon_view);
	
	gtk_container_add (GTK_CONTAINER(icon_view), widget);
	if (kiba_icon_view_get_active_count(icon_view) < KIBA_ICON_VIEW_MAX_ICONS(view_win)) {
		/* we cant just modify the just created icon, because the position in the list may switched */
		kiba_icon_view_show_n_icons(icon_view, KIBA_ICON_VIEW_MAX_ICONS(view_win));
	}
	else {
		if (!GTK_WIDGET_MAPPED(view_win->arrow_box))
			gtk_widget_show (view_win->arrow_box);
	}
}

/**
 * kiba_icon_view_win_remove:
 * @view_win: The KibaIconViewWin instance.
 * widget: The widget to remove.
 *
 * Removes widget from view_win's grid.
 */
void kiba_icon_view_win_remove (KibaIconViewWin *view_win,
			        GtkWidget       *widget) {
	KibaIconView *icon_view;
	gint active, before, after;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW_WIN(view_win));
	
	icon_view = KIBA_ICON_VIEW(view_win->icon_view);
	
	gtk_container_remove (GTK_CONTAINER(icon_view), widget);
	
	active = kiba_icon_view_get_active_count(icon_view);
	before = kiba_icon_view_get_before_count(icon_view);
	after = kiba_icon_view_get_after_count(icon_view);
	if (active <= KIBA_ICON_VIEW_MAX_ICONS(view_win) - kiba_icon_view_get_columns(icon_view)) {
		if (after > 0) {
			kiba_icon_view_hide_first_row(icon_view);
		}
		else if (before > 0) {
			kiba_icon_view_hide_last_row(icon_view);
		}
	}
	
	/* resize immediately */
	GtkAllocation alloc;
	alloc = view_win->icon_view->allocation;
	gtk_widget_queue_resize (view_win->icon_view);
	gtk_widget_size_allocate(view_win->icon_view, &alloc);
}

static void icon_view_close_cb (GtkWidget      *widget,
				GdkEventButton *button,
				gpointer        user_data) {
	KibaIconViewWin *view_win = user_data;
	
	gtk_widget_destroy (view_win->win);
}

/**
Class Functions
**/

static void kiba_icon_view_win_class_init (KibaIconViewWinClass *class) {
}

static void kiba_icon_view_win_init (KibaIconViewWin *view_win) {
	
}

/**
 * kiba_icon_view_win_new:
 * @position: The position of the boxes for custom widgets.
 * @columns: The number of columns in the grid.
 * @rows: The number of rows in the grid.
 * @close_button: TRUE if a close button should be added, otherwise FALSE
 *
 * Creates and returns a KibaIconViewWin instance.
 *
 * Returns: The KibaIconViewWin instance created.
 */
KibaIconViewWin* kiba_icon_view_win_new (KibaIconViewBoxPosition position,
					 gint                    columns,
					 gint                    rows,
					 gboolean                close_button) {
	KibaIconViewWin *view_win;
	GtkWidget *vbox;
	GtkWidget *mainbox;
	GtkWidget *arrow;
	GtkWidget *image;
	GtkWidget *event_box;
	GtkWidget *arrow_box;
	
	view_win = g_object_new (KIBA_TYPE_ICON_VIEW_WIN, NULL);
	
	view_win->rows = rows;
	
	view_win->win = kiba_info_win_new();
	gtk_window_set_decorated (GTK_WINDOW(view_win->win), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(view_win->win), FALSE);
	
	mainbox = gtk_hbox_new (FALSE, 0);
	view_win->left_box   = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(mainbox), view_win->left_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->left_box);
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show (vbox);
	view_win->right_box  = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(mainbox), view_win->right_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->right_box);
	
	view_win->top_box    = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), view_win->top_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->top_box);
	view_win->bottom_box = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_end (GTK_BOX(vbox), view_win->bottom_box, FALSE, FALSE, 0);
	gtk_widget_show (view_win->bottom_box);
	
	view_win->icon_view = kiba_icon_view_new();
	kiba_icon_view_set_columns (KIBA_ICON_VIEW(view_win->icon_view), columns);
	gtk_widget_show (view_win->icon_view);
	gtk_box_pack_start (GTK_BOX(vbox), view_win->icon_view, FALSE, FALSE, 0);
	
	if (position == KIBA_ICON_VIEW_BOX_POSITION_RIGHT)
		arrow_box = view_win->right_box;
	else
		arrow_box = view_win->left_box;
	
	view_win->arrow_box = gtk_vbox_new (FALSE, 0);
	
	if (close_button) {
		event_box = gtk_event_box_new ();
		gtk_event_box_set_visible_window (GTK_EVENT_BOX(event_box), FALSE);
		image = gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_widget_set_usize (event_box, 24, -1);
		gtk_container_add (GTK_CONTAINER (event_box), image);
		gtk_box_pack_start (GTK_BOX(view_win->arrow_box), event_box, FALSE, FALSE, 5);
		gtk_widget_show (image);
		gtk_widget_show (event_box);
		g_signal_connect (G_OBJECT(event_box), "button-press-event",
			G_CALLBACK (icon_view_close_cb), view_win);
	}
	
	event_box = gtk_event_box_new ();
	gtk_event_box_set_visible_window (GTK_EVENT_BOX(event_box), FALSE);
	view_win->up_arrow = gtk_image_new_from_stock (GTK_STOCK_GO_UP, GTK_ICON_SIZE_MENU);
	gtk_misc_set_alignment (GTK_MISC(view_win->up_arrow), 0.5, 0.95);
	gtk_widget_set_usize (event_box, 24, -1);
	gtk_container_add (GTK_CONTAINER (event_box), view_win->up_arrow);
	gtk_box_pack_start (GTK_BOX(view_win->arrow_box), event_box, TRUE, TRUE, 0);
	gtk_widget_show (event_box);
	g_signal_connect (G_OBJECT(event_box), "button-release-event",
		G_CALLBACK (icon_view_arrow_release_cb), view_win);
	
	event_box = gtk_event_box_new ();
	gtk_event_box_set_visible_window (GTK_EVENT_BOX(event_box), FALSE);
	view_win->down_arrow = gtk_image_new_from_stock (GTK_STOCK_GO_DOWN, GTK_ICON_SIZE_MENU);
	gtk_misc_set_alignment (GTK_MISC(view_win->down_arrow), 0.5, 0.05);
	gtk_widget_set_usize (event_box, 24, -1);
	gtk_container_add (GTK_CONTAINER (event_box), view_win->down_arrow);
	gtk_box_pack_start (GTK_BOX(view_win->arrow_box), event_box, TRUE, TRUE, 0);
	gtk_widget_show (event_box);
	g_signal_connect (G_OBJECT(event_box), "button-release-event",
		G_CALLBACK (icon_view_arrow_release_cb), view_win);
	
	if (close_button)
		gtk_widget_show (view_win->arrow_box);
	gtk_box_pack_start (GTK_BOX(arrow_box), view_win->arrow_box, TRUE, TRUE, 0);
	
	gtk_container_add (GTK_CONTAINER(view_win->win), mainbox);
	gtk_widget_show (mainbox);
	
	g_signal_connect (G_OBJECT(view_win->win), "button-release-event",
		G_CALLBACK (kiba_icon_view_win_button_release_cb), view_win);
	g_signal_connect_after (G_OBJECT(view_win->win), "destroy",
		G_CALLBACK (kiba_icon_view_win_destroy_cb), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "size-allocate",
		G_CALLBACK(kiba_icon_view_size_allocate_cb), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "children-before-toggled",
		G_CALLBACK(__toggle_down_arrow), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "children-after-toggled",
		G_CALLBACK(__toggle_up_arrow), view_win);
	g_signal_connect(G_OBJECT (view_win->icon_view), "size-allocate",
		G_CALLBACK(kiba_icon_view_size_allocate_cb), view_win);
	
	return view_win;
}
