/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-icon-view
 * @short_description: Icon grid.
 * @see_also:#KibaIconViewWin
 * @stability: Unstable
 *
 * Container realizing a icon grid.
 */

#include "config.h"

#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "kiba-icon-view.h"
#include "kiba-cairo-utils.h"
#include "kiba-types.h"
#include "kiba-image.h"

#define KEYPRESS_CLEAR_DELAY 1000

#define KIBA_ICON_VIEW_FIRST_ACTIVE(view) icon_view->icons_before_count
#define KIBA_ICON_VIEW_FIRST_AFTER(view) KIBA_ICON_VIEW_FIRST_ACTIVE(view)+icon_view->active_icon_count
#define KIBA_ICON_VIEW_LAST_BEFORE(view) KIBA_ICON_VIEW_FIRST_ACTIVE(view)-1
#define KIBA_ICON_VIEW_LAST_ACTIVE(view) KIBA_ICON_VIEW_FIRST_AFTER(view)-1

struct _KibaIconView {
	GtkBox box;
	
	GList *children;        /**< list contains all KibaIconViewChild's */
	gint icons_before_count;
	gint active_icon_count;
	gint icons_after_count;
	
	GtkWidget *selected_child;   /**< The selected child */
	gint columns;          /**< Grid columns */
	gboolean fixed;
	gboolean prelight_ontop;
	gboolean prelightable;
	gboolean selected_image;
	
	GCompareFunc sort_func;
	
	gint keypress_clear_id;
	char *keypress_prefix;
};

struct _KibaIconViewClass {
	GtkBoxClass  parent_class;
	
	/* signals */
	void (* child_activated) (KibaIconView *icon_view,
			          GtkWidget    *child);
	void (* child_selected) (KibaIconView *icon_view,
			         GtkWidget    *child);
	void (* child_unselected) (KibaIconView *icon_view,
			           GtkWidget    *child);
	void (* children_before_toggled) (KibaIconView *icon_view,
                                      gboolean children_before);
	void (* children_after_toggled) (KibaIconView *icon_view,
                                     gboolean childrens_after);
};

G_DEFINE_TYPE (KibaIconView, kiba_icon_view, GTK_TYPE_BOX)

/* signal ids */
enum {
	SIG_CHILD_ACTIVATED = 0,
	SIG_CHILD_SELECTED,
	SIG_CHILD_UNSELECTED,
	SIG_CHILDREN_BEFORE_TOGGLED,
	SIG_CHILDREN_AFTER_TOGGLED,
	SIG_KIBA_ICON_VIEW_LAST
};

static guint kiba_icon_view_signals[SIG_KIBA_ICON_VIEW_LAST] = { 0 };

/*** private methods ***/

static gint __clear_keypress_prefix (gpointer data) {
	KibaIconView *icon_view = data;
	if (icon_view->keypress_prefix) {
		g_free(icon_view->keypress_prefix);
		icon_view->keypress_prefix = NULL;
	}
	icon_view->keypress_clear_id = 0;
	return FALSE;
}

static void __append_keypress_prefix (KibaIconView *icon_view,
                                      const char *prefix) {
	if (icon_view->keypress_prefix) {
		char *old_prefix = icon_view->keypress_prefix;
		icon_view->keypress_prefix = g_strdup_printf("%s%s", old_prefix, prefix);
		g_free (old_prefix);
	} else {
		icon_view->keypress_prefix = g_strdup(prefix);
	}
	
	if (icon_view->keypress_clear_id)
		g_source_remove(icon_view->keypress_clear_id);
	icon_view->keypress_clear_id = g_timeout_add (KEYPRESS_CLEAR_DELAY, __clear_keypress_prefix, icon_view);
}

static void __change_after_count(KibaIconView *icon_view,
                                 gboolean increase) {
	if (increase) {
		icon_view->icons_after_count++;
		if (icon_view->icons_after_count == 1)
			g_signal_emit (icon_view,
				kiba_icon_view_signals[SIG_CHILDREN_AFTER_TOGGLED],
				0, TRUE);
	} else {
		icon_view->icons_after_count--;
		if (icon_view->icons_after_count == 0)
			g_signal_emit (icon_view,
				kiba_icon_view_signals[SIG_CHILDREN_AFTER_TOGGLED],
				0, FALSE);
	}
}

static void __change_before_count(KibaIconView *icon_view,
                                  gboolean increase) {
	if (increase) {
		icon_view->icons_before_count++;
		if (icon_view->icons_before_count == 1)
			g_signal_emit (icon_view,
				kiba_icon_view_signals[SIG_CHILDREN_BEFORE_TOGGLED],
				0, TRUE);
	} else {
		icon_view->icons_before_count--;
		if (icon_view->icons_before_count == 0)
			g_signal_emit (icon_view,
				kiba_icon_view_signals[SIG_CHILDREN_BEFORE_TOGGLED],
				0, FALSE);
	}
}

static GtkLabel* __get_label_child (GtkWidget *widget) {
	if (GTK_IS_LABEL(widget)) {
		return GTK_LABEL(widget);
	}
	if (GTK_IS_CONTAINER(widget)) {
		GtkLabel *label = NULL;
		GList *children, *l;
		
		children = gtk_container_get_children(GTK_CONTAINER(widget));
		for (l = children; l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			label = __get_label_child(child);
			if (label) break;
		}
		g_list_free(children);
		
		return label;
	}
	
	return NULL;
}

static KibaIconViewChild* __get_child_for_widget (KibaIconView *icon_view,
                                                  GtkWidget *widget) {
	GList *l;
	
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *child = l->data;
		
		if (child->widget == widget)
			return child;
	}
	
	return NULL;
}

static void __select_container_images (KibaIconView *icon_view,
                                       GtkContainer *container) {
	GList *l;
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		if (KIBA_IS_IMAGE(child)) {
			kiba_image_select (KIBA_IMAGE(child), TRUE);
			icon_view->selected_image = TRUE;
		}
		else if (GTK_IS_CONTAINER(child)) {
			__select_container_images (icon_view, GTK_CONTAINER(child));
		}
	}
}

static void __mark_container_images (KibaIconView *icon_view,
                                     GtkContainer *container) {
	GList *l;
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		if (KIBA_IS_IMAGE(child)) {
			if (GTK_WIDGET(container) == icon_view->selected_child) {
				kiba_image_select (KIBA_IMAGE(child), TRUE);
				icon_view->selected_image = TRUE;
			}
			else if (kiba_image_is_selected(KIBA_IMAGE(child))) {
				kiba_image_select (KIBA_IMAGE(child), FALSE);
			}
		}
		else if (GTK_IS_CONTAINER(child)) {
			if (GTK_WIDGET(container) == icon_view->selected_child)
				__select_container_images (icon_view, GTK_CONTAINER(child));
			else
				__mark_container_images (icon_view, GTK_CONTAINER(child));
		}
	}
}

/* default method to sort the icons */
static gint __sort_by_label (gconstpointer aicon,
                             gconstpointer bicon) {
	const KibaIconViewChild *achild = aicon;
	const KibaIconViewChild *bchild = bicon;
	GtkLabel *alabel = __get_label_child(achild->widget);
	GtkLabel *blabel = __get_label_child(bchild->widget);
	if (!alabel) return -1;
	if (!blabel) return 1;
	const char *atext = gtk_label_get_text(alabel);
	const char *btext = gtk_label_get_text(blabel);
	return kiba_utils_sort_string_list(atext, btext);
}

static gboolean kiba_icon_view_expose_cb (GtkWidget *widget,
                                          GdkEventExpose *expose) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	cairo_t *cr = NULL;
	gint border_width;
	float radius;
	
	border_width = 5;
	radius = 3.0f;
	
	if (!GTK_IS_WIDGET(icon_view->selected_child))
		return FALSE;
	
	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_translate (cr,
		icon_view->selected_child->allocation.x,
		icon_view->selected_child->allocation.y);
	cairo_paint_shape (cr,
		KIBA_ROUNDED, KIBA_ROUNDED, KIBA_ROUNDED, KIBA_ROUNDED,
		border_width/2,
		border_width/2,
		icon_view->selected_child->allocation.width - border_width/2,
		icon_view->selected_child->allocation.height - border_width/2,
		radius);
	cairo_set_source_rgba (cr, 0, 0, 0, 0.8);
	cairo_set_line_width (cr, border_width);
	cairo_stroke_preserve(cr);
	cairo_set_source_rgba (cr, 1, 1, 1, 0.6);
	cairo_fill(cr);
	
	cairo_destroy (cr);
	
	return FALSE;
}

static gboolean kiba_icon_view_expose_event (GtkWidget *widget,
                                             GdkEventExpose *expose) {
	GtkContainer *container = GTK_CONTAINER(widget);
	GList *l;
	gboolean expose_outside_view;
	
	expose_outside_view = FALSE;
	
	if (expose->area.x < widget->allocation.x) {
		expose->area.x = widget->allocation.x;
		expose_outside_view = TRUE;
	}
	if (expose->area.x + expose->area.width > widget->allocation.x + widget->allocation.width) {
		expose->area.height = widget->allocation.x + widget->allocation.width - expose->area.x;
		expose_outside_view = TRUE;
	}
	if (expose->area.y < widget->allocation.y) {
		expose->area.y = widget->allocation.y;
		expose_outside_view = TRUE;
	}
	if (expose->area.y + expose->area.height > widget->allocation.y + widget->allocation.height) {
		expose->area.height = widget->allocation.y + widget->allocation.height - expose->area.y;
		expose_outside_view = TRUE;
	}
	if (expose_outside_view) {
		gdk_region_destroy (expose->region);
		expose->region = gdk_region_rectangle (&expose->area);
	}
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		gtk_container_propagate_expose (container, child,  expose);
	}
	
	return FALSE;
}

static gboolean __child_press_cb (GtkWidget *widget,
                                                   GdkEventButton *event,
                                                   gpointer data) {
	KibaIconView *icon_view = data;
	
	if (event->button == 2)
		return FALSE;
	
	if (event->type == GDK_2BUTTON_PRESS)
		kiba_icon_view_activate_child (icon_view, widget);
	else
		kiba_icon_view_select_child (icon_view, widget);
	
	return TRUE;
}

static void kiba_icon_view_children_size_allocate (GtkWidget *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	GtkContainer *container;
	GtkAllocation obj_allocation;
	GList *l;
	gint i;
	gint columns;
	gint spacing;
	gint border_width;
	gint height;
	gboolean child_allocation_changed;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(widget));
	spacing = gtk_box_get_spacing (GTK_BOX(widget));
	columns = kiba_icon_view_get_columns (KIBA_ICON_VIEW(widget));
	
	container = GTK_CONTAINER(widget);
	
	child_allocation_changed = FALSE;
	obj_allocation.x = spacing + border_width + widget->allocation.x;
	obj_allocation.y = spacing + border_width + widget->allocation.y;
	
	i = 1;
	height = 0;
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *view_child = l->data;
		GtkWidget *child = view_child->widget;
		GtkRequisition req;
		
		if (view_child->state != KIBA_ICON_VIEW_HIDE_BEFORE)
			continue;
		
		gtk_widget_size_request (child, &req);
		
		height = MAX(height, req.height + spacing);
		
		if (i == columns) {
			obj_allocation.y -= height;
			i = 0;
		}
		i++;
	}
	if (i > 1)
		obj_allocation.y -= height;
	
	i = 1;
	height = 0;
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *view_child = l->data;
		GtkWidget *child = view_child->widget;
		GtkRequisition req;
		
		gtk_widget_size_request (child, &req);
		
		obj_allocation.width  = req.width;
		obj_allocation.height = req.height;
		
		height = MAX(height, req.height + spacing);
		
		view_child->allocation = obj_allocation;
		child->allocation.width = req.width;
		child->allocation.height = req.height;
		
		gtk_widget_size_allocate (child, &obj_allocation);
		
		if (i == columns) {
			obj_allocation.y += height;
			obj_allocation.x = spacing + border_width + widget->allocation.x;
			i = 0;
		}
		else
			obj_allocation.x += obj_allocation.width + spacing;
		
		i++;
	}
}

/*** Public methods ***/

/**
 * kiba_icon_view_get_columns:
 * @icon_view: The KibaIconView instance.
 *
 * Returns the number of columns.
 *
 * Returns: The number of columns.
 */
gint kiba_icon_view_get_columns (KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), 0);
	
	return icon_view->columns;
}

/**
 * kiba_icon_view_set_columns:
 * @icon_view: The KibaIconView instance.
 * @columns: The number of columns.
 *
 * Sets the number of columns.
 */
void kiba_icon_view_set_columns (KibaIconView *icon_view,
                                 gint columns) {
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	g_return_if_fail (columns > 0);
	g_return_if_fail (columns < 20);
	
	icon_view->columns = columns;
	
	gtk_widget_queue_resize (GTK_WIDGET(icon_view));
}

/**
 * kiba_icon_view_activate_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child to activate.
 *
 * Activates child.
 */
void kiba_icon_view_activate_child (KibaIconView *icon_view,
                                    GtkWidget *child) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	children = gtk_container_get_children(GTK_CONTAINER(icon_view));
	g_return_if_fail (g_list_find(children, child) != NULL);
	
	if (icon_view->selected_child != child) {
		icon_view->selected_child = child;
		gtk_widget_queue_draw (child);
	}
	
	g_signal_emit (icon_view,
		kiba_icon_view_signals[SIG_CHILD_ACTIVATED],
		0, child);
}

/**
 * kiba_icon_view_hide_first_row:
 * @icon_view: The #KibaIconView instance.
 *
 * Hides the first active row, showing the first unshown row after the last active row.
 */
void kiba_icon_view_hide_first_row(KibaIconView *icon_view) {
	gint i;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	for (i = 0; i < icon_view->columns; i++) {
		KibaIconViewChild *active, *after;
		
		active = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_FIRST_ACTIVE(icon_view));
		if (!active) break;
		after = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_FIRST_AFTER(icon_view));
		if (!after) break;
		
		kiba_icon_view_show_child (icon_view, active->widget, KIBA_ICON_VIEW_HIDE_BEFORE);
		kiba_icon_view_show_child (icon_view, after->widget, KIBA_ICON_VIEW_SHOW);
	}
}

/**
 * kiba_icon_view_hide_last_row:
 * @icon_view: The #KibaIconView instance.
 *
 * Hides the last active row, showing the first unshown row before the first active row.
 */
void kiba_icon_view_hide_last_row(KibaIconView *icon_view) {
	gint i;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	//FIXME: there are maybe less the 4 items in the last row
	for (i = 0; i < icon_view->columns; i++) {
		KibaIconViewChild *active, *last_before;
		
		active = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_LAST_ACTIVE(icon_view));
		if (!active) break;
		last_before = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_LAST_BEFORE(icon_view));
		if (!last_before) break;
		
		kiba_icon_view_show_child (icon_view, active->widget, KIBA_ICON_VIEW_HIDE_AFTER);
		kiba_icon_view_show_child (icon_view, last_before->widget, KIBA_ICON_VIEW_SHOW);
	}
}

/**
 * kiba_icon_view_select_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child to select.
 *
 * Selects a icon view child.
 */
void kiba_icon_view_select_child (KibaIconView *icon_view,
                                  GtkWidget *child) {
	KibaIconViewChild *view_child;
	GList *children, *l;
	GtkWidget *old_selected;
	gint i;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	view_child = __get_child_for_widget (icon_view, child);
	g_return_if_fail(view_child != NULL);
	
	children = gtk_container_get_children(GTK_CONTAINER(icon_view));
	g_return_if_fail (g_list_find(children, child) != NULL);
	
	if (icon_view->selected_child == child)
		return;
	
	old_selected = icon_view->selected_child;
	icon_view->selected_child = child;
	
	if (GTK_IS_WIDGET(old_selected)) {
		gtk_widget_queue_draw (old_selected);
		
		g_signal_emit (icon_view,
			kiba_icon_view_signals[SIG_CHILD_UNSELECTED],
			0, old_selected);
	}
	gtk_widget_queue_draw (child);
	
	icon_view->selected_image = FALSE;
	__mark_container_images(icon_view, GTK_CONTAINER(icon_view));
	
	gtk_widget_grab_focus(GTK_WIDGET(icon_view));
	
	/* make sure the selected widget is in the visable area */
	while (view_child->state == KIBA_ICON_VIEW_HIDE_BEFORE) {
		kiba_icon_view_hide_last_row(icon_view);
	}
	while (view_child->state == KIBA_ICON_VIEW_HIDE_AFTER) {
		kiba_icon_view_hide_first_row(icon_view);
	}
	
	g_signal_emit (icon_view,
		kiba_icon_view_signals[SIG_CHILD_SELECTED],
		0, child);
}

/**
 * kiba_icon_view_get_selected_child:
 * @icon_view: The KibaIconView instance.
 *
 * Returns the selected child.
 *
 * Returns: The selcted KibaIconView child.
 */
GtkWidget* kiba_icon_view_get_selected_child (KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), NULL);
	
	return icon_view->selected_child;
}

/**
 * kiba_icon_view_show_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child.
 * @state: The show state.
 *
 * Shows/Hides child.
 */
void kiba_icon_view_show_child (KibaIconView *icon_view,
                                GtkWidget *child,
                                KibaIconViewShowState state) {
	KibaIconViewChild *view_child;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	g_return_if_fail (GTK_IS_WIDGET(child));
	
	view_child = __get_child_for_widget (icon_view, child);
	g_return_if_fail(view_child != NULL);
	
	switch (state) {
		case KIBA_ICON_VIEW_SHOW:
			if (view_child->state == KIBA_ICON_VIEW_HIDE_BEFORE)
				__change_before_count(icon_view, FALSE);
			else if (view_child->state == KIBA_ICON_VIEW_HIDE_AFTER)
				__change_after_count(icon_view, FALSE);
			if (view_child->state != KIBA_ICON_VIEW_SHOW) {
				icon_view->active_icon_count++;
				gtk_widget_queue_resize (GTK_WIDGET(icon_view));
				gtk_widget_show (view_child->widget);
			}
			break;
		case KIBA_ICON_VIEW_HIDE_BEFORE:
			if (view_child->state == KIBA_ICON_VIEW_SHOW) {
				icon_view->active_icon_count--;
				gtk_widget_queue_resize (GTK_WIDGET(icon_view));
			}
			else if (view_child->state == KIBA_ICON_VIEW_HIDE_AFTER)
				__change_after_count(icon_view, FALSE);
			if (view_child->state != KIBA_ICON_VIEW_HIDE_BEFORE) {
				__change_before_count(icon_view, TRUE);
				gtk_widget_hide (view_child->widget);
			}
			break;
		case KIBA_ICON_VIEW_HIDE_AFTER:
			if (view_child->state == KIBA_ICON_VIEW_SHOW) {
				icon_view->active_icon_count--;
				gtk_widget_queue_resize (GTK_WIDGET(icon_view));
			}
			else if (view_child->state == KIBA_ICON_VIEW_HIDE_BEFORE)
				__change_before_count(icon_view, FALSE);
			if (view_child->state != KIBA_ICON_VIEW_HIDE_AFTER) {
				__change_after_count(icon_view, TRUE);
				gtk_widget_hide (view_child->widget);
			}
			break;
		default:
			g_assert_not_reached();
			break;
	}
	
	view_child->state = state;
}

void kiba_icon_view_show_n_icons (KibaIconView *icon_view,
                                  gint n) {
	gint i;
	/* make sure KIBA_ICON_VIEW_MAX_ICONS(view_win) icons are shown before hiding items */
	for (i = 0; i < n; i++) {
		KibaIconViewChild *view_icon = g_list_nth_data(icon_view->children, icon_view->icons_before_count + i);
		if (!view_icon) break;
		if (view_icon->state != KIBA_ICON_VIEW_SHOW)
			kiba_icon_view_show_child (icon_view, view_icon->widget, KIBA_ICON_VIEW_SHOW);
	}
}

gint kiba_icon_view_get_active_count(KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), 0);
	return icon_view->active_icon_count;
}
gint kiba_icon_view_get_before_count(KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), 0);
	return icon_view->icons_before_count;
}
gint kiba_icon_view_get_after_count(KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), 0);
	return icon_view->icons_after_count;
}

/**
 * kiba_icon_view_set_prelight_ontop:
 * @icon_view: The KibaIconView instance.
 * @ontop: TRUE if the prelight should be ontop, otherwise FALSE.
 *
 * Sets wether the prelight is ontop.
 */
void kiba_icon_view_set_prelight_ontop (KibaIconView *icon_view,
                                        gboolean ontop) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	if (icon_view->prelight_ontop == ontop)
		return;
	
	icon_view->prelight_ontop = ontop;
	
	if (!icon_view->prelightable)
		return;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon_view),
		G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	
	if (icon_view->prelight_ontop) {
		g_signal_connect_after(G_OBJECT (icon_view), "expose-event",
			G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	}
}

/**
 * kiba_icon_view_set_prelightable:
 * @icon_view: The KibaIconView instance.
 * @prelightable: TRUE for prelightable children, otherwise FALSE.
 *
 * Sets wether the icon view children are prelightable.
 */
void kiba_icon_view_set_prelightable (KibaIconView *icon_view,
                                      gboolean prelightable) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	if (icon_view->prelightable == prelightable)
		return;
	
	icon_view->prelightable = prelightable;
	
	if (!prelightable) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(icon_view),
			G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	}
	else {
		if (icon_view->prelight_ontop) {
			g_signal_connect_after(G_OBJECT (icon_view), "expose-event",
				G_CALLBACK(kiba_icon_view_expose_cb), NULL);
		}
	}
}

/**
 * kiba_icon_view_set_sort_method:
 * @icon_view: The KibaIconView instance.
 * @method: The method to sort the icons.
 *
 * Sets custom method to sort the icons used by @icon_view.
 */
void kiba_icon_view_set_sort_method (KibaIconView *icon_view,
                                     GCompareFunc method) {
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	icon_view->sort_func = method;
}

/*** Class implementations ***/

/**
 * gtk_container_add implementation.
 **/
static void kiba_icon_view_add (GtkContainer *container,
                                GtkWidget *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(container);
	KibaIconViewChild *child;
	
	child = g_new0 (KibaIconViewChild, 1);
	child->allocation = widget->allocation;
	child->widget = gtk_widget_ref(widget);
	child->state = KIBA_ICON_VIEW_HIDE_AFTER;
	__change_after_count(icon_view, TRUE);
	
	if (GTK_WIDGET_MAPPED(child->widget))
		gtk_widget_hide(child->widget);
	
	icon_view->children = g_list_append (icon_view->children, child);
	
	/* sort the children */
	if (icon_view->sort_func != NULL) {
		GList *l, *unsortedCopy;
		gboolean orderChanged = FALSE;
		gint i;
		
		unsortedCopy = g_list_copy(icon_view->children);
		icon_view->children = g_list_sort(icon_view->children, icon_view->sort_func);
		
		i = 0;
		for (l = icon_view->children; l != NULL; l = l->next) {
			KibaIconViewChild *achild = l->data;
			KibaIconViewChild *bchild = g_list_nth_data(unsortedCopy, i);
			
			if (achild != bchild) {
				orderChanged = TRUE;
				break;
			}
			
			i++;
		}
		g_list_free(unsortedCopy);
		
		if (orderChanged) {
			i = 0;
			for (l = icon_view->children; l != NULL; l = l->next) {
				KibaIconViewChild *view_child = l->data;
				
				if (i < icon_view->icons_before_count) {
					view_child->state = KIBA_ICON_VIEW_HIDE_BEFORE;
					if (GTK_WIDGET_MAPPED(view_child->widget))
						gtk_widget_hide (view_child->widget);
				} else if (i < icon_view->icons_before_count + icon_view->active_icon_count) {
					view_child->state = KIBA_ICON_VIEW_SHOW;
					if (!GTK_WIDGET_MAPPED(view_child->widget))
						gtk_widget_show (view_child->widget);
				} else {
					view_child->state = KIBA_ICON_VIEW_HIDE_AFTER;
					if (GTK_WIDGET_MAPPED(view_child->widget))
						gtk_widget_hide (view_child->widget);
				}
				
				gtk_widget_queue_resize_no_redraw(view_child->widget);
				view_child->widget->allocation.x = -1;
				view_child->widget->allocation.y = -1;
				
				i++;
			}
			kiba_icon_view_children_size_allocate (GTK_WIDGET(icon_view));
		}
	}
	
	(* GTK_CONTAINER_CLASS (kiba_icon_view_parent_class)->add) (container, widget);
	
	if (kiba_icon_view_get_selected_child(icon_view) == NULL) {
		kiba_icon_view_select_child (icon_view, child->widget);
	}
	
	g_signal_connect (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(__child_press_cb), icon_view);
}

/**
 * gtk_container_remove implementation.
 **/
static void kiba_icon_view_remove (GtkContainer *container,
                                   GtkWidget *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(container);
	KibaIconViewChild *child;
	
	child = __get_child_for_widget (icon_view, widget);
	if (!child) return;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		  G_CALLBACK(__child_press_cb), icon_view);
	
	if (widget == kiba_icon_view_get_selected_child(icon_view)) {
		GList *needle;
		
		icon_view->selected_child = NULL;
		
		needle = g_list_find (icon_view->children, child);
		
		if (needle) {
			KibaIconViewChild *nextChild = NULL;
			if (needle->next) {
				nextChild = needle->next->data;
			} else if (needle->prev) {
				nextChild = needle->prev->data;
			}
			if (nextChild)
				kiba_icon_view_select_child (icon_view, nextChild->widget);
		}
	}
	
	icon_view->children = g_list_remove (icon_view->children, child);
	
	switch (child->state) {
		case KIBA_ICON_VIEW_SHOW: {
			KibaIconViewChild *new_active = NULL;
			GList *l;
			
			icon_view->active_icon_count--;
			
			if (icon_view->icons_after_count > 0) {
				new_active = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_FIRST_AFTER(icon_view));
				__change_after_count(icon_view, FALSE);
			} else if (icon_view->icons_before_count > 0) {
				new_active = g_list_nth_data(icon_view->children, KIBA_ICON_VIEW_LAST_BEFORE(icon_view));
				__change_before_count(icon_view, FALSE);
			}
			
			if (new_active) {
				new_active->state = KIBA_ICON_VIEW_SHOW;
				icon_view->active_icon_count++;
				gtk_widget_show (new_active->widget);
				kiba_icon_view_show_child (icon_view, new_active->widget, KIBA_ICON_VIEW_SHOW);
			}
			
			for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
				GtkWidget *child_widget = l->data;
				gtk_widget_queue_resize_no_redraw(child_widget);
				child_widget->allocation.x = -1;
				child_widget->allocation.y = -1;
			}
			kiba_icon_view_children_size_allocate (GTK_WIDGET(icon_view));
			
			break;
		}
		case KIBA_ICON_VIEW_HIDE_BEFORE:
			__change_before_count(icon_view, FALSE);
			break;
		case KIBA_ICON_VIEW_HIDE_AFTER:
			__change_after_count(icon_view, FALSE);
			break;
		default:
			g_assert_not_reached();
			break;
	}
	
	gtk_widget_unref(child->widget);
	g_free (child);
	
	(* GTK_CONTAINER_CLASS (kiba_icon_view_parent_class)->remove) (container, widget);
}

/**
 * gtk_widget_key_press_event implementation
 **/
static gboolean kiba_icon_view_key_press_event (GtkWidget *widget,
                                                GdkEventKey *event) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	GtkWidget *active_child = icon_view->selected_child;
	GList *active_needle, *l;
	
	for (active_needle = icon_view->children; active_needle != NULL; active_needle = active_needle->next) {
		KibaIconViewChild *child = active_needle->data;
		if (active_child == child->widget)
			break;
	}
	
	switch (event->keyval) {
		case GDK_KP_Enter:
		case GDK_Return:
			kiba_icon_view_activate_child (icon_view, active_child);
			return TRUE;
// 		case GDK_Escape:
// 			gtk_widget_destroy (widget);
// 			return TRUE;
		case GDK_KP_Left:
		case GDK_Left: {
			KibaIconViewChild *left_child;
			
			if (active_needle->prev) {
				left_child = active_needle->prev->data;
			} else {
				left_child = g_list_last(icon_view->children)->data;
			}
			kiba_icon_view_select_child (icon_view, left_child->widget);
			return TRUE;
		}
		case GDK_KP_Right:
		case GDK_Right: {
			KibaIconViewChild *right_child;
			
			if (active_needle->next) {
				right_child = active_needle->next->data;
			} else {
				right_child = g_list_first(icon_view->children)->data;
			}
			kiba_icon_view_select_child (icon_view, right_child->widget);
			return TRUE;
		}
		case GDK_KP_Up:
		case GDK_Up: {
			KibaIconViewChild *up_child;
			gint i = 0;
			
			for (l = active_needle; l != NULL; l = l->prev) {
				up_child = l->data;
				
				if (l->prev == NULL) {
					l = g_list_last(icon_view->children);
				}
				
				i++;
				if (i > icon_view->columns)
					break;
			}
			kiba_icon_view_select_child (icon_view, up_child->widget);
			return TRUE;
		}
		case GDK_KP_Down:
		case GDK_Down: {
			KibaIconViewChild *down_child;
			gint i = 0;
			
			for (l = active_needle; l != NULL; l = l->next) {
				down_child = l->data;
				
				if (l->prev == NULL) {
					l = g_list_first(icon_view->children);
				}
				
				i++;
				if (i > icon_view->columns)
					break;
			}
			kiba_icon_view_select_child (icon_view, down_child->widget);
			return TRUE;
		}
		case GDK_KP_Begin:
		case GDK_Begin: { /**< FIXME: this does not match "Pos 1" key here */
			KibaIconViewChild *first_child;
			
			first_child = g_list_first(icon_view->children)->data;
			kiba_icon_view_select_child (icon_view, first_child->widget);
			return TRUE;
		}
		case GDK_KP_End:
		case GDK_End: {
			KibaIconViewChild *last_child;
			
			last_child = g_list_last(icon_view->children)->data;
			kiba_icon_view_select_child (icon_view, last_child->widget);
			return TRUE;
		}
		case GDK_KP_Page_Up:
		case GDK_Page_Up:
			kiba_icon_view_hide_last_row(icon_view);
			return TRUE;
		case GDK_KP_Page_Down:
		case GDK_Page_Down:
			kiba_icon_view_hide_first_row(icon_view);
			return TRUE;
		case GDK_KP_Delete:
		case GDK_Delete: //TODO remove currently active icon
			break;
		default: { /* search for items beginning with the pressed character */
			gchar *character = gdk_keyval_name(event->keyval);
			
			if (!character || strlen(character) != 1)
				break;
			
			__append_keypress_prefix(icon_view, character);
			
			for (l = icon_view->children; l != NULL; l = l->next) {
				KibaIconViewChild *view_child = l->data;
				GtkLabel *view_label;
				const char* labelText;
				
				/* FIXME: maybe we should use another method to get the label :) */
				view_label = __get_label_child(view_child->widget);
				if (!view_label) continue;
				
				labelText = gtk_label_get_text(view_label);
				if (labelText && g_str_has_prefix(labelText, icon_view->keypress_prefix)) {
					kiba_icon_view_select_child (icon_view, view_child->widget);
					return TRUE;
				}
			}
		}
	}
	
	return FALSE;
}

/**
 * gtk_widget_scroll_event implementation.
 **/
static gboolean kiba_icon_view_scroll_event (GtkWidget *widget,
                                             GdkEventScroll *event) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	
	if (event->direction == GDK_SCROLL_DOWN) {
		kiba_icon_view_hide_last_row(icon_view);
	}
	else if (event->direction == GDK_SCROLL_UP) {
		kiba_icon_view_hide_first_row(icon_view);
	}
	
	return TRUE;
}

/**
 * gtk_widget_size_allocate implementation.
 **/
static void kiba_icon_view_size_allocate (GtkWidget *widget,
                                          GtkAllocation *allocation) {
	widget->allocation = *allocation;
	
	kiba_icon_view_children_size_allocate (widget);
}

/**
 * gtk_widget_size_request implementation.
 **/
static void kiba_icon_view_size_request (GtkWidget *widget,
                                         GtkRequisition *requisition) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	GList *l;
	gint columns;
	gint i;
	gint spacing;
	gint border_width;
	gint row_height, row_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(widget));
	spacing = gtk_box_get_spacing (GTK_BOX(widget));
	columns = kiba_icon_view_get_columns (KIBA_ICON_VIEW(widget));
	
	requisition->width  = row_height = 0;
	requisition->height = row_width = spacing;
	
	i = 1;
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *view_child = l->data;
		GtkWidget *child = view_child->widget;
		GtkRequisition child_requisition;
		
		if (view_child->state != KIBA_ICON_VIEW_SHOW)
			continue;
		
		gtk_widget_size_request (child, &child_requisition);
		
		row_height = MAX(row_height, child_requisition.height + spacing);
		row_width += child_requisition.width + spacing;
		
		if (i == columns) {
			requisition->height += row_height;
			requisition->width   = MAX(row_width, requisition->width);
			
			i = 0;
			row_width = spacing;
			row_height = 0;
		}
		else if (KIBA_ICON_VIEW(widget)->fixed && !l->next) {
			row_width += (columns-i)*(child_requisition.width + spacing);
			requisition->width   = MAX(row_width, requisition->width);
		}
		
		i++;
	}
	requisition->width   = MAX(row_width, requisition->width);
	requisition->height += row_height;
	
	requisition->width  += 2*border_width;
	requisition->height += 2*border_width;
}

/**
 * gtk_widget_unrealize implementation
 **/
static void kiba_icon_view_unrealize (GtkWidget *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	GList *l;
	
	for (l = icon_view->children; l != NULL; l = icon_view->children) {
		KibaIconViewChild *child = l->data;
		icon_view->children = g_list_remove(icon_view->children, l->data);
		gtk_widget_unref(child->widget);
		g_free(child);
	}
	g_list_free(icon_view->children);
	
	(* GTK_WIDGET_CLASS (kiba_icon_view_parent_class)->unrealize) (widget);
}

/*** Class init/fini ***/

static void kiba_icon_view_class_init (KibaIconViewClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
	
	widget_class->key_press_event = kiba_icon_view_key_press_event;
	widget_class->size_allocate = kiba_icon_view_size_allocate;
	widget_class->size_request = kiba_icon_view_size_request;
	widget_class->expose_event = kiba_icon_view_expose_event;
	widget_class->unrealize = kiba_icon_view_unrealize;
	widget_class->scroll_event = kiba_icon_view_scroll_event;
	
	container_class->add = kiba_icon_view_add;
	container_class->remove = kiba_icon_view_remove;
	
	/* signals */
	kiba_icon_view_signals[SIG_CHILD_ACTIVATED] =
		g_signal_new ("child-activated",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_activated),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_icon_view_signals[SIG_CHILD_SELECTED] =
		g_signal_new ("child-selected",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_selected),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_icon_view_signals[SIG_CHILD_UNSELECTED] =
		g_signal_new ("child-unselected",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_unselected),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_icon_view_signals[SIG_CHILDREN_BEFORE_TOGGLED] =
		g_signal_new ("children-before-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, children_before_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	kiba_icon_view_signals[SIG_CHILDREN_AFTER_TOGGLED] =
		g_signal_new ("children-after-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, children_after_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
}

static void kiba_icon_view_init (KibaIconView *icon_view) {
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (icon_view),
			GTK_NO_WINDOW |
			GTK_CAN_FOCUS);
	
	gtk_widget_add_events (GTK_WIDGET (icon_view),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	icon_view->children = NULL;
	icon_view->selected_child = NULL;
	
	icon_view->columns = 4;
	icon_view->fixed = FALSE;
	icon_view->prelight_ontop = TRUE;
	icon_view->prelightable = TRUE;
	icon_view->keypress_clear_id = 0;
	icon_view->keypress_prefix = NULL;
	
	icon_view->sort_func = __sort_by_label;
	
	icon_view->icons_before_count = 0;
	icon_view->active_icon_count = 0;
	icon_view->icons_after_count = 0;
	
	gtk_container_set_border_width (GTK_CONTAINER(icon_view), 0);
	gtk_box_set_spacing (GTK_BOX(icon_view), 10);
}

/**
 * kiba_icon_view_new:
 *
 * Creates and returns a KibaIconView instance.
 *
 * Returns: The KibaIconView instance created.
 */
GtkWidget* kiba_icon_view_new (void) {
	return g_object_new (KIBA_TYPE_ICON_VIEW, NULL);
}
