/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-image
 * @short_description: Widget showing an image.
 * @see_also:
 * @stability: Unstable
 *
 * #KibaImage uses a cairo surface buffer to render an image.
 */

#include "../config.h"
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-image.h>

#define MIN_IMAGE_SIZE 18

#define KIBA_IMAGE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_IMAGE, KibaImagePrivate))

typedef struct _KibaImagePrivate KibaImagePrivate;

struct _KibaImagePrivate {
	GdkWindow *event_window;
	guint      width;
	guint      height;
};

G_DEFINE_TYPE (KibaImage, kiba_image, GTK_TYPE_WIDGET)

/**
 * kiba_image_set_alpha:
 * @image: The KibaImage instance.
 * @alpha: The KibaImage alpha.
 *
 * Sets image alpha.
 */
void kiba_image_set_alpha (KibaImage *image,
			   float     alpha) {
	g_return_if_fail (KIBA_IS_IMAGE(image));
	g_return_if_fail (alpha >= 0.0f);
	g_return_if_fail (alpha <= 1.0f);
	
	if (image->alpha == alpha)
		return;
	
	image->alpha = alpha;
	
	gtk_widget_queue_resize (GTK_WIDGET(image));
}

/**
 * kiba_image_get_alpha:
 * @image: The KibaImage instance.
 *
 * Returns image alpha.
 *
 * Returns: The KibaImage alpha.
 */
float kiba_image_get_alpha (KibaImage *image) {
	g_return_val_if_fail (KIBA_IS_IMAGE(image), FALSE);
	
	return image->alpha;
}

/**
 * kiba_image_set_surface:
 * @image: The KibaImage instance.
 * @surface: The surface to use.
 * @width: The width of the surface.
 * @height: The height of the surface.
 *
 * Sets the surface used for rendering.
 */
void kiba_image_set_surface (KibaImage       *image,
			     cairo_surface_t *surface,
			     gint             width,
			     gint             height) {
	g_return_if_fail (KIBA_IS_IMAGE(image));
	g_return_if_fail (surface != NULL);
	g_return_if_fail (width  > 0);
	g_return_if_fail (height > 0);
	
	if (image->surface == surface)
		return;
	
	if (image->surface)
		cairo_surface_destroy (image->surface);
	image->surface = cairo_surface_reference (surface);
	image->surface_width  = width;
	image->surface_height = height;
	
	gtk_widget_queue_resize (GTK_WIDGET(image));
}

static void kiba_image_draw_pixbuf_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	GdkPixbuf *pixbuf = data;
	
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
}

/**
 * kiba_image_set_pixbuf:
 * @image: The KibaImage instance.
 * @pixbuf: The pixbuf to use.
 * @width: The width of the pixbuf.
 * @height: The height of the pixbuf.
 *
 * Sets the pixbuf for the surface.
 */
void kiba_image_set_pixbuf (KibaImage  *image,
			    GdkPixbuf  *pixbuf,
			    gint        width,
			    gint        height) {
	cairo_t *cr;
	
	g_return_if_fail (KIBA_IS_IMAGE(image));
	g_return_if_fail (GDK_IS_PIXBUF(pixbuf));
	g_return_if_fail (width  > 0);
	g_return_if_fail (height > 0);
	
	image->surface_width  = width;
	image->surface_height = height;
	
	cr = kiba_win_cairo_create (kiba_get_win(kiba_get()));
	cairo_update_surface_at_size (cr, &image->surface, width, height,
			kiba_image_draw_pixbuf_surface, pixbuf);
	cairo_destroy (cr);
}

/**
 * kiba_image_set_size:
 * @image: The KibaImage instance.
 * @width: The width of the KibaImage.
 * @height: The height of the KibaImage.
 *
 * Sets the image size.
 */
void kiba_image_set_size (KibaImage *image,
			  gint       width,
			  gint       height) {
	GtkWidget *widget;
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	g_return_if_fail (GTK_IS_WIDGET(image));
	g_return_if_fail (width  > 0);
	g_return_if_fail (height > 0);
	g_return_if_fail (width  < 200);
	g_return_if_fail (height < 200);
	
	priv->width = width;
	priv->height = height;
	
// 	widget = GTK_WIDGET(image);
// 	widget->requisition.width  = width;
// 	widget->requisition.height = height;
	
	gtk_widget_queue_resize (widget);
}

/**
 * kiba_image_is_selected:
 * @image: The KibaImage instance.
 *
 * Returns wether the image is selected.
 *
 * Returns: TRUE if the KIBA_IMAGE_STATE_SELECTED flag is set, otherwise FALSE is returned.
 */
gboolean kiba_image_is_selected (KibaImage *image) {
	g_return_val_if_fail (KIBA_IS_IMAGE(image), FALSE);
	
	return (image->state == KIBA_IMAGE_STATE_SELECTED);
}

/**
 * kiba_image_select:
 * @image: The KibaImage instance.
 * @select: TRUE to set the KIBA_IMAGE_STATE_SELECTED flag, otherwise  FALSE to unset the flag.
 *
 * Selects image.
 */
void kiba_image_select (KibaImage      *image,
			gboolean        select) {
	g_return_if_fail (KIBA_IS_IMAGE(image));
	
	if (select)
		image->state = KIBA_IMAGE_STATE_SELECTED;
	else
		image->state = KIBA_IMAGE_STATE_NONE;
	
	gtk_widget_queue_draw (GTK_WIDGET(image));
}

/**
Class Functions
**/

static gboolean kiba_image_expose_event (GtkWidget      *widget,
				         GdkEventExpose *expose) {
	KibaImage *image = KIBA_IMAGE (widget);
	
	cairo_t *cr = NULL;
	
	if (!image->surface)
		return FALSE;
	
	if (!GTK_WIDGET_MAPPED(widget))
		return FALSE;
	
	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_translate (cr, widget->allocation.x, widget->allocation.y);
// 	cairo_scale (cr,
// 		(float) widget->allocation.width/image->surface_width,
// 		(float) widget->allocation.height/image->surface_height);
	cairo_translate (cr,
		(widget->allocation.width-image->surface_width)/2,
		(widget->allocation.height-image->surface_height)/2);
	cairo_set_source_surface (cr, image->surface, 0, 0);
	cairo_paint_with_alpha (cr, image->alpha);
	
	if (image->state == KIBA_IMAGE_STATE_SELECTED) {
		cairo_set_source_rgba (cr,
			1.0,
			1.0,
			1.0,
			image->alpha * 0.6);
		cairo_mask_surface (cr, image->surface, 0, 0);
	}
	
	cairo_destroy (cr);
	
	return FALSE;
}

static void kiba_image_screen_changed (GtkWidget *widget,
				       GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

static void kiba_image_size_allocate (GtkWidget      *widget,
				      GtkAllocation  *allocation) {
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	if (GTK_WIDGET_CLASS (kiba_image_parent_class)->size_allocate)
		GTK_WIDGET_CLASS (kiba_image_parent_class)->size_allocate (widget, allocation);
	
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	gdk_window_move_resize (priv->event_window,
				widget->allocation.x,
				widget->allocation.y,
				widget->allocation.width,
				widget->allocation.height);
}

static void kiba_image_size_request (GtkWidget      *widget,
				     GtkRequisition *requisition) {
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	requisition->width  = priv->width;
	requisition->height = priv->height;
	
}

static void kiba_image_map (GtkWidget *widget) {
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	if (GTK_WIDGET_CLASS (kiba_image_parent_class)->map)
		GTK_WIDGET_CLASS (kiba_image_parent_class)->map (widget);
	
	gdk_window_show (priv->event_window);
}

static void kiba_image_unmap (GtkWidget *widget) {
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	gdk_window_hide (priv->event_window);
	
	if (GTK_WIDGET_CLASS (kiba_image_parent_class)->unmap)
		GTK_WIDGET_CLASS (kiba_image_parent_class)->unmap (widget);
}

static void kiba_image_unrealize (GtkWidget *widget) {
	KibaImage *image = KIBA_IMAGE(widget);
	KibaImagePrivate *priv;
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);

	if (image->surface)
		cairo_surface_destroy (image->surface);
	
	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	if (GTK_WIDGET_CLASS (kiba_image_parent_class)->unrealize)
		GTK_WIDGET_CLASS (kiba_image_parent_class)->unrealize (widget);
}

static void kiba_image_realize (GtkWidget *widget) {
	GdkWindowAttr attributes;
	gint attributes_mask;
	KibaImagePrivate *priv;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	priv = KIBA_IMAGE_GET_PRIVATE(widget);
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window);
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	
	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	gdk_window_set_user_data (priv->event_window, widget);
	gdk_window_register_dnd (priv->event_window);
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
}

static void kiba_image_class_init (KibaImageClass *class) {
	GtkWidgetClass    *widget_class;
	
	widget_class                  = GTK_WIDGET_CLASS (class);
	
	widget_class->realize         = kiba_image_realize;
	widget_class->unrealize       = kiba_image_unrealize;
	widget_class->map             = kiba_image_map;
	widget_class->unmap           = kiba_image_unmap;
	widget_class->expose_event    = kiba_image_expose_event;
	widget_class->screen_changed  = kiba_image_screen_changed;
	widget_class->size_request    = kiba_image_size_request;
	widget_class->size_allocate   = kiba_image_size_allocate;
	
	g_type_class_add_private (class, sizeof (KibaImagePrivate));
}

static void kiba_image_init (KibaImage *image) {
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (image), GTK_NO_WINDOW);
	
	image->surface = NULL;
	image->surface_width  = 0;
	image->surface_height = 0;
	
	image->alpha = 1.0f;
	
	image->state = KIBA_IMAGE_STATE_NONE;
	
	kiba_image_screen_changed (GTK_WIDGET (image), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (image), TRUE);
}

/**
 * kiba_image_new:
 *
 * Creates and returns a KibaImage Instance.
 *
 * Returns: The KibaImage instance created.
 */
GtkWidget* kiba_image_new (void) {
	return g_object_new (KIBA_TYPE_IMAGE, NULL);
}
