/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-info-win
 * @short_description: Transparent toplevel window.
 * @see_also:
 * @stability: Unstable
 *
 * Transparent toplevel window.
 */

#include "../config.h"
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-macros.h"
#include "kiba-prefs.h"
#include "kiba-cairo-utils.h"
#include "kiba-info-win.h"

G_DEFINE_TYPE (KibaInfoWin, kiba_info_win, GTK_TYPE_WINDOW)

typedef struct {
	gboolean draw;
	
	KibaRGBA *gradient;
	gint numColors;
	KibaRGBA border_color;   /**< Border color */
	
	KibaCornerShape topleft_shape;
	KibaCornerShape topright_shape;
	KibaCornerShape bottomright_shape;
	KibaCornerShape bottomleft_shape;
	
	float corner_radius;
	
	union {
		KibaBackgroundStyle style;  /**< The background style */
		char *file;   /**< The pattern file path */
	} background;
}KibaInfoWinSettings;

static KibaInfoWinSettings *info_win_settings = NULL;
static GList *info_wins = NULL;

static void __redraw_all (void) {
	GList *l;
	for (l = info_wins; l != NULL; l = l->next) {
		KibaInfoWin *info_win = l->data;
		gtk_widget_queue_draw(GTK_WIDGET(info_win));
	}
}

static void __prefs_gradient_cb (const char *option_id,
                                 gpointer value,
                                 gpointer data) {
	KibaInfoWinSettings *settings = data;
	GSList *list = value;
	gint length = g_slist_length(list);
	KibaRGBA *gradient = g_new0(KibaRGBA, length);
	
	kiba_prefs_string_list_to_rgba_array(list, gradient);
	if (settings->gradient)
		g_free (settings->gradient);
	settings->gradient = gradient;
	settings->numColors = length;
	
	__redraw_all();
}

static void __prefs_shape_cb  (const char *option_id,
                               gpointer value,
                               gpointer data) {
	KibaInfoWinSettings *settings = data;
	KibaCornerShape corner;
	corner =  kiba_prefs_edge_style_from_string ((char*) value);
	
	if (strcmp (option_id, "popup_corner_topleft") == 0)
		settings->topleft_shape = corner;
	else if (strcmp (option_id, "popup_corner_topright") == 0)
		settings->topright_shape = corner;
	else if (strcmp (option_id, "popup_corner_bottomright") == 0)
		settings->bottomright_shape = corner;
	else if (strcmp (option_id, "popup_corner_bottomleft") == 0)
		settings->bottomleft_shape = corner;
	
	__redraw_all();
}

static void __border_color_cb (const char *option_id,
                               gpointer value,
                               gpointer data) {
	KibaInfoWinSettings *settings = data;
	settings->border_color = kiba_prefs_rgba_from_string ((char*) value);
	__redraw_all();
}

static void __prefs_border_width_cb (const char *option_id,
                                     gpointer value,
                                     gpointer data) {
	GList *l;
	for (l = info_wins; l != NULL; l = l->next)
		gtk_container_set_border_width (GTK_CONTAINER(l->data), GPOINTER_TO_INT (value));
	__redraw_all();
}

static void __prefs_draw_cb (const char *option_id,
                             gpointer value,
                             gpointer data) {
	KibaInfoWinSettings *settings = data;
	settings->draw = (gboolean) GPOINTER_TO_INT (value);
	__redraw_all();
}

static void __prefs_background_cb (const char *option_id,
                                   gpointer value,
                                   gpointer data) {
	KibaInfoWinSettings *settings = data;
	const char *valueString = (char*) value;
	if (g_str_has_suffix (valueString, ".png") || g_str_has_suffix (valueString, ".svg"))
		settings->background.file = (char*) value;
	else
		settings->background.style = kiba_prefs_background_style_from_string (valueString);
	__redraw_all();
}

static void __prefs_corner_radius_cb (const char *option_id,
                                      gpointer value,
                                      gpointer data) {
	KibaInfoWinSettings *settings = data;
	settings->corner_radius = (float)g_ascii_strtod((char*) value, NULL);
	__redraw_all();
}

static void __size_allocate_cb (GtkWidget *widget,
                                GtkAllocation *allocation,
                                gpointer data) {
// 	if (allocation->width != widget->allocation.width || allocation->height != widget->allocation.height)
		gtk_widget_queue_draw(widget);
}

/*** Public methods ***/

KibaRGBA kiba_get_info_win_border_color () {
	if (!info_win_settings)
		return (KibaRGBA) {0.0f,0.0f,0.0f,1.0f};
	return info_win_settings->border_color;
}

/**
Class Functions
**/

static gboolean kiba_info_win_expose_event (GtkWidget      *widget,
				            GdkEventExpose *expose) {
	KibaInfoWin *info_win = KIBA_INFO_WIN(widget);
	KibaInfoWinSettings *settings = info_win_settings;
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern = NULL;
	GtkWidget *child = NULL;
	gint border_width;
	
	border_width = gtk_container_get_border_width(GTK_CONTAINER(widget));
	
	child = gtk_bin_get_child (GTK_BIN (widget));
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	cr = gdk_cairo_create (widget->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	/* clear the window to be transparent */
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	
	if (settings->draw) {
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		
		/* draw shape for the info window */
		cairo_paint_shape (cr,
			settings->topleft_shape, settings->topright_shape,
			settings->bottomleft_shape, settings->bottomright_shape,
			border_width/2,
			border_width/2,
			widget->allocation.width - border_width/2,
			widget->allocation.height - border_width/2,
			settings->corner_radius);
		
		if (settings->background.style == KIBA_GRADIENT)
			pattern = cairo_get_gradient_pattern (border_width/2,
				border_width/2,
				widget->allocation.width - border_width,
				widget->allocation.height - border_width,
				settings->gradient, settings->numColors);
		else {
			pattern = cairo_get_png_pattern (settings->background.file, 1.0);
		}
		
		cairo_set_source (cr, pattern);
		cairo_fill_preserve (cr);
		cairo_pattern_destroy (pattern);
		
		cairo_set_source_rgba (cr,
			settings->border_color.r,
			settings->border_color.g,
			settings->border_color.b,
			settings->border_color.a);
		cairo_set_line_width (cr, border_width);
		cairo_stroke (cr);
	}
	
	cairo_destroy (cr);
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return FALSE;
}

static void kiba_info_win_screen_changed (GtkWidget *widget,
					  GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

static void kiba_info_win_class_init (KibaInfoWinClass *class) {
	GtkWidgetClass  *widget_class;
	
	widget_class  = GTK_WIDGET_CLASS (class);
	
	widget_class->expose_event = kiba_info_win_expose_event;
	widget_class->screen_changed  = kiba_info_win_screen_changed;
}

static void __destroy_cb(GtkWidget *widget,
                         gpointer data) {
	info_wins = g_list_remove (info_wins, widget);
	if (g_list_length(info_wins) == 0) {
		KibaPrefsFile *file;
		
		file = kiba_prefs_get_general_file(kiba_prefs_get());
		kiba_prefs_remove_callback_by_data (file, info_win_settings);
	
		if (info_win_settings->gradient)
			g_free (info_win_settings->gradient);
		g_free(info_win_settings);
		info_win_settings = NULL;
	}
}

static void kiba_info_win_init (KibaInfoWin *info_win) {
	KibaScreen *screen;
	KibaPrefsFile *file;
	
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (info_win), GTK_NO_WINDOW);
	
	screen = kiba_screen_get();
	
	info_wins = g_list_append(info_wins, info_win);
	
	gtk_widget_add_events (GTK_WIDGET (info_win),
		GDK_EXPOSURE_MASK |
		GDK_FOCUS_CHANGE_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	kiba_info_win_screen_changed (GTK_WIDGET (info_win), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (info_win), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (info_win), FALSE);
	
	file = kiba_prefs_get_general_file(kiba_prefs_get());
	if (!info_win_settings) {
		info_win_settings = g_new0(KibaInfoWinSettings, 1);
		kiba_prefs_add_callback (file, "popup_draw",
			__prefs_draw_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_background",
			__prefs_background_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_corner_radius",
			__prefs_corner_radius_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_border_color",
			__border_color_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_topleft",
			__prefs_shape_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_topright",
			__prefs_shape_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_bottomleft",
			__prefs_shape_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_bottomright",
			__prefs_shape_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_gradient",
			__prefs_gradient_cb, info_win_settings, TRUE);
		kiba_prefs_add_callback (file, "popup_border_width",
			__prefs_border_width_cb, info_win_settings, TRUE);
	} else {
		gint border_width;
		border_width = kiba_prefs_get_int(file, "popup_border_width", NULL);
		gtk_container_set_border_width (GTK_CONTAINER(info_win), border_width);
	}
	
	g_signal_connect(G_OBJECT(info_win), "destroy",
		G_CALLBACK(__destroy_cb), NULL);
	g_signal_connect(G_OBJECT (info_win), "size-allocate",
		G_CALLBACK(__size_allocate_cb), NULL);
}

/**
 * kiba_info_win_new:
 *
 * Creates and returns a KibaInfoWin Instance.
 *
 * Returns: The KibaInfoWin Instance created.
 */
GtkWidget* kiba_info_win_new () {
	return g_object_new (KIBA_TYPE_INFO_WIN, NULL);
}
