/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Daniel    Bessler    <danielb@kiba-dock.org>
 *              Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
*/

/**
 * SECTION:kiba-key-file
 * @short_description: KeyFile methods.
 * @see_also:#GKeyFile
 * @stability: Unstable
 *
 * KeyFile methods.
 */

#include <string.h>
#include <libintl.h>

#include "kiba-key-file.h"
#include "kiba-utils.h"
#include "kiba-globals.h"

/**
 * kiba_key_file_launch:
 * @keyfile: The GKeyFile instance.
 *
 * Launches a GKeyFile.
 */
void kiba_key_file_launch (GKeyFile  *keyfile) {
	static const char *keys [] = { "Exec",
				       "URL",
				       "Path"
				     };
	
	gint i;
	
	g_return_if_fail (keyfile != NULL);
	
	for (i = 0; i < G_N_ELEMENTS (keys); i++) {
		char *value;
		
		value = g_key_file_get_value (keyfile, "Desktop Entry", keys [i], NULL);
		if (value == NULL)
			continue;
		
		char command[strlen(value)+23];
		
		switch (i) {
			case 0:
				if (kiba_launch_programm(value))
					return;
				break;
			case 1:
			case 2:
				sprintf(command, "%s %s", file_manager_global, value);
				if (kiba_launch_programm(command))
					return;
				break;
		}
	}
}

/**
 * kiba_key_file_launch_at_path:
 * @path: The URI to the KeyFile.
 *
 * Launches a key file.
 */
void kiba_key_file_launch_at_path (const char *path) {
	static const char file_uri_prefix[] = "file://";
	
	GKeyFile *keyfile;
	const char *filename;
	
	if (g_str_has_prefix (path, file_uri_prefix))
		filename = path + strlen (file_uri_prefix);
	else
		filename = path;
	
	keyfile = g_key_file_new ();
	if (!g_key_file_load_from_file (keyfile, filename, 0, NULL))
		return;
	
	kiba_key_file_launch (keyfile);
	g_key_file_free (keyfile);
}

/**
 * kiba_key_file_get_desktop_entry_at_path:
 * @filename: The Path to the Desktop File.
 * @key: The Key to look for.
 *
 * Get a Desktop File Value for the Key Name.
 *
 * Returns: The Desktop File Value.
 */
char* kiba_key_file_get_desktop_entry_at_path (const char *filename,
					       const char *key) {
	static const char desktop_entry_group[] = "Desktop Entry";
	
	GKeyFile *file;
	GError *error = NULL;
	char *value, *translated_value;
	
	g_return_val_if_fail(filename != NULL, NULL);
	g_return_val_if_fail(key != NULL, NULL);
	g_return_val_if_fail(g_str_has_suffix(filename, ".desktop"), NULL);
	g_return_val_if_fail(g_file_test (filename, G_FILE_TEST_IS_REGULAR), NULL);
	
	file = g_key_file_new ();
	if (!g_key_file_load_from_file (file, filename, 0, &error)) {
		return NULL;
	}
	g_clear_error (&error);
	error = NULL;
	
	translated_value = g_key_file_get_locale_string (file,
					desktop_entry_group,
					key,
					NULL,
					&error);
	g_clear_error (&error);
	error = NULL;
	if (!translated_value) {
		g_key_file_free (file);
		value = g_key_file_get_value (file, desktop_entry_group, key, &error);
		g_clear_error (&error);
		error = NULL;
		return value;
	}
	else {
		g_key_file_free (file);
		return translated_value;
	}
	
	return NULL;
}

/**
 * kiba_key_file_get_desktop_entry:
 * @file: The GKeyFile.
 * @key: The Key to look for.
 *
 * Gets a Desktop File Value from a GKeyFile.
 *
 * Returns: The Desktop File Value.
 */
char* kiba_key_file_get_desktop_entry (GKeyFile   *file,
				       const char *key) {
	static const char desktop_entry_group[] = "Desktop Entry";
	
	GError     *error = NULL;
	char       *value, *translated_value;
	
	translated_value = g_key_file_get_locale_string (file,
					desktop_entry_group,
					key,
					NULL,
					&error);
	if (!translated_value) {
		value = g_key_file_get_value (file, desktop_entry_group, key, &error);
		return value;
	}
	else
		return translated_value;
	
	return NULL;
}

/**
 * kiba_key_file_write_desktop_entry:
 * @desktop_file: The path to the .desktop file.
 * @desktop_entry: The key to look for.
 * @new_value: The new value for the key.
 *
 * Writes Desktop File Value for the Key Name in Desktop File.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_key_file_write_desktop_entry (const char *desktop_file,
					    const char *desktop_entry,
					    const char *new_value) {
	static const char file_uri_prefix[] = "file://";
	static const char desktop_entry_group[] = "Desktop Entry";
	
	GKeyFile   *key_file;
	GError     *error = NULL;
	const char *filename;
	
	if (g_str_has_prefix (desktop_file, file_uri_prefix))
		filename = desktop_file + strlen (file_uri_prefix);
	else
		filename = desktop_file;
	
	key_file = g_key_file_new ();
	if (!g_key_file_load_from_file (key_file, filename, 0, &error))
		return FALSE;
	g_clear_error (&error);
	error = NULL;
	
	/* { the function does not work here :/
		g_key_file_set_value (key_file,
			desktop_entry_group,
			desktop_entry,
			new_value);
	} */
	{
		FILE *file;
		char *content;
		gsize file_length;
		
		content = g_key_file_to_data (key_file, &file_length, &error);
		g_clear_error (&error);
		error = NULL;
		if (!content) {
			g_key_file_free (key_file);
			return FALSE;
		}
		
		file = fopen (desktop_file, "w");
		if (file == NULL) {
			g_free (content);
			g_key_file_free (key_file);
			return FALSE;
		}
		
		char *old_value;
		
		old_value = g_key_file_get_value (key_file,
			desktop_entry_group, desktop_entry, &error);
		g_clear_error (&error);
		error = NULL;
		
		if (old_value) {
			char *needle, *begin, *end;
			gint changed_line_len;
			
			needle = g_strdup_printf ("\n%s=", desktop_entry);
			end = g_strrstr (content+1, needle);
			g_free (needle);
			
			changed_line_len = strlen (desktop_entry) + strlen(old_value) + 1;
			end += changed_line_len + 1;
			g_free (old_value);
			
			if (strlen (end) == 1)
				end = NULL;
			
			begin = g_strndup (content+1,
					file_length - changed_line_len -
					(end != NULL ? strlen(end) : 0) - 2);
			
			fprintf (file,
				"%s\n"
				"%s=%s\n"
				"%s",
				begin,
				desktop_entry, new_value,
				end ? end : "");
			
			g_free (begin);
		}
		/* writing new key */
		else {
			fprintf (file,
				"%s"
				"%s=%s\n",
				content,
				desktop_entry, new_value);
		}
		fclose (file);
		g_free (content);
	}
	
	g_key_file_free (key_file);
	
	return TRUE;
}

/**
 * kiba_key_file_write_locale_desktop_entry:
 * @desktop_file: The path to the .desktop file.
 * @desktop_entry: The key to look for.
 * @new_value: The new value for the key.
 * Writes localized Key.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_key_file_write_locale_desktop_entry (const char *desktop_file,
						   const char *desktop_entry,
						   const char *new_value) {
	char *locale_entry;
	char *locale;
	const char *messages_locale;
	
	messages_locale = setlocale (LC_MESSAGES, NULL);
	locale = g_strrstr(messages_locale, "@");
	if (locale != NULL)
		locale = g_strndup (messages_locale,
				strlen(messages_locale) - strlen(locale));
	else
		locale = g_strdup (messages_locale);
	
	locale_entry = g_strdup_printf ("%s[%s]", desktop_entry, locale);
	g_free (locale);
	
	return kiba_key_file_write_desktop_entry (desktop_file, locale_entry, new_value);
}

/**
 * kiba_key_file_add_missing_desktop_entries:
 * @uri: The Path to the Key File.
 * @key_entry_group: The Entry Group to look for.
 * @fields: The Key Array of wanted Keys in the File.
 * @elements: Element number of the Key Array.
 *
 * Adds all missing Keys from the Key Array to the Key File.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_key_file_add_missing_desktop_entries (const char  *uri,
						    const char  *key_entry_group,
						    const char **fields,
						    int          elements) {
	static const char file_uri_prefix[] = "file://";
	
	GKeyFile   *file;
	GError     *error = NULL;
	const char *filename;
	char       *s;
	int         i;
	
	if (g_str_has_prefix (uri, file_uri_prefix))
		filename = uri + strlen (file_uri_prefix);
	else
		filename = uri;
	
	file = g_key_file_new ();
	if (!g_key_file_load_from_file (file, filename, 0, &error))
		return FALSE;
	g_clear_error (&error);
	error = NULL;
	
	for (i = 0; i < elements; i++) {
		s = g_key_file_get_value (file, key_entry_group, fields[i], &error);
	
		g_clear_error (&error);
		error = NULL;
		
		if (s) {
			g_free (s);
			continue;
		}
		
		if (strcmp (fields[i], "InactiveIcon") == 0) {
			char *active_icon;
			
			active_icon = g_key_file_get_value (file, key_entry_group, "Icon", &error);
			g_clear_error (&error);
			error = NULL;
			
			if (active_icon == NULL)
				continue;
			
			kiba_key_file_write_desktop_entry (filename, "InactiveIcon", active_icon);
			g_free (active_icon);
		}
		else if (strcmp (fields[i], "Class") == 0) {
			kiba_key_file_write_desktop_entry (filename, "Class", "NotYetLaunched");
		}
		else if (strcmp (fields[i], "LauncherGroup") == 0) {
			static const struct { const char *name; } groups[] = {
				{ "WebBrowser" },    { "Filemanager" }, { "AudioVideo" },
				{ "Accessibility" }, { "Viewer" },      { "Game" },
				{ "Office" },        { "Network" },     { "Utility" },
				{ "Development" },   { "Prefs" },       { "System" },
				{ "GNOME" },         { "X-XFCE" }
			};
	
			char *categories = NULL, *group = NULL;
			int   fnord;
	
			categories = g_key_file_get_value (file, key_entry_group,
						"Categories", &error);
			g_clear_error (&error);
			error = NULL;
			
			if (categories) {
				for (fnord = 0; fnord < G_N_ELEMENTS (groups); fnord++) {
					if (g_strrstr (categories, groups[fnord].name)) {
						group = g_strdup (groups[fnord].name);
						break;
					}
				}
			}
	
			if (group == NULL) {
				group = g_key_file_get_value (file, key_entry_group, "Name", &error);
				g_clear_error (&error);
				error = NULL;
			}
				
			if (group != NULL) {
				kiba_key_file_write_desktop_entry (filename,
					"LauncherGroup", group);
				g_free (group);
			}
		}
		else if (strcmp (fields[i], "GroupMember") == 0) {
			kiba_key_file_write_desktop_entry (filename, "GroupMember", "1");
		}
	}
	
	g_key_file_free (file);
	
	return TRUE;
}
