/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include <locale.h>

#include "kiba.h"
#include "kiba-plugin-loader.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "kiba-globals.h"
#include "kiba-surface-factory.h"
#include "prefs-dbus.h"

#include "prefs-private.h"
#include "surface-factory-private.h"
#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#ifdef USE_KDE
	const char *file_manager_global = "konqueror";
#else
#ifdef USE_XFCE
	const char *file_manager_global = "thunar";
#else
	const char *file_manager_global = "nautilus --no-desktop";
#endif
#endif

int main (int argc, char *argv[]) {
	Kiba *kiba;
	KibaPluginLoader *loader;
#ifdef USE_LIBSN
	KibaSn *sn;
#endif
	char *pluginPath = NULL;
	char *dockPluginPath, *globalPluginPath;
	gint i;
	
#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, DATADIR"/locale");
	textdomain (PACKAGE);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
#endif
	
	setlocale (LC_NUMERIC, "C");  /* use decimal point instead of comma */
	setlocale (LC_CTYPE, "");
	setlocale (LC_MESSAGES, "");
	
	/* parsing command line parameters */
	for (i = 0; i < argc; i++) {
		if (g_strrstr (argv[i], "--help") != NULL ||
		    g_strrstr (argv[i], "-h") != NULL) {
			fprintf (stderr, "Usage: kiba-dock [--verbose] [--plugin-path plugin_path] ...\n");
			return EXIT_SUCCESS;
		}
		/* enables debugging */
		else if (g_strrstr (argv[i], "--verbose") != NULL ||
			 g_strrstr (argv[i], "-v") != NULL)
			kiba_debug_init ();
		else if (i < argc + 1) {
			if (g_strrstr (argv[i], "--plugin-path") != NULL) {
				pluginPath = argv[i+1];
				if (!g_file_test (pluginPath, G_FILE_TEST_EXISTS)) {
					fprintf (stderr, "Invalid Directory in --plugin-path argument: '%s'\n", pluginPath);
					pluginPath = NULL;
				}
			}
		}
	}
	
#ifdef HAVE_SVG
	DEBUG_INFO("Kiba was build with SVG support\n");
#else
	DEBUG_INFO("Kiba was build without SVG support\n");
#endif
#ifdef USE_GLITZ
	DEBUG_INFO("Kiba was build with GLITZ support\n");
#else
	DEBUG_INFO("Kiba was build without GLITZ support\n");
	#ifdef HAVE_GLITZ
	#ifndef CAIRO_HAS_GLITZ_SURFACE
		DEBUG_MISC("Your current cairo installation does not have glitz Support.\n");
		DEBUG_MISC("Cairo does not include glitz support by default.\n"
			"To use glitz in kiba you must recompile cairo with '--enable-glitz'.\n");
	#endif
	#endif
#endif
#ifdef USE_SDL
	DEBUG_INFO("Kiba was build with SDL support\n");
#else
	DEBUG_INFO("Kiba was build without SDL support\n");
#endif
	
	if (!pluginPath)
		pluginPath = LIBDIR "/kiba-dock";
	if (!g_file_test (pluginPath, G_FILE_TEST_IS_DIR)) {
		DEBUG_ERROR (
				"Failed to locate Plugins at '%s'\n"
				"Please install the Plugins at '" LIBDIR "/kiba-dock' "
				"or use the '--plugin-path' command line parameter.\n",
				pluginPath);
		return EXIT_FAILURE;
	}
	
	DEBUG_MISC("Initialling Gtk+ Library\n");
//TODO: is this already needed for the thumbnail-factory thread?
// 	g_thread_init (NULL);
// 	gdk_threads_init ();
// 	gdk_threads_enter ();
	
	gtk_disable_setlocale();
	gtk_init (&argc, &argv);
	
#ifdef HAVE_SVG
	DEBUG_MISC("Initiall RSVG Library\n");
	rsvg_init ();
#endif
	
	/* show tips dialog */
// 	kiba_dialog_tips ();
	
	/* make the plugins at pluginPath loadable */
	loader = kiba_plugin_loader_get();
	globalPluginPath = g_strdup_printf ("%s/global_plugins", pluginPath);
	if (!kiba_plugin_loader_initiall_plugins_at_path (loader, KIBA_PLUGIN_GLOBAL, globalPluginPath)) {
		g_object_unref (G_OBJECT(loader));
		g_free(globalPluginPath);
		return EXIT_FAILURE;
	}
	g_free(globalPluginPath);
	dockPluginPath = g_strdup_printf ("%s/dock_plugins", pluginPath);
	if (!kiba_plugin_loader_initiall_plugins_at_path (loader, KIBA_PLUGIN_DOCK, dockPluginPath)) {
		g_object_unref (G_OBJECT(loader));
		g_free(dockPluginPath);
		return EXIT_FAILURE;
	}
	g_free(dockPluginPath);
	
	/* set up the preferences database */
	kiba_prefs_get ();
	/* set up the dbus interface */
	prefs_dbus_get();
	
#ifdef USE_LIBSN
	sn = kiba_sn_get ();
#endif
	
	/* creating the docks */
	kiba = kiba_get();
	
	/* enter the GTK mainloop */
	gtk_main();
// 	gdk_threads_leave ();
	
	
	DEBUG_MISC("Destroying docks..\n");
	g_object_unref (G_OBJECT(kiba));
	
	DEBUG_MISC("Cleaning up..\n");
	g_object_unref (G_OBJECT(loader));
#ifdef USE_LIBSN
	g_object_unref(G_OBJECT(sn));
#endif
	prefs_dbus_fini ();
	kiba_screen_destroy (kiba_screen_get());
	kiba_prefs_fini ();
	kiba_surface_factory_destroy();
	DEBUG_MISC("Cleaned up!\n");

#ifdef HAVE_SVG
	rsvg_term ();
#endif
	
	return EXIT_SUCCESS;
}
