/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *  Note   : based on some gnome-panel functions
*/

#include <string.h>
#include <gdk/gdkx.h>

#include "kiba-surface-factory.h"
#include "kiba-image.h"
#include "kiba-menu-items.h"
#include "kiba-key-file.h"
#include "kiba-utils.h"
#include "kiba-x11.h"

static GtkTooltips *panel_tooltips = NULL;

static void drag_begin_menu_cb (GtkWidget      *widget,
				GdkDragContext *context) {
	/* FIXME: workaround for a possible gtk+ bug
	 *    See bugs #92085(gtk+) and #91184(panel) for details.
	 */
	gtk_tooltips_disable (panel_tooltips);
}

/* This is a _horrible_ hack to have this here. This needs to be added to the
 * GTK+ menuing code in some manner.
 */
static void drag_end_menu_cb (GtkWidget      *widget,
			      GdkDragContext *context) {
  GtkWidget *xgrab_shell;
  GtkWidget *parent;

  /* Find the last viewable ancestor, and make an X grab on it
   */
  parent = widget->parent;
  xgrab_shell = NULL;

  /* FIXME: workaround for a possible gtk+ bug
   *    See bugs #92085(gtk+) and #91184(panel) for details.
   */
//   if (panel_global_config_get_tooltips_enabled ())
    gtk_tooltips_enable (panel_tooltips);

  while (parent)
    {
      gboolean viewable = TRUE;
      GtkWidget *tmp = parent;
      
      while (tmp)
	{
	  if (!GTK_WIDGET_MAPPED (tmp))
	    {
	      viewable = FALSE;
	      break;
	    }
	  tmp = tmp->parent;
	}
      
      if (viewable)
	xgrab_shell = parent;
      
      parent = GTK_MENU_SHELL (parent)->parent_menu_shell;
    }
  
  if (xgrab_shell && !GTK_MENU(xgrab_shell)->torn_off)
    {
      GdkCursor *cursor = gdk_cursor_new (GDK_ARROW);

      if ((gdk_pointer_grab (xgrab_shell->window, TRUE,
			     GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			     GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
			     GDK_POINTER_MOTION_MASK,
			     NULL, cursor, GDK_CURRENT_TIME) == 0))
	{
	  if (gdk_keyboard_grab (xgrab_shell->window, TRUE,
				 GDK_CURRENT_TIME) == 0)
	    GTK_MENU_SHELL (xgrab_shell)->have_xgrab = TRUE;
	  else
	    {
	      gdk_pointer_ungrab (GDK_CURRENT_TIME);
	    }
	}

      gdk_cursor_unref (cursor);
    }
}

static void drag_data_get_string_cb (GtkWidget        *widget,
				     GdkDragContext   *context,
				     GtkSelectionData *selection_data,
				     guint             info,
				     guint             time,
				     const char       *string) {
	gtk_selection_data_set (selection_data,
				selection_data->target, 8, (guchar *)string,
				strlen(string));
}

GtkTooltips* kiba_get_tooltips () {
	if (!panel_tooltips)
		panel_tooltips = gtk_tooltips_new ();
	
	return panel_tooltips;
}

void kiba_menu_item_setup_uri_drag (GtkWidget   *menuitem,
				    const char  *icon,
				    const char  *uri) {
	static GtkTargetEntry menu_item_targets[] = {
		{ "text/uri-list", 0, 0 }
	};

	gtk_drag_source_set (menuitem,
			     GDK_BUTTON1_MASK|GDK_BUTTON2_MASK,
			     menu_item_targets, 1,
			     GDK_ACTION_COPY);

	if (icon != NULL)
		gtk_drag_source_set_icon_name (menuitem, icon);
	
	g_signal_connect (G_OBJECT (menuitem), "drag_begin",
			  G_CALLBACK (drag_begin_menu_cb), NULL);
	if (uri)
		g_signal_connect_data (G_OBJECT (menuitem), "drag_data_get",
			       G_CALLBACK (drag_data_get_string_cb),
			       g_strdup (uri),
			       (GClosureNotify)g_free,
			       0 /* connect_flags */);
	g_signal_connect (G_OBJECT (menuitem), "drag_end",
			  G_CALLBACK (drag_end_menu_cb), NULL);
}

static void __image_destroy_cb(GtkWidget *image, gpointer data) {
	cairo_surface_t *surface = data;
	kiba_surface_factory_remove_surface(surface);
}

static void __surface_loaded(cairo_surface_t *surface, gint size, gpointer data) {
	GtkWidget *image_menu_item = data;
	GtkWidget *image;
	
	image = kiba_image_new ();
	kiba_image_set_size(KIBA_IMAGE(image), size, size);
	kiba_image_set_surface(KIBA_IMAGE(image), surface, size, size);
	
	g_object_set_data_full (G_OBJECT (image_menu_item),
				"Panel:Image",
				g_object_ref (image),
				(GDestroyNotify) g_object_unref);
 
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (image_menu_item), image);
	
	g_signal_connect (G_OBJECT(image), "destroy",
		G_CALLBACK(__image_destroy_cb), surface);
	
	gtk_widget_show (image);
}

void kiba_load_menu_image_deferred (GtkWidget   *image_menu_item,
				    GtkIconSize  icon_size,
				    const char  *stock_id,
				    const char  *image_filename,
				    const char  *fallback_image_filename) {
	int icon_height = 24;
	
	gtk_icon_size_lookup (icon_size, NULL, &icon_height);
	
	if (image_filename && kiba_surface_factory_load_icon (image_filename, icon_height, __surface_loaded, image_menu_item))
		return;
	
	if (stock_id && kiba_surface_factory_load_icon (stock_id, icon_height, __surface_loaded, image_menu_item))
		return;
	
	if (fallback_image_filename && kiba_surface_factory_load_icon (fallback_image_filename, icon_height, __surface_loaded, image_menu_item))
		return;
}

void kiba_menu_item_setup (GtkWidget   *menuitem,
			   GtkIconSize  icon_size,
			   GtkWidget   *image,
			   const char  *title,
			   gboolean     create_invisible_mnemonic) {
	GtkWidget *label;

	if (create_invisible_mnemonic) {

		label = g_object_new (GTK_TYPE_ACCEL_LABEL, NULL);
		gtk_label_set_text_with_mnemonic (GTK_LABEL (label), title);

		gtk_label_set_pattern (GTK_LABEL (label), "");

		gtk_accel_label_set_accel_widget (GTK_ACCEL_LABEL (label),
						  menuitem);
	} else
		label = gtk_label_new_with_mnemonic (title);

	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	gtk_widget_show (label);
	
	gtk_container_add (GTK_CONTAINER (menuitem), label);
	
	if (image) {
		g_object_set_data_full (G_OBJECT (menuitem),
					"Panel:Image",
					g_object_ref (image),
					(GDestroyNotify) g_object_unref);
		gtk_widget_show (image);
		if (GTK_IS_IMAGE_MENU_ITEM (menuitem))
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem), image);
	}

	gtk_widget_show (menuitem);
}

void kiba_menu_item_setup_with_icon (GtkWidget   *item,
				     GtkIconSize  icon_size,
				     const char  *icon_name,
				     const char  *stock_id,
				     const char  *title,
				     gboolean     create_invisible_mnemonic) {
	if (icon_name || stock_id)
		kiba_load_menu_image_deferred (item, icon_size,
						stock_id, icon_name, NULL);

	kiba_menu_item_setup (item, icon_size, NULL, title, create_invisible_mnemonic);
}

void kiba_menu_item_create_from_desktop (GtkWidget *menu,
					 char      *path,
					 char      *force_name) {
	static const char file_uri_prefix[] = "file://";
	
	GKeyFile *key_file;
	gboolean loaded;
	GtkWidget *item;
	char *path_freeme;
	char *full_path;
	char *uri;
	char *icon;
	char *name;
	char *comment;

	path_freeme = NULL;

	key_file = g_key_file_new ();

	if (g_path_is_absolute (path)) {
		loaded = g_key_file_load_from_file (key_file, path,
						    G_KEY_FILE_NONE, NULL);
		full_path = path;
	} else {
		char *lookup_file;
		char *desktop_path;

		if (!g_str_has_suffix (path, ".desktop")) {
			desktop_path = g_strconcat (path, ".desktop", NULL);
		} else {
			desktop_path = path;
		}

		lookup_file = g_strconcat ("applications", G_DIR_SEPARATOR_S,
					   desktop_path, NULL);
		loaded = g_key_file_load_from_data_dirs (key_file, lookup_file,
							 &path_freeme,
							 G_KEY_FILE_NONE,
							 NULL);
		full_path = path_freeme;
		g_free (lookup_file);

		if (desktop_path != path)
			g_free (desktop_path);
	}

	if (!loaded) {
		g_key_file_free (key_file);
		if (path_freeme)
			g_free (path_freeme);
		return;
	}

	icon = kiba_key_file_get_desktop_entry (key_file, "Icon");
	comment = kiba_key_file_get_desktop_entry (key_file, "Comment");

	if (!force_name || strlen (force_name) == 0)
		name = kiba_key_file_get_desktop_entry (key_file, "Name");
	else
		name = g_strdup (force_name);

	item = gtk_image_menu_item_new ();
	kiba_menu_item_setup_with_icon (item, GTK_ICON_SIZE_MENU,
				   icon, NULL, name, TRUE);

	if (comment != NULL)
		gtk_tooltips_set_tip (kiba_get_tooltips(), item, comment, NULL);

	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	g_signal_connect_data (item, "activate",
			       G_CALLBACK (kiba_menu_item_activate_desktop_file),
			       g_strdup (full_path),
			       (GClosureNotify) g_free, 0);
	g_signal_connect (G_OBJECT (item), "button_press_event",
			  G_CALLBACK (kiba_menu_item_dummy_button_press_event), NULL);
	
	if (g_str_has_prefix (full_path, file_uri_prefix))
		kiba_menu_item_setup_uri_drag (item, full_path, icon);
	else {
		while (strlen (full_path) > 3 && full_path[1] == '/')
			full_path++;
		
		uri = g_strdup_printf ("%s%s", file_uri_prefix, full_path);
		kiba_menu_item_setup_uri_drag (item, uri, icon);
		g_free (uri);
	}

	g_key_file_free (key_file);

	if (icon)
		g_free (icon);

	if (name)
		g_free (name);

	if (comment)
		g_free (comment);

	if (path_freeme)
		g_free (path_freeme);
}

gboolean kiba_menu_item_dummy_button_press_event (GtkWidget      *menuitem,
						  GdkEventButton *event) {
// 	if (event->button == 3)
// 		return TRUE;
	
	return FALSE;
}

void kiba_menu_item_activate_desktop_file (GtkWidget  *menuitem,
					   const char *path) {
	kiba_key_file_launch_at_path (path);
}

GtkWidget* kiba_menu_add_separator (GtkWidget *menu) {
	GtkWidget *menuitem;
	
	menuitem = gtk_separator_menu_item_new ();
	gtk_widget_set_sensitive (menuitem, FALSE);
	gtk_widget_show (menuitem);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);

	return menuitem;
}
