
/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-object
 * @short_description: Base class for objects on #KibaDock.
 * @see_also: #KibaWin,#KibaDock
 * @stability: Unstable
 *
 * Abstract base class for objects on #KibaDock.
 * #KibaObject is responsible for scaling/positiong objects on #KibaDock.
 */

#include "config.h"

#include <string.h>
#include <math.h>

#include "kiba.h"
#include "kiba-object.h"
#include "kiba-drawable-object.h"
#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-macros.h"
#include "kiba-x11.h"

#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#define KIBA_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_OBJECT, KibaObjectPrivate))

G_DEFINE_ABSTRACT_TYPE (KibaObject, kiba_object, GTK_TYPE_BIN)

typedef struct {
	GdkWindow *event_window;
	
	char *group;    /**< dock group */

	gint destroy_id;
	
	float scale_x; /* surface scale factor */
	float scale_y; /* surface scale factor */
	
	guint workTodo;
} KibaObjectPrivate;

/* signal ids */
enum {
	SIG_URGENT_TOGGLED,
	SIG_WORKING_TOGGLED,
	SIG_ACTIVE_TOGGLED,
	SIG_DRAGGING_TOGGLED,
	SIG_KIBA_OBJECT_LAST
};

/* propertie ids */
enum {
	_PROP_GROUP = 1,
};

static guint kiba_object_signals[SIG_KIBA_OBJECT_LAST] = { 0 };

/*** Private Methods ***/

static gint __destroy_timeout (gpointer data) {
	KibaObject *object = data;
	KibaObjectPrivate *priv;
	GtkWidget *dock;
	float old_factor;
	
	dock = GTK_WIDGET(object)->parent;
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (KIBA_IS_DRAWABLE_OBJECT(object))
		kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object));
	old_factor = object->destroy_factor;
	object->destroy_factor -= 0.05f;
	
	if (object->position == KIBA_OBJECT_ON_DOCK && KIBA_DOCK(dock)->scroll_offset > 0) {
		KIBA_DOCK(dock)->scroll_offset -= (GTK_WIDGET(object)->allocation.width -
			object->destroy_factor*GTK_WIDGET(object)->allocation.width)/4;
		KIBA_DOCK(dock)->scroll_offset = MAX(KIBA_DOCK(dock)->scroll_offset, 0);
	}
	
	kiba_object_update_scale_factor (object);
	
	if (object->destroy_factor <= 0.0f) {
		object->destroy_factor = 0.0f;
		if (GTK_WIDGET(object)->parent)
			gtk_container_remove (GTK_CONTAINER(dock), GTK_WIDGET(object));
		else
			gtk_widget_destroy (GTK_WIDGET(object));
		return FALSE;
	} else {
		gtk_widget_queue_resize_no_redraw (GTK_WIDGET(object));
	}
	
	return TRUE;
}

static gint __scale_timeout (gpointer data) {
	KibaObject *object = data;
	KibaObjectPrivate *priv;
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (priv->destroy_id)
		return FALSE;
	
	object->destroy_factor += 0.05f;
	object->destroy_factor = MIN(1.0f, object->destroy_factor);
	
	kiba_object_update_scale_factor (object);
	gtk_widget_queue_resize_no_redraw (GTK_WIDGET(object));
	
	return (object->destroy_factor != 1.0f);
}

static void __realize_cb (GtkWidget *widget,
                          gpointer data) {
	KibaObject *object = KIBA_OBJECT(widget);
	KibaWin *win = KIBA_WIN(widget->parent->parent);
	Kiba *kiba = kiba_get();
	
	g_return_if_fail(GTK_WIDGET_REALIZED(widget));
	
	kiba_title_add_object (kiba_get_title(kiba), object);
	
	kiba_progress_add_object (win->progress, object);
	
	g_timeout_add (20, __scale_timeout, object);
}

/*** Public methods ***/

/**
 * kiba_object_get_state:
 * @object: The #KibaObject instance.
 * @state: The state flag.
 *
 * Checks if state flag is set.
 *
 * Returns: Wether @state is set.
 */
gboolean kiba_object_get_state (KibaObject *object,
                                KibaObjectState state) {
	g_return_val_if_fail(KIBA_IS_OBJECT(object), FALSE);
	return GET_FLAG(object->state, state);
}

/**
 * kiba_object_set_state:
 * @object: The #KibaObject instance.
 * @state: The state flag.
 *
 * Sets state flag.
 */
void kiba_object_set_state (KibaObject *object,
                            KibaObjectState state) {
	KibaObjectPrivate *priv;
	
	g_return_if_fail(KIBA_IS_OBJECT(object));
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (state == KIBA_OBJECT_WORKING)
		priv->workTodo++;
	
	if (GET_FLAG(object->state, state))
		return;
	SET_FLAG(object->state, state);
	
	switch (state) {
		case KIBA_OBJECT_WORKING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_WORKING_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_ACTIVE: {
			GtkWidget *widget;
			KibaWin *win;
			
			widget = GTK_WIDGET(object);
			win = KIBA_WIN(widget->parent->parent);
			
			kiba_triangle_add_widget (win->triangle, widget);
			
			g_signal_emit (object,
				kiba_object_signals[SIG_ACTIVE_TOGGLED],
				0, TRUE);
			break;
		}
		case KIBA_OBJECT_URGENT: {
			g_signal_emit (object,
				kiba_object_signals[SIG_URGENT_TOGGLED],
				0, TRUE);
		}
		case KIBA_OBJECT_DRAGGING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_DRAGGING_TOGGLED],
				0, TRUE);
		}
		case KIBA_OBJECT_PRELIGHTED: {
			if (KIBA_IS_DRAWABLE_OBJECT(object))
				kiba_drawable_object_queue_draw(KIBA_DRAWABLE_OBJECT(object));
		}
	}
}

/**
 * kiba_object_get_state:
 * @object: The #KibaObject instance.
 * @state: The state flag.
 *
 * Unsets state flag.
 */
void kiba_object_unset_state (KibaObject *object,
                              KibaObjectState state) {
	KibaObjectPrivate *priv;
	
	g_return_if_fail(KIBA_IS_OBJECT(object));
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (state == KIBA_OBJECT_WORKING) {
		priv->workTodo--;
		if (priv->workTodo > 0)
			return;
	}
	
	if (!GET_FLAG(object->state, state))
		return;
	DEL_FLAG(object->state, state);
	
	switch (state) {
		case KIBA_OBJECT_WORKING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_WORKING_TOGGLED],
				0, FALSE);
			break;
		}
		case KIBA_OBJECT_ACTIVE: {
			GtkWidget *widget;
			KibaWin *win;
			
			widget = GTK_WIDGET(object);
			win = KIBA_WIN(widget->parent->parent);
			
			kiba_triangle_remove_widget (win->triangle, widget);
			
			g_signal_emit (object,
				kiba_object_signals[SIG_ACTIVE_TOGGLED],
				0, FALSE);
		}
		case KIBA_OBJECT_URGENT: {
			g_signal_emit (object,
				kiba_object_signals[SIG_URGENT_TOGGLED],
				0, FALSE);
		}
		case KIBA_OBJECT_DRAGGING: {
			g_signal_emit (object,
				kiba_object_signals[SIG_DRAGGING_TOGGLED],
				0, FALSE);
		}
		case KIBA_OBJECT_PRELIGHTED: {
			if (KIBA_IS_DRAWABLE_OBJECT(object))
				kiba_drawable_object_queue_draw(KIBA_DRAWABLE_OBJECT(object));
		}
	}
}

/**
 * kiba_object_set_tooltip_text:
 * @object: The KibaObject instance.
 *
 */
void kiba_object_set_tooltip_text(KibaObject *object,
                                  const char *text) {
	g_return_if_fail(KIBA_IS_OBJECT(object));
#if GTK_CHECK_VERSION(2,12,0) == TRUE
	gtk_widget_set_tooltip_text(GTK_WIDGET(object), text);
#endif
}

/**
 * kiba_object_launch:
 * @object: The KibaObject Instance.
 * @command: The command to launch.
 *
 * The method executes the given command.
 * Additional it passes the command to libsn, to monitor the launch process.
 *
 * Returns: The PID.
 */
int kiba_object_launch(KibaObject *object,
                       const char *command) {
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
#ifdef USE_LIBSN
	/* monitor this launch with libsn */
	kiba_sn_launch(command, G_OBJECT(object));
#endif
	return kiba_launch_programm(command);
}

/**
 * kiba_object_get_destroy_factor:
 * @object: The #kibaObject instance.
 *
 * Returns the current destroy factor.
 *
 * Returns: The destroy factor.
 */
float kiba_object_get_destroy_factor (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), 1.0f);
	
	return object->destroy_factor;
}

/**
 * kiba_object_set_icon_geometry:
 * @object: The KibaObject instance.
 * @window: The KibaWindow to set the _NET_WM_ICON property on.
 *
 * Sets the _NET_WM_ICON property on the window, for minimze/restore animations.
 */
void kiba_object_set_icon_geometry (KibaObject *object,
                                    KibaWindow *window) {
	GtkWidget *dock;
	GtkAllocation geometry;
	gint fX, fY;
	
	g_return_if_fail (KIBA_IS_WINDOW(window));
	
	dock = GTK_WIDGET(object)->parent;
	
	fX = (int) object->alloc.x;
	fY = (int) object->alloc.y;
	
	geometry.width = GTK_WIDGET(object)->allocation.width;
	geometry.height = GTK_WIDGET(object)->allocation.height;
	geometry.x = dock->parent->allocation.x + fX;
	geometry.y = dock->parent->allocation.y + fY;
	
	kiba_window_set_minimize_restore_point(window, geometry.x, geometry.y, geometry.width, geometry.height);
}

/**
 * kiba_object_show_title:
 * @widget: The KibaObject instance.
 *
 * Shows the object title.
 */
void kiba_object_show_title (GtkWidget *widget) {
	Kiba *kiba;
	KibaTitle *title;
	
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	kiba = kiba_get();
	title = kiba_get_title(kiba);
	
	if (!kiba_title_get_flag(title, KIBA_TITLE_DRAWABLE))
		return;
	
	kiba_title_show_object (title, KIBA_OBJECT(widget));
}

/**
 * kiba_object_hide_title:
 * @widget: The KibaObject instance.
 *
 * Hides the object title.
 */
void kiba_object_hide_title (GtkWidget *widget) {
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	Kiba *kiba = kiba_get();
	KibaTitle *title = kiba_get_title(kiba);
	
	kiba_title_hide_object (title, KIBA_OBJECT(widget));
}

/**
 * kiba_object_set_title:
 * @object: The KibaObject instance.
 * @text: The object title.
 *
 * Sets the object title's text.
 */
void kiba_object_set_title (KibaObject *object,
                            const char *text) {
	g_return_if_fail (KIBA_IS_OBJECT (object));
	Kiba *kiba = kiba_get();
	KibaTitle *title = kiba_get_title(kiba);
	
	kiba_title_set_object_text (title, object, text);
}

/**
 * kiba_object_get_title:
 * @object: The KibaObject instance.
 *
 * Returns the object title.
 *
 * Returns: The object title.
 */
const char* kiba_object_get_title (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT (object), NULL);
	Kiba *kiba = kiba_get();
	KibaTitle *title = kiba_get_title(kiba);
	
	return kiba_title_get_object_text(title, object);
}

/**
 * kiba_object_get_destroying:
 * @object: The #KibaObject instance.
 *
 * Returns wether the object is about to be destroyed.
 *
 * Returns: Wether the object is about to be destroyed.
 */
gboolean kiba_object_get_destroying (KibaObject *object) {
	KibaObjectPrivate *priv;
	
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	return priv->destroy_id != 0;
}

/**
 * kiba_object_set_requisition_factor:
 * @object: The KibaObject instance.
 * @scale_x: The KibaObject widget width scale factor.
 * @scale_y: The KibaObject widget height scale factor.
 *
 * Sets size scale factors of the object.
 * The actual object width is (*object->size)*object->scale_x, ignoring the custom scale factors.
 */
void kiba_object_set_requisition_factor (KibaObject *object,
                                         float scale_x,
                                         float scale_y) {
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	GtkWidget *widget = GTK_WIDGET(object);
	GtkRequisition req;
	GtkAllocation alloc;
	
	if (object->scale_x == scale_x || object->scale_y == scale_y)
		return;
	
	object->scale_x = scale_x;
	object->scale_y = scale_y;
	
	/* resize immediately */
	gtk_widget_queue_resize_no_redraw(widget);
	alloc = widget->allocation;
	gtk_widget_size_request(widget, &req);
	if (req.width || req.height) {
		alloc.width = req.width;
		alloc.height = req.height;
	}
	gtk_widget_size_allocate (widget, &alloc);
}

/**
 * kiba_object_destroy:
 * @object: The KibaObject instance.
 *
 * Destroys the KibaObject, scaling the object down and unrealizing it.
 */
void kiba_object_destroy (KibaObject *object) {
	KibaObjectPrivate *priv;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (!priv->destroy_id)
		priv->destroy_id = g_timeout_add (20, __destroy_timeout, object);
	
	if (kiba_object_get_state(object, KIBA_OBJECT_ACTIVE))
		kiba_object_unset_state(object, KIBA_OBJECT_ACTIVE);
	
	if (KIBA_IS_DRAWABLE_OBJECT(object))
		KIBA_DRAWABLE_OBJECT(object)->paint = NULL;
}

/**
 * kiba_object_remove_custom_offset_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom offset function.
 */
void kiba_object_remove_custom_offset_function_by_data (KibaObject *object,
                                                        gpointer user_data) {
	KibaObjectCustomOffsetData *data;
	GList *l;
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_offset_data = g_list_remove (object->custom_offset_data, l->data);
			
			return;
		}
	}
}

/**
 * kiba_object_add_custom_offset_function:
 * @object: The KibaObject instance.
 * @place_func: The surface offset method.
 * @max_func: The method to get the maximal offset.
 * @user_data: The user data.
 *
 * Adds offset function, for custom surface positions.
 */
void kiba_object_add_custom_offset_function (KibaObject *object,
                                             KibaObjectCustomOffsetFunc place_func,
                                             KibaObjectCustomMaxScaleFunc max_func,
                                             gpointer user_data) {
	KibaObjectCustomOffsetData *data;
	GList *l;
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->place_func == place_func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomOffsetData, 1);
	data->place_func = place_func;
	data->max_func = max_func;
	data->data = user_data;
	
	object->custom_offset_data = g_list_append (object->custom_offset_data, data);
}

/**
 * kiba_object_remove_custom_scale_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom scale function.
 */
void kiba_object_remove_custom_scale_function_by_data (KibaObject *object,
						       gpointer    user_data) {
	KibaObjectCustomScaleData *data;
	GList *l;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_scale_funcs = g_list_remove (object->custom_scale_funcs, l->data);
			
			return;
		}
	}
	
}

/**
 * kiba_object_add_custom_scale_function:
 * @object: The KibaObject instance.
 * @scale_func: The scale method.
 * @max_func: The method to return the maximal scale factor.
 * @scale_widget: TRUE to scale the KibaObject widget, otherwise FALSE to scale only the surface.
 * @user_data The user data.
 *
 * Adds scale function, for custom scale animations.
 */
void kiba_object_add_custom_scale_function (KibaObject *object,
                                            KibaObjectCustomScaleFunc scale_func,
                                            KibaObjectCustomMaxScaleFunc max_func,
                                            gboolean scale_widget,
                                            gpointer user_data) {
	KibaObjectCustomScaleData *data;
	GList *l;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->scale_func == scale_func) {
			data->data = user_data;
			data->scale_widget = scale_widget;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomScaleData, 1);
	data->scale_func = scale_func;
	data->max_func = max_func;
	data->scale_widget = scale_widget;
	data->data = user_data;
	
	object->custom_scale_funcs = g_list_append (object->custom_scale_funcs, data);
}

/**
 * kiba_object_is_on_dock:
 * @object: The #KibaObject instance.
 *
 * Returns whether object is on the dock.
 *
 * Returns: TRUE if the object is on the dock, otherwise FALSE is returned.
 **/
gboolean kiba_object_is_on_dock (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return object->position == KIBA_OBJECT_ON_DOCK;
}

/**
 * kiba_object_update_scale_factor:
 * @object: The #KibaObject instance.
 *
 * Applys custom scale factor changes to the object size.
 *
 * Returns: TRUE if the scale factor changed, otherwise FALSE is returned.
 */
gboolean kiba_object_update_scale_factor (KibaObject *object) {
	KibaObjectPrivate *priv;
	GtkWidget *dock;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	priv = KIBA_OBJECT_GET_PRIVATE(object);
	
	scale_x *= object->destroy_factor;
	scale_y *= object->destroy_factor;
	
	/* add current scale factor of all custom scale methods */
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		float func_scale_x = 1.0f;
		float func_scale_y = 1.0f;
		
		data->scale_func (object, &func_scale_x, &func_scale_y, data->data);
		
		scale_x *= func_scale_x;
		scale_y *= func_scale_y;
	}
	
	/* factor didnt changed */
	if (scale_x == priv->scale_x && scale_y == priv->scale_y)
		return FALSE;
	
	priv->scale_x = scale_x;
	priv->scale_y = scale_y;
	
	if (KIBA_IS_DRAWABLE_OBJECT(object))
		kiba_drawable_object_queue_draw (KIBA_DRAWABLE_OBJECT(object));
	
	return TRUE;
}

/**
 * kiba_object_get_scale_factor:
 * @object: The #KibaObject instance.
 * @scale_x_ret_ The width scale factor return.
 * @scale_y_ret_ The height scale factor return.
 *
 * Returns the surface scale factor.
 */
void kiba_object_get_scale_factor (KibaObject *object,
                                   float *scale_x_ret,
                                   float *scale_y_ret) {
	KibaObjectPrivate *priv;
	g_return_if_fail (KIBA_IS_OBJECT(object));
	priv = KIBA_OBJECT_GET_PRIVATE(object);
	*scale_x_ret = priv->scale_x;
	*scale_y_ret = priv->scale_y;
}

/*** Class Implementations ***/

/**
 * gtk_widget_size_request implementation.
 **/
static void kiba_object_size_request (GtkWidget *widget,
                                      GtkRequisition *requisition) {
	KibaObject *object = KIBA_OBJECT(widget);
	GtkWidget *child = gtk_bin_get_child (GTK_BIN(widget));
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	GList *l;
	
	scale_x = object->scale_x;
	scale_y = object->scale_y;
	requisition->width  = object->size.width;
	requisition->height = object->size.height;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		/* this scale method just scales the surface, not the widget */
		if (!data->scale_widget) continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		scale_x *= custom_factor_x;
		scale_y *= custom_factor_y;
	}
	
	requisition->width = (int) (requisition->width*scale_x);
	requisition->height = (int) (requisition->height*scale_y);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void kiba_object_unrealize (GtkWidget *widget) {
	KibaObject *object = KIBA_OBJECT(widget);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	Kiba *kiba = kiba_get();
	GList *l;
	
	if (priv->destroy_id)
		g_source_remove (priv->destroy_id);
	
	for (l = object->custom_scale_funcs; l != NULL; l = object->custom_scale_funcs) {
		KibaObjectCustomScaleData *data = l->data;
		
		g_free (data);
		object->custom_scale_funcs = g_list_remove (object->custom_scale_funcs, l->data);
	}
	g_list_free (object->custom_scale_funcs);
	
	for (l = object->custom_offset_data; l != NULL; l = object->custom_offset_data) {
		KibaObjectCustomOffsetData *data = l->data;
		
		g_free (data);
		object->custom_offset_data = g_list_remove (object->custom_offset_data, l->data);
	}
	g_list_free (object->custom_offset_data);
	
	kiba_title_remove_object (kiba_get_title(kiba), object);
	
	(* GTK_WIDGET_CLASS (kiba_object_parent_class)->unrealize) (widget);
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_object_realize (GtkWidget *widget) {
	/* this method must be implemented by the subclass */
	g_warning ("KibaObjectClass::realize not implemented for `%s'",
		g_type_name (G_TYPE_FROM_INSTANCE (widget)));
}

/**
 * g_object_set_property implementation.
 **/
static void kiba_object_set_property (GObject *object,
                                      guint property_id,
                                      const GValue *value,
                                      GParamSpec *pspec) {
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_GROUP:
			priv->group = g_value_dup_string (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/**
 * g_object_get_property implementation.
 **/
static void kiba_object_get_property (GObject *object,
                                      guint property_id,
                                      GValue *value,
                                      GParamSpec *pspec) {
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_GROUP: 
			g_value_set_string (value, priv->group);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/*** Class init ***/

static void kiba_object_class_init (KibaObjectClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaObjectPrivate));

	/* GObject method implementations */
	gobject_class->set_property = kiba_object_set_property;
	gobject_class->get_property = kiba_object_get_property;
	
	/* GtkWidget method implementations */
	widget_class->realize = kiba_object_realize;
	widget_class->unrealize = kiba_object_unrealize;
	widget_class->size_request = kiba_object_size_request;
	
	/* properties */
	g_object_class_install_property 
		(gobject_class,
		 _PROP_GROUP,
		 g_param_spec_string ("group",
		 "Group",
		 "Group on the Dock",
		 NULL,
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	
	/* signals */
	kiba_object_signals[SIG_URGENT_TOGGLED] =
		g_signal_new ("urgent-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, urgent_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	kiba_object_signals[SIG_ACTIVE_TOGGLED] =
		g_signal_new ("active-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, active_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	kiba_object_signals[SIG_WORKING_TOGGLED] =
		g_signal_new ("working-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, working_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	kiba_object_signals[SIG_DRAGGING_TOGGLED] =
		g_signal_new ("dragging-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, dragging_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
}

static void kiba_object_init (KibaObject *object) {
	static const GtkTargetEntry targets[] = {
		{ "text/plain", 0, 0 }
	};
	
	KibaObjectPrivate *priv;
	
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	gtk_widget_add_events (GTK_WIDGET (object),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	object->position = KIBA_OBJECT_ON_DOCK;
	object->scale_x = 1.0f;
	object->scale_y = 1.0f;
	object->alloc = (KibaAllocation) {0.0f, 0.0f, 0.0f, 0.0f};
	object->size = (GtkRequisition) {28, 28};
	object->destroy_factor = 0.01f;
	
	object->custom_scale_funcs = NULL;
	
	priv->workTodo = 0;
	priv->scale_x = 1.0f;
	priv->scale_y = 1.0f;
	
	gtk_drag_dest_set (GTK_WIDGET (object),
			GTK_DEST_DEFAULT_ALL,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
	gtk_drag_dest_add_uri_targets (GTK_WIDGET (object));
	
	/* catching the realize event, since we dont do the realizing in this class */
	g_signal_connect_after(G_OBJECT(object), "realize",
		G_CALLBACK(__realize_cb), NULL);
}
