/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-plugin-loader
 * @short_description: Loads modules.
 * @see_also:#KibaPlugin
 * @stability: Unstable
 *
 * Loads modules.
 */

#include "../config.h"

#include <string.h>
#include <libintl.h>

#include "kiba.h"
#include "kiba-plugin-loader.h"
#include "kiba-plugin.h"
#include "kiba-debug.h"
#include "kiba-dialog.h"
#include "kiba-utils.h"

#include "prefs-private.h"
#include "kiba-private.h"

#define KIBA_PLUGIN_LOADER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_PLUGIN_LOADER, KibaPluginLoaderPrivate))

#define _(string)  gettext (string)

G_DEFINE_TYPE (KibaPluginLoader, kiba_plugin_loader, G_TYPE_OBJECT)

struct _KibaPluginService {
	KibaPlugin *plugin;
	const char *id;
	const DBusGMethodInfo *method;
};

typedef struct {
	GList *pluginPaths; /* list of pathes to plugins */
	GList *plugins; /* list of loaded plugins, remove plugins from here _only_ in the dispose handler */
	GList *services; /* list of KibaPluginService's */
	
	GList *dependencyDialogs;
	GList *dependOnDialogs;
}KibaPluginLoaderPrivate;

typedef struct {
	GtkWidget *dialog;
	KibaPlugin *plugin; /* the plugin that was about to load/unload */
	GList *dependencies; /* list of dependencies that needs to be loaded/unloaded */
	KibaPrefsFile *configFile;
}KibaDependencyDialogData;

/* signals */
enum {
	SIG_PLUGIN_SERVICE_REGISTERED = 0,
	SIG_PLUGIN_SERVICE_UNREGISTERED,
	SIG_PLUGIN_LOADER_LAST
};

static guint kiba_plugin_loader_signals[SIG_PLUGIN_LOADER_LAST] = { 0 };

static KibaPluginLoader *plugin_loader_instance = NULL;

/*** Private methods ***/

/** Dialogs **/

static KibaDependencyDialogData* __get_dialog_for_plugin (KibaPlugin *plugin,
                                                          GList *dialogList) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	GList *l;
	for (l = dialogList; l != NULL; l = l->next) {
		KibaDependencyDialogData *data = l->data;
		if (data->plugin == plugin)
			return data;
	}
	return NULL;
}

static char* __dep_list_to_string (GList *deps) {
	char *ret = NULL;
	GList *l;
	
	for (l = deps; l != NULL; l = l->next) {
		KibaPluginDependencyData *pluginID = l->data;
		if (!ret)
			ret = g_strdup(pluginID->id);
		else {
			char *buf = ret;
			ret = g_strdup_printf("%s, %s", buf, pluginID->id);
			g_free(buf);
		}
	}
	
	return ret;
}

static char* __format_depend_on_message (GList *deps) {
	char *dependencyList;
	char *ret;
	
	dependencyList = __dep_list_to_string(deps);
	ret = g_strdup_printf (_("Do you want to disable following plugins: '%s'?"), dependencyList);
	g_free(dependencyList);
	
	return ret;
}

static char* __format_dependency_message (GList *deps) {
	char *dependencyList;
	char *ret;
	
	dependencyList = __dep_list_to_string(deps);
	ret = g_strdup_printf (_("Do you want to enable following plugins: '%s'?"), dependencyList);
	g_free(dependencyList);
	
	return ret;
}

static void __dependency_dialog_destroyed (GtkWidget *widget,
                                           gpointer data) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	KibaDependencyDialogData *pluginData = data;
	GList *l;
	
	if (g_list_find(priv->dependencyDialogs, pluginData))
		priv->dependencyDialogs = g_list_remove(priv->dependencyDialogs, pluginData);
	else if (g_list_find(priv->dependOnDialogs, pluginData))
		priv->dependOnDialogs = g_list_remove(priv->dependencyDialogs, pluginData);
	
	for (l = pluginData->dependencies; l != NULL; l = pluginData->dependencies) {
		KibaPluginDependencyData *foo = l->data;
		g_free(foo->id);
		g_free(foo);
		pluginData->dependencies = g_list_remove(pluginData->dependencies, l->data);
	}
	g_list_free(pluginData->dependencies);
	g_free(pluginData);
}

static void __dependency_dialog_response (GtkDialog *dialog,
                                          gint response_id,
                                          gpointer data) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	KibaDependencyDialogData *pluginData = data;
	Kiba *kiba = kiba_get();
	
	switch (response_id) {
		case GTK_RESPONSE_CANCEL:
			gtk_widget_destroy(GTK_WIDGET(dialog));
			break;
		case GTK_RESPONSE_YES: {
			KibaPrefs *prefs = kiba_prefs_get();
			KibaPrefsFile *file = pluginData->configFile;
			KibaPlugin *plugin = pluginData->plugin;
			KibaPlugin *firstPlugin;
			KibaWin *win = kiba_get_plugin_win(kiba, plugin);
			GList *l, *failedDeps = NULL, *depDialogs = NULL;
			char *configKey;
			
			for (l = pluginData->dependencies; l != NULL; l = l->next) {
				KibaPluginDependencyData *dep = l->data;
				KibaPlugin *depPlugin;
				KibaPrefsFile *depFile = dep->file;
				
				depPlugin = kiba_get_plugin_for_id (kiba, dep->file, dep->id);
				if (depPlugin && !depPlugin->loaded) {
					configKey = g_strdup_printf("%s_enable", dep->id);
					kiba_prefs_set_option(depFile, configKey, "bool", "TRUE");
					if (depPlugin->loaded)
						kiba_prefs_write_to_config_file(depFile, configKey, "bool", "TRUE");
					g_free(configKey);
				}
				
				/* the plugin was loaded due to calling kiba_prefs_set_option above, we can exit */
				if (plugin->loaded)
					return;
				
				if (!depPlugin)
					failedDeps = g_list_append(failedDeps, dep->id);
				else if (!depPlugin->loaded) {
					KibaDependencyDialogData *dialog = __get_dialog_for_plugin(depPlugin, priv->dependencyDialogs);
					
					if (!dialog)
						failedDeps = g_list_append(failedDeps, dep->id);
					else
						depDialogs = g_list_append(depDialogs, dep->id);
				}
			}
			
			if (failedDeps != NULL) { /* some plugins failed to load */
				char *failedList = kiba_utils_glist_to_string(failedDeps);
				DEBUG_ERROR ("Failed to load plugin '%s': Could not load '%s'\n",
					G_TYPE_MODULE(plugin)->name, failedList);
				g_free(failedList);
				
				for (l = failedDeps; l != NULL; l = failedDeps)
					failedDeps = g_list_remove(failedDeps, l->data);
				g_list_free(failedDeps);
			} else if (depDialogs != NULL) { /* some plugins popped up a dependency dialog */
				char *message = __format_dependency_message(pluginData->dependencies);
				gtk_message_dialog_format_secondary_text (
					GTK_MESSAGE_DIALOG(pluginData->dialog), message);
				g_free(message);
				
				for (l = depDialogs; l != NULL; l = depDialogs)
					depDialogs = g_list_remove(depDialogs, l->data);
				g_list_free(depDialogs);
				//we keep the dialog waiting what happens with the other dialog(s)
				//TODO test me
				break;
			} else { /* everything is fine :) */
				configKey = g_strdup_printf("%s_enable", G_TYPE_MODULE(plugin)->name);
				kiba_prefs_set_option(file, configKey, "bool", "TRUE");
				g_free(configKey);
			}
			
			gtk_widget_destroy(GTK_WIDGET(dialog));
			
			break;
		}
		default:
			break;
	}
}

static void __depend_on_dialog_response (GtkDialog *dialog,
                                          gint response_id,
                                          gpointer data) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	KibaDependencyDialogData *pluginData = data;
	
	switch (response_id) {
		case GTK_RESPONSE_CANCEL:
			gtk_widget_destroy(GTK_WIDGET(dialog));
			break;
		case GTK_RESPONSE_YES: {
			Kiba *kiba = kiba_get();
			KibaPrefs *prefs = kiba_prefs_get();
			KibaPrefsFile *file = kiba_get_plugin_config_file(kiba, pluginData->plugin);
			KibaPlugin *sourcePlugin = pluginData->plugin;
			KibaWin *win = kiba_get_plugin_win(kiba, sourcePlugin);
			GList *l, *failedDeps = NULL, *depDialogs = NULL;
			char *configKey;
			
			for (l = pluginData->dependencies; l != NULL; l = l->next) {
				KibaPluginDependencyData *depData = l->data;
				KibaPlugin *depPlugin;
				KibaPrefsFile *depFile = depData->file;
				
				depPlugin = kiba_get_plugin_for_id (kiba, depData->file, depData->id);
				if (!depPlugin) continue;
				
				if (depPlugin->loaded) {
					configKey = g_strdup_printf("%s_enable", depData->id);
					kiba_prefs_set_option(depFile, configKey, "bool", "FALSE");
					if (depPlugin->loaded) /* save changes to the config file */
						kiba_prefs_write_to_config_file(depFile, configKey, "bool", "FALSE");
					g_free(configKey);
				}
				
				/* the plugin was unloaded due to calling kiba_prefs_set_option above, we can exit */
				if (!sourcePlugin->loaded)
					return;
				
				if (depPlugin->loaded) {
					KibaDependencyDialogData *dialog = __get_dialog_for_plugin(depPlugin, priv->dependOnDialogs);
					
					if (!dialog)
						failedDeps = g_list_append(failedDeps, depData->id);
					else if (depPlugin->loaded)
						depDialogs = g_list_append(depDialogs, dialog);
				}
			}
			
			if (failedDeps != NULL) { /* some plugins failed to unload */
				char *failedList = kiba_utils_glist_to_string(failedDeps);
				DEBUG_ERROR ("Failed to unload plugin '%s': Could not unload '%s'\n",
					G_TYPE_MODULE(sourcePlugin)->name, failedList);
				g_free(failedList);
				
				for (l = failedDeps; l != NULL; l = failedDeps)
					failedDeps = g_list_remove(failedDeps, l->data);
				g_list_free(failedDeps);
			} else if (depDialogs) { /* some plugins popped up a dependency dialog */
				char *message = __format_depend_on_message(pluginData->dependencies);
				gtk_message_dialog_format_secondary_text (
					GTK_MESSAGE_DIALOG(pluginData->dialog), message);
				g_free(message);
				
				for (l = depDialogs; l != NULL; l = depDialogs)
					depDialogs = g_list_remove(depDialogs, l->data);
				g_list_free(depDialogs);
				//we keep the dialog waiting what happens with the other dialog(s)
				break;
			} else { /* everything is fine :) */
				configKey = g_strdup_printf("%s_enable", G_TYPE_MODULE(sourcePlugin)->name);
				kiba_prefs_set_option(file, configKey, "bool", "FALSE");
				g_free(configKey);
			}
			
			gtk_widget_destroy(GTK_WIDGET(dialog));
			
			break;
		}
		default:
			break;
	}
}

void kiba_plugin_loader_popup_dependency_dialog (KibaPlugin *plugin,
                                                 GList *dependencies,
                                                 KibaPrefsFile *configFile) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	KibaDependencyDialogData *callback_data;
	KibaPluginDependencyData *depData;
	char *title, *message;
	
	callback_data = __get_dialog_for_plugin(plugin, priv->dependencyDialogs);
	if (callback_data) {
		GList *l;
		for (l = dependencies; l != NULL; l = dependencies) {
			GList *deps;
			gboolean depAdded = FALSE;
			KibaPluginDependencyData *depData = l->data;
			dependencies = g_list_remove(dependencies, l->data);
			for (deps = callback_data->dependencies; deps != NULL; deps = deps->next) {
				KibaPluginDependencyData *addedDepData = deps->data;
				if (strcmp(depData->id, addedDepData->id) == 0 && addedDepData->type == depData->type) {
					depAdded = TRUE;
					break;
				}
			}
			if (!depAdded)
				callback_data->dependencies = g_list_append(callback_data->dependencies, l->data);
		}
		g_list_free(dependencies);
		
		char *message = __format_dependency_message(callback_data->dependencies);
		gtk_message_dialog_format_secondary_text (
			GTK_MESSAGE_DIALOG(callback_data->dialog), message);
		g_free(message);
	} else {
		callback_data = g_new0(KibaDependencyDialogData, 1);
		callback_data->dependencies = dependencies;
		callback_data->plugin = plugin;
		callback_data->configFile = configFile;
		title = g_strdup_printf (_("'%s' plugin requires a disabled dependencies"), G_TYPE_MODULE(plugin)->name);
		message = __format_dependency_message(dependencies);
		callback_data->dialog = kiba_dialog_confirm(title, message);
		g_free(title);
		g_free(message);
		
		g_signal_connect (G_OBJECT(callback_data->dialog), "response",
			G_CALLBACK(__dependency_dialog_response), callback_data);
		g_signal_connect (G_OBJECT(callback_data->dialog), "destroy",
			G_CALLBACK (__dependency_dialog_destroyed), callback_data);
		
		priv->dependencyDialogs = g_list_append(priv->dependencyDialogs, callback_data);
	}
}

void kiba_plugin_loader_popup_depend_on_dialog (KibaPlugin *plugin,
                                                GList *dependencies,
                                                KibaPrefsFile *file) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	KibaDependencyDialogData *data;
	char *title, *message;
	
	data = __get_dialog_for_plugin(plugin, priv->dependOnDialogs);
	if (data) {
		GList *l;
		for (l = dependencies; l != NULL; l = dependencies) {
			GList *deps;
			gboolean depAdded = FALSE;
			KibaPluginDependencyData *depData = l->data;
			dependencies = g_list_remove(dependencies, l->data);
			for (deps = data->dependencies; deps != NULL; deps = deps->next) {
				KibaPluginDependencyData *addedDepData = deps->data;
				if (strcmp(depData->id, addedDepData->id) == 0 && addedDepData->type == depData->type) {
					depAdded = TRUE;
					break;
				}
			}
			if (!depAdded)
				data->dependencies = g_list_append(data->dependencies, l->data);
		}
		g_list_free(dependencies);
		
		char *message = __format_depend_on_message(data->dependencies);
		gtk_message_dialog_format_secondary_text (
			GTK_MESSAGE_DIALOG(data->dialog), message);
		g_free(message);
	} else {
		data = g_new0(KibaDependencyDialogData, 1);
		data->dependencies = dependencies;
		data->plugin = plugin;
		data->configFile = file;
		title = g_strdup_printf (_("Some loaded plugins depend on the '%s' plugin"),
				G_TYPE_MODULE(plugin)->name);
		message = __format_depend_on_message(dependencies);
		data->dialog = kiba_dialog_confirm(title, message);
		g_free(title);
		g_free(message);
		
		g_signal_connect (G_OBJECT(data->dialog), "response",
			G_CALLBACK(__depend_on_dialog_response), data);
		g_signal_connect (G_OBJECT(data->dialog), "destroy",
			G_CALLBACK (__dependency_dialog_destroyed), data);
		
		gtk_window_present(GTK_WINDOW(data->dialog));
		
		priv->dependOnDialogs = g_list_append(priv->dependOnDialogs, data);
	}
}

/** service helper **/

static const char *get_method_data (const DBusGObjectInfo *object,
                                    const DBusGMethodInfo *method) {
	return object->data + method->data_offset;
}

static void __free_value_array(GValueArray *array) {
	while (array->n_values > 0) {
		GValue *value = g_value_array_get_nth (array, 0);
		g_value_array_remove (array, 0);
		if (G_IS_VALUE(value))
			g_value_unset(value);
	}
	g_value_array_free(array);
}

static void __plugin_dbus_registered (KibaPlugin *plugin,
                                      KibaPluginLoader *loader) {
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	gint i;
	
	for (i = 0; i < plugin->dbus_info->n_method_infos; i++) {
		KibaPluginService *service;
		const DBusGMethodInfo *method;
		const char *methodData;
		const char *methodID;
		
		method = &(plugin->dbus_info->method_infos[i]);
		methodData = get_method_data(plugin->dbus_info, method);
		methodID = &(methodData[1]);
		
		/* FIXME: is there a better way to obtain the method Name? */
		while (methodID[0] != '\0') methodID++;
		methodID++;
		
		service = g_new0(KibaPluginService, 1);
		service->plugin = plugin;
		service->id = methodID;
		service->method = method;
		
		priv->services = g_list_append(priv->services, service);
		
		g_signal_emit (loader, kiba_plugin_loader_signals[SIG_PLUGIN_SERVICE_REGISTERED], 0, service);
	}
}

/*** Public Functions ***/

/**
 * kiba_plugin_loader_get_service:
 * @loader: The #KibaPluginLoader instance.
 * @request_plugin: The #KibaPlugin that requests the service, or NULL.
 * @seviceID: The service ID.
 *
 * Get a plugin service.
 *
 * Returns: A KibaPluginService instance, or NULL if no service is associated to @serviceID.
 */
KibaPluginService* kiba_plugin_loader_get_service (KibaPluginLoader *loader,
                                                   KibaPlugin *request_plugin,
                                                   const char *serviceID) {
	KibaPluginLoaderPrivate *priv;
	Kiba *kiba = kiba_get();
	KibaWin *win = NULL;
	GList *l;
	
	if (request_plugin)
		win = kiba_get_plugin_win(kiba, request_plugin);
	
	priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	for (l = priv->services; l != NULL; l = l->next) {
		KibaPluginService *service = l->data;
		KibaWin *serviceWin = NULL;
		
		if (win)
			serviceWin = kiba_get_plugin_win(kiba, service->plugin);
		
		/* skip services offered for other windows */
		if (kiba_plugin_type(service->plugin) == KIBA_PLUGIN_DOCK && win != serviceWin)
			continue;
		
		if (strcmp(service->id, serviceID) == 0)
			return service;
	}
	
	return NULL;
}

/**
 * kiba_plugin_loader_invoke_service:
 * @loader: The #KibaPluginLoader instance.
 * @service: The KibaPluginService to invoke.
 * @value_array: A GValueArray holding the service parameter.
 *
 * Invokes service associated to @service with the given parameters @value_array.
 * Note that the first GObject* parameter and the last GError** parameter are added to @value_array here.
 * @value_array is freed in this method, to allow passing kiba_utils_value_array_init() as parameter.
 *
 * Returns: TRUE on success, otherwhise FALSE is returned.
 */
gboolean kiba_plugin_loader_invoke_service (KibaPluginLoader *loader,
                                            KibaPluginService *service,
                                            GValueArray *value_array) {
	GValue return_value = {0,};
	GValue error_value = {0,};
	GError *error = NULL;
	GClosure closure;
	gboolean success;
	GList *l, *plugins;
	
	/* This is evil.  We do this to work around the fact that
	 * the generated glib marshallers check a flag in the closure object
	 * which we don't care about.  We don't need/want to create
	 * a new closure for each invocation.
	 */
	memset (&closure, 0, sizeof (closure));
	
	/* prepend KibaPlugin as first parameter */
	g_value_array_prepend (value_array, NULL);
	g_value_init (g_value_array_get_nth (value_array, 0), G_TYPE_OBJECT);
	g_value_set_object (g_value_array_get_nth (value_array, 0), service->plugin);
	
	/* append GError parameter */
	g_value_init (&error_value, G_TYPE_POINTER);
	g_value_set_pointer(&error_value, &error);
	g_value_array_append (value_array, &error_value);
	
	/* initiall return value to TRUE */
	g_value_init (&return_value, G_TYPE_BOOLEAN);
	g_value_set_boolean(&return_value, TRUE);
	
	/* invoke the service */
	service->method->marshaller (&closure, &return_value,
				value_array->n_values,
				value_array->values,
				NULL, service->method->function);
	
	if (error != NULL) {
		g_error_free(error);
		success = FALSE;
	} else {
		success = g_value_get_boolean(&return_value);
	}
	
	/* cleanup */
	g_value_unset(&return_value);
	__free_value_array(value_array);
	
	return TRUE;
}

/**
 * kiba_plugin_loader_for_all_plugins:
 * @loader: The KibaPluginLoader instance.
 * @for_all_func: The Function to launch for every KibaPlugin.
 * @data: The UserData for the Function.
 *
 * Launches function for all loaded KibaPlugin's.
 */
void kiba_plugin_loader_for_all_plugins (KibaPluginLoader *loader,
                                         GFunc for_all_func,
                                         void  *data) {
	KibaPluginLoaderPrivate *priv;
	GList *l;
	g_return_if_fail(KIBA_IS_PLUGIN_LOADER(loader));
	priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	for (l = priv->plugins; l != NULL; l = l->next)
		for_all_func (KIBA_PLUGIN(l->data), data);
}

/**
 * kiba_plugin_loader_get_plugins:
 * @loader: The KibaPluginLoader instance.
 *
 * Returns an internal copy of a list holding #KibaPlugin's.
 *
 * Returns: List of #KibaPlugin's.
 */
GList* kiba_plugin_loader_get_plugins (KibaPluginLoader *loader) {
	KibaPluginLoaderPrivate *priv;
	g_return_val_if_fail(KIBA_IS_PLUGIN_LOADER(loader), NULL);
	priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	return priv->plugins;
}

/**
 * kiba_plugin_loader_get_plugin_paths:
 * @loader: The KibaPluginLoader instance.
 *
 * Returns an internal copy of plugin pathes.
 *
 * Returns: List of plugin pathes.
 */
GList* kiba_plugin_loader_get_plugin_paths (KibaPluginLoader *loader) {
	KibaPluginLoaderPrivate *priv;
	g_return_val_if_fail(KIBA_IS_PLUGIN_LOADER(loader), NULL);
	priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	return priv->pluginPaths;
}

/**
 * kiba_plugin_loader_initiall_plugins_at_path:
 * @loader: The KibaPluginLoader instance.
 * @plugin_path: The path to load plugins from.
 *
 * Loads all plugins from plugin_path.
 *
 * Returns: TRUE on success.
 */
gboolean kiba_plugin_loader_initiall_plugins_at_path (KibaPluginLoader *loader,
                                                      KibaPluginType type,
                                                      const char *plugin_path) {
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	GDir *dir;
	GError *error = NULL;
	char *file_name;
	
	DEBUG_INFO("Initialling plugins at '%s'\n", plugin_path);
	
	dir = g_dir_open (plugin_path, 0, &error);
	if (error != NULL) {
		DEBUG_ERROR ("Failed to open Plugin Directory '%s': %s\n", plugin_path, error->message);
		g_error_free (error);
		return FALSE;
	}
	
	while (TRUE) {
		KibaPluginLoaderPluginData *data;
		const char *file_name = g_dir_read_name (dir);
		GList *l;
		
		/* quit loop if there are no more files */
		if (!file_name) break;
		
		if (!g_str_has_suffix (file_name, ".so")) continue;
		
		gboolean added = FALSE;
		for (l = priv->pluginPaths; l != NULL; l = l->next) {
			data = l->data;
			if (g_str_has_suffix(data->path, file_name)) {
				DEBUG_ERROR("Failed to initiall plugin '%s': another plugin with the same name was initialled before!\n", file_name);
				added = TRUE;
				break;
			}
		}
		if (added) continue;
		
		data = g_new0 (KibaPluginLoaderPluginData, 1);
		data->path = g_strdup_printf("%s/%s", plugin_path, file_name);
		data->type = type;
		priv->pluginPaths = g_list_append(priv->pluginPaths, data);
	}
	g_dir_close (dir);
	
	return TRUE;
}

/**
 * kiba_plugin_loader_new_plugin:
 * @loader: The #KibaPluginLoader instance.
 * @plugin_path: The path to load plugin from.
 *
 * Initialls plugin at @plugin_path.
 *
 * Returns: The #KibaPlugin created.
 */
KibaPlugin* kiba_plugin_loader_new_plugin (KibaPluginLoader *loader,
                                           const char *pluginPath) {
	GList *l;
	KibaPlugin *plugin;
	KibaPluginLoaderPluginData *data;
	gboolean valid;
	
	g_return_val_if_fail(KIBA_IS_PLUGIN_LOADER(loader), NULL);
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	
	valid = FALSE;
	for (l = priv->pluginPaths; l != NULL; l = l->next) {
		data = l->data;
		if (strcmp(pluginPath, data->path) == 0) {
			valid = TRUE;
			break;
		}
	}
	g_return_val_if_fail(valid == TRUE, NULL);
	
	plugin = kiba_plugin_new (data->path, data->type);
	priv->plugins = g_list_append(priv->plugins, plugin);
	
	g_signal_connect(G_OBJECT(plugin), "dbus-registered",
		G_CALLBACK(__plugin_dbus_registered), loader);
	
	return plugin;
}

/**
 * kiba_plugin_loader_load_plugin:
 * @loader: The #KibaPluginLoader instance.
 * @plugin: The #KibaPlugin to load.
 *
 * Loads @plugin.
 */
gboolean kiba_plugin_loader_load_plugin (KibaPluginLoader *loader,
                                         KibaPlugin *plugin) {
	g_return_if_fail(KIBA_IS_PLUGIN_LOADER(loader));
	
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	GList *l;
	
	g_return_val_if_fail(g_list_find(priv->plugins, plugin) != NULL, FALSE);
	
	/* load the plugin */
	if (!g_type_module_use (G_TYPE_MODULE(plugin))) {
		kiba_plugin_loader_unload_plugin(loader, plugin);
		return FALSE;
	}
	
	/* update dependency dialogs, maybe there is a dialog waiting for this plugin being loaded */
	for (l = priv->dependencyDialogs; l != NULL; l = l->next) {
		KibaDependencyDialogData *data = l->data;
		gboolean depRemoved = FALSE;
		GList *plugs;
		
		for (plugs = data->dependencies; plugs != NULL; plugs = plugs->next) {
			KibaPluginDependencyData *dep = plugs->data;
			
			if (strcmp(G_TYPE_MODULE(plugin)->name, dep->id) == 0) {
				data->dependencies = g_list_remove (data->dependencies, dep);
				g_free(dep->id);
				g_free(dep);
				depRemoved = TRUE;
				break;
			}
		}
		
		if (!depRemoved) continue;
		
		if (g_list_length(data->dependencies) == 0) {
			data->dependencies = NULL;
			gtk_dialog_response (GTK_DIALOG(data->dialog), GTK_RESPONSE_YES);
			if (!priv->dependencyDialogs)
				break;
			else
				l = priv->dependencyDialogs;
		} else {
			char *message = __format_dependency_message(data->dependencies);
			gtk_message_dialog_format_secondary_text (
				GTK_MESSAGE_DIALOG(data->dialog), message);
			g_free(message);
		}
	}
	
	return TRUE;
}

/**
 * kiba_plugin_loader_unload_plugin:
 * @loader: The #KibaPluginLoader instance.
 * @plugin: The #KibaPlugin to unload.
 *
 * Unloads @plugin.
 */
void kiba_plugin_loader_unload_plugin (KibaPluginLoader *loader,
                                       KibaPlugin *plugin) {
	g_return_if_fail(KIBA_IS_PLUGIN_LOADER(loader));
	g_return_if_fail(KIBA_IS_PLUGIN(plugin));
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	GList *l;
	
	if (plugin->loaded) {
		/* we need to destroy the plugin types before calling g_type_module_unuse,
		to decrease the module counter, otherwise kiba_plugin_unload would not be called */
		if (plugin->vtable->unload)
			plugin->vtable->unload (plugin);
		g_type_module_unuse (G_TYPE_MODULE(plugin));
	}
	
	for (l = priv->services; l != NULL; l = l->next) {
		KibaPluginService *service = l->data;
		
		if (service->plugin != plugin) continue;
		
		if (l->next) l = l->next;
		else l = NULL;
		
		g_signal_emit (loader, kiba_plugin_loader_signals[SIG_PLUGIN_SERVICE_UNREGISTERED], 0, service);
		
		priv->services = g_list_remove(priv->services, service);
		g_free(service);
		
		if (l == NULL) break;
	}
	
	plugin->vtable = NULL;
	plugin->options = NULL;
	plugin->option_elements = 0;
	plugin->dependencies = NULL;
	plugin->num_dependencies = 0;
	
	/* update depend on dialogs, maybe there is a dialog waiting for this plugin being unloaded */
	for (l = priv->dependOnDialogs; l != NULL; l = l->next) {
		KibaDependencyDialogData *data = l->data;
		gboolean depRemoved = FALSE;
		GList *plugs;
		
		for (plugs = data->dependencies; plugs != NULL; plugs = plugs->next) {
			KibaPluginDependencyData *dep = plugs->data;
			
			if (strcmp(G_TYPE_MODULE(plugin)->name, dep->id) == 0) {
				data->dependencies = g_list_remove (data->dependencies, dep);
				g_free(dep->id);
				g_free(dep);
				depRemoved = TRUE;
				break;
			}
		}
		
		if (!depRemoved) continue;
		
		if (g_list_length(data->dependencies) == 0) {
			data->dependencies = NULL;
			gtk_dialog_response (GTK_DIALOG(data->dialog), GTK_RESPONSE_YES);
			if (!priv->dependOnDialogs)
				break;
			else
				l = priv->dependOnDialogs;
		} else {
			char *message = __format_depend_on_message(data->dependencies);
			gtk_message_dialog_format_secondary_text (
				GTK_MESSAGE_DIALOG(data->dialog), message);
			g_free(message);
		}
	}
}

/**
Class Functions
**/

static void kiba_plugin_loader_dispose (GObject *object) {
	KibaPluginLoader *loader = KIBA_PLUGIN_LOADER(object);
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	GList *l;
	
	for (l = priv->plugins; l != NULL; l = priv->plugins) {
		KibaPlugin *plugin = l->data;
		priv->plugins = g_list_remove(priv->plugins, l->data);
		if (plugin->loaded)
			kiba_plugin_loader_unload_plugin(loader, plugin);
		//g_object_unref(G_OBJECT(plugin)); //plugins cant be destroyed, because they extend from GTypeModule
	}
	g_list_free(priv->plugins);
	
	for (l = priv->pluginPaths; l != NULL; l = priv->pluginPaths) {
		KibaPluginLoaderPluginData *data = l->data;
		priv->pluginPaths = g_list_remove(priv->pluginPaths, l->data);
		g_free(data->path);
		g_free(data);
	}
	g_list_free(priv->pluginPaths);
	
	for (l = priv->dependencyDialogs; l != NULL; l = priv->dependencyDialogs) {
		KibaDependencyDialogData *data = l->data;
		gtk_widget_destroy(data->dialog);
	}
	g_list_free(priv->dependencyDialogs);
	
	for (l = priv->dependOnDialogs; l != NULL; l = priv->dependOnDialogs) {
		KibaDependencyDialogData *data = l->data;
		gtk_widget_destroy(data->dialog);
	}
	g_list_free(priv->dependOnDialogs);
	
	G_OBJECT_CLASS(kiba_plugin_loader_parent_class)->dispose (object);
}

static void kiba_plugin_loader_class_init (KibaPluginLoaderClass *class) {
	GObjectClass *object_class = G_OBJECT_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaPluginLoaderPrivate));
	
	object_class->dispose = kiba_plugin_loader_dispose;
	
	/* signals */
	kiba_plugin_loader_signals[SIG_PLUGIN_SERVICE_REGISTERED] =
		g_signal_new ("service-registered",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaPluginLoaderClass, service_registered),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_POINTER);
	kiba_plugin_loader_signals[SIG_PLUGIN_SERVICE_UNREGISTERED] =
		g_signal_new ("service-unregistered",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaPluginLoaderClass, service_unregistered),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_POINTER);
}

static void kiba_plugin_loader_init (KibaPluginLoader *loader) {
	KibaPluginLoaderPrivate *priv = KIBA_PLUGIN_LOADER_GET_PRIVATE(loader);
	
	priv->pluginPaths = NULL;
	priv->plugins = NULL;
	priv->dependencyDialogs = NULL;
	priv->dependOnDialogs = NULL;
	priv->services = NULL;
}

static KibaPluginLoader* kiba_plugin_loader_new (void) {
	return g_object_new (KIBA_TYPE_PLUGIN_LOADER, NULL);
}

/**
 * kiba_plugin_loader_get:
 *
 * Creates/returns the KibaPluginLoader instance.
 *
 * Returns: The KibaPluginLoader instance.
 */
KibaPluginLoader* kiba_plugin_loader_get (void) {
	if (!plugin_loader_instance)
		plugin_loader_instance = kiba_plugin_loader_new();
	
	return plugin_loader_instance;
}
