/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-plugin
 * @short_description: Module Type.
 * @see_also:#KibaPluginLoader
 * @stability: Unstable
 *
 * Module Type.
 */

#include "../config.h"
#include <math.h>
#include <string.h>
#include <dlfcn.h>
#include <libintl.h>

#include "kiba.h"
#include "kiba-private.h"
#include "kiba-dialog.h"
#include "kiba-plugin.h"
#include "kiba-plugin-loader.h"
#include "kiba-types.h"
#include "kiba-debug.h"

#define KIBA_PLUGIN_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_PLUGIN, KibaPluginPrivate))

#define _(string)  gettext (string)

G_DEFINE_TYPE (KibaPlugin, kiba_plugin, G_TYPE_TYPE_MODULE)

typedef struct _KibaPluginPrivate KibaPluginPrivate;

struct _KibaPluginPrivate {
	KibaPluginType type; /**< global or dock plugin? */
	char *lib_path; /**< the library path */
	
	void *handle;   /* opened library, or NULL */
	
	gboolean dbus_registered; /* flag indicating if the plugin object is registered on the bus */
};

/* signal ids */
enum {
	SIG_UNLOADED = 0,
	SIG_LOADED,
	SIG_DBUS_REGISTERED,
	SIG_PLUGIN_LAST
};

/* propertie ids */
enum {
	_PROP_LIB = 1,
	_PROP_TYPE,
};

static guint plugin_signals[SIG_PLUGIN_LAST] = { 0 };

static GList *registerdPlugins = NULL;

KibaPlugin* __get_plugin_at_path (const char *path,
                                  KibaPrefsFile *file) {
	Kiba *kiba = kiba_get();
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	GList *l, *plugins;
	
	plugins = kiba_plugin_loader_get_plugins(loader);
	for (l = plugins; l != NULL; l = l->next) {
		KibaPlugin *plugin = l->data;
		KibaPrefsFile *plugFile;
		
		plugFile = kiba_get_plugin_config_file (kiba, plugin);
		if (plugFile != file)
			continue;
		
		if (strcmp(path, kiba_plugin_get_lib_path(plugin)) != 0)
			continue;
		
		return plugin;
	}
	
	return NULL;
}

static void __popup_conflict_dialog (KibaPlugin *plugin,
                                     const char *conflictID) {
	char *title;
	title = g_strdup_printf (_("Failed to load '%s' plugin"), G_TYPE_MODULE(plugin)->name);
	kiba_dialog_error(title, _("Plugin is conflicts with '%s' plugin, both can't be enabled at the same time."), conflictID);
	g_free(title);
}

/*** public methods ***/

void kiba_plugin_set_up_dbus (KibaPlugin *plugin,
                              const DBusGObjectInfo *info) {
	KibaPluginPrivate  *priv;
	KibaPluginClass *class;
	DBusGConnection *connection;
	DBusGProxy *proxy;
	guint32 ret;
	GError *error = NULL;
	char *namespace;
	char *path;
	
	g_return_if_fail(KIBA_IS_PLUGIN(plugin));
	g_return_if_fail(plugin->dbus_info == NULL);
	
	priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	plugin->dbus_info = info;
	
	class = KIBA_PLUGIN_GET_CLASS(plugin);
	
	/* Get the connection and ensure the name is not used yet */
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (connection == NULL) {
		g_error_free (error);
		return;
	}
	
	proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	namespace = g_strdup_printf (KIBA_PLUGIN_DBUS_NAMESPACE ".%s", G_TYPE_MODULE(plugin)->name);
	if (!org_freedesktop_DBus_request_name (proxy, namespace, 0, &ret, &error)) {
		g_error_free (error);
		g_free(namespace);
		return;
	}
	g_free(namespace);
	
	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (class), info);
	
	/* Register the app on the bus */
	if (!priv->dbus_registered) {
		path = g_strdup_printf (KIBA_PLUGIN_DBUS_PATH "/%s", G_TYPE_MODULE(plugin)->name);
		dbus_g_connection_register_g_object (connection, path, G_OBJECT (plugin));
		g_free(path);
		priv->dbus_registered = TRUE;
	}
	
	g_signal_emit (plugin, plugin_signals[SIG_DBUS_REGISTERED], 0);
}

gboolean kiba_plugin_register (KibaPlugin *plugin) {
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), FALSE);

	/* If this plugin has been registered already then exit */
	if (g_list_find(registerdPlugins, plugin))
		return TRUE;
	
	if (plugin->vtable->load == NULL || plugin->vtable->unload == NULL) {
		DEBUG_ERROR ("'%s' plugin is not loadable: "
			"The plugin does not implement all required functions\n", G_TYPE_MODULE(plugin)->name);
		return FALSE;
	}
	
	registerdPlugins = g_list_append(registerdPlugins, plugin);
	
	return TRUE;
}

/**
 * kiba_plugin_type:
 * @plugin: The #KibaPlugin instance.
 *
 * Get type of @plugin
 *
 * Return: The plugin type.
 */
KibaPluginType kiba_plugin_type(KibaPlugin *plugin) {
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), KIBA_PLUGIN_DOCK);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	return priv->type;
}

/**
 * kiba_plugin_get_lib_path:
 * @plugin: The #KibaPlugin instance.
 *
 * Get library path of @plugin
 *
 * Return: The plugin library path.
 */
const char* kiba_plugin_get_lib_path(KibaPlugin *plugin) {
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), NULL);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	return priv->lib_path;
}

/*** Class implementations ***/

static void kiba_plugin_unload (GTypeModule *module) {
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPlugin *plugin = KIBA_PLUGIN(module);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	gint i;
	
	g_signal_emit (plugin, plugin_signals[SIG_UNLOADED], 0);
	
	if (g_list_find(registerdPlugins, plugin))
		registerdPlugins = g_list_remove (registerdPlugins, plugin);
	
	KibaPrefsFile *configFile = kiba_get_plugin_config_file(kiba_get(), plugin);
	kiba_prefs_remove_options (configFile, plugin->options);
	
	dlclose (priv->handle);
	priv->handle = NULL;
	plugin->loaded = FALSE;
	
	//FIXME unregister the dbus info
	plugin->dbus_info = NULL;
	plugin->vtable = NULL;
	plugin->options = NULL;
	plugin->option_elements = 0;
	plugin->dependencies = NULL;
	plugin->num_dependencies = 0;
	
	DEBUG_MISC("'%s' Plugin unloaded\n", G_TYPE_MODULE(plugin)->name);
}

static gboolean kiba_plugin_load (GTypeModule *module) {
	typedef gboolean (*PluginInitFunc) (KibaPlugin*);
	
	Kiba *kiba = kiba_get();
	KibaPrefs *prefs = kiba_prefs_get();
	KibaPrefsFile *configFile;
	KibaWin *win;
	KibaPluginLoader *loader = kiba_plugin_loader_get();
	KibaPlugin *plugin = KIBA_PLUGIN(module);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	PluginInitFunc plugin_init_func;
	gboolean dependenciesMet = TRUE;
	void *handle = NULL;
	char *dlsym_error = NULL;
	GList *pluginPaths, *plugins;
	GList *l;
	gint i;
	
	plugin_init_func = NULL;
	
	/* Load The library */
	handle = dlopen (priv->lib_path, RTLD_LAZY);
	if (!handle) {
		DEBUG_ERROR("failed to open library '%s': %s\n",
			priv->lib_path, dlerror());
		return FALSE;
	}
	
	dlerror();
	plugin_init_func = (PluginInitFunc) dlsym(handle, "kiba_init_plugin");
	dlsym_error = dlerror();
	if (dlsym_error) {
		DEBUG_ERROR ("'%s' is not usable because the "
					"'kiba_init_plugin' symbol could not be "
					"found.  Does the plugin call the "
					"KIBA_INIT_PLUGIN() macro?\n", priv->lib_path);
		dlclose (handle);
		return FALSE;
	}
	
	if (!plugin_init_func (plugin)) {
		dlclose (handle);
		DEBUG_ERROR("failed to register Plugin '%s'\n", priv->lib_path);
		return FALSE;
	}
	
	if (plugin->vtable->loadable_flag == KIBA_PLUGIN_ONCE_LOADABLE) {
		for (l = registerdPlugins; l != NULL; l = l->next) {
			KibaPlugin *plug = l->data;
			if (plug == plugin) continue;
			if (strcmp(kiba_plugin_get_lib_path(plug), priv->lib_path) == 0) {
				DEBUG_ERROR("Failed to load plugin at '%s': The plugin can only be loaded once!\n", priv->lib_path);
				dlclose (handle);
				return FALSE;
			}
		}
	}
	
	priv->handle = handle;
	
	configFile = kiba_get_plugin_config_file(kiba, plugin);
	win = kiba_get_plugin_win(kiba, plugin);
	
	/* check for conflicting plugins */
	for (i = 0; i < plugin->num_conflicts; i++) {
		KibaPluginConflict conflict = plugin->conflicts[i];
		KibaPrefsFile *conflictFile;
		KibaPlugin *plug;
		
		if (conflict.type == KIBA_PLUGIN_DOCK)
			conflictFile = configFile;
		else
			conflictFile = kiba_prefs_get_general_file(prefs);
		
		plug = kiba_get_plugin_for_id(kiba, conflictFile, conflict.id);
		if (!plug || !plug->loaded) continue;
		
		__popup_conflict_dialog(plugin, conflict.id);
		dlclose (handle);
		return FALSE;
	}
	/* check if there are plugins running mentioning a conflict with this one */
	plugins = kiba_plugin_loader_get_plugins(loader);
	for (l = plugins; l != NULL; l = l->next) {
		KibaPlugin *plug = l->data;
		
		for (i = 0; i < plug->num_conflicts; i++) {
			KibaPluginConflict conflict = plug->conflicts[i];
			
			if (strcmp(conflict.id, G_TYPE_MODULE(plugin)->name) != 0)
				continue;
			if (conflict.type == KIBA_PLUGIN_DOCK && win != kiba_get_plugin_win(kiba, plug))
				continue;
			
			__popup_conflict_dialog(plugin, G_TYPE_MODULE(plug)->name);
			dlclose (handle);
			return FALSE;
		}
	}
	
	/* check for depending plugins */
	pluginPaths = kiba_plugin_loader_get_plugin_paths(loader);
	for (i = 0; i < plugin->num_dependencies; i++) {
		KibaPluginDependency needed_plugin = plugin->dependencies[i];
		KibaPlugin *plug;
		KibaPrefsFile *depFile;
		gboolean pluginFound = FALSE;
		char *enableKey;
		char *suffix;
		GList *l;
		
		if (needed_plugin.type == KIBA_PLUGIN_DOCK)
			depFile = configFile;
		else
			depFile = kiba_prefs_get_general_file(prefs);
		
		plug = kiba_get_plugin_for_id(kiba, depFile, needed_plugin.id);
		if (plug) { /* the needed plugin is already added to the loader */
			if (!plug->loaded) {
				KibaPluginDependencyData *depData;
				depData = g_new0(KibaPluginDependencyData, 1);
				depData->id = g_strdup(needed_plugin.id);
				depData->type = needed_plugin.type;
				depData->file = depFile;
				kiba_plugin_loader_popup_dependency_dialog(plugin, g_list_append(NULL, depData), configFile);
				dependenciesMet = FALSE;
			}
			continue;
		}
		
		dependenciesMet = FALSE;
		
		suffix = g_strdup_printf ("%s.so", needed_plugin.id);
		/* add the plugin and check if it is enabled */
		for (l = pluginPaths; l != NULL; l = l->next) {
			KibaPluginLoaderPluginData *pathData = l->data;
			
			if (pathData->type != needed_plugin.type)
				continue;
			
			if (g_str_has_suffix(pathData->path, suffix)) {
				KibaPlugin *depPlugin;
				KibaPrefsFile *depFile;
				
				if (needed_plugin.type == KIBA_PLUGIN_GLOBAL) {
					kiba_add_general_plugin(pathData->path);
					depFile = kiba_prefs_get_general_file(prefs);
				}
				else {
					kiba_add_dock_plugin(kiba, win, pathData->path);
					depFile = configFile;
				}
				
				depPlugin = __get_plugin_at_path(pathData->path, depFile);
				
				if (!depPlugin) {
					pluginFound = FALSE;
				} else if (depPlugin && !depPlugin->loaded) {
					KibaPluginDependencyData *depData;
					depData = g_new0(KibaPluginDependencyData, 1);
					depData->id = g_strdup(needed_plugin.id);
					depData->type = needed_plugin.type;
					depData->file = depFile;
					kiba_plugin_loader_popup_dependency_dialog(plugin, g_list_append(NULL, depData), configFile);
					pluginFound = TRUE;
				}
				
				break;
			}
		}
		g_free (suffix);
		
		if (!pluginFound) { /* a needed plugin could not be found, this plugin can not be loaded */
			char *title;
			title = g_strdup_printf (_("Failed to load '%s' plugin"), G_TYPE_MODULE(plugin)->name);
			kiba_dialog_error(title, _("Unable to find dependency plugin '%s', you dont have such a plugin installed."), needed_plugin.id);
			g_free(title);
			continue;
		}
	}
	
	if (!dependenciesMet) {
		dlclose (handle);
		return FALSE;
	}
	
	/* Initiall Plugin Options */
	if (plugin->options)
		kiba_prefs_add_options (configFile, plugin->options, plugin->option_elements);
	
	/* Load Plugin */
	plugin->loaded = plugin->vtable->load (plugin);
	if (!plugin->loaded) {
		if (plugin->options)
			kiba_prefs_remove_options (configFile, plugin->options);
		dlclose (handle);
		DEBUG_ERROR ("'%s' is not loadable\n", priv->lib_path);
		return FALSE;
	}
	
	g_signal_emit (plugin, plugin_signals[SIG_LOADED], 0);
	
	DEBUG_MISC("'%s' Plugin loaded\n", G_TYPE_MODULE(plugin)->name);
	
	return TRUE;
}

/**
 * g_object_set_property implementation.
 **/
static void kiba_plugin_set_property (GObject *object,
                                      guint property_id,
                                      const GValue *value,
                                      GParamSpec *pspec) {
	KibaPlugin *self = KIBA_PLUGIN(object);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_LIB: {
			priv->lib_path = g_value_dup_string (value);
			break;
		}
		case _PROP_TYPE: {
			priv->type = g_value_get_int (value);
			break;
		}
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/**
 * g_object_get_property implementation.
 **/
static void kiba_plugin_get_property (GObject *object,
                                      guint property_id,
                                      GValue *value,
                                      GParamSpec *pspec) {
	KibaPlugin *self = KIBA_PLUGIN(object);
	KibaPluginPrivate *priv = KIBA_PLUGIN_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_LIB:
			g_value_set_string (value, priv->lib_path);
			break;
		case _PROP_TYPE:
			g_value_set_int (value, priv->type);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/*** Class init/fini ***/

static void kiba_plugin_class_init (KibaPluginClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GTypeModuleClass *module_class = G_TYPE_MODULE_CLASS (class);
	
	/* GObject method implementations */
	gobject_class->set_property = kiba_plugin_set_property;
	gobject_class->get_property = kiba_plugin_get_property;
	
	module_class->load = kiba_plugin_load;
	module_class->unload = kiba_plugin_unload;
	
	g_type_class_add_private (gobject_class, sizeof (KibaPluginPrivate));
	
	/* properties */
	g_object_class_install_property 
		(gobject_class,
		 _PROP_LIB,
		 g_param_spec_string ("lib",
		 "Lib",
		 "Library path",
		 NULL,
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	g_object_class_install_property 
		(gobject_class,
		 _PROP_TYPE,
		 g_param_spec_int ("type",
		 "Type",
		 "Plugin type",
		 0,
		 KIBA_PLUGIN_TYPE_LAST-1,
		 0,
		 G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE));
	
	/* signals */
	plugin_signals[SIG_UNLOADED] = g_signal_new ("unloaded",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginClass, unloaded),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	plugin_signals[SIG_LOADED] = g_signal_new ("loaded",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginClass, loaded),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	plugin_signals[SIG_DBUS_REGISTERED] = g_signal_new ("dbus-registered",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginClass, dbus_registered),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
}

static void kiba_plugin_init (KibaPlugin *plugin) {
	KibaPluginPrivate  *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	
	plugin->enabled = FALSE;
	plugin->loaded = FALSE;
	priv->handle = NULL;
	plugin->dbus_info = NULL;
	priv->dbus_registered = FALSE;
}

/**
 * kiba_plugin_new:
 * @library_path: The path to the Plugin Library.
 *
 * Creates and returns a new KibaPlugin instance.
 *
 * Returns: The KibaPlugin instance created.
 */
KibaPlugin* kiba_plugin_new (const char *library_path,
                             KibaPluginType type) {
	KibaPlugin  *plugin;
	gchar *so_name;
	
	g_return_val_if_fail(library_path != NULL, NULL);
	
	if (!g_file_test(library_path, G_FILE_TEST_EXISTS))
		return NULL;
	
	DEBUG_MISC("Found plugin library at '%s'\n", library_path);
	
	plugin = g_object_new (KIBA_TYPE_PLUGIN,
				"lib", library_path,
				"type", type,
				NULL);
	
	so_name = g_strrstr (library_path, "/") + 4;
	char name[strlen (so_name)];
	sprintf (name, "%s", so_name);
	name[strlen (so_name)-3] = '\0';
	g_type_module_set_name (G_TYPE_MODULE(plugin), name);
	
	return plugin;
}
