/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <math.h>

#include "kiba-dock.h"
#include "kiba-separator.h"
#include "dock-private.h"

#define SMALL_BORDERWIDTH(width) (2*width)/3

static void __draw_podest_hat (KibaDock *dock,
						    cairo_t  *cr) {
	const KibaLayoutData *d;
	gint border_width;
	gint small_border_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));
	small_border_width = SMALL_BORDERWIDTH(border_width);
	
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		if (d->valignment == 2) { /* dock at the bottom edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* bottom-left point */
				2*small_border_width,
				dock->surface_data.height - border_width);
			cairo_line_to (cr, /* top-left point */
				small_border_width/2 + dock->shape.corner_radius,
				small_border_width/2);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - small_border_width/2 - dock->shape.corner_radius,
				small_border_width/2);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - 2*small_border_width,
				dock->surface_data.height - border_width);
			cairo_line_to (cr, /* bottom-left point */
				2*small_border_width,
				dock->surface_data.height - border_width);
		}
		else { /* dock at the top edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-left point */
				2*small_border_width,
				border_width);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - 2*small_border_width,
				border_width);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - small_border_width/2 - dock->shape.corner_radius,
				dock->surface_data.height);
			cairo_line_to (cr, /* bottom-left point */
				small_border_width/2 + dock->shape.corner_radius,
				dock->surface_data.height);
			cairo_line_to (cr, /* top-left point */
				2*small_border_width,
				border_width);
		}
	}
	else {
		if (d->halignment == 0) { /* dock at the left edge */
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-left point */
				border_width,
				2*small_border_width);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width,
				small_border_width/2 + dock->shape.corner_radius);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width,
				dock->surface_data.height - small_border_width/2 - dock->shape.corner_radius);
			cairo_line_to (cr, /* bottom-left point */
				border_width,
				dock->surface_data.height - 2*small_border_width);
			cairo_line_to (cr, /* top-left point */
				border_width,
				2*small_border_width);
		}
		else {
			cairo_set_line_width (cr, small_border_width);
			cairo_move_to (cr, /* top-right point */
				dock->surface_data.width - border_width,
				2*small_border_width);
			cairo_line_to (cr, /* bottom-right point */
				dock->surface_data.width - border_width,
				dock->surface_data.height - 2*small_border_width);
			cairo_line_to (cr, /* bottom-left point */
				0,
				dock->surface_data.height - small_border_width/2 - dock->shape.corner_radius);
			cairo_line_to (cr, /* top-left point */
				0,
				small_border_width/2 + dock->shape.corner_radius);
			cairo_line_to (cr, /* top-right point */
				dock->surface_data.width - border_width,
				2*small_border_width);
		}
	}
	
	cairo_close_path (cr);
}

static void __draw (KibaDock *dock,
					 cairo_t  *cr,
					 gpointer  data) {
	const KibaLayoutData *d;
	cairo_pattern_t  *pattern;
	gint border_width;
	gint small_border_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(dock));
	small_border_width = SMALL_BORDERWIDTH(border_width);
	
	d = &kiba_layout_data[dock->d.position];
	
	cairo_save (cr);
		cairo_set_line_width (cr, border_width);
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		
		cairo_set_source_rgba (cr,
			dock->style.border_rgba.r, dock->style.border_rgba.g,
			dock->style.border_rgba.b, dock->style.border_rgba.a);
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			if (d->valignment == 2) { /* dock at the bottom edge */
				cairo_move_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* bottom-left point */
					small_border_width,
					dock->surface_data.height - small_border_width);
				cairo_line_to (cr, /* top-left point */
					small_border_width + dock->shape.corner_radius,
					small_border_width);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - small_border_width - dock->shape.corner_radius,
					small_border_width);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - small_border_width,
					dock->surface_data.height - small_border_width);
				cairo_stroke (cr);
			}
			else { /* dock at the top edge */
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2 + dock->shape.corner_radius,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2 - dock->shape.corner_radius,
					dock->surface_data.height - border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_stroke (cr);
			}
		}
		else {
			if (d->halignment == 0) { /* dock at the left edge */
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-left point */
					border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-right point */
					dock->surface_data.width - small_border_width,
					border_width/2 + dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - small_border_width,
					dock->surface_data.height - border_width/2 - dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
			}
			else {
				cairo_move_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_line_to (cr, /* bottom-right point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
				
				cairo_set_line_width (cr, small_border_width);
				cairo_move_to (cr, /* top-right point */
					dock->surface_data.width - border_width/2,
					border_width/2);
				cairo_line_to (cr, /* top-left point */
					small_border_width,
					border_width/2 + dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					small_border_width,
					dock->surface_data.height - border_width/2 - dock->shape.corner_radius);
				cairo_line_to (cr, /* bottom-left point */
					dock->surface_data.width - border_width/2,
					dock->surface_data.height - border_width/2);
				cairo_stroke (cr);
			}
		}
	cairo_restore (cr);
	
	cairo_save (cr);
		__draw_podest_hat (dock, cr);
		
		pattern = kiba_dock_get_pattern (dock);
		if (d->anchor.y == 0)
			cairo_rotate (cr, M_PI);
		else if (d->anchor.x == 0)
			cairo_rotate (cr, M_PI/2);
		else if (d->anchor.x == 4)
			cairo_rotate (cr, 3*M_PI/2);
		cairo_set_source (cr, pattern);
		cairo_fill (cr);
		cairo_pattern_destroy (pattern);
	cairo_restore (cr);
}

static __draw_separator_line (KibaDock *dock, GtkWidget *separator,
                              cairo_t *cr,
                              const KibaLayoutData *d,
                              gint fX, gint fY,
                              gint coord1,
                              gint coord2,
                              gint border_width) {
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		if (d->valignment == 2) { /* dock at the bottom edge */
			cairo_move_to (cr, coord1 + (coord1 - coord2)/2,
				fY + dock->d.height + 2);
			cairo_rel_line_to (cr, coord2 - coord1,
				-dock->d.height - border_width - 2);
		}
		else { /* dock at the top edge */
			cairo_move_to (cr, coord1 + (coord1 - coord2)/2, fY - 2);
			cairo_rel_line_to (cr, coord2 - coord1,
				dock->d.height + border_width + 2);
		}
	}
	else if (d->halignment == 0) { /* dock at the left edge */
		cairo_move_to (cr, fX - 2, coord1 + (coord1 - coord2)/2);
		cairo_rel_line_to (cr,
			dock->d.height + border_width + 2,
			coord2 - coord1);
	}
	else { /* dock at the right edge */
		cairo_move_to (cr, fX + dock->d.height + 2,
			coord1 + (coord1 - coord2)/2);
		cairo_rel_line_to (cr,
			-dock->d.height - border_width - 2,
			coord2 - coord1);
	}
}

static void __separator_draw (KibaDock *dock,
					           cairo_t  *cr,
					           gpointer  data) {
	GtkWidget *widget = GTK_WIDGET(dock);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	gint border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	GList *l;
	gint dock_size;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		dock_size = dock->surface_data.width -
			2*dock->shape.corner_radius - 2*border_width;
	}
	else {
		dock_size = dock->surface_data.height -
			2*dock->shape.corner_radius - 2*border_width;
	}
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		GtkWidget *separator = child->widget;
		gint coord1, coord2;
		float fX, fY;
		
		if (!KIBA_IS_SEPARATOR(separator))
			continue;
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			fX = (float) KIBA_OBJECT(separator)->alloc.x;
			fY = widget->allocation.y + (d->valignment/2)*(widget->allocation.height - dock->d.height);
			coord1 = fX + separator->allocation.width/2;
			coord2 = (dock_size*(coord1 - GTK_WIDGET(dock)->allocation.x))/dock->surface_data.width +
				(dock->surface_data.width - dock_size)/2 + GTK_WIDGET(dock)->allocation.x;
		}
		else {
			fY = (float) KIBA_OBJECT(separator)->alloc.y;
			fX = widget->allocation.x + (d->halignment/2)*(widget->allocation.width - dock->d.height);
			coord1 = fY + separator->allocation.height/2;
			coord2 = (dock_size*(coord1 - GTK_WIDGET(dock)->allocation.y))/dock->surface_data.height +
				(dock->surface_data.height - dock_size)/2 + GTK_WIDGET(dock)->allocation.y;
		}
		
		cairo_save(cr);
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				cairo_set_line_width (cr, KIBA_OBJECT(separator)->alloc.width - 3*border_width);
			else
				cairo_set_line_width (cr, KIBA_OBJECT(separator)->alloc.height - 3*border_width);
			cairo_set_operator(cr, CAIRO_OPERATOR_CLEAR);
			cairo_set_source_rgba (cr,0,0,0,0);
			__draw_separator_line(dock, separator, cr, d, fX, fY, coord1, coord2, border_width);
			cairo_stroke(cr);
		cairo_restore (cr);
		cairo_set_source_rgba (cr,
			dock->style.border_rgba.r,
			dock->style.border_rgba.g,
			dock->style.border_rgba.b,
			dock->style.border_rgba.a);
		cairo_set_line_width (cr, border_width);
		cairo_save(cr);
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				cairo_translate(cr, (KIBA_OBJECT(separator)->alloc.width - 2*border_width)/2, 0);
			else
				cairo_translate(cr, 0, (KIBA_OBJECT(separator)->alloc.height - 2*border_width)/2);
			__draw_separator_line(dock, separator, cr, d, fX, fY, coord1, coord2, border_width);
			cairo_stroke(cr);
		cairo_restore (cr);
		cairo_save(cr);
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
				cairo_translate(cr, -(KIBA_OBJECT(separator)->alloc.width - 2*border_width)/2, 0);
			else
				cairo_translate(cr, 0, -(KIBA_OBJECT(separator)->alloc.height - 2*border_width)/2);
			__draw_separator_line(dock, separator, cr, d, fX, fY, coord1, coord2, border_width);
			cairo_stroke(cr);
		cairo_restore (cr);

		gint size = coord1 - coord2;
		size = MAX(size, -size) +  4*border_width;
		if (kiba_separator_get_size(KIBA_SEPARATOR(separator)) != size)
			kiba_separator_set_size(KIBA_SEPARATOR(separator), size);
	}
}

static void __clip_func (KibaDock *dock,
					      cairo_t  *cr,
					      gpointer  data) {
	const KibaLayoutData *d;
	gint x, y, width, height;
	gint border_width;
	
	border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	
	x = GTK_WIDGET(dock)->allocation.x;
	y = GTK_WIDGET(dock)->allocation.y;
	width = GTK_WIDGET(dock)->allocation.width;
	height = GTK_WIDGET(dock)->allocation.height;
	
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		cairo_rectangle (cr,
			x, y - ((d->valignment-2)/2)*border_width,
			width, height - (d->valignment/2)*border_width);
	else
		cairo_rectangle (cr,
			x - ((d->halignment-2)/2)*border_width, y,
			width - (d->halignment/2)*border_width, height);
	cairo_clip (cr);
}

void kiba_dock_podest_shape_activate (KibaDock *dock) {
	dock->shape.draw_func = __draw;
	dock->shape.separator_draw_func = __separator_draw;
	dock->shape.clip_func = __clip_func;
	dock->shape.draw_data = NULL;
	dock->shape.triangle_ontop = FALSE;
}
