/*
 *  Copyright (C) 2007 Daniel Bessler  <danielb@settings-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@settings-dock.org>
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include "kiba.h"
#include "kiba-debug.h"
#include "kiba-prefs.h"
#include "prefs-dbus.h"
#include "marshallers.h"

#define KIBA_SETTINGS_DBUS_PATH  "/org/kiba/dock/KibaSettings"
#define KIBA_SETTINGS_DBUS_NAMESPACE  "org.kiba.dock.KibaSettings"

static PrefsDbus *dbus_instance = NULL;

G_DEFINE_TYPE (PrefsDbus, prefs_dbus, G_TYPE_OBJECT)

enum {
	INT_OPTION_CHANGED = 0,
	DOUBLE_OPTION_CHANGED,
	BOOL_OPTION_CHANGED,
	CHAR_OPTION_CHANGED,
	LIST_OPTION_CHANGED,
	SIG_LAST_SIGNAL
};

static guint prefs_dbus_signals[SIG_LAST_SIGNAL] = { 0 };

/*** Methods ***/

static gboolean prefs_dbus_add_dock_config_file (PrefsDbus *dbus,
                                                 const gchar *config_path,
                                                 GError  **error) {
	Kiba *kiba = kiba_get();
	kiba_add_config_file(kiba, config_path);
	return TRUE;
}

static gboolean prefs_dbus_remove_dock_config_file (PrefsDbus *dbus,
                                                    const gchar *config_path,
                                                    GError  **error) {
	Kiba *kiba = kiba_get();
	kiba_remove_config_file(kiba, config_path);
	return TRUE;
}

static gboolean prefs_dbus_set_int_option (PrefsDbus   *dbus,
                                           const gchar *config_path,
                                           const gchar *option_id,
                                           gint value,
                                           GError  **error) {
	KibaPrefsFile *file = kiba_prefs_get_file(config_path);
	gchar *string;
	
	string = g_strdup_printf ("%i\n", value);
	kiba_prefs_set_option (file, option_id, "int", string);
	g_free (string);
	
	g_signal_emit (dbus,
		prefs_dbus_signals[INT_OPTION_CHANGED],
		0, config_path, option_id, value);
	
	return TRUE;
}

static gboolean prefs_dbus_set_double_option (PrefsDbus *dbus,
                                              const gchar *config_path,
                                              const gchar *option_id,
                                              gdouble value,
                                              GError **error) {
	KibaPrefsFile *file = kiba_prefs_get_file(config_path);
	char buf [G_ASCII_DTOSTR_BUF_SIZE];
	gchar *string = NULL;
	
	string = g_ascii_dtostr (buf, sizeof(buf), value);
	kiba_prefs_set_option (file, option_id, "double", string);
	
	g_signal_emit (dbus,
		prefs_dbus_signals[DOUBLE_OPTION_CHANGED],
		0, config_path, option_id, value);
	
	return TRUE;
}

static gboolean prefs_dbus_set_bool_option (PrefsDbus   *dbus,
                                            const gchar *config_path,
                                            const gchar *option_id,
                                            gboolean     value,
                                            GError     **error) {
	KibaPrefsFile *file = kiba_prefs_get_file(config_path);
	kiba_prefs_set_option (file, option_id, "bool", (value==TRUE ? "TRUE" : "FALSE"));
	
	g_signal_emit (dbus,
		prefs_dbus_signals[BOOL_OPTION_CHANGED],
		0, config_path, option_id, value);
	
	return TRUE;
}

static gboolean prefs_dbus_set_char_option (PrefsDbus *dbus,
                                            const gchar *config_path,
                                            const gchar *option_id,
                                            const gchar *value,
                                            GError **error) {
	KibaPrefsFile *file = kiba_prefs_get_file(config_path);
	kiba_prefs_set_option (file, option_id, "char", value);
	
	g_signal_emit (dbus,
		prefs_dbus_signals[CHAR_OPTION_CHANGED],
		0, config_path, option_id, value);
	
	return TRUE;
}

static gboolean prefs_dbus_set_list_option (PrefsDbus *dbus,
                                            const gchar *config_path,
                                            const gchar *option_id,
                                            const gchar *value,
                                            GError     **error) {
	KibaPrefsFile *file = kiba_prefs_get_file(config_path);
	kiba_prefs_set_option (file, option_id, "list", value);
	
	g_signal_emit (dbus,
		prefs_dbus_signals[LIST_OPTION_CHANGED],
		0, config_path, option_id, value);
	
	return TRUE;
}

/*** DBus Init ***/

static gboolean kiba_prefs_dbus_init (PrefsDbus *dbus) {
	DBusGConnection *connection;
	DBusGProxy *proxy;
	guint32 ret;
	GError *error = NULL;
	
	/* Get the connection and ensure the name is not used yet */
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (connection == NULL) {
		g_error_free (error);
		return FALSE;
	}
	
	proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
					DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	if (!org_freedesktop_DBus_request_name (proxy, KIBA_SETTINGS_DBUS_NAMESPACE, 0, &ret, &error)) {
		g_error_free (error);
		return FALSE;
	}
	
	if (ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
		return FALSE;
	}
	
	/* Register the app on the bus */
	dbus_g_connection_register_g_object (connection,
		KIBA_SETTINGS_DBUS_PATH, G_OBJECT (dbus));
	
	return TRUE;
}

#include "prefs-dbus-glue.h"

/*** Class Init ***/

static void prefs_dbus_class_init (PrefsDbusClass *class) {
	prefs_dbus_signals[INT_OPTION_CHANGED] =
		g_signal_new ("int-option-changed",
			      G_OBJECT_CLASS_TYPE (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PrefsDbusClass, int_option_changed),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_STRING_INT,
			      G_TYPE_NONE, 3,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
	prefs_dbus_signals[DOUBLE_OPTION_CHANGED] =
		g_signal_new ("double-option-changed",
			      G_OBJECT_CLASS_TYPE (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PrefsDbusClass, double_option_changed),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_STRING_DOUBLE,
			      G_TYPE_NONE, 3,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_DOUBLE);
	prefs_dbus_signals[BOOL_OPTION_CHANGED] =
		g_signal_new ("bool-option-changed",
			      G_OBJECT_CLASS_TYPE (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PrefsDbusClass, bool_option_changed),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_STRING_BOOLEAN,
			      G_TYPE_NONE, 3,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
	prefs_dbus_signals[CHAR_OPTION_CHANGED] =
		g_signal_new ("char-option-changed",
			      G_OBJECT_CLASS_TYPE (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PrefsDbusClass, char_option_changed),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_STRING_STRING,
			      G_TYPE_NONE, 3,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	prefs_dbus_signals[LIST_OPTION_CHANGED] =
		g_signal_new ("list-option-changed",
			      G_OBJECT_CLASS_TYPE (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (PrefsDbusClass, list_option_changed),
			      NULL, NULL,
			      _kiba_marshal_VOID__STRING_STRING_STRING,
			      G_TYPE_NONE, 3,
			      G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	
	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (class),
					&dbus_glib_prefs_dbus_object_info);
}

static void prefs_dbus_init (PrefsDbus *dbus) {
	kiba_prefs_dbus_init(dbus);
}

static PrefsDbus* prefs_dbus_new (void) {
	return g_object_new (PREFS_TYPE_DBUS, NULL);
}

PrefsDbus* prefs_dbus_get (void) {
	if (!dbus_instance)
		dbus_instance = prefs_dbus_new();
	
	return dbus_instance;
}

void prefs_dbus_fini (void) {
	PrefsDbus *dbus = prefs_dbus_get();
	DBusGProxy *proxy;
	DBusGConnection *connection;
	GError *error = NULL;
	
	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (connection == NULL)
		return;
	
	proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
				DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	org_freedesktop_DBus_release_name (proxy, KIBA_SETTINGS_DBUS_NAMESPACE, 0, &error);
	
	if (dbus)
		g_object_unref (dbus);
}

/*** Public Functions ***/

void prefs_dbus_set_int (const gchar *configPath,
                         const gchar *option_id,
                         gint value) {
	PrefsDbus *dbus = prefs_dbus_get();
	
	g_signal_emit (dbus,
		prefs_dbus_signals[INT_OPTION_CHANGED],
		0, configPath, option_id, value);
}

void prefs_dbus_set_double (const gchar *configPath,
                            const gchar *option_id,
                            gdouble value) {
	PrefsDbus *dbus = prefs_dbus_get();
	
	g_signal_emit (dbus,
		prefs_dbus_signals[DOUBLE_OPTION_CHANGED],
		0, configPath, option_id, value);
}

void prefs_dbus_set_bool (const gchar *configPath,
                          const gchar *option_id,
                          gboolean value) {
	PrefsDbus *dbus = prefs_dbus_get();
	
	g_signal_emit (dbus,
		prefs_dbus_signals[BOOL_OPTION_CHANGED],
		0, configPath, option_id, value);
}

void prefs_dbus_set_char (const gchar *configPath,
                          const gchar *option_id,
                          const gchar *value) {
	PrefsDbus *dbus = prefs_dbus_get();
	
	g_signal_emit (dbus,
		prefs_dbus_signals[CHAR_OPTION_CHANGED],
		0, configPath, option_id, value);
}

void prefs_dbus_set_list (const gchar *configPath,
                          const gchar *option_id,
                          const gchar *value) {
	PrefsDbus *dbus = prefs_dbus_get();
	
	g_signal_emit (dbus,
		prefs_dbus_signals[LIST_OPTION_CHANGED],
		0, configPath, option_id, value);
}
