/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Daniel    Bessler    <danielb@kiba-dock.org>
 *              Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
*/

/**
 * SECTION:kiba-prefs-schema
 * @short_description: XML Schema parser.
 * @see_also:#kiba-prefs
 * @stability: Unstable
 *
 * XML Schema parser.
 */

#include "../config.h"

#include <string.h>
#include <libintl.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "kiba-prefs-schema.h"
#include "kiba-prefs.h"
#include "kiba-debug.h"

#define GENERAL_SCHEMAS DATADIR"/kiba-dock/config_schemas/general"
#define PLUGIN_SCHEMAS DATADIR"/kiba-dock/config_schemas/plugins"
#define DOCK_SCHEMAS DATADIR"/kiba-dock/config_schemas/dock"

static void __free_member(SchemaMember *member) {
	if (member->widget_name) g_free(member->widget_name);
	if (member->title) g_free(member->title);
	if (member->type) g_free(member->type);
	if (member->range) g_free(member->range);
	if (member->value) g_free(member->value);
	if (member->default_value) g_free(member->default_value);
	if (member->description) g_free(member->description);
	g_free(member);
}

/** XML schema parser **/

static SchemaMember* create_option_member (KibaSchema *schema, SchemaSubSection *subsection) {
	SchemaMember *ret;
	ret = g_new0 (SchemaMember, 1);
	ret->section_name = subsection->section_name;
	ret->subsection_name = subsection->subsection_name;
	ret->widget_name = NULL;
	ret->title = NULL;
	ret->type = NULL;
	ret->range = NULL;
	ret->value = NULL;
	ret->default_value = NULL;
	ret->description = NULL;
	ret->schema = schema;
	subsection->options = g_list_append(subsection->options, ret);
	
	return ret;
}

static SchemaSubSection* __get_subsection_for_id (SchemaSection *section,
                                               const gchar   *subsection) {
	SchemaSubSection *ret;
	GList *l;
	
	for (l = section->subsections; l != NULL; l = l->next) {
		SchemaSubSection *foo = l->data;
		
		if (strcmp (g_ascii_strdown (foo->subsection_name, strlen (foo->subsection_name)),
			        g_ascii_strdown (subsection, strlen (subsection))) == 0)
			return foo;
	}
	
	ret = g_new0 (SchemaSubSection, 1);
	ret->subsection_name = g_strdup (subsection);
	ret->section_name = section->section_name;
	section->subsections = g_list_append(section->subsections, ret);
	
	return ret;
}

static SchemaSection* __get_section_for_id (KibaSchema *schema,
                                            const char *section) {
	SchemaSection *ret;
	GList *l;
	
	for (l = schema->sections; l != NULL; l = l->next) {
		SchemaSection *foo = l->data;
		
		if (strcmp (g_ascii_strdown (foo->section_name_untranslated, strlen (foo->section_name)),
			        g_ascii_strdown (section, strlen (section))) == 0)
			return foo;
	}
	
	ret = g_new0 (SchemaSection, 1);
	ret->section_name = g_strdup (section);
	ret->section_name_untranslated = ret->section_name;
	ret->icon_list = NULL;
	ret->description = NULL;
	schema->sections = g_list_append(schema->sections, ret);
	
	return ret;
}

static void __get_option_from_node (KibaSchema *schema,
                                    SchemaSubSection *subsection,
                                    const char *locale,
                                    xmlNode *a_element) {
	SchemaMember *member = create_option_member (schema, subsection);
	xmlNode *cur_node   = NULL;
	
	for (cur_node = a_element; cur_node; cur_node = cur_node->next) {
		xmlChar *lang;
		
		if (cur_node->type != XML_ELEMENT_NODE)
			continue;
		
		if (!cur_node->children || !cur_node->children->content)
			continue;
		
		lang = xmlGetProp(cur_node, "lang");
		
		if (!strcmp (cur_node->name, XML_NAME))
			member->widget_name = g_strdup (cur_node->children->content);
		else if (!strcmp (cur_node->name, XML_TITLE) && (!lang || g_strrstr (locale, lang))) {
			if (member->title)
				g_free (member->title);
			member->title = g_strdup (cur_node->children->content);
		}
		else if (!strcmp (cur_node->name, XML_TYPE))
			member->type = g_strdup (cur_node->children->content);
		else if (!strcmp (cur_node->name, XML_RANGE))
			member->range = g_strdup (cur_node->children->content);
		else if (!strcmp (cur_node->name, XML_DEFAULT))
			member->default_value = g_strdup (cur_node->children->content);
		else if (!strcmp (cur_node->name, XML_DESCRIPTION) && (!lang || g_strrstr (locale, lang))) {
			if (member->description)
				g_free (member->description);
			member->description = g_strdup (cur_node->children->content);
		}
	}
	
	if (!member->type || !member->widget_name) {
		subsection->options = g_list_remove(subsection->options, member);
		__free_member (member);
	}
	else if (!member->title)
		member->title = g_strdup (member->widget_name);
}

static void __get_subsection_from_node (KibaSchema *schema,
                                        SchemaSection *section,
                                        const char *locale,
                                        xmlNode *a_element) {
	SchemaSubSection *subsection = NULL;
	xmlNode *cur_node   = NULL;
	
	for (cur_node = a_element; cur_node; cur_node = cur_node->next) {
		xmlChar *lang;
	
		if (cur_node->type != XML_ELEMENT_NODE)
			continue;
		
		if (!cur_node->children || !cur_node->children->content)
			continue;
		
		lang = xmlGetProp(cur_node, "lang");
		
		/* first we need the sec_name key to identify the section */
		if (!subsection) {
			if (!strcmp (cur_node->name, XML_SUBSECTION_NAME)) {
				subsection = __get_subsection_for_id (section,
							cur_node->children->content);
				cur_node = a_element;
			}
			continue;
		}
		/* get localized name */
		else if (!strcmp (cur_node->name, XML_SUBSECTION_NAME) &&
			 lang && g_strrstr (locale, lang)) {
			g_free (subsection->subsection_name);
			subsection->subsection_name = NULL;
			subsection->subsection_name = g_strdup (cur_node->children->content);
		}
		else if (!strcmp (cur_node->name, XML_OPTION))
			__get_option_from_node (schema,
				subsection,
				locale,
				cur_node->children);
	}
}

static SchemaSection* __get_section_from_node (KibaSchema *schema,
                                     const char *locale,
                                     xmlNode *node) {
	SchemaSection *section = NULL;
	xmlNode *cur_node = NULL;
	
	for (cur_node = node; cur_node; cur_node = cur_node->next) {
		xmlChar *lang;
		
		if (cur_node->type != XML_ELEMENT_NODE)
			continue;
		
		if (!cur_node->children || !cur_node->children->content)
			continue;
		
		lang = xmlGetProp(cur_node, "lang");
		
		/* first we need the sec_name key to identify the section */
		if (!section) {
			if (!strcmp (cur_node->name, XML_SECTION_NAME)) {
				section = __get_section_for_id (schema, cur_node->children->content);
				cur_node = node;
			}
			continue;
		}
		/* get localized name */
		else if (!strcmp (cur_node->name, XML_SECTION_NAME) && lang && g_strrstr (locale, lang)) {
			section->section_name_untranslated = section->section_name;
			section->section_name = g_strdup (cur_node->children->content);
		}
		else if (!strcmp (cur_node->name, XML_SECTION_ICON))
			section->icon_list = kiba_prefs_get_list_from_string (cur_node->children->content);
		else if (!strcmp (cur_node->name, XML_SECTION_DESC) && (!lang || g_strrstr (locale, lang))) {
			if (section->description)
				g_free (section->description);
			section->description = g_strdup (cur_node->children->content);
		}
		else if (!strcmp (cur_node->name, XML_SUBSECTION))
			__get_subsection_from_node (schema,
				section,
				locale,
				cur_node->children);
	}
	
	return section;
}

static void __get_sections (KibaSchema *schema,
                            const char *locale,
                            xmlNode *node,
                            KibaSchemaType type) {
	xmlNode *sectionNode = NULL;
	
	for (sectionNode = node; sectionNode; sectionNode = sectionNode->next) {
		if (sectionNode->type == XML_ELEMENT_NODE && strcmp (sectionNode->name, XML_SECTION) == 0) {
			SchemaSection *section;
			section = __get_section_from_node (schema, locale, sectionNode->children);
			section->type = type;
		}
		/* recursively call function for none section elements, since they may have section childs */
		else
			__get_sections (schema, locale, sectionNode->children, type);
	}
}

static void __get_schema (KibaSchema  *schema,
                          const char  *locale,
                          const char  *file,
                          KibaSchemaType type) {
	xmlDoc  *doc = NULL;
	xmlNode *root_element = NULL;
	
	/*parse the file and get the DOM */
	doc = xmlReadFile(file, NULL, 0);
	if (doc == NULL) {
		DEBUG_ERROR("could not parse schema file '%s'\n", file);
		return;
	}
	
	root_element = xmlDocGetRootElement(doc);
	__get_sections (schema, locale, root_element, type);
	xmlFreeDoc(doc);
}

/** **/

static void __get_schemas_from_dir (KibaSchema *schema,
				  const gchar *locale,
				  const gchar *schema_path,
				  KibaSchemaType type) {
	GDir *schema_dir;
	gchar *file_name = NULL;
	
	/* open the directory for reading */
	schema_dir = g_dir_open (schema_path, 0, NULL);
	if (!schema_dir) return;
	
	/* if the dir is empty*/
	if (g_dir_read_name(schema_dir) == NULL) return;
	g_dir_rewind (schema_dir);
	
	/* parse all schema files in schema_dir */
	for (file_name = (gchar *) g_dir_read_name(schema_dir); file_name != NULL;
	     file_name = (gchar *) g_dir_read_name(schema_dir)) {
		char  *path;
		
		if (!g_str_has_suffix (file_name, XML_FILE_SUFFIX))
			continue;
		
		path = g_strdup_printf ("%s/%s", schema_path, file_name);
		__get_schema (schema, locale, path, type);
		g_free (path);
	}
	g_dir_close (schema_dir);
}

/*** Public Methods ***/

GList* kiba_prefs_schema_get_options(KibaSchema *schema) {
	GList *sections, *subsections, *options;
	GList *ret = NULL;
	
	for (sections = schema->sections; sections != NULL; sections = sections->next) {
		SchemaSection *section = sections->data;
		for (subsections = section->subsections; subsections != NULL; subsections = subsections->next) {
			SchemaSubSection *subsection = subsections->data;
			for (options = subsection->options; options != NULL; options = options->next) {
				SchemaMember *schemaOption = options->data;
				KibaOption *option = g_new0(KibaOption, 1);
				option->id = schemaOption->widget_name;
				option->type = schemaOption->type;
				if (schemaOption->default_value)
					option->value = schemaOption->default_value;
				else
					option->value = "(null)";
				ret = g_list_append(ret, option);
			}
		}
	}
	
	return ret;
}

void kiba_prefs_schema_set_options_from_config(KibaSchema *schema, KibaPrefsFile *file) {
	GList *sections, *subsections, *options;
	for (sections = schema->sections; sections != NULL; sections = sections->next) {
		SchemaSection *section = sections->data;
		for (subsections = section->subsections; subsections != NULL; subsections = subsections->next) {
			SchemaSubSection *subsection = subsections->data;
			for (options = subsection->options; options != NULL; options = options->next) {
				SchemaMember *option = options->data;
				char *newValue = kiba_prefs_get_option_from_file (file, option->widget_name, option->type);
				if (option->value)
					g_free(option->value);
				if (newValue)
					option->value = newValue;
				else if (option->default_value)
					option->value = g_strdup (option->default_value);
				else
					option->value = g_strdup ("(null)");
			}
		}
	}
}

/*** Init ***/

/**
 * kiba_prefs_get_schema:
 *
 * Creates and returns a new KibaSchema.
 *
 * Returns: The created KibaSchema.
 */
KibaSchema* kiba_prefs_schema_get (KibaSchemaType type) {
	KibaSchema *schema;
	gchar *userSchemas;
	gchar *locale;
	
	schema = g_new0 (KibaSchema, 1);
	
	locale = setlocale (LC_ALL, NULL);
	switch (type) {
		case KIBA_SCHEMA_GENERAL:
			if (g_file_test(GENERAL_SCHEMAS, G_FILE_TEST_IS_DIR))
				__get_schemas_from_dir (schema, locale, GENERAL_SCHEMAS, KIBA_SCHEMA_GENERAL);
			//TODO user schema?
			break;
		case KIBA_SCHEMA_DOCK:
			if (g_file_test(DOCK_SCHEMAS, G_FILE_TEST_IS_DIR))
				__get_schemas_from_dir (schema, locale, DOCK_SCHEMAS, KIBA_SCHEMA_DOCK);
			if (g_file_test(PLUGIN_SCHEMAS, G_FILE_TEST_IS_DIR))
				__get_schemas_from_dir (schema, locale, PLUGIN_SCHEMAS, KIBA_SCHEMA_PLUGIN);
			userSchemas = g_strdup_printf ("%s/.kiba-dock/config_schemas", g_get_home_dir());
			if (g_file_test(userSchemas, G_FILE_TEST_IS_DIR))
				__get_schemas_from_dir (schema, locale, userSchemas, KIBA_SCHEMA_USER);
			g_free (userSchemas);
			break;
	}
	
	return schema;
}

void kiba_prefs_schema_fini(KibaSchema *schema) {
	GList *sections, *subsections, *options;
	GSList *icons;
	
	for (sections = schema->sections; sections != NULL; sections = schema->sections) {
		SchemaSection *section = sections->data;
		for (subsections = section->subsections; subsections != NULL; subsections = section->subsections) {
			SchemaSubSection *subsection = subsections->data;
			for (options = subsection->options; options != NULL; options = subsection->options) {
				SchemaMember *option = options->data;
				subsection->options = g_list_remove(subsection->options, option);
				__free_member(option);
			}
			section->subsections = g_list_remove(section->subsections, subsection);
			g_list_free(subsection->options);
			if (subsection->subsection_name)
				g_free(subsection->subsection_name);
			g_free(subsection);
		}
		schema->sections = g_list_remove(schema->sections, section);
		g_list_free(section->subsections);
		if (section->section_name)
			g_free(section->section_name);
		if (section->description)
			g_free(section->description);
		for (icons = section->icon_list; icons != NULL; icons = section->icon_list) {
			char *data = icons->data;
			section->icon_list = g_slist_remove(section->icon_list, icons->data);
			g_free(data);
		}
		g_slist_free(section->icon_list);
		g_free(section);
	}
	g_list_free(schema->sections);
}
