/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Daniel    Bessler    <danielb@kiba-dock.org>
 *              Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
*/

/**
 * SECTION:kiba-prefs
 * @short_description: Kiba-Dock preferences.
 * @see_also:#kiba-prefs-xml
 * @stability: Unstable
 *
 * Kiba-Dock preferences.
 */

#include "../config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <libintl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <libxml/parser.h>
#include <libxml/xmlwriter.h>

#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "kiba-utils.h"
#include "kiba-debug.h"
#include "prefs-dbus.h"
#include "prefs-private.h"

static GQuark kiba_prefs_error_quark (void) {
	static GQuark q = 0;
	if (q == 0)
		q = g_quark_from_static_string ("kiba-prefs-quark");
	return q;
}
#define KIBA_PREFS_ERROR kiba_prefs_error_quark ()

typedef struct {
	KibaPrefCallback func;
	gpointer data;
}KibaOptionCallback;

typedef struct {
	PrefMember member; /**< Array of PrefMember's */
	gpointer value; /**< Value pointer */
	GList *callbacks;  /**< List of callbacks for the setting */
}KibaPrefOption;

typedef struct {
	PrefMember  *member; /**< Array of PrefMember's */
	gint member_elements; /**< Number of member elements */
	GList *options; /**< List with all KibaPrefOption's */
}KibaPrefGroup;

struct _KibaPrefsFile {
	char *path;
	GList *options;
	GList *groups; /**< List with all KibaPrefGroup's */
};

struct _KibaPrefs {
	GList *configFiles; /**< list of KibaPrefsFile's */
	
	KibaPrefsFile *generalFile; /**< file with general settings */
	KibaPrefsFile *globalFile; /**< file with global dock settings, these settings overwrite settings made in configFiles */
};

static KibaPrefs *kiba_prefs = NULL;

/*** Private methods ***/

static void __free_options(GList *options) {
	GList *l;
	for (l = options; l != NULL; l = options) {
		KibaOption *optionData = l->data;
		options = g_list_remove(options, optionData);
		g_free(optionData->id);
		g_free(optionData->type);
		if (optionData->value)
			g_free(optionData->value);
		g_free(optionData);
	}
	g_list_free(options);
}

static void __free_prefs_option (KibaPrefOption *option) {
	GList *l;
	
	switch (option->member.type) {
		case KIBA_DOUBLE:
		case KIBA_STRING: {
			if (option->value != NULL)
				g_free(option->value);
			break;
		}
		case KIBA_LIST: {
			GSList *l;
			if (!option->value) break;
			for (l = (GSList*) option->value; l != NULL; l = (GSList*) option->value) {
				char *data = l->data;
				g_free(data);
				option->value = g_slist_remove((GSList*) option->value, l->data);
			}
			g_slist_free(option->value);
			break;
		}
	}
	
	for (l = option->callbacks; l != NULL; l = option->callbacks) {
		KibaOptionCallback *cb = l->data;
		option->callbacks = g_list_remove(option->callbacks, l->data);
		g_free(cb);
	}
}

static void __free_file (KibaPrefsFile *file) {
	GList *l;
	
	__free_options(file->options);
	
	for (l = file->groups; l != NULL; l = file->groups) {
		KibaPrefGroup *group = l->data;
		GList *options;
		file->groups = g_list_remove(file->groups, l->data);
		
		for (options = group->options; options != NULL; options = group->options) {
			KibaPrefOption *option = options->data;
			group->options = g_list_remove(group->options, options->data);
			__free_prefs_option(option);
		}
		
		g_free(group);
	}
	
	g_free(file->path);
	g_free(file);
}

static gboolean __file_contains_option (KibaPrefsFile *file,
                                        const char *id) {
	GList *l;
	for (l = file->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0)
			return TRUE;
	}
	return FALSE;
}

static KibaPrefsFile* __get_file_at_path (KibaPrefs *prefs,
                                          const char *configPath) {
	GList *l;
	for (l = prefs->configFiles; l != NULL; l = l->next) {
		KibaPrefsFile *file = l->data;
		if (strcmp(file->path, configPath) == 0)
			return file;
	}
	if (prefs->globalFile && strcmp(prefs->globalFile->path, configPath) == 0)
		return prefs->globalFile;
	if (prefs->generalFile && strcmp(prefs->generalFile->path, configPath) == 0)
		return prefs->generalFile;
	return NULL;
}

static KibaPrefOption* __get_prefs_option (KibaPrefsFile *file,
                                           const char *id) {
	GList *l;
	
	for (l = file->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (strcmp (option->member.id, id) == 0)
				return option;
		}
	}
	
	return NULL;
}

static void __set_option_from_string (KibaPrefOption *option,
				    const char     *new_value) {
	switch (option->member.type) {
		case KIBA_DOUBLE:
		case KIBA_STRING: {
			if (option->value != NULL)
				g_free(option->value);
			option->value = (gpointer) (g_strdup(new_value));
			break;
		}
		case KIBA_INT: {
			option->value = GINT_TO_POINTER(atoi (new_value));
			break;
		}
		case KIBA_BOOLEAN: {
			option->value = GINT_TO_POINTER(strcmp (new_value, "TRUE") == 0);
			break;
		}
		case KIBA_LIST: {
			char *value_buf;
			
			if (option->value) {
				GSList *l;
				for (l = (GSList*) option->value; l != NULL; l = (GSList*) option->value) {
					char *data = l->data;
					g_free(data);
					option->value = g_slist_remove((GSList*) option->value, l->data);
				}
				g_slist_free(option->value);
				option->value = NULL;
			}
			
			value_buf = (char*) new_value;
			while (strlen(value_buf)) {
				int i;
				char *value;
				
				value = g_strdup (value_buf);
				for (i = 1; i < strlen(value)-1; i++) {
					if (value[i] == ',') {
						value[i] = '\0';
						break;
					}
				}
				value_buf += strlen(value);
				if (!strlen(value_buf))
					value[strlen(value)-1] = '\0';
				
				option->value = (gpointer) g_slist_append ((GSList*) option->value, g_strdup(value+1));
				g_free(value);
			}
			break;
		}
	}
}

/** XML coinfig parsing/writing **/

/* read options from xml config file */
static GList* __get_options (KibaPrefs *settings,
                             const char *configPath) {
	xmlDoc *doc;
	xmlNode *root_element;
	xmlNode *settingNode;
	GList *ret;
	
	if (!g_file_test(configPath, G_FILE_TEST_EXISTS))
		return NULL;
	
	/*parse the file and get the DOM */
	doc = xmlReadFile(configPath, NULL, 0);
	if (doc == NULL) {
		DEBUG_ERROR("could not parse config file '%s'", configPath);
		return NULL;
	}
	
	ret = NULL;
	
	root_element = xmlDocGetRootElement(doc);
	for (settingNode = root_element->children; settingNode; settingNode = settingNode->next) {
		if (settingNode->type != XML_ELEMENT_NODE) continue;
		struct _xmlAttr *attr;
		KibaOption *option;
		char *id = NULL, *type = NULL;
		
		/* get the option id */
		for (attr = settingNode->properties; attr; attr = attr->next) {
			if (strcmp(attr->name, "id") == 0)
				id = g_strdup(attr->children->content);
			else if (strcmp(attr->name, "type") == 0)
				type = g_strdup(attr->children->content);
		}
		if (!id) {
			if (type) g_free(type);
			continue;
		} else if (!type)
			continue;
		
		option = g_new0(KibaOption, 1);
		option->type = type;
		option->id = id;
		if (settingNode->children && settingNode->children->content)
			option->value = g_strdup(settingNode->children->content);
		else
			option->value = g_strdup("(null)");
		
		ret = g_list_append(ret, option);
	}
	
	xmlFreeDoc(doc);
	
	return ret;
}

/* write options to the xml config file */
static void __write_options (KibaPrefs *settings,
                             KibaPrefsFile *file) {
	xmlTextWriter *writer;
	GList *l;
	
	writer = xmlNewTextWriterFilename(file->path, 0);
	
	xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
		/* start root element */
		xmlTextWriterStartElement(writer, BAD_CAST "config");
		/* add options */
		for (l = file->options; l != NULL; l = l->next) {
			KibaOption *optionData = l->data;
			xmlTextWriterWriteString(writer, "\n\t");
			/* add option node */
			xmlTextWriterStartElement(writer, BAD_CAST "option");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "id", BAD_CAST optionData->id);
				xmlTextWriterWriteAttribute(writer, BAD_CAST "type", BAD_CAST optionData->type);
				if (optionData->value)
					xmlTextWriterWriteRaw(writer, BAD_CAST optionData->value);
				else
					xmlTextWriterWriteRaw(writer, BAD_CAST "(null)");
			xmlTextWriterEndElement(writer);
		}
		xmlTextWriterWriteString(writer, "\n");
	/* end root element and write the file */
	xmlTextWriterEndDocument(writer);
	
	/* clean up */
	xmlFreeTextWriter(writer);
}

/*** Public methods ***/

/**
 * kiba_prefs_edge_style_from_string:
 * @string: The corner style string.
 *
 * Returns a edge style from string.
 *
 * Returns: The corner style.
 */
gint kiba_prefs_edge_style_from_string (const char *string) {
	const struct { KibaCornerShape style; const char *string; } corner_data[]= {
		{ KIBA_ROUNDED, "rounded" },
		{ KIBA_MITERED, "mitered" },
		{ KIBA_SQUARED, "squared" },
		{ KIBA_NOTCHED, "notched" }
	};
	
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS(corner_data); i++) {
		if (strcmp (corner_data[i].string, string) == 0)
			return corner_data[i].style;
	}
	
	return KIBA_SQUARED;
}

/**
 * kiba_prefs_background_style_from_string:
 * @string: The style string.
 *
 * Returns a background style from string.
 *
 * Returns: The style.
 */
gint kiba_prefs_background_style_from_string (const char *string) {
	const struct { KibaBackgroundStyle style; const char *string; } bg_data[]= {
		{ KIBA_LINEAR, "linear" },
		{ KIBA_RADIAL, "radial" },
		{ KIBA_GRADIENT, "gradient" }
	};
	
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS(bg_data); i++) {
		if (strcmp (bg_data[i].string, string) == 0)
			return bg_data[i].style;
	}
	
	return KIBA_NONE;
}

/**
 * kiba_prefs_get_type_from_string:
 * @string: The type string.
 *
 * Returns the option type from string.
 *
 * Returns: The type.
 */
KibaType kiba_prefs_get_type_from_string (const char *string) {
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS (kiba_type_data); i++) {
		if (strcmp (kiba_type_data[i].pref_type, string) == 0)
			return i;
	}
	
	return KIBA_INVALID;
}

/**
 * kiba_prefs_get_position_from_string:
 * @string: The position string.
 *
 * Returns a position from string.
 *
 * Returns: The position.
 */
KibaPosition kiba_prefs_get_position_from_string (const char *string) {
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++) {
		if (strcmp (kiba_layout_data[i].name, string) == 0)
			return i;
	}
	
	return KIBA_POSITION_BOTTOM_EDGE;
}

/**
 * kiba_prefs_rgba_from_string:
 * @string: The color string.
 *
 * Returns a Color from string.
 *
 * Returns: The color.
 */
KibaRGBA kiba_prefs_rgba_from_string (const char *string) {
	KibaRGBA rgba;
	int red, green, blue, opacity;
	
	sscanf (string, "#%02X%02X%02X%02X", &red, &green, &blue, &opacity);
	rgba.r = (float) red   / 256.0;
	rgba.g = (float) green / 256.0;
	rgba.b = (float) blue  / 256.0;
	rgba.a = (float) opacity  / 256.0;
	
	return rgba;
}

/**
 * kiba_prefs_get_pointer:
 * @file: The KibaPrefsFile instance.
 * @id: The option id.
 * @error: The GError return, or NULL.
 *
 * Returns value pointer from the database.
 *
 * Returns: The pointer to the database entry, or NULL.
 */
gpointer kiba_prefs_get_pointer (KibaPrefsFile *file,
                                 const char *id,
                                 GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	g_return_val_if_fail (file != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		gpointer ret;
		
		ret = kiba_prefs_get_pointer(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return option->value;
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return NULL;
}

/**
 * kiba_prefs_get_int:
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @error: The GError return, or NULL.
 *
 * Gets Int Value from Kiba Preferences Database.
 *
 * Returns: The Value, or -1 if none Setting was found.
 */
gint kiba_prefs_get_int (KibaPrefsFile *file,
                         const char *id,
                         GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, 0);
	g_return_val_if_fail (file != NULL, 0);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		gint ret;
		
		ret = kiba_prefs_get_int(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return GPOINTER_TO_INT (option->value);
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return -1;
}

/**
 * kiba_prefs_get_bool:
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @error: The GError return, or NULL.
 *
 * Gets Bool Value from Kiba Preferences Database.
 *
 * Returns: The Value, or FALSE if none Setting was found.
 */
gboolean kiba_prefs_get_bool (KibaPrefsFile *file,
                              const char *id,
                              GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, FALSE);
	g_return_val_if_fail (file != NULL, FALSE);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		gboolean ret;
		
		ret = kiba_prefs_get_bool(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return (gboolean)(GPOINTER_TO_INT (option->value));
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return FALSE;
}

/**
 * kiba_prefs_get_float:
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @error: The GError return, or NULL.
 *
 * Gets Double Value from Kiba Preferences Database.
 *
 * Returns: The Value, or 0.0 if none Setting was found.
 */
float kiba_prefs_get_float (KibaPrefsFile *file,
                            const char *id,
                            GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, 0.0f);
	g_return_val_if_fail (file != NULL, 0.0f);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		float ret;
		
		ret = kiba_prefs_get_float(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return g_ascii_strtod((char*) (option->value), NULL);
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return 0.0f;
}

/**
 * kiba_prefs_get_string:
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @error: The GError return, or NULL.
 *
 * Gets String Value from Kiba Preferences Database.
 *
 * Returns: The Value, or NULL if none Setting was found.
 */
gchar* kiba_prefs_get_string (KibaPrefsFile *file,
                              const char *id,
                              GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	g_return_val_if_fail (file != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		gchar *ret;
		
		ret = kiba_prefs_get_string(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return (char*) option->value;
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return NULL;
}

/**
 * kiba_prefs_get_list:
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @error: The GError return, or NULL.
 *
 * Gets GList Value from Kiba Preferences Database.
 *
 * Returns: GList, or NULL if none Setting was found.
 */
GSList* kiba_prefs_get_list (KibaPrefsFile *file,
                             const char *id,
                             GError **error) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	g_return_val_if_fail (file != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	/* first try global setting */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		GError *globalError = NULL;
		GSList *ret;
		
		ret = kiba_prefs_get_list(prefs->globalFile, id, &globalError);
		if (globalError == NULL) return ret;
		g_clear_error(&globalError);
	}
	
	if (__file_contains_option(file, id)) {
		KibaPrefOption *option = __get_prefs_option(file, id);
		if (option) return (GSList*) option->value;
	}
	
	 g_set_error (error, KIBA_PREFS_ERROR, 0,
		 "Option '%s' not found in file '%s'", id, file->path);
	
	return NULL;
}

void kiba_prefs_remove_list_item (KibaPrefsFile *file,
                                  const char *list_id,
                                  char *rem_item) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_if_fail (rem_item != NULL);
	g_return_if_fail (file != NULL);
	
	prefs = kiba_prefs_get ();
	
	KibaPrefOption *option = __get_prefs_option(file, list_id);
	if (option)
		option->value = g_list_remove (option->value, rem_item);
}

/**
 * kiba_prefs_append_to_list:
 * @file: The KibaPrefsFile instance.
 * @list_id: The list Name.
 * @new_item: The New Item to append.
 *
 * Appends Item to list and write List-Option to Config File.
 */
void kiba_prefs_append_to_list (KibaPrefsFile *file,
                                const char *list_name,
                                const char *new_item) {
	GSList *list = kiba_prefs_get_list (file, list_name, NULL);
	GSList *l;
	char *value = g_strdup_printf ("[%s", new_item);
	char *val_buf;
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		
		if (strcmp (item, new_item) == 0)
			return;
	}
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf  = NULL;
		
		if (!item || !g_strrstr (item, ";"))
			continue;
		
		buf = g_strdup (value);
		g_free (value);
		value = NULL;
		value = g_strdup_printf ("%s,%s", buf, item);
	}
	
	val_buf = value;
	value = g_strdup_printf ("%s]", val_buf);
	
	kiba_prefs_write_to_config_file (file, list_name, "list", value);
	g_free (value);
	g_free (val_buf);
}

/**
 * kiba_prefs_remove_from_list:
 * @file: The KibaPrefsFile instance.
 * @list_id: The list Name.
 * @remove_item: The Item to remove.
 *
 * Removes Item from list and write List-Option to Config File.
 */
void kiba_prefs_remove_from_list (KibaPrefsFile *file,
                                  const char *list_name,
                                  const char *remove_item) {
	GSList *l;
	char *value = g_strdup ("[");
	char *val_buf;
	GSList *list;
	
	list = kiba_prefs_get_list (file, list_name, NULL);
	
	for (l = g_slist_copy(list); l != NULL; l = l->next) {
		char *item = l->data;
		
		if (strcmp (item, remove_item) == 0) {
			list = g_slist_remove (list, item);
			kiba_prefs_remove_list_item (file, list_name, item);
		}
	}
	g_slist_free (l);
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf  = NULL;
		
		if (!g_strrstr (item, ";"))
			continue;
		
		buf = g_strdup (value);
		g_free (value);
		value = NULL;
		
		if (strcmp (buf, "[") != 0)
			value = g_strdup_printf ("%s,%s", buf, item);
		else
			value = g_strdup_printf ("%s%s", buf, item);
	}
	
	val_buf = value;
	value = g_strdup_printf ("%s]", val_buf);
	
	kiba_prefs_write_to_config_file (file, list_name, "list", value);
	g_free (val_buf);
	g_free (value);
}

GSList* kiba_prefs_get_list_from_string (const char *data) {
	GSList *ret = NULL;
	char *value, *buf;
	
	value = g_strndup (data, strlen (data) - 1);
	buf = g_strdup (g_strrstr (value, ","));
	if (!buf)
		ret = g_slist_append (ret, g_strdup (value+1));
	
	while (buf) {
		char *written, *missing;
	
		ret = g_slist_append (ret, g_strdup (buf+1));
		
		written = g_strrstr (value, buf);
		if (!written)
			break;
		
		missing = g_strndup (value, strlen (value) - strlen (written));
		g_free (buf);
		buf = NULL;
		buf = g_strdup (g_strrstr (missing, ","));
		
		if (!buf)
			ret = g_slist_append (ret, g_strdup (missing+1));
		
		g_free (missing);
	}
	
	g_free (value);
	g_free (buf);
	
	return ret;
}

char* kiba_prefs_list_to_string(GList *list) {
	GList *l;
	char *ret;
	char *buf;
	
	if (!list) return g_strdup("[]");
	
	ret = g_strdup_printf("[%s", (char*) list->data);
	for (l = list->next; l != NULL; l = l->next) {
		buf = ret;
		ret = g_strdup_printf("%s,%s", buf, (char*) l->data);
		g_free (buf);
	}
	buf = ret;
	ret = g_strdup_printf("%s]", buf);
	g_free (buf);
	
	return ret;
}


static void __involve_callbacks (KibaPrefsFile *file,
                                 const char *id,
                                 KibaType type,
                                 gpointer value) {
	GList *l;
	
	KibaPrefOption *option = __get_prefs_option(file, id);
	if (!option) return;
	
	for (l = option->callbacks; l != NULL; l = l->next) {
		KibaOptionCallback *cb = l->data;
		cb->func (id, value, cb->data);
	}
}

/**
 * kiba_prefs_set_option:
 * @file: The KibaPrefsFile instance.
 * @id: The Pref Name.
 * @type: The Pref Type.
 * @value: The Pref Value.
 *
 * Sets New Value for Preferences Member and call the changed_callback.
 */
void kiba_prefs_set_option (KibaPrefsFile *file,
                            const char *id,
                            const char *type,
                            const char *value) {
	GList *l;
	KibaPrefs *prefs;
	KibaType kibaType;
	gpointer newValue = NULL;
	
	g_return_if_fail (id != NULL);
	g_return_if_fail (type != NULL);
	g_return_if_fail (value != NULL);
	g_return_if_fail(file != NULL);
	
	prefs = kiba_prefs_get();
	
	kibaType = kiba_prefs_get_type_from_string(type);
	
	/* remove global setting, this is probably not the best way/place to do this */
	if (file == prefs->globalFile && strcmp(type, "char") == 0 && strcmp(value, "REMOVE") == 0) {
		for (l = file->options; l != NULL; l = l->next) {
			KibaOption *optionData = l->data;
			if (strcmp(id, optionData->id) == 0 && strcmp(type, optionData->type) == 0) {
				if (optionData->value)
					g_free(optionData->value);
				g_free(optionData);
				file->options = g_list_remove(file->options, l->data);
				break;
			}
		}
		KibaPrefOption *globalOption = __get_prefs_option(file, id);
		if (globalOption) {
			switch (globalOption->member.type) {
				case KIBA_DOUBLE:
				case KIBA_STRING: {
					if (globalOption->value != NULL)
						g_free(globalOption->value);
					globalOption->value = NULL;
					break;
				}
				case KIBA_BOOLEAN:
				case KIBA_INT: {
					globalOption->value = NULL;
					break;
				}
				case KIBA_LIST: {
					if (globalOption->value) {
						GSList *sl;
						for (sl = (GSList*) globalOption->value; sl != NULL; sl = (GSList*) globalOption->value) {
							char *data = sl->data;
							g_free(data);
							globalOption->value = g_slist_remove((GSList*) globalOption->value, sl->data);
						}
						g_slist_free(globalOption->value);
					}
					globalOption->value = NULL;
					break;
				}
			}
		}
		
		for (l = prefs->configFiles; l != NULL; l = l->next) {
			KibaPrefsFile *dockFile = l->data;
			KibaPrefOption *option = __get_prefs_option(dockFile, id);
			if (!option) continue;
			__involve_callbacks(dockFile, id, kibaType, option->value);
		}
		
		return;
	}
	
	for (l = file->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0 && strcmp(type, optionData->type) == 0) {
			g_free(optionData->value);
			optionData->value = g_strdup(value);
			break;
		}
	}
	
	KibaPrefOption *option = __get_prefs_option(file, id);
	if (option) {
		__set_option_from_string (option, value);
		newValue = option->value;
	} else {
		return;
	}
	
	if (file == prefs->globalFile) {
		/* involving callbacks for all files */
		for (l = prefs->configFiles; l != NULL; l = l->next) {
			KibaPrefsFile *dockFile = l->data;
			
			__involve_callbacks(dockFile, id, kibaType, newValue);
		}
	} else {
		if (file != prefs->generalFile) {
			for (l = prefs->globalFile->options; l != NULL; l = l->next) {
				KibaOption *optionData = l->data;
				if (strcmp(id, optionData->id) == 0 && strcmp(type, optionData->type) == 0) {
					return;
				}
			}
		}
		
		__involve_callbacks(file, id, kibaType, newValue);
	}
}

/**
 * kiba_prefs_add_callback:
 * @file: The KibaPrefsFile instance.
 * @id: The setting id.
 * @callback: The setting changed callback.
 * @data: The user data.
 * @launch: TRUE to call the callback on adding, otherwise FALSE is returned.
 *
 * Adds value changed callback for option.
 */
void kiba_prefs_add_callback (KibaPrefsFile *file,
                              const char *id,
                              KibaPrefCallback callback,
                              gpointer data,
                              gboolean launch) {
	KibaPrefs *prefs;
	KibaPrefOption *option;
	KibaPrefOption *globalOption;
	KibaOptionCallback *cb;
	GList *l;
	
	g_return_if_fail (id != NULL);
	g_return_if_fail(file != NULL);
	
	prefs = kiba_prefs_get ();
	
	option = __get_prefs_option(file, id);
	if (!option) {
		DEBUG_WARNING("Failed to add callback for option '%s': The option cant be found!\n", id);
		return;
	}
	
	cb = g_new0 (KibaOptionCallback, 1);
	cb->func = callback;
	cb->data = data;
	
	if (launch) {
		gpointer valuePointer;
		valuePointer = option->value;
		if (file != prefs->generalFile && file != prefs->globalFile) {
			/* using global option if there is one */
			if (__file_contains_option(prefs->globalFile, id)) {
				globalOption = __get_prefs_option(prefs->globalFile, id);
				valuePointer = globalOption->value;
			}
		}
		callback (id, valuePointer, data);
	}
		
	option->callbacks = g_list_append (option->callbacks, cb);
}

/**
 * kiba_prefs_remove_callback_by_data:
 * @file: The KibaPrefsFile instance.
 * @data: The user data.
 *
 * Removes value changed callback.
 */
void kiba_prefs_remove_callback_by_data (KibaPrefsFile *file,
                                         gpointer data) {
	KibaPrefs *prefs;
	GList  *l;
	
	g_return_if_fail (data != NULL);
	g_return_if_fail(file != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = file->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			GList *cbs;
			
			for (cbs = option->callbacks; cbs != NULL; cbs = cbs->next) {
				KibaOptionCallback *cb = cbs->data;
				
				if (cb->data == data) {
					if (cbs->prev)
						cbs = cbs->prev;
					else
						cbs = option->callbacks;
					option->callbacks = g_list_remove (option->callbacks, cb);
					g_free (cb);
					cb = NULL;
				}
			}
		}
	}
}

/**
 * kiba_prefs_remove_options:
 * @file: The KibaPrefsFile instance.
 * @options: An array of PrefMember's to remove.
 *
 * Removes previously added Options from the Preferences Database.
 */
void kiba_prefs_remove_options (KibaPrefsFile *file,
                                PrefMember *options) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_if_fail (options != NULL);
	g_return_if_fail(file != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = file->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
	
		if (options == group->member) {
			GList *li;
			
			file->groups = g_list_remove (file->groups, group);
			
			for (li = group->options; li != NULL; li = group->options) {
				KibaPrefOption *option = li->data;
				
				group->options = g_list_remove (group->options, option);
				switch (option->member.type) {
					case KIBA_DOUBLE:
					case KIBA_STRING:
						g_free(option->value);
						break;
					case KIBA_LIST: {
						GSList *sl;
						for (sl = (GSList*) option->value; sl != NULL; sl = (GSList*) option->value) {
							char *data = sl->data;
							g_free(data);
							option->value = g_slist_remove((GSList*) option->value, sl->data);
						}
						g_slist_free(option->value);
						break;
					}
					case KIBA_INT:
					case KIBA_BOOLEAN:
					default:
						break;
				}
				
				g_free (option);
			}
			g_list_free (group->options);
			g_free (group);
			break;
		}
	}
	
	/* also remove this options for the global file */
	if (file != prefs->globalFile && file != prefs->generalFile) {
		kiba_prefs_remove_options (prefs->globalFile, options);
	}
}

/**
 * kiba_prefs_add_options:
 * @file: The KibaPrefsFile instance.
 * @options: An array of PrefMember's.
 * @option_elements: Element number of Option Struct.
 *
 * Initialls Options in the Preferences Database.
 */
void kiba_prefs_add_options (KibaPrefsFile *file,
                             PrefMember *options,
                             gint option_elements) {
	KibaPrefs *settings;
	KibaPrefGroup *group;
	GList *l;
	gint i = 0;
	
	g_return_if_fail (options != NULL);
	g_return_if_fail(file != NULL);
	
	settings = kiba_prefs_get ();
	
	group = g_new0 (KibaPrefGroup, 1);
	group->member = options;
	group->member_elements = option_elements;
	group->options = NULL;
	
	for (i = 0; i < option_elements; i++) {
		KibaPrefOption *option;
		KibaOption *optionData = NULL;
		
		for (l = file->options; l != NULL; l = l->next) {
			KibaOption *foo = l->data;
			if (strcmp(foo->id, options[i].id) != 0)
				continue;
			if (kiba_prefs_get_type_from_string(foo->type) != options[i].type)
				continue;
			optionData = foo;
		}
		/* option not found in config file */
// 		if (!optionData)
// 			continue;
		
		option = g_new0 (KibaPrefOption, 1);
		option->member = options[i];
		option->callbacks = NULL;
		option->value = NULL;
		if (optionData)
			__set_option_from_string(option, optionData->value);
		
		group->options = g_list_append (group->options, option);
	}
	
	file->groups = g_list_append (file->groups, group);
	
	/* also add this options for the global file */
	if (file != settings->globalFile && file != settings->generalFile) {
		kiba_prefs_add_options (settings->globalFile, options, option_elements);
	}
}

/**
 * kiba_prefs_write_to_config_file:
 * @file: The KibaPrefsFile instance.
 * @id: The Pref Name.
 * @type: The Pref Type.
 * @value: The Pref Value.
 *
 * Writes Option to Config File.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_prefs_write_to_config_file (KibaPrefsFile *file,
                                          const char *id,
                                          const char *type,
                                          const char *value) {
	KibaPrefs *settings = kiba_prefs_get();
	GList *l;
	gboolean isNewOption;
	
	g_return_if_fail(id != NULL);
	g_return_if_fail(type != NULL);
	g_return_if_fail(value != NULL);
	g_return_if_fail(file != NULL);
	
	/* get options */
	isNewOption = TRUE;
	for (l = file->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0 &&
		    strcmp(type, optionData->type) == 0) {
			g_free(optionData->value);
			optionData->value = g_strdup(value);
			isNewOption = FALSE;
			break;
		}
	}
	if (isNewOption) {
		KibaOption *newOption = g_new0(KibaOption, 1);
		newOption->id = g_strdup(id);
		newOption->type = g_strdup(type);
		newOption->value = g_strdup(value);
		file->options = g_list_append(file->options, newOption);
	}
	
	__write_options(settings, file);
}

/**
 * kiba_prefs_remove_from_config_file:
 * @file: The KibaPrefsFile instance.
 * @id: The Pref Name.
 * @type: The Pref Type.
 *
 * Removes Option to Config File.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_prefs_remove_from_config_file (KibaPrefsFile *file,
                                             const char *id,
                                             const char *type) {
	KibaPrefs *settings = kiba_prefs_get();
	GList *l;
	gboolean isNewOption;
	
	g_return_if_fail(id != NULL);
	g_return_if_fail(type != NULL);
	g_return_if_fail(file != NULL);
	
	/* get options */
	isNewOption = TRUE;
	for (l = file->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0 && strcmp(type, optionData->type) == 0) {
			g_free(optionData->value);
			g_free(optionData);
			file->options = g_list_remove(file->options, optionData);
			isNewOption = FALSE;
			break;
		}
	}
	if (isNewOption) {
		return FALSE;
	}
	
	__write_options(settings, file);
}

/**
 * kiba_prefs_get_option_from_file:
 *
 * Gets String Value from Kiba Config File.
 * @file: The KibaPrefsFile instance.
 * @id: The Option Name to watch for.
 * @type: The Type of the Option.
 *
 * Returns: The Value as String, or NULL if none Setting was found.
 */
char* kiba_prefs_get_option_from_file (KibaPrefsFile *file,
                                       const char *id,
                                       const char *type) {
	KibaPrefs *prefs = kiba_prefs_get ();
	GList *l;
	char *ret = NULL;
	
	g_return_val_if_fail(file != NULL, NULL);
	
	for (l = file->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp (optionData->id, id) == 0 && strcmp (optionData->type, type) == 0) {
			ret = g_strdup(optionData->value);
			break;
		}
	}
	
	return ret;
}

/**
 * kiba_prefs_update_config_file:
 * @prefs: The KibaPrefs instance.
 * @file: The KibaPrefsFile instance.
 *
 * Update config file with schema files.
 */
void kiba_prefs_update_config_file(KibaPrefs *prefs,
                                   KibaPrefsFile *file) {
	KibaSchema *schema;
	KibaSchemaType type;
	GList *schemaOptions, *l;
	gboolean configChanged = FALSE;
	
	g_return_if_fail(file != NULL);
	g_return_if_fail(file != prefs->globalFile);
	
	if (file == prefs->generalFile)
		type = KIBA_SCHEMA_GENERAL;
	else
		type = KIBA_SCHEMA_DOCK;
	schema = kiba_prefs_schema_get(type);
	schemaOptions = kiba_prefs_schema_get_options(schema);
	
	//TODO: remove options from config that do not appear in the schema.
	// but currently there are some used options that do not appear in the schema (for example 'dock_groups')
	
	/* add new options from schema */
	for (l = schemaOptions; l != NULL; l = l->next) {
		KibaOption *schemaOption = l->data;
		gboolean newOption = TRUE;
		GList *li;
		
		/* check if $schemaOption is allready added to the config */
		for (li = file->options; li != NULL; li = li->next) {
			KibaOption *option = li->data;
			
			if (strcmp (schemaOption->id, option->id) == 0 &&
			    strcmp (schemaOption->type, option->type) == 0) {
				newOption = FALSE;
				break;
			}
		}
		if (!newOption) continue;
		
		/* adding $schemaOption to the config file */
		KibaOption *configOption = g_new0(KibaOption, 1);
		configOption->id = g_strdup(schemaOption->id);
		configOption->type = g_strdup(schemaOption->type);
		configOption->value = g_strdup(schemaOption->value);
		file->options = g_list_append(file->options, configOption);
		
		configChanged = TRUE;
	}
	
	if (configChanged)
		__write_options(prefs, file);
	
	kiba_prefs_schema_fini(schema);
}

/**
 * kiba_prefs_get_file:
 * @configPath: Path to the config file.
 *
 * Returns KibaPrefsFile for @configPath.
 *
 * Returns: A KibaPrefsFile.
 */
KibaPrefsFile* kiba_prefs_get_file (const char *configPath) {
	KibaPrefs *prefs = kiba_prefs_get();
	
	g_return_val_if_fail(configPath != NULL, NULL);
	
	return __get_file_at_path(prefs, configPath);
}

/**
 * kiba_prefs_get_file_path:
 * @file: The KibaPrefsFile instance.
 *
 * Returns path to @file.
 *
 * Returns: A path to @file.
 */
const char* kiba_prefs_get_file_path (KibaPrefsFile *file) {
	g_return_val_if_fail(file != NULL, NULL);
	
	return file->path;
}

KibaPrefsFile* kiba_prefs_add_file_with_type (const char *configPath,
                                              KibaPrefsFileType type) {
	KibaPrefs *prefs = kiba_prefs_get();
	KibaPrefsFile *file;
	
	g_return_val_if_fail(configPath != NULL, NULL);
	g_return_val_if_fail(g_str_has_suffix(configPath, ".xml"), NULL);
	
	file = __get_file_at_path(prefs, configPath);
	/* if the file allready was added */
	if (file) return file;
	
	if (!g_file_test(configPath, G_FILE_TEST_EXISTS)) {
		FILE *foo = fopen(configPath, "w");
		fprintf(foo, "<?xml version=\"1.0\"?>\n<config>\n</config>\n");
		fclose (foo);
	}
	
	file = g_new0(KibaPrefsFile, 1);
	file->path = g_strdup(configPath);
	file->groups = NULL;
	file->options = __get_options(kiba_prefs, configPath);
	
	switch (type) {
		case KIBA_PREFS_FILE_GENERAL:
			prefs->generalFile = file;
			break;
		case KIBA_PREFS_FILE_GLOBAL:
			prefs->globalFile = file;
			break;
		case KIBA_PREFS_FILE_INDIVIDUAL:
			prefs->configFiles = g_list_append(prefs->configFiles, file);
			break;
		default:
			g_assert_not_reached();
			break;
	}
	
	return file;
}


/**
 * kiba_prefs_add_config_file:
 * @configPath: Path to the config file to add.
 *
 * Add config file at @configFile to the database.
 */
KibaPrefsFile* kiba_prefs_add_config_file (const char *configPath) {
	return kiba_prefs_add_file_with_type(configPath, KIBA_PREFS_FILE_INDIVIDUAL);
}

KibaPrefsFile* kiba_prefs_get_general_file (KibaPrefs *prefs) {
	g_return_val_if_fail(prefs != NULL, NULL);
	return prefs->generalFile;
}

KibaPrefsFile* kiba_prefs_get_global_file (KibaPrefs *prefs) {
	g_return_val_if_fail(prefs != NULL, NULL);
	return prefs->globalFile;
}

gboolean kiba_prefs_has_global_option (KibaPrefsFile *file,
                                       const char *id) {
	KibaPrefs *prefs = kiba_prefs_get();
	if (file == prefs->globalFile)
		return TRUE;
	if (__file_contains_option(prefs->globalFile, id))
		return TRUE;
	return FALSE;
}

/*** Initi/Fini ***/

/**
 * kiba_prefs_fini:
 *
 * Frees allocated Memory.
 */
void kiba_prefs_fini (void) {
	KibaPrefs *prefs = kiba_prefs_get();
	GList *l;
	
	for (l = prefs->configFiles; l != NULL; l = prefs->configFiles) {
		KibaPrefsFile *file = l->data;
		prefs->configFiles = g_list_remove(prefs->configFiles, file);
		__free_file(file);
	}
	g_list_free(prefs->configFiles);
	
	__free_file(prefs->generalFile);
	__free_file(prefs->globalFile);
	
	g_free (prefs);
}

static void kiba_prefs_init () {
	kiba_prefs = g_new0 (KibaPrefs, 1);
	kiba_prefs->configFiles = NULL;
	kiba_prefs->generalFile = NULL;
	kiba_prefs->globalFile = NULL;
}

/**
 * kiba_prefs_get:
 *
 * Creates and returns a new KibaPrefs object.
 *
 * Returns: The KibaPrefs object created.
 */
KibaPrefs* kiba_prefs_get () {
	if (!kiba_prefs)
		kiba_prefs_init();
	
	return kiba_prefs;
}
