/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-progress
 * @short_description: Shows progress on #KibaObject's.
 * @see_also:#KibaObject
 * @stability: Unstable
 *
 * Shows progress on #KibaObject's.
 */

#include "../config.h"
#include <math.h>

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-progress.h"
#include "kiba-macros.h"

#define PROGRESS_SIZE 20
#define PROGRESS_STEP 2
#define PROGRESS_STOP_VALUE 100
#define PROGRESS_BORDER_SIZE 2

typedef struct {
	KibaProgress *progress; /**< Pointer to the KibaProgress instance */
	
	KibaObject *object;
	
	GdkRectangle area; /**< The progress area on the window */
	
	gint value; /**< current destination value of the process (<=100) */
	gint current_value; /**< current value of the process (<=value) */
	gint value_change_id;
}KibaProgressObject;

struct _KibaProgress {
	GObject parent;
	
	GHashTable *objects;
	
	KibaProgressStyle style;   /**< The progress background style */
	cairo_surface_t *surface; /**< The progress surface */
};

struct _KibaProgressClass {
	GObjectClass parent_class;
};

G_DEFINE_TYPE (KibaProgress, kiba_progress, G_TYPE_OBJECT)

/*** Private methods ***/

static void __remove_object (gpointer key,
                             gpointer value,
                             gpointer user_data) {
	KibaProgress *progress = user_data;
	KibaObject *object = key;
	
	kiba_progress_remove_object(progress, object);
}

static void __object_queue_redraw (gpointer key,
                                   gpointer value,
                                   gpointer user_data) {
	KibaProgressObject *progress_object = value;
	KibaObject *object = key;
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *dock = widget->parent;
	g_return_if_fail(dock != NULL);
	GtkWidget *win = dock->parent;
	g_return_if_fail(win != NULL);
	
	if (object->position != KIBA_OBJECT_ON_DOCK)
		return;
	
	kiba_win_invalidate_area (win, progress_object->area);
}

static void __update_object_position (KibaProgressObject *progress_object) {
	KibaObject *object = progress_object->object;
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *dock = widget->parent;
	const KibaLayoutData *d;
	GdkRectangle new_position;
	gint fX, fY;
	
	if (!dock)
		d = &kiba_layout_data[KIBA_POSITION_BOTTOM_EDGE];
	else
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	fX = widget->allocation.x;
	fY = widget->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		new_position.x = fX +
			(widget->allocation.width - progress_object->area.width) / 2 + widget->allocation.width/2;
		new_position.y = (d->valignment/2) * (fY +
			(widget->allocation.height - progress_object->area.height) / 2 + widget->allocation.height/2);
	}
	else {
		new_position.y = fY +
			(widget->allocation.height - progress_object->area.height) / 2 + widget->allocation.height/2;
		new_position.x = (d->halignment/2) * (fX +
			(widget->allocation.width - progress_object->area.width) / 2 + widget->allocation.width/2);
	}
	
	if (new_position.x != progress_object->area.x || new_position.y != progress_object->area.y) {
		__object_queue_redraw (object, progress_object, NULL);
		
		progress_object->area.x = new_position.x;
		progress_object->area.y = new_position.y;
		
		__object_queue_redraw (object, progress_object, NULL);
	}
}

static gint __animate_value_changes (gpointer data) {
	KibaProgressObject *progress_object = data;
	KibaObject *object = progress_object->object;
	
	if (progress_object->value > progress_object->current_value) {
		progress_object->current_value += PROGRESS_STEP;
		if (progress_object->value < progress_object->current_value)
			progress_object->current_value = progress_object->value;
	}
	else if (progress_object->value < progress_object->current_value) {
		progress_object->current_value -= PROGRESS_STEP;
		if (progress_object->value > progress_object->current_value)
			progress_object->current_value = progress_object->value;
	}
	
	__object_queue_redraw (object, progress_object, NULL);
	
	if (progress_object->current_value >= PROGRESS_STOP_VALUE) {
		progress_object->current_value = 0;
		progress_object->value = 0;
	}
	
	if (progress_object->current_value == progress_object->value) {
		progress_object->value_change_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static void __object_size_allocate_cb (GtkWidget *widget,
                                       GtkAllocation *allocation,
                                       gpointer user_data) {
	KibaProgressObject *progress_object = user_data;
	__update_object_position (progress_object);
}

/** Rendering **/

static void __draw_surface (cairo_t *cr,
                            cairo_surface_t *surface,
                            gint width,
                            gint height,
                            void *data) {
	KibaProgress *progress = data;
	gint radius;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	radius = PROGRESS_SIZE/2;
	
	cairo_arc (cr, radius, radius, radius - (PROGRESS_BORDER_SIZE/2), 0, 2*M_PI);
	cairo_close_path (cr);
	cairo_set_source_rgba (cr,
		progress->style.bg_rgba.r,
		progress->style.bg_rgba.g,
		progress->style.bg_rgba.b,
		progress->style.bg_rgba.a);
	cairo_fill_preserve (cr);
	
	cairo_set_source_rgba (cr,
		progress->style.rgba.r,
		progress->style.rgba.g,
		progress->style.rgba.b, 1);
	cairo_set_line_width (cr, PROGRESS_BORDER_SIZE);
	cairo_stroke (cr);
}

static void __update_surface (KibaProgress *progress) {
	Kiba *kiba = kiba_get();
	KibaWin *win = kiba_get_win(kiba);
	cairo_t *cr;
	
	cr = kiba_win_cairo_create (win);
	cairo_update_surface_at_size (cr, &progress->surface,
		PROGRESS_SIZE, PROGRESS_SIZE,
		__draw_surface, progress);
	cairo_destroy (cr);
}

static void __draw_object (gpointer key,
                           gpointer value,
                           gpointer user_data) {
	KibaProgressObject *progress_object = value;
	KibaObject *object = key;
	cairo_t *cr = user_data;
	gint radius;
	
	if (!kiba_progress_object_is_running(progress_object->progress, object))
		return;
	
	if (object->position != KIBA_OBJECT_ON_DOCK)
		return;
	
	radius = PROGRESS_SIZE/2;
	
	cairo_save (cr);
		cairo_translate (cr, progress_object->area.x, progress_object->area.y);
		
		cairo_set_source_surface (cr, progress_object->progress->surface, 0, 0);
		cairo_paint (cr);
		
		//FIXME: this causes heavy cpu load, i guess cairo does not like arcs much
		{
			cairo_move_to (cr, radius, radius);
			cairo_arc (cr,
				radius,
				radius,
				radius,
				M_PI/2,
				(90 + 360*progress_object->current_value/PROGRESS_STOP_VALUE) * (M_PI/180));
				
			cairo_set_source_rgba (cr,
				progress_object->progress->style.rgba.r,
				progress_object->progress->style.rgba.g,
				progress_object->progress->style.rgba.b,
				progress_object->progress->style.rgba.a);
			cairo_fill (cr);
		}
	cairo_restore (cr);
}

static void __object_destroy_cb (GtkWidget *widget,
                                 gpointer data) {
	KibaProgress *progress = data;
	KibaObject *object = KIBA_OBJECT(widget);
	kiba_progress_remove_object(progress, object);
}

static void __value_destroy_func (gpointer value) {
	KibaProgressObject *progress_object = value;
	
	if (progress_object->value_change_id) {
		g_source_remove(progress_object->value_change_id);
	}
	
	if (KIBA_IS_OBJECT(progress_object->object)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(progress_object->object),
			G_CALLBACK(__object_size_allocate_cb), progress_object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(progress_object->object),
			G_CALLBACK(__object_destroy_cb), progress_object->progress);
	}
	
	g_free (progress_object);
}

/*** Public methods ***/

/**
 * kiba_progress_object_add:
 * @progress: The #KibaProgress instance.
 * @object: The #KibaObject to add.
 *
 * Adds progress object.
 * The progress object is removed automatical when @object is destroyed.
 */
void kiba_progress_add_object (KibaProgress *progress,
                               KibaObject *object) {
	KibaProgressObject *progress_object;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	progress_object = g_hash_table_lookup (progress->objects, object);
	g_return_if_fail (progress_object == NULL);
	
	progress_object = g_new0 (KibaProgressObject, 1);
	progress_object->object = object;
	progress_object->area.width  = PROGRESS_SIZE;
	progress_object->area.height = PROGRESS_SIZE;
	progress_object->value = 0;
	progress_object->current_value = 0;
	progress_object->value_change_id = 0;
	progress_object->progress = progress;
	
	g_hash_table_insert (progress->objects, object, progress_object);
	
	__update_object_position (progress_object);
	
	g_signal_connect (G_OBJECT(object), "size-allocate",
		G_CALLBACK(__object_size_allocate_cb), progress_object);
	g_signal_connect_after (G_OBJECT(object), "size-allocate",
		G_CALLBACK(__object_size_allocate_cb), progress_object);
	g_signal_connect (G_OBJECT(object), "destroy",
		G_CALLBACK(__object_destroy_cb), progress);
}

/**
 * kiba_progress_object_remove:
 * @progress: The #KibaProgress instance.
 * @object: The #KibaObject to remove.
 *
 * Removes progress object.
 */
void kiba_progress_remove_object (KibaProgress *progress,
                                  KibaObject *object) {
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	g_hash_table_remove(progress->objects, object);
}

/**
 * kiba_progress_object_is_running:
 * @progress: The #KibaProgress instance.
 * @object: The #KibaObject.
 *
 * Return if a progress for a KibaProgress Instance is running.
 *
 * Returns: TRUE  if a progress is running, otherwise FALSE is returned.
 */
gboolean kiba_progress_object_is_running (KibaProgress *progress,
                                          KibaObject *object) {
	KibaProgressObject *progress_object;
	
	g_return_val_if_fail (KIBA_IS_PROGRESS(progress), FALSE);
	
	progress_object = g_hash_table_lookup (progress->objects, object);
	g_return_val_if_fail (progress_object != NULL, FALSE);
	
	return progress_object->current_value != 0;
}

/**
 * kiba_progress_object_set_info:
 * @progress: The #KibaProgress instance.
 * @object: The #KibaObject to set the value for.
 * @value: The new progress value (must be <= 100).
 *
 * Set progress info for a KibaProgressObject instance.
 */
void kiba_progress_object_set_value (KibaProgress *progress,
                                     KibaObject *object,
                                     gint value) {
	KibaProgressObject *progress_object;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	g_return_if_fail (value <= PROGRESS_STOP_VALUE);
	
	progress_object = g_hash_table_lookup (progress->objects, object);
	g_return_if_fail (progress_object != NULL);
	
	progress_object->value = value;
	
	if (!progress_object->value_change_id)
		progress_object->value_change_id = g_timeout_add (20, __animate_value_changes, progress_object);
}

/**
 * kiba_progress_get_style:
 * @progress: The #KibaProgress instance.
 *
 * Returns: The progress style.
 */
KibaProgressStyle kiba_progress_get_style (KibaProgress *progress) {
	g_return_val_if_fail (KIBA_IS_PROGRESS(progress),
		((KibaProgressStyle) {{0.0f,0.0f,0.0f,0.0f},{0.0f,0.0f,0.0f,0.0f}}));
	return progress->style;
}

/**
 * kiba_progress_set_style:
 * @progress: The KibaProgress instance.
 * @style: The style to set.
 *
 * Sets progress background style.
 */
void kiba_progress_set_style (KibaProgress *progress,
                              KibaProgressStyle style) {
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	progress->style = style;
	
	__update_surface (progress);
	g_hash_table_foreach (progress->objects, __object_queue_redraw, NULL);
}

/**
 * kiba_progress_draw:
 * @progress: The KibaProgress Instance.
 * @cr: The Drawing Context.
 *
 * Draws all running progresses to the drawing context.
 */
void kiba_progress_draw (KibaProgress *progress,
                         cairo_t *cr) {
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	g_hash_table_foreach (progress->objects, __draw_object, cr);
}

/*** class implementations ***/

static void kiba_progress_dispose (GObject *object) {
	KibaProgress *progress = KIBA_PROGRESS(object);
	
	if (progress->surface)
		cairo_surface_destroy (progress->surface);
	
	g_hash_table_remove_all (progress->objects);
	g_hash_table_unref(progress->objects);
	
	G_OBJECT_CLASS(kiba_progress_parent_class)->dispose (object);
}

/*** Class init ***/

static void kiba_progress_class_init (KibaProgressClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS(class);
	
	gobject_class->dispose = kiba_progress_dispose;
}

static void kiba_progress_init (KibaProgress *progress) {
	progress->style = (KibaProgressStyle) {
		{ 0.0f, 0.0f, 0.0f, 0.7f },
		{ 1.0f, 1.0f, 1.0f, 0.8f }
	};
	progress->surface = NULL;
	
	progress->objects = g_hash_table_new_full (
		NULL, /**<< using default hash function **/
		g_direct_equal, /**<< comparing pointers **/
		NULL,
		__value_destroy_func);
}

/**
 * kiba_progress_new:
 *
 * Create a KibaProgress Instance.
 *
 * Returns: The KibaProgress Instance created.
 */
KibaProgress* kiba_progress_new (void) {
	return g_object_new (KIBA_TYPE_PROGRESS, NULL);
}
