/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *
*/

/**
 * SECTION:kiba-screen
 * @short_description: Object representing the physical screen.
 * @see_also:#KibaWindow,#GdkScreen
 * @stability: Unstable
 *
 * Object representing the physical screen.
 */

#include <string.h>
#include <gdk/gdkx.h>
#include <X11/Xatom.h>

#include "kiba-x11.h"
#include "kiba-window.h"
#include "kiba-screen.h"
#include "kiba-prefs.h"

#define KIBA_SCREEN_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_SCREEN, KibaScreenPrivate))

G_DEFINE_TYPE (KibaScreen, kiba_screen, G_TYPE_OBJECT)

typedef struct _KibaScreenPrivate KibaScreenPrivate;
typedef struct _KibaScreenOptions KibaScreenOptions;

Atom _NET_CLIENT_LIST_STACKING;

struct _KibaScreenPrivate {
	GdkWindow     *topmost_window;
	
	gint           workspace;
	GdkPoint       viewport;
};

enum {
	SIG_WINDOW_OPENED = 0,
	SIG_WINDOW_CLOSED,
	SIG_WITHDRAWN_TOGGLED,
	SIG_STACK_CHANGED,
	SIG_VIEWPORT_CHANGED,
	SIG_WORKSPACE_CHANGED,
	SIG_SCREEN_CHANGED,
	SIG_SCREEN_LAST
};

static guint screen_signals[SIG_SCREEN_LAST] = { 0 };
static KibaScreen *default_screen = NULL;

static GdkWindow *get_gdk_root_window ();
static GdkFilterReturn x_root_callback (GdkXEvent *gdkxevent,
                                        GdkEvent  *event,
                                        gpointer   data);

/**
Helper Functions
**/

/**
 * kiba_screen_set_monitor:
 * @screen: The KibaScreen.
 * @monitor: The monitor number.
 *
 * Sets the monitor for the screen geometry.
 */
void kiba_screen_set_monitor (KibaScreen *screen,
			      gint        monitor) {
	g_return_if_fail (KIBA_IS_SCREEN(screen));
	
	screen->monitor = monitor;
	
	if (gdk_screen_get_n_monitors (screen->gdk_screen) >= monitor) {
		screen->geometry.x = 0;
		screen->geometry.y = 0;
		screen->geometry.width = gdk_screen_get_width (screen->gdk_screen);
		screen->geometry.height = gdk_screen_get_height (screen->gdk_screen);
	}
	else
		gdk_screen_get_monitor_geometry (screen->gdk_screen,
					monitor,
					&screen->geometry);
}

static gboolean kiba_screen_skip_hints (GdkWindow *window) {
	char *state = kiba_get_wintype (gdk_display, window);
	if (state == NULL)
		return TRUE;
	
	if (strcmp (state, "_NET_WM_WINDOW_TYPE_NORMAL") == 0 ||
	    strcmp (state, "_NET_WM_WINDOW_TYPE_DESKTOP") == 0) {
		g_free (state);
		return TRUE;
	}
	
	g_free (state);
	return FALSE;
}

/**
 * kiba_screen_get_stack:
 * @screen: The KibaScreen to get the stack from.
 *
 * Returns a list with all KibaWindow's managed by the screen.
 */
GList* kiba_screen_get_stack (KibaScreen *screen) {
	Screen *screen_x11;
	GList *ret = NULL;
	Atom type_return;
	gint format_return;
	gulong nitems_return;
	gulong bytes_after_return;
	guchar *data = NULL;
	
// 	root_stack = gdk_screen_get_window_stack (screen->gdk_screen);
	
	g_return_val_if_fail (GDK_IS_SCREEN(screen->gdk_screen), NULL);
	g_return_val_if_fail (gdk_x11_screen_supports_net_wm_hint (screen->gdk_screen,
				gdk_atom_intern ("_NET_CLIENT_LIST_STACKING", FALSE)), NULL);
	
	screen_x11 = gdk_x11_screen_get_xscreen (screen->gdk_screen);
	
	if (XGetWindowProperty (gdk_display, RootWindowOfScreen(screen_x11),
		kiba_get_atom (gdk_display, "_NET_CLIENT_LIST_STACKING"),
		0, G_MAXLONG, False, XA_WINDOW, &type_return, &format_return, &nitems_return,
		&bytes_after_return, &data) == Success) {

		if ((type_return == XA_WINDOW) && (format_return == 32) &&
		    (data) && (nitems_return > 0)) {
			
			GdkNativeWindow *stack = (GdkNativeWindow *) data;
			GdkWindow *win;
			int i;
			
			for (i = 0; i < nitems_return; i++) {
				win = gdk_window_foreign_new_for_display (gdk_x11_lookup_xdisplay (gdk_display), stack[i]);
				
				if (win != NULL &&  kiba_screen_skip_hints(win))
					ret = g_list_append (ret, win);
			}
		}
	}
	
	if (data)
		XFree (data);
	
	return ret;
}

static void kiba_window_state_changed_cb(KibaWindow *window, gint states_changed, gpointer data) {
	KibaScreen *screen = KIBA_SCREEN_DEFAULT;

	if (window->state & (1 << KIBA_WINDOW_CLOSED)) {
		g_signal_emit (screen,
			screen_signals[SIG_WINDOW_CLOSED],
			0, window);
		
		screen->windows = g_list_remove (screen->windows, window);
		g_object_unref (window);
	} else if (states_changed & (1 << KIBA_WINDOW_WITHDRAWN)) {
		g_signal_emit (screen,
			screen_signals[SIG_WITHDRAWN_TOGGLED],
			0, window);
	}
}

static gint kiba_screen_check_window_stack (gpointer data) {
	KibaScreen *screen = data;
	KibaScreenPrivate *priv;
	GList *root_stack, *l;
	GdkWindow *topmost;
	GdkPoint active_viewport;
	int num_viewports_x, num_viewports_y, num_workspaces;
	
	priv = KIBA_SCREEN_GET_PRIVATE (screen);
	
	root_stack = kiba_screen_get_stack (screen);
	
	/* check for changed workspace */
	num_workspaces = kiba_num_workspaces ();
	if (num_workspaces > 1) {
		int active_workspace;
			
		active_workspace = kiba_get_active_workspace_number (gdk_display);
			
		if (active_workspace + 1 != priv->workspace) {
			g_signal_emit (screen,
				screen_signals[SIG_WORKSPACE_CHANGED],
				0, active_workspace + 1);
			
			priv->workspace = active_workspace + 1;
		}
	}
	
	/* check for changed viewport */
	kiba_get_num_viewports (&num_viewports_x, &num_viewports_y);
	active_viewport.x = num_viewports_x - 1;
	active_viewport.y = num_viewports_y - 1;
	kiba_get_avtive_viewport (&active_viewport.x, &active_viewport.y);
	if (active_viewport.x + 1 != priv->viewport.x) {
		active_viewport.x++;
		active_viewport.y++;
		
		g_signal_emit (screen,
			screen_signals[SIG_VIEWPORT_CHANGED],
			0, &active_viewport);
		
		priv->viewport = active_viewport;
	}
	
	/* creating new KibaWin instances for opened windows */
	for (l = root_stack; l != NULL; l = l->next) {
		GdkWindow *gdk_window = l->data;
		KibaWindow *window;
			
		if (kiba_screen_get_window (screen, gdk_window) != NULL)
			continue;
			
		window = kiba_window_new (gdk_window);
		
		g_signal_connect_after(G_OBJECT(window), "state-changed",
		  G_CALLBACK(kiba_window_state_changed_cb), NULL);
		
		g_signal_emit (screen,
			screen_signals[SIG_WINDOW_OPENED],
			0, window);
		
		screen->windows = g_list_append (screen->windows, window);
	}
	
	/* check if stack changed */
	topmost = g_list_nth_data (root_stack, g_list_length (root_stack) - 1);
	if (topmost != priv->topmost_window) {
		KibaWindow *topmost_window;
		
		topmost_window = kiba_screen_get_window (screen, topmost);
		
		g_signal_emit (screen,
			screen_signals[SIG_STACK_CHANGED],
			0, topmost_window);
			
		priv->topmost_window = topmost;
	}
	
	return TRUE;
}

/**
Public Functions
**/

/**
 * kiba_screen_for_each_window:
 * @screen: The KibaScreen.
 * @for_each_func: (gpointer data, GdkWindow *window); @data The Userdata @window The GdkWindow (Topmost Window is used first)
 * @data: The Userdata
 *
 * Runs Functions for every GdkWindow in KibaScreen's stack
 */
void kiba_screen_for_each_window (KibaScreen *screen,
				  GFunc       for_each_func,
				  gpointer    data) {
	GList *l;
	
	g_return_if_fail (KIBA_IS_SCREEN(screen));
	
	for (l = screen->windows; l != NULL; l = l->next) {
		KibaWindow *window = l->data;
		
		for_each_func (data, window);
	}
}

/**
 * kiba_screen_get_window:
 * @screen: The KibaScreen.
 * @gdk_window: The GdkWindow to search for in KibaScreen's stack.
 *
 * Runs Functions for every GdkWindow in KibaScreen's stack
 *
 * Returns: The KibaWindow associated with gdk_window, or NULL if none found.
 */
KibaWindow* kiba_screen_get_window (KibaScreen *screen,
				    GdkWindow  *gdk_window) {
	GList *l;
	
	g_return_val_if_fail (KIBA_IS_SCREEN(screen), NULL);
	
	for (l = screen->windows; l != NULL; l = l->next) {
		KibaWindow *window = l->data;
		
		if (window->gdk_window == gdk_window)
			return window;
	}
	
	return NULL;
}

static void screen_size_changed_cb (GdkScreen *screen,
				    gpointer   data) {
	KibaScreen *kiba_screen = data;
	
	kiba_screen_set_monitor (kiba_screen, kiba_screen->monitor);
}

static void default_gdk_screen_changed (GdkDisplayManager *display_manager,
				        GdkDisplay        *display,
				        gpointer           user_data) {
	KibaScreen *screen = user_data;
	KibaScreenPrivate *priv;
	GdkScreen *default_screen;
	
	priv = KIBA_SCREEN_GET_PRIVATE (screen);
	default_screen = gdk_display_get_default_screen(display);
	
	if (display != gdk_display_get_default() || default_screen == screen->gdk_screen)
		return;
	
	screen->gdk_screen = default_screen;
	
	if (!display)
		return;
	
	kiba_screen_set_monitor (screen, 0);
	
	g_signal_connect (G_OBJECT(screen->gdk_screen), "size-changed",
		G_CALLBACK(screen_size_changed_cb), screen);
	screen_size_changed_cb (default_screen, screen);
	
	g_signal_emit (screen,
		screen_signals[SIG_SCREEN_CHANGED],
		0, default_screen);
	
	priv->topmost_window = NULL;
}

/**
Class Functions
**/

static void kiba_screen_class_init (KibaScreenClass *class) {
	g_type_class_add_private (class, sizeof (KibaScreenPrivate));
	
	screen_signals[SIG_WINDOW_OPENED] =
		g_signal_new ("window-opened",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, window_opened),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_WINDOW);
	
	screen_signals[SIG_WINDOW_CLOSED] =
		g_signal_new ("window-closed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, window_closed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_WINDOW);
	
	screen_signals[SIG_WITHDRAWN_TOGGLED] =
		g_signal_new ("withdrawn-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, withdrawn_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_WINDOW);
	
	screen_signals[SIG_STACK_CHANGED] =
		g_signal_new ("active-window-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, active_window_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_WINDOW);
	
	screen_signals[SIG_WORKSPACE_CHANGED] =
		g_signal_new ("workspace-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, workspace_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	
	screen_signals[SIG_VIEWPORT_CHANGED] =
		g_signal_new ("viewport-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, viewport_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			GDK_TYPE_RECTANGLE);
	
	screen_signals[SIG_SCREEN_CHANGED] =
		g_signal_new ("screen-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaScreenClass, screen_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			GDK_TYPE_SCREEN);
	
	_NET_CLIENT_LIST_STACKING = kiba_get_atom (gdk_display,
	  "_NET_CLIENT_LIST_STACKING");
}

static void kiba_screen_init (KibaScreen *screen) {
	KibaScreenPrivate *priv = KIBA_SCREEN_GET_PRIVATE (screen);
	GdkDisplayManager *display_manager;
	
	display_manager = gdk_display_manager_get ();
	
	screen->windows = NULL;
	screen->monitor = 0;
	screen->geometry = (GdkRectangle) {
		0, 0, 0, 0
	};
	
	priv->topmost_window = NULL;
	priv->workspace = 1;
	priv->viewport = (GdkPoint) {
		0, 0
	};
	
	default_gdk_screen_changed (display_manager,
				gdk_display_get_default(),
				(gpointer) screen);
	g_signal_connect (G_OBJECT(display_manager), "display-opened",
		G_CALLBACK(default_gdk_screen_changed), screen);
	
	kiba_show_interest_in_events(DefaultRootWindow(gdk_display));
	gdk_window_add_filter (get_gdk_root_window(), x_root_callback, screen);
}

/**
 * get_gdk_root_window:
 * A convenience function to get the GDKWindow corresponding to the root X Window.
 *
 * Returns: The GDKWindow corresponding to the root X Window.
 */
static GdkWindow *get_gdk_root_window () {
	return gdk_window_foreign_new_for_display (gdk_x11_lookup_xdisplay (gdk_display), DefaultRootWindow (gdk_display));
}

/**
 * x_root_callback:
 * @gdkxevent: The X event to filter
 * @event: The GDK event to which the X event will be translated
 * @data: The corresponding KibaScreen
 *
 * Callback from X, the windowing system, when the corresponding window changes
 *
 * Returns: GDK_FILTER_CONTINUE (event not handled, continue processing)
 */
static GdkFilterReturn x_root_callback (GdkXEvent *gdkxevent,
                                        GdkEvent  *event,
                                        gpointer   screen) {
	XEvent *xevent = gdkxevent;
	
	if (xevent->type == PropertyNotify &&
	  xevent->xproperty.atom == _NET_CLIENT_LIST_STACKING)
		kiba_screen_check_window_stack (screen);
	
	return GDK_FILTER_CONTINUE;
}

static KibaScreen* kiba_screen_new () {
	return g_object_new (KIBA_TYPE_SCREEN, NULL);
}

/**
 * kiba_screen_get:
 *
 * Returns and/or Createsd a KibaScreen Instnace.
 *
 * Returns: The KibaScreen object.
 */
KibaScreen* kiba_screen_get () {
	if (!default_screen)
		default_screen = kiba_screen_new ();
	
	return default_screen;
}

/**
 * kiba_screen_destroy:
 * @screen: The KibaScreen.
 *
 * Destroys and frees KibaScreen.
 */
void kiba_screen_destroy (KibaScreen *screen) {
	KibaScreenPrivate  *priv;
	GList *l;
	
	if (!default_screen)
		return;
	
	priv = KIBA_SCREEN_GET_PRIVATE (screen);
	
	for (l = screen->windows; l != NULL; l = screen->windows) {
		KibaWindow *window = l->data;
		
		screen->windows = g_list_remove (screen->windows, l->data);
		g_object_unref (window);
	}
	g_list_free (screen->windows);
	
	gdk_window_remove_filter (get_gdk_root_window (), x_root_callback, screen);
	g_object_unref (screen);
}
