/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-separator
 * @short_description: Separator object on #KibaDock.
 * @see_also:#KibaWin,#KibaDock
 * @stability: Unstable
 *
 * Separator object on #KibaDock.
 */

#include "kiba-separator.h"
#include "kiba-dock.h"

struct _KibaSeparator {
	KibaFixedObject  parent;
	
	gint size; /* used for clipping only */
};

struct _KibaSeparatorClass {
	KibaFixedObjectClass parent_class;
};

G_DEFINE_TYPE (KibaSeparator, kiba_separator, KIBA_TYPE_FIXED_OBJECT)

static void __set_size (KibaSeparator *separator) {
	GtkWidget *widget = GTK_WIDGET(separator);
	GtkWidget *dock = widget->parent;
	const KibaLayoutData *d;
	GdkRectangle area;
	GtkWidget *win = NULL;
	
	if (!dock) return;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (dock) win = dock->parent;
	if (win) {
		gint width, height, x, y;
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			width = MAX(separator->size, widget->allocation.width);
			x = widget->allocation.x - (width - widget->allocation.width)/2;
			height = widget->allocation.height;
			y = widget->allocation.y;
		} else {
			height =  MAX(separator->size, widget->allocation.height);
			y = widget->allocation.y - (height - widget->allocation.height)/2;
			width = widget->allocation.width;
			x = widget->allocation.x;
		}
		area = (GdkRectangle) {x, y, width, height};
		kiba_win_invalidate_area (win, area);
	}
}

/*** public methods ***/

void kiba_separator_set_size (KibaSeparator *separator,
                              gint size) {
	g_return_if_fail(KIBA_IS_SEPARATOR(separator));
	
	__set_size(separator);
	separator->size = size;
	__set_size(separator);
}

gint kiba_separator_get_size (KibaSeparator *separator) {
	g_return_val_if_fail(KIBA_IS_SEPARATOR(separator), 0);
	return separator->size;
}

/*** custom implementations ***/

/**
 * gtk_widget_size_allocate implementation.
 **/
static void kiba_separator_size_allocate (GtkWidget *widget,
                                          GtkAllocation *allocation) {
	KibaSeparator *separator = KIBA_SEPARATOR(widget);
	
	if (allocation->x != widget->allocation.x || allocation->width != widget->allocation.width ||
	    allocation->y != widget->allocation.y || allocation->height != widget->allocation.height) {
		__set_size(separator);
			(* GTK_WIDGET_CLASS (kiba_separator_parent_class)->size_allocate) (widget, allocation);
		__set_size(separator);
	}
}

/*** class init/fini ***/

static void kiba_separator_class_init (KibaSeparatorClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	widget_class->size_allocate = kiba_separator_size_allocate;
}

static void kiba_separator_init (KibaSeparator *separator) {
	separator->size = -1;
}

/**
 * kiba_separator_new:
 *
 * Creates and returns a KibaSeparator Instance.
 *
 * Returns: The KibaSeparator object created.
 */
GtkWidget* kiba_separator_new (const char *group) {
	GtkWidget *widget =  g_object_new (KIBA_TYPE_SEPARATOR,
                                 "group", group, NULL);
	return widget;
}
