/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"

#include <string.h>
#if 0
#include <pthread.h>
#endif

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-utils.h"
#include "kiba-inline-pixbufs.h"
#include "kiba-surface-factory.h"

typedef struct {
	GList *callbacks;
	cairo_surface_t *surface;
	gint size;
	gint referenceCount;
	char *uri;
} KibaSurfaceInfo;

/* List of currently used surfaces. */
static GList *surfaces = NULL;

#if 0
typedef struct {
	KibaSurfaceFactoryCallback func;
	gpointer userdata;
}KibaSurfaceFactoryCallbackData;

/* Our mutex used when accessing data shared between the main thread and the
   surface thread, i.e. the surface_thread_is_running flag and the
   surfaces_to_make list. */
static pthread_mutex_t surfaces_mutex = PTHREAD_MUTEX_INITIALIZER;
/* The currently surfaceed icon. it also exists in the surfaces_to_make list
 * to avoid adding it again. Lock surfaces_mutex when accessing this. */
static KibaSurfaceInfo *currently_surfacing = NULL;
/* The list of KibaSurfaceInfo structs containing information about the
   surfaces we are making. Lock surfaces_mutex when accessing this. */
static volatile GList *surfaces_to_make = NULL;
/* List of surfaces that need to invoke the callbacks.
   Lock surfaces_mutex when accessing this. */
static volatile GList *surfaces_to_load = NULL;
/* A flag to indicate whether a surface thread is running, so we don't
   start more than one. Lock surfaces_mutex when accessing this. */
static volatile gboolean surface_thread_is_running = FALSE;

static gint surface_load_timeout_id = 0;
#endif

/*** Private methods ***/

static void __free_surface_info (KibaSurfaceInfo *info) {
	g_list_free(info->callbacks);
	g_free(info->uri);
	g_free(info);
}

/* __draw_surface is invoked in a separate thread. */
static void __draw_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaSurfaceInfo *info = data;
	GdkPixbuf *pixbuf = NULL;
	GtkIconInfo *iconInfo = NULL;
#ifdef USE_SVG
	RsvgHandle *handle = NULL;
#endif
	
#ifdef USE_SVG
	if (kiba_load_svg (info->uri, &info->size, &handle)) {
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
		return;
	}
#endif
	if (kiba_load_icon (info->uri, info->size, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}
	
	iconInfo = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), GTK_STOCK_MISSING_IMAGE, info->size, 0);
	if (iconInfo != NULL) {
		const char *path;
		
		path = gtk_icon_info_get_filename (iconInfo);
		gtk_icon_info_free(iconInfo);
		pixbuf = gdk_pixbuf_new_from_file_at_size (path, info->size, -1, NULL);
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
	}
}

#if 0
/* walk through surfaces_to_load list and invoke callbacks, this is done in the main thread */
static gint __load_surfaces (gpointer data) {
	GList *l;
	
	for (l = (GList*) surfaces_to_load; l != NULL; l = (GList*) surfaces_to_load) {
		KibaSurfaceInfo *info = l->data;
		GList *li;
		
		pthread_mutex_lock (&surfaces_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		surfaces_to_load = g_list_remove ((GList*) surfaces_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&surfaces_mutex);
		
		surfaces = g_list_append(surfaces, info);
		
		for (li = info->callbacks; li != NULL; li = info->callbacks) {
			KibaSurfaceFactoryCallbackData *data = li->data;
			data->func(info->surface, info->size, data->userdata);
			info->callbacks = g_list_remove(info->callbacks, li->data);
		}
	}
	
	if (surfaces_to_make != NULL)
		return TRUE;
	
	surface_load_timeout_id = 0;
	return FALSE;
}

/* surface_thread is invoked as a separate thread to make surfaces. */
static gpointer __thread_start (gpointer data) {
	KibaSurfaceInfo *info = NULL;
	KibaWin *win = kiba_get_win(kiba_get());
	GList *l;
	
	/* We loop until there are no more thumbails to make, at which point
	   we exit the thread. */
	for (;;) {
		pthread_mutex_lock (&surfaces_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		
		/* Pop the last surface we just made off the head of the
		   list and free it. I did this here so we only have to lock
		   the mutex once per surface, rather than once before
		   creating it and once after.
		   Don't pop the surface off the queue if the original file
		   mtime of the request changed. Then we need to redo the surface.
		*/
		if (currently_surfacing) {
			g_assert (info == currently_surfacing);
			surfaces_to_make = g_list_remove ((GList*) surfaces_to_make, currently_surfacing);
		}
		currently_surfacing = NULL;
		
		/* If there are no more surfaces to make, reset the
		   surface_thread_is_running flag, unlock the mutex, and
		   exit the thread. */
		if (surfaces_to_make == NULL) {
			surface_thread_is_running = FALSE;
			pthread_mutex_unlock (&surfaces_mutex);
			pthread_exit (NULL);
		}
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&surfaces_mutex);
		
		/* Get the next one to make. We leave it on the list until it
		   is created so the main thread doesn't add it again while we
		   are creating it. */
		info = surfaces_to_make->data;
		currently_surfacing = info;
		
		/* Create the surface. */
		gdk_threads_enter ();
		cairo_t *cr = kiba_win_cairo_create (win);
		cairo_update_surface_at_size (cr, &info->surface,
				info->size, info->size, __draw_surface, info);
		cairo_destroy (cr);
		gdk_threads_leave ();
		
		pthread_mutex_lock (&surfaces_mutex);
		/*********************************
		 * MUTEX LOCKED
		 *********************************/
		surfaces_to_load = g_list_append((GList*) surfaces_to_load, info);
		/*********************************
		 * MUTEX UNLOCKED
		 *********************************/
		pthread_mutex_unlock (&surfaces_mutex);
	}
}
#endif

static KibaSurfaceInfo* __get_surface_info_from_path(const char *iconPath) {
	GList *l;
	for (l = (GList*) surfaces; l != NULL; l = l->next) {
		KibaSurfaceInfo *data = l->data;
		if (strcmp(data->uri, iconPath) == 0)
			return data;
	}
	return NULL;
}

static KibaSurfaceInfo* __get_surface_info_from_surface(cairo_surface_t *surface) {
	GList *l;
	for (l = surfaces; l != NULL; l = l->next) {
		KibaSurfaceInfo *data = l->data;
		if (data->surface == surface)
			return data;
	}
	return NULL;
}

/*** Public methods ***/

/**
 * kiba_surface_factory_unref_surface:
 * @surface: The cairo surface to destroy.
 *
 * Decrease reference count of @surface by one, and remove it from the factory if the reference counter reached 0.
 */
void kiba_surface_factory_remove_surface (cairo_surface_t *surface) {
	KibaSurfaceInfo *data;
	
	data = __get_surface_info_from_surface(surface);
	if (!data) return;
	
	data->referenceCount--;
	if (data->referenceCount > 0)
		return;
	
	//TODO remove the surface after a small delay, it maybe reused in a small amount of time
	surfaces = g_list_remove(surfaces, data);
	cairo_surface_destroy(data->surface);
	__free_surface_info(data);
}

/**
 * kiba_surface_factory_load_icon:
 * @icon: The icon to load.
 * @size: Size for @icon.
 * @func: Callback to invoke after the surface was loaded.
 * @userdata: Userdata for @func.
 *
 * Load @icon to surface in a separate thread (if the icon isnt allready added to the factory)
 * and invoke @func in the main thread.
 * Please note that @func may is not called immediately.
 */
gboolean kiba_surface_factory_load_icon (const char *icon,
                                         gint size,
                                         KibaSurfaceFactoryCallback func,
                                         gpointer userdata) {
	KibaWin *win = kiba_get_win(kiba_get());
	KibaSurfaceInfo *data;
	GList *l;
	char *iconPath;
	
	g_return_if_fail(func != NULL);
	
	iconPath = kiba_get_uri_for_icon(icon, size);
	if (!iconPath) return FALSE;
	
	data = __get_surface_info_from_path(iconPath);
	if (data && data->size == size) {
		data->referenceCount++;
		func(data->surface, size, userdata);
		g_free(iconPath);
		return TRUE;
	}
	
#if 0
	pthread_attr_t thread_attributes;
	pthread_t surface_thread;
	
	KibaSurfaceFactoryCallbackData *callbackData = g_new(KibaSurfaceFactoryCallbackData, 1);
	callbackData->func = func;
	callbackData->userdata = userdata;
	
	/* check if the surface is in the to make list */
	for (l = (GList*) surfaces_to_make; l != NULL; l = l->next) {
		data = l->data;
		if (strcmp(iconPath, data->uri) == 0 && data->size == size) {
			data->referenceCount++;
			/* data->callbacks is only used in the main thread, so we can savely access it here ;) */
			data->callbacks = g_list_append (data->callbacks, callbackData);
			return TRUE;
		}
	}
	
	data = g_new0(KibaSurfaceInfo, 1);
	data->uri = iconPath;
	data->size = size;
	data->surface = NULL;
	data->callbacks = g_list_append (NULL, callbackData);
	data->referenceCount = 1;
	
	pthread_mutex_lock (&surfaces_mutex);
	/*********************************
	 * MUTEX LOCKED
	 *********************************/
	surfaces_to_make = g_list_append ((GList*) surfaces_to_make, data);
	/*********************************
	 * MUTEX UNLOCKED
	 *********************************/
	pthread_mutex_unlock (&surfaces_mutex);
	
	if (!surface_load_timeout_id) {
		surface_load_timeout_id = g_timeout_add (100, (GSourceFunc) __load_surfaces, NULL);
	}
	
	if (surface_thread_is_running)
		return TRUE;
	
	/* We create the thread in the detached state, as we don't need/want
	   to join with it at any point. */
	pthread_attr_init (&thread_attributes);
	pthread_attr_setdetachstate (&thread_attributes,
				     PTHREAD_CREATE_DETACHED);
#ifdef _POSIX_THREAD_ATTR_STACKSIZE
	pthread_attr_setstacksize (&thread_attributes, 128*1024);
#endif
	
	/* We set a flag to indicate the thread is running, so we don't create
	   a new one. We don't need to lock a mutex here, as the surface
	   thread isn't running yet. And we know we won't create the thread
	   twice, as we also check __thread_starter_id before
	   scheduling this idle function. */
	surface_thread_is_running = TRUE;
	pthread_create (&surface_thread, &thread_attributes, __thread_start, NULL);
#else
	data = g_new0(KibaSurfaceInfo, 1);
	data->uri = iconPath;
	data->size = size;
	data->surface = NULL;
	data->callbacks = NULL;
	data->referenceCount = 1;
	
	cairo_t *cr = kiba_win_cairo_create (win);
	cairo_update_surface_at_size (cr, &data->surface,
			data->size, data->size, __draw_surface, data);
	cairo_destroy (cr);
	
	func(data->surface, data->size, userdata);
#endif
	
	return TRUE;
}

void kiba_surface_factory_destroy() {
	KibaSurfaceInfo *data;
	GList *l;
	
#if 0
	if (surface_thread_is_running) {
		surface_thread_is_running = FALSE;
		pthread_exit (NULL);
	}
	
	for (l = (GList*) surfaces_to_make; l != NULL; l = (GList*) surfaces_to_make) {
		data = l->data;
		__free_surface_info(data);
		surfaces_to_make = g_list_remove((GList*) surfaces_to_make, l->data);
	}
#endif
	
	for (l = surfaces; l != NULL; l = l->next) {
		surfaces = g_list_remove(surfaces, l->data);
		data = l->data;
		data->referenceCount = 1;
		kiba_surface_factory_remove_surface(data->surface);
	}
	g_list_free(surfaces);
}
