/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-title
 * @short_description: Title for #KibaObject's.
 * @see_also:#KibaObject
 * @stability: Unstable
 *
 * Title for #KibaObject's.
 */

/*
 * FIXME: if you switch very fast the hovered object, the cpu is about ~30% higher then without title here.
 *        its related to window resizing, the only fix i can imagine is to use a window big enough for each object title.
 *        or each object must have its own title window,
 *        but thats overhead since two objects shouldnt show the title window at the same time.
 */

#include <math.h>
#include <string.h>
#include <glib.h>

#include "kiba.h"
#include "kiba-title.h"
#include "kiba-screen.h"
#include "kiba-cairo-utils.h"
#include "kiba-debug.h"
#include "kiba-macros.h"

#define BACKGROUND_STEPSIZE  0.02f
#define FADE_FPS 100
#define TITLE_OBJECT_DISTANCE 9
#define MAX_FPS 100

#define KIBA_TITLE_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_TITLE, KibaTitlePrivate))

/**
 * Title state.
 **/
typedef enum {
	TITLE_HIDING = 0,
	TITLE_HIDDEN,
	TITLE_SHOWING,
	TITLE_SHOWN
}KibaTitleState;

typedef struct {
	KibaTitle *title;  /**< Pointer to the KibaTitle instance */
	GtkWidget *widget;  /**< Pointer to the KibaObject instance */
	
	CairoSurfaceData surface_data; /**< The title surface */
	PangoLayout *pango_layout; /**< The pango layout of current text */
	char *text;  /**< The current text */
}KibaTitleObject;

struct _KibaTitle {
	GtkWindow parent;
	
	GHashTable *objects;
	KibaTitleObject *active_object;
	
	gint flags; /**< Title flags */
	KibaTitleState state; /**< The title state */
	KibaTitleStyle style; /**< the background style */
	char *font; /**< The title font */
	gint hide_delay; /**< The delay to hide the title */
	
	float fade_alpha;
	gint scale_id;
	
	gint redraw_id;
	gboolean redraw_flag;
	
	GdkPixmap *pixmap; /**< offscreen pixmap */
};

struct _KibaTitleClass {
	GtkWindowClass parent_class;
};

G_DEFINE_TYPE (KibaTitle, kiba_title, GTK_TYPE_WINDOW)

/*** private methods ***/

static void __resize (KibaTitle *title) {
	GtkAllocation allocation = (GtkAllocation) {0,0,0,0};
	gtk_widget_size_allocate (GTK_WIDGET(title), &allocation);
}

static cairo_t* __cairo_create (KibaTitle *title) {
	GtkWidget *widget;
	cairo_t *cr;
	
	if (!GTK_WIDGET_REALIZED(title))
		return NULL;
	
	widget = GTK_WIDGET(title);
	
	cr = gdk_cairo_create (title->pixmap);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
	
	return gdk_cairo_create (widget->window);
}

static gint __redraw (gpointer data) {
	KibaTitle *title = data;
	KibaTitleObject *object = title->active_object;
	GtkWidget *widget = GTK_WIDGET(title);
	KibaDock *dock;
	const KibaLayoutData *d;
	float fade_factor;
	cairo_t *cr;
	gint i;
	
	if (!object || !PANGO_IS_LAYOUT(object->pango_layout))
		return;
	
	dock = KIBA_DOCK(object->widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	fade_factor = title->fade_alpha < 0.8f ? title->fade_alpha + 0.2f : 1.0f;
		if (GDK_IS_PIXMAP(title->pixmap))
			g_object_unref(title->pixmap);
		title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
	
	cr = __cairo_create(title);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return;
	
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	
	if (GET_FLAG(title->flags, KIBA_TITLE_SCALING)) {
		cairo_translate (cr,
			-((fade_factor-1.0f) * widget->allocation.width) / 2.0f,
			-((fade_factor-1.0f) * widget->allocation.height) / 2.0f);
		cairo_scale (cr, fade_factor, fade_factor);
	}
	
	cairo_save (cr);
	/* rotate the title to be vertical */
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		cairo_rotate (cr, (d->halignment + 1)*M_PI/2.0f);
		if (d->halignment == 2)
			cairo_translate (cr, -widget->allocation.height, 0);
		else
			cairo_translate (cr, 0, -widget->allocation.width);
	}
	
	cairo_translate (cr,
		title->style.border_width + 5,
		title->style.border_width + 1);
	
	/* layout the shadow */
	cairo_set_source_rgba (cr,
			title->fade_alpha*title->style.shadow_rgba.r,
			title->fade_alpha*title->style.shadow_rgba.g,
			title->fade_alpha*title->style.shadow_rgba.b,
			title->fade_alpha*title->style.shadow_rgba.a);
	for (i = 0; i < 4; i++) {
		cairo_save (cr);
			cairo_translate (cr, (1 - (i&2)), (1 - 2*(i&1)));
			pango_cairo_show_layout (cr, object->pango_layout);
		cairo_restore (cr);
	}
	
	/* layout the text */
	cairo_set_source_rgba (cr, title->fade_alpha*title->style.rgba.r,
		title->fade_alpha*title->style.rgba.g,
		title->fade_alpha*title->style.rgba.b,
		title->fade_alpha*title->style.rgba.a);
	cairo_save (cr);
		pango_cairo_show_layout (cr, object->pango_layout);
	cairo_restore (cr);
	cairo_restore (cr);
	
	/* draw highlight */
	if (GET_FLAG(title->flags, KIBA_TITLE_HIGHLIGHTING)) {
		cairo_save (cr);
			cairo_pattern_t *pat;
			cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
			
			static float title_highlight_offset = -1.0f;
			title_highlight_offset += 0.01f;
			if (title_highlight_offset > 1.0f)
				title_highlight_offset = -1.0f;
			
			pat = cairo_pattern_create_linear (widget->allocation.width * title_highlight_offset, 0.0,
					widget->allocation.width * (title_highlight_offset + 0.7), 0.0);
			cairo_pattern_add_color_stop_rgba (pat, 0.9, 1.0, 1.0, 1.0, 0.0);
			cairo_pattern_add_color_stop_rgba (pat, 0.3,
				title->fade_alpha*title->style.highlightColor.r,
				title->fade_alpha*title->style.highlightColor.g,
				title->fade_alpha*title->style.highlightColor.b,
				title->fade_alpha*title->style.highlightColor.a);
			cairo_pattern_add_color_stop_rgba (pat, 0.0, 1.0, 1.0, 1.0, 0.0);
			cairo_set_source (cr, pat);
			cairo_paint(cr);
			cairo_pattern_destroy (pat);
			
			/* queue draw as long as the widget is visible */
			if (title->fade_alpha > 0.0f)
				gtk_widget_queue_draw(GTK_WIDGET(title));
		cairo_restore (cr);
	}
	
	/* draw background */
	cairo_save (cr);
		cairo_set_operator (cr, CAIRO_OPERATOR_DEST_OVER);
		cairo_scale (cr,
			(float) widget->allocation.width / object->surface_data.width,
			(float) widget->allocation.height / object->surface_data.height);
		cairo_set_source_surface (cr, object->surface_data.surface, 0.0, 0.0);
		cairo_paint_with_alpha (cr, title->fade_alpha);
	cairo_restore (cr);
	
	cairo_destroy(cr);
	
	GdkGC *gc;
	gc = gdk_gc_new (title->pixmap);
	gdk_draw_drawable (widget->window,
			   gc,
			   title->pixmap,
			   0, 0, 0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	gdk_gc_unref (gc);
	
	title->redraw_id = 0;
	
	return FALSE;
}

static void __draw_object_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaTitleObject *object = data;
	cairo_pattern_t *pattern = NULL;
	
	cairo_paint_shape (cr,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_width/2,
		object->title->style.border_width/2,
		object->surface_data.width - object->title->style.border_width/2,
		object->surface_data.height - object->title->style.border_width/2,
		5);
	
	if (object->title->style.background.style == KIBA_GRADIENT)
		pattern = cairo_get_gradient_pattern (object->title->style.border_width/2,
				object->title->style.border_width/2,
				object->surface_data.width - object->title->style.border_width/2,
				object->surface_data.height - object->title->style.border_width/2,
				object->title->style.gradient, object->title->style.numColors);
	else
		pattern = cairo_get_png_pattern (object->title->style.background.file, 1.0);
	if (pattern == NULL)
		return;
	
	cairo_set_source (cr, pattern);
	cairo_fill_preserve (cr);
	cairo_pattern_destroy (pattern);
	
	cairo_set_source_rgba (cr,
		object->title->style.border_rgba.r,
		object->title->style.border_rgba.g,
		object->title->style.border_rgba.b,
		object->title->style.border_rgba.a);
	cairo_set_line_width (cr, object->title->style.border_width);
	cairo_stroke (cr);
}

static void __remove_objects (gpointer key,
                              gpointer value,
                              gpointer user_data) {
	KibaTitle *title = user_data;
	KibaObject *object = key;
	kiba_title_remove_object(title, key);
}

static void __update_object_pango_layout (gpointer key,
                                          gpointer value,
                                          gpointer user_data) {
	KibaTitle *title = user_data;
	KibaTitleObject *object = value;
	cairo_t *cr = NULL;
	cr = __cairo_create (title);
	cairo_update_pango_layout (object->text,
		title->font,
		cr,
		&object->pango_layout);
	cairo_destroy (cr);
}

static void __update_object_surface (gpointer key,
                                     gpointer value,
                                     gpointer user_data) {
	KibaTitle *title = user_data;
	KibaTitleObject *object = value;
	KibaDock *dock;
	cairo_t *cr;
	const KibaLayoutData *d;
	
	dock = KIBA_DOCK(object->widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	struct {
		gint  *w,  *h;
	} magic[2] = {
		{ &object->surface_data.width,  &object->surface_data.height },
		{ &object->surface_data.height, &object->surface_data.width }
	};
	
	if (!PANGO_IS_LAYOUT(object->pango_layout)) {
		*magic[d->orientation].w = 100;
		*magic[d->orientation].h = 40;
	}
	else {
		PangoRectangle  log;
		
		pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
		
		*magic[d->orientation].w = log.width + 10 + 2*object->title->style.border_width;
		*magic[d->orientation].h = log.height + 2 + 2*object->title->style.border_width;
	}
	
	cr = __cairo_create (object->title);
	cairo_update_surface (cr, &object->surface_data, __draw_object_surface, object);
	cairo_destroy (cr);
}

static gint __fade_timeout (gpointer data) {
	KibaTitle *title = data;
	
	title->redraw_flag = TRUE;
	gtk_widget_queue_draw (GTK_WIDGET(title));
	
	if (title->state == TITLE_HIDING) {
		title->fade_alpha -= 1.0f / title->hide_delay;
		
		if (title->fade_alpha <= 0.0f) {
			title->state = TITLE_HIDDEN;
// 			title->active_object = NULL;
			title->scale_id = 0;
			gtk_widget_hide(GTK_WIDGET(title));
			return FALSE;
		}
	}
	else {
		title->scale_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static void __object_size_allocate_cb (GtkWidget     *widget,
					      GtkAllocation *allocation,
					      gpointer       user_data) {
	KibaTitleObject *object = user_data;
	
	if (!kiba_title_object_is_shown(object->title, KIBA_OBJECT(object->widget)))
		return;
	
	object->title->redraw_flag = TRUE;
	gtk_widget_queue_resize (GTK_WIDGET(object->title));
	__resize(object->title);
}

static void __object_destroy_cb (GtkWidget *widget,
                                 gpointer data) {
	KibaObject *object = KIBA_OBJECT(widget);
	KibaTitle *title = data;
	kiba_title_remove_object(title, object);
}

static void __value_destroy_func (gpointer value) {
	KibaTitleObject *title_object = value;
	
	if (KIBA_IS_OBJECT(title_object->widget)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(title_object->widget),
			  G_CALLBACK(__object_size_allocate_cb), title_object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(title_object->widget),
			  G_CALLBACK(__object_destroy_cb), title_object->title);
	}
	
	if (title_object->surface_data.surface)
		cairo_surface_destroy (title_object->surface_data.surface);
	if (title_object->pango_layout)
		g_object_unref (title_object->pango_layout);
	if (title_object->text)
		g_free (title_object->text);
	
	g_free (title_object);
}

/*** Public methods ***/

/**
 * kiba_title_object_add:
 *
 * Adds title object.
 */
void kiba_title_add_object (KibaTitle *title,
                            KibaObject *object) {
	KibaTitleObject *title_object;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	title_object = g_hash_table_lookup (title->objects, object);
	g_return_if_fail(title_object == NULL);
	
	title_object = g_new0 (KibaTitleObject, 1);
	title_object->widget = GTK_WIDGET(object);
	title_object->pango_layout = NULL;
	title_object->text = NULL;
	title_object->surface_data.surface = NULL;
	title_object->surface_data.width = 40;
	title_object->surface_data.height = 40;
	title_object->title = title;
	
	g_hash_table_insert (title->objects, object, title_object);
	
	g_signal_connect_after (G_OBJECT(object), "size-allocate",
		G_CALLBACK(__object_size_allocate_cb), title_object);
	g_signal_connect_after (G_OBJECT(object), "destroy",
		G_CALLBACK(__object_destroy_cb), title);
}

/**
 * kiba_title_remove_object:
 *
 * Removes title object.
 */
void kiba_title_remove_object (KibaTitle *title,
                               KibaObject *object) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	g_hash_table_remove(title->objects, object);
}

/**
 * kiba_title_show_object:
 *
 * Shows title object.
 */
void kiba_title_show_object (KibaTitle *title,
                             KibaObject *object) {
	KibaDock *dock;
	const KibaLayoutData *d;
	gboolean show_widget;
	KibaTitleObject *title_object;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	title_object = g_hash_table_lookup (title->objects, object);
	g_return_if_fail (title_object != NULL);
	
	if (title_object->pango_layout == NULL)
		return;
	
	dock = KIBA_DOCK(title_object->widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	if (!title->active_object)
		show_widget = TRUE;
	
	title->active_object = title_object;
	title_object->title->state = TITLE_SHOWING;
	
	if (title->scale_id) {
		g_source_remove(title->scale_id);
		title->scale_id = 0;
	}
	title->fade_alpha = 1.0f;
	title_object->title->state = TITLE_SHOWN;
	
	if (show_widget)
		gtk_widget_show(GTK_WIDGET(title_object->title));
	
	title->redraw_flag = TRUE;
	gtk_widget_queue_resize(GTK_WIDGET(title_object->title));
	__resize(title_object->title);
}

/**
 * kiba_title_hide_object:
 *
 * Hides title object.
 */
void kiba_title_hide_object (KibaTitle *title,
                             KibaObject *object) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	g_return_if_fail (KIBA_IS_OBJECT(object));
	KibaTitleObject *title_object;
	
	title_object = g_hash_table_lookup (title->objects, object);
	g_return_if_fail (title_object != NULL);
	
	if (title_object != title->active_object) {
		return;
	}
	
	if (!title->scale_id)
		title->scale_id = g_timeout_add (1000/FADE_FPS, __fade_timeout, title_object->title);
	title_object->title->state = TITLE_HIDING;
}

/**
 * kiba_title_object_is_shown:
 *
 * Returns wether the title is shown.
 */
gboolean kiba_title_object_is_shown (KibaTitle *title,
                                     KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_TITLE(title), FALSE);
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	KibaTitleObject *title_object;
	
	title_object = g_hash_table_lookup (title->objects, object);
	g_return_val_if_fail (title_object != NULL, FALSE);
	
	return (title_object == title->active_object);
}

/**
 * kiba_title_set_object_text:
 *
 * Sets title text.
 */
void kiba_title_set_object_text (KibaTitle *title,
                                 KibaObject *object,
                                 const char *text) {
	KibaDock *dock;
	const KibaLayoutData *d;
	gint diff;
	cairo_t *cr = NULL;
	PangoRectangle log;
	KibaTitleObject *title_object;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	title_object = g_hash_table_lookup (title->objects, object);
	
	g_return_if_fail (title_object != NULL);
	g_return_if_fail (text != NULL);
	
	dock = KIBA_DOCK(title_object->widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	//TODO: maximal text length
	
	if (title_object->text) {
		if (strcmp (title_object->text, text) == 0)
			return;
		diff = strlen (title_object->text) - strlen (text);
		g_free (title_object->text);
	}
	else
		diff = -5;
	title_object->text = g_strdup (text);
	
	cr = __cairo_create (title_object->title);
	cairo_update_pango_layout (title_object->text,
		title_object->title->font,
		cr,
		&title_object->pango_layout);
	cairo_destroy (cr);

	if (kiba_title_object_is_shown (title, object)) {
		title->redraw_flag = TRUE;
		gtk_widget_queue_resize (GTK_WIDGET(title_object->title));
		__resize(title_object->title);
	}
	
	if (diff < -4 || diff > 4)
		__update_object_surface (object, title_object, title);
}

/**
 * kiba_title_get_object_text:
 *
 * Returns title text.
 */
char* kiba_title_get_object_text (KibaTitle *title,
                                  KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_TITLE(title), NULL);
	g_return_val_if_fail (KIBA_IS_OBJECT(object), NULL);
	KibaTitleObject *title_object;
	
	title_object = g_hash_table_lookup (title->objects, object);
	g_return_val_if_fail (title_object != NULL, NULL);
	
	return title_object->text;
}

/**
 * kiba_title_set_font:
 * title: The #KibaTitle instance.
 *
 * Sets title font.
 */
void kiba_title_set_font (KibaTitle *title,
                          char *font) {
	GtkWidget *widget;
	GList *l;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	widget = GTK_WIDGET(title);
	
	title->font = font;
	
	/* update pango layouts for the new font */
	g_hash_table_foreach (title->objects, __update_object_pango_layout, title);
	
	title->redraw_flag = TRUE;
	gtk_widget_queue_resize (GTK_WIDGET(title));
	__resize(title);
}

/**
 * kiba_title_set_flag:
 * title: The #KibaTitle instance.
 *
 * Sets title flag.
 */
void kiba_title_set_flag (KibaTitle *title,
                          KibaTitleFlag flag) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	if (GET_FLAG(title->flags, flag))
		return;
	
	SET_FLAG(title->flags, flag);
	
	switch (flag) {
		default:
			break;
	}
}

/**
 * kiba_title_unset_flag:
 * title: The #KibaTitle instance.
 *
 * Unsets title flag.
 */
void kiba_title_unset_flag (KibaTitle *title,
                            KibaTitleFlag flag) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	if (!GET_FLAG(title->flags, flag))
		return;
	
	DEL_FLAG(title->flags, flag);
	
	switch (flag) {
		default:
			break;
	}
}

gboolean kiba_title_get_flag (KibaTitle *title,
                              KibaTitleFlag flag) {
	g_return_val_if_fail(KIBA_IS_TITLE(title), FALSE);
	return GET_FLAG(title->flags, flag);
}

/**
 * kiba_title_set_style:
 * title: The #KibaTitle instance.
 *
 * Sets the background style.
 */
void kiba_title_set_style (KibaTitle *title,
                           KibaTitleStyle style) {
	GList *l;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	title->style = style;
	
	g_hash_table_foreach (title->objects, __update_object_surface, title);
	
	if (kiba_title_is_shown (title)) {
		gtk_widget_queue_draw(GTK_WIDGET(title));
		title->redraw_flag = TRUE;
	}
}

/**
 * kiba_title_is_shown:
 * title: The #KibaTitle instance.
 *
 * Returns wether the title is shown.
 */
gboolean kiba_title_is_shown (KibaTitle *title) {
	g_return_val_if_fail (KIBA_IS_TITLE(title), FALSE);
	
	return title->active_object != NULL;
}

KibaTitleStyle* kiba_title_get_style (KibaTitle *title) {
	g_return_val_if_fail(KIBA_IS_TITLE(title), NULL);
	return &title->style;
}

void kiba_title_set_hide_delay (KibaTitle *title,
                                gint hide_delay) {
	g_return_if_fail(KIBA_IS_TITLE(title));
	title->hide_delay = hide_delay;
}

/*** Widget Class Implementations ***/

/*
 * gtk_widget_expose implementation.
 */
static gboolean kiba_title_expose_event (GtkWidget *widget,
                                         GdkEventExpose *event) {
	KibaTitle *title = KIBA_TITLE(widget);
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	if (!title->redraw_id && title->redraw_flag) {
		title->redraw_flag = FALSE;
		title->redraw_id = g_timeout_add (1000/MAX_FPS, __redraw, widget);
	}
	
	return FALSE;
}

/*
 * gtk_widget_screen_changed implementation.
 */
static void kiba_title_screen_changed (GtkWidget *widget,
                                       GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
// 	title->redraw_flag = TRUE;
}

/*
 * gtk_widget_size_allocate implementation.
 */
static void kiba_title_size_allocate (GtkWidget *widget,
                                      GtkAllocation *allocation)  {
	KibaTitle *title = KIBA_TITLE(widget);
	KibaTitleObject *object = title->active_object;
	KibaDock *dock;
	const KibaLayoutData *d;
	gboolean size_changed = FALSE;
	
	if (!object)
		return;
	
	dock = KIBA_DOCK(object->widget->parent);
	d = &kiba_layout_data[dock->d.position];
	
	if (object && PANGO_IS_LAYOUT(object->pango_layout)) {
		GtkWidget *dock = object->widget->parent;
		PangoRectangle log;
		const KibaLayoutData *d;
		
		pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
		
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			allocation->height = log.width + 10 + 2*title->style.border_width;
			allocation->width = log.height + 2 + 2*title->style.border_width;
		}
		else {
			allocation->width = log.width + 10 + 2*title->style.border_width;
			allocation->height = log.height + 2 + 2*title->style.border_width;
	 	}
	} else {
		allocation->width = 0;
		allocation->height = 0;
	}
	
	allocation->x = (int) KIBA_OBJECT(object->widget)->alloc.x +
		GTK_WIDGET(dock)->parent->allocation.x;
	allocation->y = (int) KIBA_OBJECT(object->widget)->alloc.y +
		GTK_WIDGET(dock)->parent->allocation.y;
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation->x += (d->halignment == 0 ? object->widget->allocation.width :
			-allocation->width) - (d->halignment - 1) * TITLE_OBJECT_DISTANCE;
		allocation->y += -(allocation->height - object->widget->allocation.height)/2;
	}
	else {
		allocation->x += -(allocation->width - object->widget->allocation.width)/2;
		allocation->y += (d->valignment == 0 ? object->widget->allocation.height :
			-allocation->height) - (d->valignment - 1) * TITLE_OBJECT_DISTANCE;
	}
	
	allocation->x = KIBA_SCREEN_DEFAULT->geometry.x + MAX(0, allocation->x);
	allocation->x = KIBA_SCREEN_DEFAULT->geometry.x + MIN(KIBA_SCREEN_DEFAULT->geometry.width - allocation->width, allocation->x);
	allocation->y = KIBA_SCREEN_DEFAULT->geometry.y + MAX(0, allocation->y);
	allocation->y = KIBA_SCREEN_DEFAULT->geometry.y + MIN(KIBA_SCREEN_DEFAULT->geometry.height - allocation->height, allocation->y);
	
	if (allocation->width != widget->allocation.width || allocation->height != widget->allocation.height)
		size_changed = TRUE;
	
	widget->allocation = *allocation;
	
	if (size_changed) {
		gdk_window_move_resize(widget->window,
			allocation->x, allocation->y,
			allocation->width, allocation->height);
		if (GDK_IS_PIXMAP(title->pixmap))
			g_object_unref(title->pixmap);
		title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
		title->redraw_flag = TRUE;
	} else {
		gdk_window_move(widget->window,
			allocation->x, allocation->y);
	}
	
	(* GTK_WIDGET_CLASS (kiba_title_parent_class)->size_allocate) (widget, allocation);
}

/*
 * gtk_widget_unrealize implementation.
 */
static void kiba_title_unrealize (GtkWidget *widget) {
	KibaTitle *title = KIBA_TITLE(widget);
	
	if (GDK_IS_PIXMAP(title->pixmap))
		g_object_unref(title->pixmap);
	
	g_hash_table_remove_all (title->objects);
	g_hash_table_unref(title->objects);
	
	(* GTK_WIDGET_CLASS (kiba_title_parent_class)->unrealize) (widget);
}

/*
 * gtk_widget_realize implementation.
 */
static void kiba_title_realize (GtkWidget *widget) {
	KibaTitle *title = KIBA_TITLE(widget);
	GtkWindow *window = GTK_WINDOW (widget);
	GdkWindowAttr attributes;
	gint attributes_mask;
	
	/* make sure the size is properly allocated */
	if (widget->allocation.x == -1 &&
	    widget->allocation.y == -1 &&
	    widget->allocation.width == 1 &&
	    widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 0, 0 };
		
		gtk_widget_size_request (widget, &requisition);
		allocation.width = requisition.width;
		allocation.height = requisition.height;
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	attributes.window_type = GDK_WINDOW_TOPLEVEL;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.title = window->title;
	attributes.wmclass_name = window->wmclass_name;
	attributes.wmclass_class = window->wmclass_class;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.visual = gtk_widget_get_visual(widget);
	attributes.colormap = gtk_widget_get_colormap(widget);
	attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	if (attributes.visual != NULL && attributes.colormap != NULL)
		attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;
	attributes_mask |= (window->title ? GDK_WA_TITLE : 0);
	attributes_mask |= (window->wmclass_name ? GDK_WA_WMCLASS : 0);
	
	widget->window = gdk_window_new (gdk_get_default_root_window (),
				&attributes, attributes_mask);
	/* catch the window events on the widget */
	gdk_window_set_user_data (widget->window, window);
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	gdk_window_set_type_hint (widget->window, GDK_WINDOW_TYPE_HINT_DOCK);
	gdk_window_set_icon_name (widget->window, "kiba-dock");
	gdk_window_stick (widget->window);
	gdk_window_set_keep_above (widget->window, TRUE);
	
	title->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
	
	title->objects = g_hash_table_new_full (
		NULL, /**<< using default hash function **/
		g_direct_equal, /**<< comparing pointers **/
		NULL,
		__value_destroy_func);
}

/*** Class init ***/

static void kiba_title_class_init (KibaTitleClass *class) {
	GtkWidgetClass *widget_class;
	
	widget_class = GTK_WIDGET_CLASS (class);
	
	widget_class->realize = kiba_title_realize;
	widget_class->unrealize = kiba_title_unrealize;
	widget_class->size_allocate = kiba_title_size_allocate;
	widget_class->expose_event = kiba_title_expose_event;
	widget_class->screen_changed = kiba_title_screen_changed;
}

static void kiba_title_init (KibaTitle *title) {
	KibaScreen *screen;
	
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (title),
			GTK_NO_WINDOW |
			GTK_DOUBLE_BUFFERED);
	
	screen = kiba_screen_get();
	
	gtk_widget_add_events (GTK_WIDGET (title), GDK_EXPOSURE_MASK);
	gtk_widget_set_app_paintable (GTK_WIDGET (title), TRUE);
	
	gtk_window_set_role (GTK_WINDOW(title), "kiba-title");
	
	SET_FLAG(title->flags, KIBA_TITLE_DRAWABLE);
	
	KibaRGBA *gradient = g_new0(KibaRGBA, 1);
	gradient[0] = (KibaRGBA) {1.0,1.0,1.0,1.0};
	title->style = (KibaTitleStyle) {
		KIBA_GRADIENT, KIBA_ROUNDED,
		2.0f, 45.0f, 5.0f,
		{ 1.0f, 1.0f, 1.0f, 0.8f },
		{ 0.0f, 0.0f, 0.0f, 0.8f },
		{ 0.8f, 0.8f, 0.8f, 0.8f }, { 0.0f, 0.0f, 0.0f, 0.8f },
		gradient, 1
	};
	title->font = "Arial bold 14";
	title->hide_delay = 50;
	title->state = TITLE_HIDDEN;
	
	title->objects = NULL;
	
	title->active_object = NULL;
	title->fade_alpha = 0.0f;
	title->scale_id = 0;
	title->redraw_id = 0;
	title->redraw_flag = TRUE;
	
	kiba_title_screen_changed(GTK_WIDGET(title), screen->gdk_screen);
}

/**
 * kiba_title_new:
 *
 * Creates and returns a new KibaTitle.
 *
 * Returns: The KibaTitle object created.
 */
KibaTitle* kiba_title_new (void) {
	return g_object_new (KIBA_TYPE_TITLE, NULL);
}
