/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *
*/

/**
 * SECTION:kiba-triangle
 * @short_description: ...
 * @see_also:#KibaObject
 * @stability: Unstable
 *
 * ...
 */

#include <math.h>

#include "kiba-triangle.h"
#include "kiba.h"
#include "kiba-win.h"
#include "kiba-utils.h"
#include "kiba-macros.h"
#include "kiba-debug.h"
#include "kiba-svg-macro.h"

#define KIBA_TRIANGLE_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_TRIANGLE, KibaTrianglePrivate))

G_DEFINE_TYPE (KibaTriangle, kiba_triangle, G_TYPE_OBJECT)

typedef struct {
	KibaTriangle *triangle;
	GtkWidget *widget;
	GdkRectangle area;
	float alpha;
	gint fade_out_id;
	gint fade_in_id;
}KibaTriangleObject;

typedef struct {
	GList     *triangle_objects; /* list of all KibaTriangleObject's */
}KibaTrianglePrivate;

static GdkRectangle kiba_triangle_get_area (KibaTriangleObject *object) {
	KibaTriangle *triangle = object->triangle;
	GdkRectangle area;
	KibaDock *dock;
	const KibaLayoutData *d;
	gint fX, fY;
	
	dock = KIBA_DOCK(KIBA_WIN_CHILD(triangle->win));
	d = &kiba_layout_data[dock->d.position];
	
	fX = object->widget->allocation.x;
	fY = object->widget->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		area.x = fX;
		area.x += (object->widget->allocation.width - TRIANGLE_ICON_SIZE)/2;
		area.y = GTK_WIDGET(dock)->allocation.y + (d->valignment/2)*
			(GTK_WIDGET(dock)->allocation.height - TRIANGLE_ICON_SIZE);
	}
	else {
		area.y = fY;
		area.y += (object->widget->allocation.height - TRIANGLE_ICON_SIZE)/2;
		area.x = GTK_WIDGET(dock)->allocation.x + (d->halignment/2)*
			(GTK_WIDGET(dock)->allocation.width - TRIANGLE_ICON_SIZE);
	}
	
	area.width = TRIANGLE_ICON_SIZE;
	area.height = TRIANGLE_ICON_SIZE;
	
	return area;
}

static void kiba_triangle_queue_draw_all (KibaTriangle *triangle) {
	KibaTrianglePrivate *priv;
	KibaTriangleObject *object;
	GList *l;
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	
	for (l = priv->triangle_objects; l != NULL; l = l->next) {
		object = l->data;
		
		if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
			kiba_win_invalidate_area (object->triangle->win, object->area);
	}
}

/**
Rendering
**/

static void kiba_triangle_draw_surface (cairo_t          *cr,
					cairo_surface_t  *surface,
					gint              width,
					gint              height,
					void             *data) {
	char *icon = data;
	GdkPixbuf *pixbuf;
	
#ifdef USE_SVG
	gint size = TRIANGLE_ICON_SIZE;
	RsvgHandle *handle = NULL;
	
	if (kiba_load_svg (icon, &size, &handle)) {
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
		return;
	}
#endif
	
	pixbuf = NULL;
	if (kiba_load_icon (icon, TRIANGLE_ICON_SIZE, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0.0, 0.0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}
	
	cairo_set_line_width (cr, 2.0);
	cairo_set_line_join  (cr, CAIRO_LINE_JOIN_ROUND);
	
	cairo_move_to (cr, 0.5*width, 0.1*height);
	cairo_line_to (cr, 0.9*width, 0.7*height);
	cairo_line_to (cr, 0.1*width, 0.7*height);
	cairo_line_to (cr, 0.5*width, 0.1*height);
	cairo_close_path (cr);
	
	cairo_set_source_rgba (cr, 1, 1, 1, 0.8);
	cairo_fill_preserve (cr);
	cairo_set_source_rgba (cr, 0.8, 0.8, 0.8, 0.6);
	cairo_stroke (cr);
}

static void kiba_triangle_update_surface (KibaTriangle *triangle,
					  const char   *icon) {
	KibaWin *default_win;
	cairo_t *cr;
	char *file;
	
	file = g_strdup (icon);
	default_win = kiba_get_win(kiba_get());
	cr = kiba_win_cairo_create (default_win);
	cairo_update_surface_at_size (cr, &triangle->surface,
		TRIANGLE_ICON_SIZE, TRIANGLE_ICON_SIZE,
		kiba_triangle_draw_surface, file);
	cairo_destroy (cr);
	g_free (file);
	
/* invert dock background
	cairo_t *surface_cr = NULL;
	KibaDock *dock = KIBA_DOCK(KIBA_WIN_CHILD(default_win));
	surface_cr = cairo_create(triangle->surface_data.surface);
	cairo_set_operator (surface_cr, CAIRO_OPERATOR_ATOP);
	cairo_set_source_rgba (surface_cr, 
		1.0-dock->style.bg_rgba_2.r,
		1.0-dock->style.bg_rgba_2.g,
		1.0-dock->style.bg_rgba_2.b,
		0.75);
	cairo_paint (surface_cr);
	cairo_destroy (surface_cr); */
}

/**
 * kiba_triangle_draw:
 *
 * Draw the triangle Surface-Buffer to a Drawing context.
 */
void kiba_triangle_draw (KibaTriangle         *triangle,
			 const KibaLayoutData *d,
			 cairo_t              *cr) {
	KibaTrianglePrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	
	if (!GET_FLAG(triangle->flags, KIBA_TRIANGLE_DRAWABLE))
		return;
	
	for (l = priv->triangle_objects; l != NULL; l = l->next) {
		KibaTriangleObject *object = l->data;
		float alpha;
		
		if (!KIBA_IS_DRAWABLE_OBJECT(object->widget))
			continue;
		
		alpha = KIBA_DRAWABLE_OBJECT(object->widget)->scroll_alpha *
			object->alpha * kiba_object_get_destroy_factor(KIBA_OBJECT(object->widget));
		
		cairo_save (cr);
		cairo_translate (cr, object->area.x, object->area.y);
		
		cairo_translate (cr, TRIANGLE_ICON_SIZE/2, TRIANGLE_ICON_SIZE/2);
		if (d->anchor.y == 0)
			cairo_rotate (cr, M_PI);
		else if (d->anchor.x == 0)
			cairo_rotate (cr, M_PI/2);
		else if (d->anchor.x == 4) 
			cairo_rotate (cr, 3*M_PI/2);
		cairo_translate (cr, -TRIANGLE_ICON_SIZE/2, -TRIANGLE_ICON_SIZE/2);
		
		cairo_set_source_surface (cr, triangle->surface, 0, 0);
		cairo_paint_with_alpha (cr, alpha);
		cairo_restore (cr);
	}
}

/**
Properties
**/

/**
 * kiba_triangle_set_icon:
 *
 * Sets the triangle icon.
 */
void kiba_triangle_set_icon (KibaTriangle *triangle,
			     const char   *icon) {
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	kiba_triangle_update_surface (triangle, icon);
	
	kiba_triangle_queue_draw_all (triangle);
}

/**
 * kiba_triangle_set_flag:
 *
 * Sets triangle flag.
 */
void kiba_triangle_set_flag (KibaTriangle *triangle,
			      gint          flag) {
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	if (GET_FLAG(triangle->flags, flag))
		return;
	
	SET_FLAG(triangle->flags, flag);
	
	switch (flag) {
		case KIBA_TRIANGLE_DRAWABLE:
			kiba_triangle_queue_draw_all(triangle);
			break;
		default:
			DEBUG_ERROR ("invalid KibaTriangle flag '%i'\n", flag);
			break;
	}
}

/**
 * kiba_triangle_unset_flag:
 *
 * Unsets triangle flag.
 */
void kiba_triangle_unset_flag (KibaTriangle *triangle,
			       gint          flag) {
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	if (!GET_FLAG(triangle->flags, flag))
		return;
	
	DEL_FLAG(triangle->flags, flag);
	
	switch (flag) {
		case KIBA_TRIANGLE_DRAWABLE:
			kiba_triangle_queue_draw_all(triangle);
			break;
		default:
			DEBUG_ERROR ("invalid KibaTriangle flag '%i'\n", flag);
			break;
	}
}

/**
Public Functions
**/

static gint fade_in_triangle (gpointer data) {
	KibaTriangleObject *object = data;
	
	object->alpha += 0.05f;
	
	if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
		kiba_win_invalidate_area (object->triangle->win, object->area);
	
	if (object->alpha >= 1.0f) {
		object->alpha = 1.0f;
		object->fade_in_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static gint fade_out_triangle (gpointer data) {
	KibaTriangleObject *object = data;
	KibaTrianglePrivate *priv;
	
	object->alpha -= 0.05f;
	
	if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
		kiba_win_invalidate_area (object->triangle->win, object->area);
	
	if (object->alpha*KIBA_OBJECT(object->widget)->destroy_factor <= 0.0f) {
		priv = KIBA_TRIANGLE_GET_PRIVATE(object->triangle);
		
		object->alpha = 0.0f;
		priv->triangle_objects = g_list_remove (priv->triangle_objects, object);
		gtk_widget_unref(object->widget);
		g_free (object);
		object = NULL;
		return FALSE;
	}
	
	return TRUE;
}

static void triangle_object_size_allocate_cb (GtkWidget     *widget,
					      GtkAllocation *allocation,
					      gpointer       user_data) {
	KibaTriangleObject *object = user_data;
	GdkRectangle old_area;
	GdkRectangle new_area;
	
	old_area = object->area;
	new_area = kiba_triangle_get_area (object);
	
	if (old_area.x == new_area.x && old_area.y == new_area.y)
		return;
	
	if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
		kiba_win_invalidate_area (object->triangle->win, object->area);
	object->area  = new_area;
	if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
		kiba_win_invalidate_area (object->triangle->win, object->area);
}

void kiba_triangle_add_widget (KibaTriangle *triangle,
				    GtkWidget    *widget) {
	KibaTrianglePrivate *priv;
	KibaTriangleObject *object = NULL;
	GList *l;
	
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	
	for (l = priv->triangle_objects; l != NULL; l = l->next) {
		KibaTriangleObject *obj = l->data;
		
		if (obj->widget == widget) {
			object = obj;
			break;
		}
	}
	
	if (object) {
		if (object->fade_out_id) {
			g_source_remove (object->fade_out_id);
			object->fade_out_id = 0;
		}
		if (!object->fade_out_id)
			object->fade_in_id = g_timeout_add (50, fade_in_triangle, object);
		return;
	}
	object = g_new0 (KibaTriangleObject, 1);
	object->triangle = triangle;
	object->widget = gtk_widget_ref(widget);
	object->alpha = 0.0;
	
	priv->triangle_objects = g_list_append (priv->triangle_objects, object);
	
	object->fade_in_id = g_timeout_add (50, fade_in_triangle, object);
	
	triangle_object_size_allocate_cb (widget, &widget->allocation, object);
	g_signal_connect_after (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(triangle_object_size_allocate_cb), object);

}

void kiba_triangle_remove_widget (KibaTriangle *triangle,
				    GtkWidget    *widget) {
	KibaTrianglePrivate *priv;
	KibaTriangleObject *object = NULL;
	GList *l;
	
	g_return_if_fail (KIBA_IS_TRIANGLE(triangle));
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	
	for (l = priv->triangle_objects; l != NULL; l = l->next) {
		KibaTriangleObject *obj = l->data;
		
		if (obj->widget == widget) {
			object = obj;
			break;
		}
	}
	
	if (!object || object->fade_out_id)
		return;
	
	if (object->fade_in_id) {
		g_source_remove (object->fade_in_id);
		object->fade_in_id = 0;
	}
	object->fade_out_id = g_timeout_add (50, fade_out_triangle, object);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		  G_CALLBACK(triangle_object_size_allocate_cb), object);
}

/**
 * kiba_triangle_object_activate:
 *
 * (De)Activates a triangle.
 */
void kiba_triangle_object_activate (KibaTriangle *triangle,
				    GtkWidget    *widget,
				    gboolean      activate) {
}

/**
Class Functions
**/

static void kiba_triangle_class_init (KibaTriangleClass *class) {
	g_type_class_add_private (class, sizeof (KibaTrianglePrivate));
}

static void kiba_triangle_init (KibaTriangle *triangle) {
	KibaTrianglePrivate *priv;
	
	triangle->surface = NULL;
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	priv->triangle_objects = NULL;
}

/**
 * kiba_triangle_new:
 * @widget: The KibaWin instance.
 *
 * Create a KibaTriangle Instance.
 *
 * Returns: The KibaTriangle Instance created.
 */
KibaTriangle* kiba_triangle_new (GtkWidget *widget) {
	KibaTriangle *triangle = g_object_new (KIBA_TYPE_TRIANGLE, NULL);
	
	triangle->win = widget;
	
	return triangle;
}

/**
 * kiba_triangle_destroy:
 * @triangle: The KibaTriangle Instance.
 *
 * Free allocated memory for a KibaTriangle Instance.
 */
void kiba_triangle_destroy (KibaTriangle *triangle) {
	KibaTrianglePrivate *priv;
	KibaTriangleObject *object;
	GList *l;
	
	priv = KIBA_TRIANGLE_GET_PRIVATE(triangle);
	
	for (l = priv->triangle_objects; l != NULL; l = priv->triangle_objects) {
		object = l->data;
		if (object->fade_out_id)
			g_source_remove (object->fade_out_id);
		if (object->fade_in_id)
			g_source_remove (object->fade_in_id);
		priv->triangle_objects = g_list_remove (priv->triangle_objects, l->data);
		gtk_widget_unref(object->widget);
		g_free (object);
	}
	g_list_free (priv->triangle_objects);
	if (triangle->surface)
		cairo_surface_destroy (triangle->surface);
	g_object_unref (triangle);
	triangle = NULL;
}
