/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-win
 * @short_description: Window container for a #KibaDock Instance.
 * @see_also: #KibaDock,#KibaObject
 * @stability: Unstable
 *
 * Window container for a #KibaDock Instance.
 */

#include "../config.h"
#include <unistd.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#if GTK_CHECK_VERSION(2,9,0) == FALSE
  #include <X11/extensions/shape.h>
#endif

#include "kiba.h"
#include "kiba-win.h"
#include "kiba-object.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-x11.h"
#include "kiba-separator.h"
#include "kiba-macros.h"
#include "kiba-cairo-utils.h"
#ifdef USE_GLITZ
  #include "glitz-helper.h"
#endif
#ifdef USE_SDL
  #include "sdl-helper.h"
#endif
#ifndef USE_GLITZ
#ifndef USE_SDL
  #define USE_OFFSCREEN_PIXMAP
#endif
#endif

#include "dock-private.h"

/* TRUE for benchmarking */
#define BENCHMARK_ENABLE FALSE
/* TRUE to see only the rerendered parts of the window */
#define CLIPPING_CHECK FALSE

#define KIBA_WIN_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_WIN, KibaWinPrivate))

#define WIN_DRAG_AREA 10000
#define DOCK_SNAP_AREA 25
#define DOCK_DRAG_ALPHA 0.4
#define MAX_FPS 100

G_DEFINE_TYPE (KibaWin, kiba_win, GTK_TYPE_WINDOW)

typedef struct _KibaWinPrivate KibaWinPrivate;

/**
 * KibaWinDragData:
 * @receive_func: the method to handle the drag data on receive events.
 * @dest_func: the method to check if receive_func is able to handle the drag.
 * @user_data: the user data pointer for the methods.
 *
 * Type to handle a drag and drop receice event.
 */
typedef struct {
	KibaDragReceiveFunc receive_func;
	KibaDragDestFunc dest_func;
	gpointer user_data;
}KibaWinDragData;

struct _KibaWinPrivate {
	/* list of damaged rectangles in queue.
	   these areas will be redrawn on the next expose event */
	GList *invalidate_areas;
	/* id for the framerate limiter timeout */
	gint redraw_id;
	
	gint move_dock_id; /* id for the dock drag timeout */
	gint check_hide_id; /* id for the timeout checking the mouse position for hiding */
	gint hide_id; /* id for the timeout hiding the dock */
	gint raise_id;
	
	GdkRectangle mouse_anchor;  /* saves last mouse position for dragging */
	GList *drag_reveive_data; /* list of custom dnd receive functions (KibaWinDragData) */
	
#ifdef USE_GLITZ
	glitz_drawable_format_t *pGlitzDrawFormat;
	glitz_drawable_t *pGlitzDrawable;
	glitz_format_t *pGlitzFormat;
#endif
#ifdef USE_SDL
	unsigned char *puc_surface;
	gint puc_width;
	gint puc_height;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	GdkPixmap *pixmap;
#endif
};

/*
 * Type used to represent a drag and drop receive method.
 */
typedef struct {
	KibaWinDragData *data;
	GList *actions;
	GdkDragContext *context;
	char *drag_data;
	char *action;
	char *icon;
}KibaDragAction;

/* signal ids */
enum {
	SIG_DOCK_HIDDEN = 0,
	SIG_DOCK_UNHIDDEN,
	SIG_KIBA_WIN_LAST
};

static guint kiba_win_signals[SIG_KIBA_WIN_LAST] = { 0 };

static gint __redraw (gpointer data);

#if BENCHMARK_ENABLE == TRUE
static gint redraw_count = 0;
#endif

/*** Helper Functions ***/

#if BENCHMARK_ENABLE == TRUE
static gint __benchmark_cb (gpointer data) {
	fprintf (stderr, "Benchmark: %i FPS\n", redraw_count);
	redraw_count = 0;
	
	return TRUE;
}
#endif

#if GTK_CHECK_VERSION(2,9,0) == FALSE
static void __do_shape_combine_mask (GdkWindow *window,
				   GdkBitmap *mask,
				   gint       x,
				   gint       y) {
	Pixmap pixmap;
	int ignore;
	int maj;
	int min;
	
	if (!XShapeQueryExtension (GDK_WINDOW_XDISPLAY (window), &ignore, &ignore))
		return;
	
	if (!XShapeQueryVersion (GDK_WINDOW_XDISPLAY (window), &maj, &min))
		return;
	
	if (maj != 1 && min < 1)
		return;
	
	if (mask)
		pixmap = GDK_DRAWABLE_XID (mask);
	else {
		x = 0;
		y = 0;
		pixmap = None;
	}
	
	XShapeCombineMask (GDK_WINDOW_XDISPLAY (window),
			GDK_DRAWABLE_XID (window),
			ShapeInput,
			x, y,
			pixmap,
			ShapeSet);
}
#endif

/*
 * Shapes the input of the window. Input Shaped means that the window is not able to receive
 * any input events in a defined area.
 * Just a rectangular area for the dock and objects is not input shaped.
 */
static void __shape_input (GtkWidget *widget) {
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	GdkBitmap *shape_bitmap;
	GtkWidget *dock;
	cairo_t *cr = NULL;
	gint x, y, width, height;
	const KibaLayoutData *d;
	GList *l;
	
	if (priv->move_dock_id)
		return;
	
	shape_bitmap = (GdkBitmap*)
		gdk_pixmap_new (NULL,
			widget->allocation.width,
			widget->allocation.height, 1);
	if (!shape_bitmap)
		return;
	
	cr = gdk_cairo_create (shape_bitmap);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS) {
		g_object_unref ((gpointer) shape_bitmap);
		return;
	}
	
	/* clear the context for a "no-event" window */
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	
	/* renders a rectangle for the dock */
	dock = KIBA_WIN_CHILD(widget);
	if (dock) {
		gint owidth = 0, oheight = 0;
		
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
		
		cairo_set_source_rgba (cr, 1, 1, 1, 1);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(dock)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			
			owidth = MAX(owidth, child->allocation.width);
			oheight = MAX(oheight, child->allocation.height);
		}
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			x = dock->allocation.x;
			width = dock->allocation.width;
			
			height = oheight + KIBA_DOCK(dock)->d.object_margin;
			y = dock->allocation.y + (d->valignment/2)*(dock->allocation.height - height);
		}
		else {
			y = dock->allocation.y;
			height = dock->allocation.height;
			
			width = owidth + KIBA_DOCK(dock)->d.object_margin;
			x = dock->allocation.x + (d->halignment/2)*(dock->allocation.width - width);
		}
		
		cairo_rectangle (cr, x, y, width, height);
		cairo_fill (cr);
	}
	cairo_destroy (cr);
	
#if GTK_CHECK_VERSION(2,9,0) == FALSE
	__do_shape_combine_mask (widget->window, NULL, 0, 0);
	__do_shape_combine_mask (widget->window, shape_bitmap, 0, 0);
#else
	gdk_window_input_shape_combine_mask (widget->window, NULL, 0, 0);
	gdk_window_input_shape_combine_mask (widget->window, shape_bitmap, 0, 0);
#endif
	
	g_object_unref ((gpointer) shape_bitmap);
}

/*
 * Returns the closest position related to the coordinates x/y.
 */
static int __find_closest_dock_position (GtkWidget *dock,
						int x, int y) {
	int i, dx, dy, min_d, d, position;
	
	position = KIBA_DOCK(dock)->d.position;
	
	min_d = WIN_DRAG_AREA;
	for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++) {
		dx = x - KIBA_SCREEN_DEFAULT->geometry.width * kiba_layout_data[i].anchor.x / 4;
		dy = y - KIBA_SCREEN_DEFAULT->geometry.height * kiba_layout_data[i].anchor.y / 4;
		
		d = dx * dx + dy * dy;
		if (d < min_d) {
			min_d    = d;
			position = i;
		}
	}
	
	return position;
}

static GtkAllocation __get_dock_allocation (GtkWidget *widget,
                                            GtkWidget *dock,
                                            GtkAllocation allocation,
                                            const KibaLayoutData *d,
                                            GList *docks) {
	GList *l;
	GtkAllocation ret;
	GtkAllocation absolutAlloc = (GtkAllocation) {
		allocation.x + widget->allocation.x,
		allocation.y + widget->allocation.y,
		allocation.width,
		allocation.height
	};
	
	ret = allocation;
	
	for (l = docks; l != NULL; l = l->next) {
		GtkWidget *dockWidget = l->data;
		const KibaLayoutData *dockD;
		
		if (dockWidget == dock)
			continue;
		
		dockD = &kiba_layout_data[KIBA_DOCK(dockWidget)->d.position];
		GtkAllocation checkAlloc = (GtkAllocation) {
			dockWidget->allocation.x + dockWidget->parent->allocation.x,
			dockWidget->allocation.y + dockWidget->parent->allocation.y,
			dockWidget->allocation.width,
			dockWidget->allocation.height
		};
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			if (checkAlloc.y < absolutAlloc.y) {
				if (checkAlloc.y + checkAlloc.height < absolutAlloc.y)
					continue;
			}
			else if (checkAlloc.y > absolutAlloc.y + KIBA_DOCK(dock)->d.height)
					continue;
			if (dockD->orientation == GTK_ORIENTATION_VERTICAL) {
				if (absolutAlloc.x < checkAlloc.x + KIBA_DOCK(dockWidget)->d.height && allocation.x >= checkAlloc.x)
					ret.x = dockWidget->allocation.x + KIBA_DOCK(dockWidget)->d.height;
				else if (absolutAlloc.x + absolutAlloc.width > checkAlloc.x + checkAlloc.width - KIBA_DOCK(dockWidget)->d.height &&
						absolutAlloc.x + absolutAlloc.width <= checkAlloc.x + checkAlloc.width)
					ret.x = widget->allocation.x + widget->allocation.width - KIBA_DOCK(dockWidget)->d.height - allocation.width;
			}
			else {
				if (checkAlloc.x < absolutAlloc.x) {
					if (checkAlloc.x + checkAlloc.width >= absolutAlloc.x)
						ret.x = dockWidget->allocation.x + dockWidget->allocation.width + 1;
				}
				else if (checkAlloc.x <= absolutAlloc.x + absolutAlloc.width) {
						ret.x = dockWidget->allocation.x - allocation.width - 1;
				}
			}
		} else {
			if (checkAlloc.x < absolutAlloc.x) {
				if (checkAlloc.x + checkAlloc.width < absolutAlloc.x)
					continue;
			}
			else if (checkAlloc.x > absolutAlloc.x + KIBA_DOCK(dock)->d.height)
					continue;
			if (dockD->orientation == GTK_ORIENTATION_HORIZONTAL) {
				if (absolutAlloc.y < checkAlloc.y + KIBA_DOCK(dockWidget)->d.height && allocation.y >= checkAlloc.y)
					ret.y = dockWidget->allocation.y + KIBA_DOCK(dockWidget)->d.height;
				else if (absolutAlloc.y + absolutAlloc.height > checkAlloc.y + checkAlloc.height - KIBA_DOCK(dockWidget)->d.height &&
						absolutAlloc.y + absolutAlloc.height <= checkAlloc.y + checkAlloc.height)
					ret.y = widget->allocation.y + widget->allocation.height - KIBA_DOCK(dockWidget)->d.height - allocation.height;
			}
			else {
				if (checkAlloc.y < absolutAlloc.y) {
					if (checkAlloc.y + checkAlloc.height >= absolutAlloc.y)
						ret.y = dockWidget->allocation.y + dockWidget->allocation.height + 1;
				}
				else if (checkAlloc.y <= absolutAlloc.y + absolutAlloc.height) {
						ret.y = dockWidget->allocation.y - allocation.height - 1;
				}
			}
		}
	}
	
	return ret;
}

/*
 * Default method to allocate the size of the dock.
 * Note: this method is part of a custom size allocation process.
 * dont call this function directly, call instead: (KibaWin*)win->allocate_children(KibaWin*, NULL)
 */
static void __resize_children (KibaWin *win,
                               gpointer data) {
	Kiba *kiba = kiba_get();
	GtkWidget *widget = GTK_WIDGET(win);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	GtkRequisition req;
	GtkAllocation allocation;
	GList *docks, *l;
	gint edge_offset;
	const KibaLayoutData *d;
	
	allocation = dock->allocation;
	
	/* get the size request from the dock */
	gtk_widget_size_request (dock, &req);
	if (req.width || req.height) {
		/* non-empty window */
		allocation.width = req.width;
		allocation.height = req.height;
	}
	
	/* snap dock at center position */
	edge_offset = (KIBA_DOCK(dock)->d.offset < DOCK_SNAP_AREA &&
		KIBA_DOCK(dock)->d.offset > -DOCK_SNAP_AREA && priv->move_dock_id ?
		0 : KIBA_DOCK(dock)->d.offset);
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDDEN)
			allocation.x = d->halignment *
				(GTK_WIDGET (widget)->allocation.width - allocation.width) / 2;
		allocation.y = edge_offset + d->valignment *
			(GTK_WIDGET (widget)->allocation.height - allocation.height) / 2;
	}
	else {
		allocation.x = edge_offset + d->halignment *
			(GTK_WIDGET (widget)->allocation.width - allocation.width) / 2;
		if (KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDDEN)
			allocation.y = d->valignment *
				(GTK_WIDGET (widget)->allocation.height - allocation.height) / 2;
	}
	
	if (KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDDEN) {
		if (allocation.x + allocation.width > GTK_WIDGET (widget)->allocation.width)
			allocation.x = GTK_WIDGET (widget)->allocation.width - allocation.width;
		else if (allocation.x <= 0)
			allocation.x = 1;
		if (allocation.y + allocation.height > GTK_WIDGET (widget)->allocation.height)
			allocation.y = GTK_WIDGET (widget)->allocation.height - allocation.height;
		else if (allocation.y <= 0)
			allocation.y = 1;
	}
	
	GtkAllocation fixedAllocation;
	
	docks = kiba_get_docks(kiba);
	
	/* first try to move the dock to a safe place */
	fixedAllocation = __get_dock_allocation (widget, dock, allocation, d, docks);
	allocation.x = fixedAllocation.x;
	allocation.y = fixedAllocation.y;
	
	//check for the dock being offscreen
	if (allocation.x + allocation.width > GTK_WIDGET (widget)->allocation.width) {
		allocation.width = GTK_WIDGET (widget)->allocation.width - allocation.x;
	}
	else if (allocation.x < 0) {
		allocation.width += allocation.x;
		allocation.x = 0;
	}
	if (allocation.y + allocation.height > GTK_WIDGET (widget)->allocation.height) {
		allocation.height = GTK_WIDGET (widget)->allocation.height - allocation.y;
	}
	else if (allocation.y < 0) {
		allocation.height += allocation.y;
		allocation.y = 0;
	}
	
	/* next resize the dock to not overlap */
	fixedAllocation = __get_dock_allocation (widget, dock, allocation, d, docks);
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		gint dx;
		dx = fixedAllocation.x - allocation.x;
		dx = MAX(dx, -dx);
		allocation.width -= dx;
		if (fixedAllocation.x > allocation.x)
			allocation.x += dx;
	} else {
		gint dy;
		dy = fixedAllocation.y - allocation.y;
		dy = MAX(dy, -dy);
		allocation.height -= dy;
		if (fixedAllocation.y > allocation.y)
			allocation.y += dy;
	}
	
	for (l = docks; l != NULL; l = docks)
		docks = g_list_remove(docks, l->data);
	g_list_free(docks);
	
	allocation.width = MAX(allocation.width, 0);
	allocation.height = MAX(allocation.height, 0);
	gtk_widget_size_allocate (dock, &allocation);
}

/*
 * FIXME: when the dock is pressed, and moved by __move_dock(),
 *        then the dock changes the position immediately for about a few pixel (wich looks uggly)
 */
static gint __move_dock (gpointer data) {
	KibaWin *win = data;
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	KibaPosition position;
	gint x, y;
	const KibaLayoutData *d;
	GList *l;
	
	gdk_window_get_pointer (GTK_WIDGET(win)->window, &x, &y, NULL);
	
	position = __find_closest_dock_position (dock,
			x + GTK_WIDGET(win)->allocation.x,
			y + GTK_WIDGET(win)->allocation.y);
	
	d = &kiba_layout_data[position];
	
	if (position != KIBA_DOCK(dock)->d.position) {
		kiba_dock_set_position (KIBA_DOCK(dock), position);
		gdk_window_get_pointer (GTK_WIDGET(win)->window, &x, &y, NULL);
	}
	else {
		if (d->orientation == GTK_ORIENTATION_VERTICAL)
			KIBA_DOCK(dock)->d.offset += y - priv->mouse_anchor.y;
		else
			KIBA_DOCK(dock)->d.offset += x - priv->mouse_anchor.x;
		win->allocate_children (win, win->allocate_data);
	}
	
	priv->mouse_anchor.x = x;
	priv->mouse_anchor.y = y;
	
	return TRUE;
}

static gint __hide_dock (gpointer data) {
	GtkWidget *win = data;
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GtkAllocation allocation;
	const KibaLayoutData *d;
	gboolean hidden = FALSE;
	GList *l;
	
	if (KIBA_DOCK(dock)->autohide.state != DOCK_HIDING) {
		priv->hide_id = 0;
		return FALSE;
	}
	if (priv->move_dock_id) {
		KIBA_DOCK(dock)->autohide.state = DOCK_UNHIDDEN;
		priv->hide_id = 0;
		return FALSE;
	}
	allocation = dock->allocation;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation.x += (d->halignment - 1)*KIBA_DOCK(dock)->autohide.speed;
		if (allocation.x < ((d->halignment-2)/2) * dock->allocation.width) {
			hidden = TRUE;
			allocation.x = (d->halignment - 1) * dock->allocation.width;
		}
		if (allocation.x > (d->halignment/2) * win->allocation.width) {
			hidden = TRUE;
			allocation.x = (d->halignment/2) * win->allocation.width;
		}
	}
	else {
		allocation.y += (d->valignment - 1)*KIBA_DOCK(dock)->autohide.speed;
		if (allocation.y < ((d->valignment-2)/2) * dock->allocation.height) {
			hidden = TRUE;
			allocation.y = ((d->valignment-2)/2) * allocation.height;
		}
		if (allocation.y > (d->valignment/2) * win->allocation.height) {
			hidden = TRUE;
			allocation.y = (d->valignment/2) * win->allocation.height;
		}
	}
	
	gtk_widget_size_allocate (dock, &allocation);
	
	/* the dock is now hidden, exiting the timeout.. */
	if (hidden) {
		priv->hide_id = 0;
		KIBA_DOCK(dock)->autohide.state = DOCK_HIDDEN;
		kiba_dock_update_strut(dock);
		
		g_signal_emit (win,
			kiba_win_signals[SIG_DOCK_HIDDEN],
			0, dock);
		
		return FALSE;
	}
	
	return TRUE;
}

static gint __unhide_dock (gpointer data) {
	GtkWidget *win = data;
	GtkWidget *dock;
	GtkAllocation allocation;
	KibaWinPrivate *priv;
	const KibaLayoutData *d;
	gboolean unhidden = FALSE;
	GList *l;
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	dock = KIBA_WIN_CHILD(win);
	
	if (priv->move_dock_id) {
		KIBA_DOCK(dock)->autohide.state = DOCK_UNHIDDEN;
		return FALSE;
	}
	
	allocation = dock->allocation;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation.x -= (d->halignment - 1) * KIBA_DOCK(dock)->autohide.speed;
		
		if (d->halignment == 2 && allocation.x < win->allocation.width - allocation.width) {
			unhidden = TRUE;
			allocation.y = win->allocation.width - allocation.width;
		}
		else if (d->halignment == 0 && allocation.x > 0) {
			unhidden = TRUE;
			allocation.x = 0;
		}
	}
	else {
		allocation.y -= (d->valignment - 1) * KIBA_DOCK(dock)->autohide.speed;
		
		if (d->valignment == 2 && allocation.y < win->allocation.height - allocation.height) {
			unhidden = TRUE;
			allocation.y = win->allocation.height - allocation.height;
		}
		else if (d->valignment == 0 && allocation.y > 0) {
			unhidden = TRUE;
			allocation.y = 0;
		}
	}
	
	gtk_widget_size_allocate (dock, &allocation);
	
	if (unhidden)
		g_signal_emit (win,
			kiba_win_signals[SIG_DOCK_UNHIDDEN],
			0, dock);
	
	return !unhidden;
}

static gint __hide_delay (gpointer data) {
	KibaDock *dock;
	
	dock = KIBA_DOCK(KIBA_WIN_CHILD(data));
	
	dock->autohide.state = DOCK_HIDING;
	g_timeout_add (20 , __hide_dock, GTK_WIDGET(data));
	
	return FALSE;
}

static gint __check_hide_dock (gpointer data) {
	GtkWidget *win = data;
	KibaWinPrivate  *priv;
	GtkWidget *dock;
	gint x, y;
	float area;
	
	dock = KIBA_WIN_CHILD(win);
	
	gdk_window_get_pointer (dock->window, &x, &y, NULL);
	x -= dock->allocation.x;
	y -= dock->allocation.y;
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	if (priv->move_dock_id)
		return TRUE;
	
	area = KIBA_DOCK(dock)->autohide.area;
	
	if (x > -area && x < dock->allocation.width + area &&
	    y > -area && y < dock->allocation.height + area) {
		if (KIBA_DOCK(dock)->autohide.state == DOCK_HIDING ||
		    KIBA_DOCK(dock)->autohide.state == DOCK_HIDDEN) {
			KIBA_DOCK(dock)->autohide.state = DOCK_UNHIDING;
			kiba_dock_update_strut(dock);
			
			g_timeout_add (20 , __unhide_dock, win);
		}
	}
	else if (!priv->hide_id && (KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDING ||
		 KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDDEN)) {
		priv->hide_id = g_timeout_add (KIBA_DOCK(dock)->autohide.delay*1000,
					__hide_delay, win);
	}
	
	return TRUE;
}

static gint __raise_window (gpointer data) {
	KibaWin *win = KIBA_WIN(data);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	
	gdk_window_set_keep_below (GTK_WIDGET (win)->window, TRUE);
	/* without that the window is getting borders !?! */
	gdk_window_set_decorations (GTK_WIDGET (win)->window, 0);
	
	priv->raise_id = 0;
	
	return FALSE;
}

/*
 * Clears the window to be transparent.
 */
static void __clear (KibaWin *win) {
	KibaWinPrivate  *priv;
	GtkWidget *widget;
	cairo_t *cr;
	
	widget = GTK_WIDGET(win);
	priv = KIBA_WIN_GET_PRIVATE (widget);
	
	cr = kiba_win_cairo_create (win);
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return;
	
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);
	cairo_destroy (cr);
	
	/* renders the offscreen drawable to the window */
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->pGlitzDrawable,
		priv->pGlitzDrawFormat);
#endif
#ifdef USE_SDL
	kiba_sdl_draw_func (&priv->puc_surface,
		widget->allocation.width,
		widget->allocation.height);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap)) {
		GdkGC *gc;
		
		gc = gdk_gc_new (priv->pixmap);
		if (GDK_IS_GC(gc)) {
			gdk_draw_drawable (widget->window,
				gc,
				priv->pixmap,
				0, 0, 0, 0,
				widget->allocation.width,
				widget->allocation.height);
			gdk_gc_unref (gc);
		}
	}
#endif
}

/*
 * Rerenders invalidate areas of the window.
 */
static gint __redraw (gpointer data) {
	GtkWidget *widget = data;
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaDock *dock = KIBA_DOCK(KIBA_WIN_CHILD(widget));
	cairo_t *cr = NULL;
	GList *l;
	const KibaLayoutData *d;
#ifndef USE_SDL
	GdkGC *gc;
	GdkRegion *clip_region;
#endif
#ifdef USE_GLITZ
	/* it seems to better for the cpu usage to clip only one big rectangle for glitz */
	GdkRectangle clip_box;
#endif
	
	if (dock)
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	else
		d = &kiba_layout_data[KIBA_POSITION_BOTTOM_EDGE];
	g_return_val_if_fail (priv->invalidate_areas != NULL, TRUE);
	
#ifndef USE_SDL
	clip_region = gdk_region_new();
	
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap)) {
		g_object_unref(priv->pixmap);
		priv->pixmap = NULL;
	}
	priv->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
	gc = gdk_gc_new (priv->pixmap);
#endif
#endif
	
	cr = kiba_win_cairo_create (KIBA_WIN(widget));
	if (cairo_status(cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
#if CLIPPING_CHECK == TRUE
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);
#endif
	
	/* clip invalidate areas */
	for (l = priv->invalidate_areas; l != NULL; l = l->next) {
		GdkRectangle *damaged_area = l->data;
		
#ifndef USE_GLITZ
		cairo_rectangle (cr, damaged_area->x, damaged_area->y,
			damaged_area->width, damaged_area->height);
#endif
#ifndef USE_SDL
		gdk_region_union_with_rect (clip_region, damaged_area);
#endif
	}
#ifdef USE_OFFSCREEN_PIXMAP
#if CLIPPING_CHECK == FALSE
	gdk_gc_set_clip_region (gc, clip_region);
#endif
#endif
#ifdef USE_GLITZ
	gdk_region_get_clipbox (clip_region, &clip_box);
	cairo_rectangle (cr, clip_box.x, clip_box.y,
		clip_box.width, clip_box.height);
#endif
#ifndef USE_SDL
	gdk_region_destroy (clip_region);
#endif
	cairo_clip (cr);
	
#if CLIPPING_CHECK == TRUE
	cairo_set_source_rgba (cr, 0, 0, 0, 1);
#else
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
#endif
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	
	/* draw dock&children */
	cairo_save (cr);
		kiba_dock_expose (dock, cr);
	cairo_restore (cr);
	
	if (GET_FLAG(dock->flags, KIBA_DOCK_DRAW_PROGRESS)) {
		cairo_save (cr);
			kiba_progress_draw (KIBA_WIN(widget)->progress, cr);
		cairo_restore (cr);
	}
	
	cairo_destroy (cr);
	
	/* rendering the offscreen drawable to the window */
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->pGlitzDrawable,
		priv->pGlitzDrawFormat);
#endif
#ifdef USE_SDL
	kiba_sdl_draw_func (&priv->puc_surface,
		widget->allocation.width,
		widget->allocation.height);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	gdk_draw_drawable (widget->window,
			   gc,
			   priv->pixmap,
			   0, 0, 0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	gdk_gc_unref (gc);
#endif
	
#if BENCHMARK_ENABLE == TRUE
	redraw_count++;
#endif
	
	priv->redraw_id = 0;
	for (l = priv->invalidate_areas; l != NULL; l = priv->invalidate_areas) {
		GdkRectangle *damaged_area = l->data;
		
		priv->invalidate_areas = g_list_remove (priv->invalidate_areas, damaged_area);
		g_free (damaged_area);
	}
	g_list_free (priv->invalidate_areas);
	priv->invalidate_areas = NULL;
	
	return FALSE;
}

/*** KibaDock signal handler ***/

/*
 * Callback for the context menu items that offer different drag and drop receive methods.
 */
static void __dnd_menu_item_activate_cb (GtkWidget *widget,
						gpointer   data) {
	KibaDragAction *dnd_action = data;
	
	dnd_action->data->receive_func (dnd_action->context,
			dnd_action->drag_data,
			dnd_action->data->user_data);
}

static void __dnd_menu_destroy_cb (GtkWidget *widget,
					  gpointer   data) {
	GList *actions = data;
	GList *l;
	
	for (l = actions; l != NULL; l = actions) {
		KibaDragAction *dnd_action = l->data;
		
		actions = g_list_remove (actions, l->data);
		g_free (dnd_action->action);
		g_free (dnd_action->icon);
		g_free (dnd_action->drag_data);
		g_object_unref (G_OBJECT(dnd_action->context));
		g_free (dnd_action);
	}
	g_list_free (actions);
}

static void __dock_drag_data_received_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     gint              x,
					     gint              y,
					     GtkSelectionData *selection_data,
					     guint             info,
					     guint             time,
					     gpointer          data) {
	KibaWin *win = KIBA_WIN (data);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	GList *actions = NULL;
	GList *l;
	
	g_strstrip ((char *) selection_data->data);
	
	/* checking for possible receive methods, adding them to actions list */
	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;
		KibaDragAction *dnd_action;
		char *action = NULL, *icon = NULL;
		
		data->dest_func (drag_context,
				(char *) selection_data->data,
				data->user_data,
				&action, &icon);
		if (!action)
			continue;
		
		dnd_action = g_new0 (KibaDragAction, 1);
		dnd_action->data = data;
		dnd_action->action = g_strdup(action);
		dnd_action->icon = (icon != NULL ? g_strdup(icon) : "");
		dnd_action->drag_data = g_strdup ((char *) selection_data->data);
		dnd_action->context = g_object_ref (G_OBJECT(drag_context));
		
		actions = g_list_append (actions, dnd_action);
	}
	
	/* displaying a menu to allow the user choosing the receive method */
	if (g_list_length (actions) > 1) {
		GtkWidget *menu;
		GtkWidget *item;
		GtkWidget *image;
		
		menu = gtk_menu_new ();
		
		for (l = actions; l != NULL; l = l->next) {
			KibaDragAction *dnd_action = l->data;
			GdkPixbuf *pixbuf = NULL;
			
			item = gtk_image_menu_item_new_with_label (dnd_action->action);
			kiba_load_icon (dnd_action->icon, 24, &pixbuf);
			if (pixbuf) {
				image = gtk_image_new_from_pixbuf (pixbuf);
				g_object_unref (pixbuf);
			}
			else {
				image = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE,
						GTK_ICON_SIZE_MENU);
			}
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (__dnd_menu_item_activate_cb), dnd_action);
			gtk_widget_show (item);
		}
		
		gtk_menu_popup (GTK_MENU(menu),
			NULL, NULL, NULL, NULL,
			3, time);
		
		/* connecting signal to free some memory */
		g_signal_connect (G_OBJECT(menu), "destroy",
			G_CALLBACK (__dnd_menu_destroy_cb), actions);
	}
	/* calling the one possible receive method */
	else if (actions) {
		KibaDragAction *dnd_action = actions->data;
		
		dnd_action->data->receive_func (dnd_action->context,
				dnd_action->drag_data,
				dnd_action->data->user_data);
		
		actions = g_list_remove (actions, actions->data);
		g_free (dnd_action->action);
		g_free (dnd_action->icon);
		g_free (dnd_action->drag_data);
		g_object_unref (G_OBJECT(dnd_action->context));
		g_free (dnd_action);
		g_list_free (actions);
	}
	
	gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

static gboolean __dock_button_press_event (GtkWidget      *widget,
					      GdkEventButton *event,
					      gpointer        data) {
	KibaWin *win  = data;
	KibaDock *dock = KIBA_DOCK (widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (win);
	
	if (event->button == 3) return FALSE;
	
	/* starting the dock drag */
	priv->mouse_anchor.x = event->x + widget->allocation.x;
	priv->mouse_anchor.y = event->y + widget->allocation.y;
	if (!priv->move_dock_id) {
		if (KIBA_DOCK(dock)->autohide.state != DOCK_UNHIDDEN)
			KIBA_DOCK(dock)->autohide.state = DOCK_UNHIDDEN;
		priv->move_dock_id = g_timeout_add (20, __move_dock, win);
	}
	
	/* setting drag alpha */
	kiba_dock_set_alpha (KIBA_DOCK(widget), DOCK_DRAG_ALPHA);
	
	return FALSE;
}

static gboolean __dock_button_release_event (GtkWidget      *widget,
					        GdkEventButton *event,
					        gpointer        data) {
	KibaWin *win  = data;
	KibaDock *dock = KIBA_DOCK (widget);
	KibaWinPrivate *priv;
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	
	/* stopping the dock drag */
	if (priv->move_dock_id) {
		g_source_remove (priv->move_dock_id);
		priv->move_dock_id = 0;
		
		__shape_input(GTK_WIDGET(win));
	}
	
	priv->mouse_anchor.x = 0;
	priv->mouse_anchor.y = 0;
	
	/* reverting alpha */
	kiba_dock_set_alpha (dock, 1.0);
	
	return FALSE;
}

static void __dock_object_count_changed (GtkContainer *container,
					    GtkWidget    *widget,
					    gpointer      user_data) {
	KibaWin *win = user_data;
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (win);
	
	if (!GTK_WIDGET_REALIZED(container))
		return;
	
	win->allocate_children (win, win->allocate_data);
}

static void __dock_position_changed (KibaDock     *dock,
					KibaPosition  position,
					gpointer      data) {
	//FIXME: looks weird
	gtk_widget_queue_resize (GTK_WIDGET(dock));
	gtk_widget_queue_resize (GTK_WIDGET(data));
}


static void __dock_autohide_toggled (KibaDock  *dock,
				        gboolean   enabled,
				        gpointer   data) {
	KibaWin *win = data;
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (win);
	
	if (enabled && !priv->check_hide_id)
		priv->check_hide_id = g_timeout_add (50, __check_hide_dock, win);
	else if (!enabled) {
		if (priv->check_hide_id) {
			g_source_remove (priv->check_hide_id);
			priv->check_hide_id = 0;
		}
		if (priv->hide_id) {
			g_source_remove (priv->hide_id);
			priv->hide_id = 0;
		}
		if (dock->autohide.state == DOCK_HIDING || dock->autohide.state == DOCK_HIDDEN) {
			dock->autohide.state = DOCK_UNHIDING;
			g_timeout_add (20 , __unhide_dock, win);
		}
	}
}

static void __dock_size_allocate_cb (GtkWidget *widget,
					GtkAllocation *allocation,
					gpointer       data) {
	if (!GTK_WIDGET_REALIZED(data))
		return;
	
	/* reloading the input shape */
	__shape_input (GTK_WIDGET(data));
}

/* GdkScreen Event */
static void __screen_size_changed_cb (GdkScreen *screen,
				    gpointer   data) {
	GtkWidget *widget = data;
	
	gtk_widget_queue_resize (widget);
}

/*** Public Methods ***/

/**
 * kiba_win_add_drag_func:
 * @win: The KibaWin Instance.
 * @receive_func: The method to receive drag sources.
 * @dest_func: The method to determine if a drag source can be received.
 * @user_data: The user data.
 *
 * Adds a drag and drop receive Method.
 * @dest_func is called first to check if @receive_func is able to receive the drag.
 */
void kiba_win_add_drag_func (KibaWin            *win,
			     KibaDragReceiveFunc receive_func,
			     KibaDragDestFunc    dest_func,
			     gpointer            user_data) {
	KibaWinPrivate *priv;
	KibaWinDragData *new_data;
	GList *l;
	
	g_return_if_fail (KIBA_IS_WIN(win));
	g_return_if_fail (receive_func != NULL);
	g_return_if_fail (dest_func != NULL);
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	
	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;
		
		if (data->receive_func == receive_func) {
			data->user_data = user_data;
			return;
		}
	}
	
	new_data = g_new0 (KibaWinDragData, 1);
	new_data->receive_func = receive_func;
	new_data->dest_func    = dest_func;
	new_data->user_data    = user_data;
	
	priv->drag_reveive_data = g_list_append (priv->drag_reveive_data, new_data);
}

/**
 * kiba_win_remove_drag_func:
 * @win: The KibaWin Instance.
 * @receive_func: The drag and drop receive method to remove.
 *
 * Removes previously added drag and drop receive function.
 */
void kiba_win_remove_drag_func (KibaWin            *win,
				KibaDragReceiveFunc receive_func) {
	KibaWinPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_WIN(win));
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	
	for (l = priv->drag_reveive_data; l != NULL; l = l->next) {
		KibaWinDragData *data = l->data;
		
		if (data->receive_func == receive_func) {
			priv->drag_reveive_data = g_list_remove (priv->drag_reveive_data, data);
			g_free (data);
			return;
		}
	}
}

/**
 * kiba_win_unset_children_allocate_func:
 * @win: The KibaWin Instance.
 * @func: The function to set.
 * @data: The userdata passed to @func.
 *
 * Sets @func as size allocation method.
 *
 * Returns: TRUE if @func is the new size allocate method, otherwise FALSE is returned.
 */
gboolean kiba_win_set_children_allocate_func (KibaWin *win,
                                              KibaWinAllocateChildrenFunc func,
                                              gpointer data) {
	if (win->allocate_children != __resize_children)
		return FALSE;
	
	win->allocate_children = func;
	win->allocate_data = data;
	
	gtk_widget_queue_resize(GTK_WIDGET(win));
	
	return TRUE;
}

/**
 * kiba_win_unset_children_allocate_func:
 * @win: The KibaWin Instance.
 * @func: The function to unset.
 *
 * Reverts to default size allocation method.
 */
gboolean kiba_win_unset_children_allocate_func (KibaWin *win,
                                                KibaWinAllocateChildrenFunc func) {
	if (win->allocate_children != func)
		return FALSE;
	
	win->allocate_children = __resize_children;
	win->allocate_data = NULL;
	
	gtk_widget_queue_resize(GTK_WIDGET(win));
	
	return TRUE;
}

/**
 * kiba_win_set_flag:
 * @win: The KibaWin Instance.
 * @prob: The property to set.
 *
 * Sets a @KibaWinProb property.
 */
void kiba_win_set_flag (KibaWin *win,
                        KibaWinProbs prob) {
	GtkWidget *child;
	
	g_return_if_fail (KIBA_IS_WIN(win));
	
	child = gtk_bin_get_child(GTK_BIN(win));
	
	if (GET_FLAG(win->probs, prob))
		return;
	
	SET_FLAG(win->probs, prob);
	
	if (!GTK_WIDGET_REALIZED(win))
		return;
	
	switch (prob) {
		case KIBA_WIN_ABOVE:
			gdk_window_set_keep_above (GTK_WIDGET (win)->window, TRUE);
			break;
		case KIBA_WIN_BELOW:
			gdk_window_set_keep_below (GTK_WIDGET (win)->window, TRUE);
			/* without that the window is getting borders !?! */
			gdk_window_set_decorations (GTK_WIDGET (win)->window, 0);
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

/**
 * kiba_win_unset_flag:
 * @win: The KibaWin Instance.
 * @prob: The property to unset.
 *
 * Unsets a @KibaWinProb property.
 */
void kiba_win_unset_flag (KibaWin *win,
                          KibaWinProbs prob) {
	GtkWidget *child;
	
	child = gtk_bin_get_child(GTK_BIN(win));
	
	g_return_if_fail (KIBA_IS_WIN(win));
	
	if (!GET_FLAG(win->probs, prob))
		return;
	
	DEL_FLAG(win->probs, prob);
	if (!GTK_WIDGET_REALIZED(win))
		return;
	
	switch (prob) {
		case KIBA_WIN_ABOVE:
			gdk_window_set_keep_above (GTK_WIDGET(win)->window, FALSE);
			break;
		case KIBA_WIN_BELOW:
			gdk_window_set_keep_below (GTK_WIDGET (win)->window, FALSE);
			
			if (child)
				kiba_dock_update_strut(child);
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

/**
 * kiba_win_invalidate_area:
 * @widget: The KibaWin Instance.
 * @area: The area that should be redrawn on next expose event.
 *
 * Invalidate @area on win and queues draw.
 */
void kiba_win_invalidate_area (GtkWidget *widget,
                               GdkRectangle area) {
	KibaWinPrivate  *priv;
	GList *l;
	GtkAllocation *new_damage;
	
	g_return_if_fail (KIBA_IS_WIN(widget));
	
	priv = KIBA_WIN_GET_PRIVATE (widget);
	
	/* avoid off-window areas */
	if (area.x < 0) area.x = 0;
	if (area.x + area.width > widget->allocation.width)
		area.width = widget->allocation.width - area.x;
	if (area.y < 0) area.y = 0;
	if (area.y + area.height > widget->allocation.height)
		area.height = widget->allocation.height - area.y;
	
	for (l = priv->invalidate_areas; l != NULL; l = l->next) {
		GdkRectangle *added_area = l->data;
		
		/* there is allready an area containing the new area, we can skip the new area */
		if (area.x >= added_area->x && area.y >= added_area->y &&
		    area.x + area.width  <= added_area->x + added_area->width &&
		    area.y + area.height <= added_area->y + added_area->height)
			return;
		/* the new area contains another added area */
		if (area.x <= added_area->x && area.y <= added_area->y &&
		    area.x + area.width  >= added_area->x + added_area->width &&
		    area.y + area.height >= added_area->y + added_area->height) {
			priv->invalidate_areas = g_list_remove (priv->invalidate_areas, added_area);
			g_free (added_area);
			added_area = NULL;
		}
	}
	
	new_damage = g_new0 (GtkAllocation, 1);
	new_damage->x = area.x;
	new_damage->y = area.y;
	new_damage->width = area.width;
	new_damage->height = area.height;
	
	priv->invalidate_areas = g_list_append (priv->invalidate_areas, new_damage);
	
	if (!priv->redraw_id)
		priv->redraw_id = g_timeout_add (1000/MAX_FPS, __redraw, widget);
}

/**
 * kiba_win_cairo_create:
 * @win: The KibaWin Instance.
 *
 * Creates a cairo context, compatible to @win.
 * 
 * Returns: the drawing context created.
 */
cairo_t* kiba_win_cairo_create (KibaWin *win) {
	KibaWinPrivate  *priv;
	GtkWidget *widget;
	cairo_t *cr;
	
	g_return_val_if_fail (KIBA_IS_WIN(win), NULL);
	
	if (!GTK_WIDGET_REALIZED(win))
		return NULL;
	
	priv = KIBA_WIN_GET_PRIVATE (win);
	widget = GTK_WIDGET(win);
	
#ifdef USE_GLITZ
	cr = kiba_glitz_get_cairo_context (priv->pGlitzDrawable,
			priv->pGlitzDrawFormat,
			priv->pGlitzFormat,
			widget->allocation.width,
			widget->allocation.height);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
#endif
#ifdef USE_SDL
	cr = cairo_get_context_from_puc_data (&priv->puc_surface,
			widget->allocation.width,
			widget->allocation.height);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	cr = gdk_cairo_create (priv->pixmap);
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS)
		return cr;
#endif
	cairo_destroy (cr);
	
	return gdk_cairo_create (widget->window);
}

/**
 * kiba_win_quit:
 * @widget:  The KibaWin Instance.
 *
 * Hide the dock.
 *
 * Returns: FALSE if the dock is hidden, otherwise TRUE is returned.
 */
gboolean kiba_win_quit (GtkWidget *widget) {
	KibaWinPrivate *priv;
	GtkWidget *dock;
	
	g_return_if_fail (KIBA_IS_WIN(widget));
	
	priv = KIBA_WIN_GET_PRIVATE (widget);
	dock = GTK_BIN(widget)->child;
	
	if (priv->check_hide_id) {
		g_source_remove (priv->check_hide_id);
		priv->check_hide_id = 0;
	}
	KIBA_DOCK(dock)->autohide.state = DOCK_HIDING;
	
	if (!__hide_dock(widget))
		return FALSE;
	
	GdkRectangle area;
	area.x = dock->allocation.x;
	area.y = dock->allocation.y;
	area.width = dock->allocation.width;
	area.height = dock->allocation.height;
	
	kiba_win_invalidate_area (widget, area);
	__redraw (widget);
	
	return TRUE;
}

/*** Widget Class Implementations ***/

/*
 * gtk_container_remove implementation.
 */
static void kiba_win_remove (GtkContainer *container,
			     GtkWidget    *widget) {
	KibaWinPrivate *priv;
	
	priv = KIBA_WIN_GET_PRIVATE (container);
	
	/* disconnecting the KibaDock event handler */
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_button_press_event), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_button_release_event), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_object_count_changed), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_position_changed), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_autohide_toggled), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_size_allocate_cb), container);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__dock_drag_data_received_cb), container);
	
	if (priv->check_hide_id) {
		g_source_remove (priv->check_hide_id);
		priv->check_hide_id = 0;
	}
	
	(* GTK_CONTAINER_CLASS (kiba_win_parent_class)->remove) (container, widget);
}

/*
 * gtk_container_add implementation.
 */
static void kiba_win_add (GtkContainer *container,
			  GtkWidget    *widget) {
	KibaWinPrivate *priv;
	
	g_return_if_fail (KIBA_IS_DOCK(widget));
	
	priv = KIBA_WIN_GET_PRIVATE (container);
	
	/* connecting to KibaDock events */
	g_signal_connect_after (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(__dock_button_press_event), container);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(__dock_button_release_event), container);
	g_signal_connect (G_OBJECT(widget), "add",
		G_CALLBACK(__dock_object_count_changed), container);
	g_signal_connect (G_OBJECT(widget), "remove",
		G_CALLBACK(__dock_object_count_changed), container);
	g_signal_connect (G_OBJECT(widget), "position-changed",
		G_CALLBACK(__dock_position_changed), container);
	g_signal_connect (G_OBJECT(widget), "autohide-toggled",
		G_CALLBACK(__dock_autohide_toggled), container);
	g_signal_connect_after (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(__dock_size_allocate_cb), container);
	g_signal_connect_after (G_OBJECT(widget), "drag-data-received",
		G_CALLBACK(__dock_drag_data_received_cb), container);
	
	if (GET_FLAG(KIBA_DOCK(widget)->flags, KIBA_DOCK_AUTOHIDE) && !priv->check_hide_id)
		priv->check_hide_id = g_timeout_add (50, __check_hide_dock, container);
	
	(* GTK_CONTAINER_CLASS (kiba_win_parent_class)->add) (container, widget);
}

/*
 * gtk_widget_expose implementation.
 */
static gboolean kiba_win_expose_event (GtkWidget      *widget,
				       GdkEventExpose *event) {
	KibaWinPrivate  *priv;
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	priv = KIBA_WIN_GET_PRIVATE (widget);
	
	kiba_win_invalidate_area (widget, event->area);
	
	/* using maximal MAX_FPS fps */
	if (!priv->redraw_id)
		priv->redraw_id = g_timeout_add (1000/MAX_FPS, __redraw, widget);
	
	return FALSE;
}

/*
 * gtk_widget_screen_changed implementation.
 */
static void kiba_win_screen_changed (GtkWidget *widget,
                                     GdkScreen *previous_screen) {
	KibaWinPrivate *priv;
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	priv = KIBA_WIN_GET_PRIVATE (widget);
	
#ifdef USE_GLITZ
	kiba_glitz_get_colormap (&priv->pGlitzDrawFormat,
		&visual,
		&colormap);
#endif
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgba_colormap (screen);
		visual   = gdk_screen_get_rgba_visual (screen);
	}
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

/*
 * gtk_widget_size_allocate implementation.
 */
static void kiba_win_size_allocate (GtkWidget     *widget,
				    GtkAllocation *allocation) {
	KibaWinPrivate  *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaWin *win = KIBA_WIN(widget);
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	const KibaLayoutData *d;
	gboolean size_changed = TRUE;
	GList *l;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		allocation->x = KIBA_SCREEN_DEFAULT->geometry.x +
			d->halignment/2 * (KIBA_SCREEN_DEFAULT->geometry.width - allocation->width);
		allocation->y = KIBA_SCREEN_DEFAULT->geometry.y;
	}
	else {
		allocation->y = KIBA_SCREEN_DEFAULT->geometry.y +
			d->valignment/2 * (KIBA_SCREEN_DEFAULT->geometry.height - allocation->height);
		allocation->x = KIBA_SCREEN_DEFAULT->geometry.x;
	}
	
	if (allocation->width  == widget->allocation.width &&
	    allocation->height == widget->allocation.height)
		size_changed = FALSE;
	
	widget->allocation = *allocation;
	
	if (GTK_BIN(widget)->child) {
		GList *l;
		
		win->allocate_children (win, win->allocate_data);
	}
	
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	if (!size_changed) {
		gdk_window_move(widget->window,
			allocation->x, allocation->y);
		return;
	}
	
#ifdef USE_GLITZ
	kiba_glitz_resize_func (priv->pGlitzDrawable, allocation->width, allocation->height);
#endif
#ifdef USE_SDL
	if (priv->puc_surface)
		g_free (priv->puc_surface);
	priv->puc_surface = cairo_create_puc_data (widget->allocation.width, widget->allocation.height);
	priv->puc_width = widget->allocation.width;
	priv->puc_height = widget->allocation.height;
	kiba_sdl_resize_func (widget->allocation.width, widget->allocation.height);
	kiba_sdl_update_surface (widget->allocation.width, widget->allocation.height);
#endif

	if (priv->redraw_id) {
		g_source_remove(priv->redraw_id);
		priv->redraw_id = 0;
	}
	
	for (l = priv->invalidate_areas; l != NULL; l = priv->invalidate_areas) {
		GdkRectangle *damaged_area = l->data;
		
		priv->invalidate_areas = g_list_remove (priv->invalidate_areas, damaged_area);
		g_free (damaged_area);
	}
	g_list_free (priv->invalidate_areas);
	priv->invalidate_areas = NULL;
	
	/* update the offscreen drawable */
	cairo_t *cr = NULL;
#ifdef USE_OFFSCREEN_PIXMAP
	GdkGC *gc;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap))
		g_object_unref(priv->pixmap);
	priv->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
	gc = gdk_gc_new (priv->pixmap);
#endif
	cr = kiba_win_cairo_create (KIBA_WIN(widget));
	if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {
		cairo_set_tolerance (cr, 0.1);
		cairo_set_source_rgba (cr, 0, 0, 0, 0);
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		cairo_paint (cr);
		
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		cairo_save (cr);
			kiba_dock_expose (KIBA_DOCK(dock), cr);
		cairo_restore (cr);
		cairo_save (cr);
			kiba_progress_draw (KIBA_WIN(widget)->progress, cr);
		cairo_restore (cr);
		
		cairo_destroy (cr);
	}
	
	/* move/resize the window */
	gdk_window_move_resize(widget->window,
		allocation->x, allocation->y,
		allocation->width, allocation->height);
	
	/* rendering the offscreen drawable to the window */
#ifdef USE_GLITZ
	kiba_glitz_draw_func (priv->pGlitzDrawable, priv->pGlitzDrawFormat);
#endif
#ifdef USE_SDL
	kiba_sdl_draw_func (&priv->puc_surface,
		widget->allocation.width, widget->allocation.height);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	gdk_draw_drawable (widget->window,
			   gc, priv->pixmap,
			   0, 0, 0, 0,
			   widget->allocation.width,
			   widget->allocation.height);
	gdk_gc_unref (gc);
#endif
}

/*
 * gtk_widget_size_request implementation.
 */
static void kiba_win_size_request (GtkWidget      *widget,
				   GtkRequisition *requisition) {
	GtkWidget *dock = KIBA_WIN_CHILD(widget);
	const KibaLayoutData *d;
	gint bounce_height;
	gint width;
	float pulse_factor;
	float offsetX = 0.0f, offsetY = 0.0f;
	GList *l;
	
	if (!GTK_WIDGET_REALIZED(widget)) {
		requisition->width = 0;
		requisition->height = 0;
		return;
	}
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	gtk_widget_size_request (dock, requisition);
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaObject *object = KIBA_OBJECT(child->widget);
		float customX = 0.0f, customY = 0.0f;
		GList *li;
		
		for (li = object->custom_offset_data; li != NULL; li = li->next) {
			float customXX = 0.0, customYY = 0.0;
			KibaObjectCustomOffsetData *data = li->data;
			
			data->max_func (object, &customXX, &customYY, data->data);
			customX = MAX(customX, customXX);
			customX = MAX(customY, customYY);
		}
		offsetX = MAX(customX, offsetX);
		offsetY = MAX(customX, offsetY);
	}
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		requisition->height = KIBA_SCREEN_DEFAULT->geometry.height;
		requisition->width += offsetX;
	}
	else {
		requisition->width = KIBA_SCREEN_DEFAULT->geometry.width;
		requisition->height += offsetY;
	}
}

static gboolean kiba_win_enter_notify_event (GtkWidget *widget,
					     GdkEventCrossing *event) {
	KibaWin *win = KIBA_WIN(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	
	if (GET_FLAG(win->probs, KIBA_WIN_BELOW)) {
		if (priv->raise_id) {
			g_source_remove(priv->raise_id);
			priv->raise_id = 0;
		}
		gdk_window_set_keep_below (GTK_WIDGET (win)->window, FALSE);
	}
	
	return FALSE;
}

static gboolean kiba_win_leave_notify_event (GtkWidget *widget,
					     GdkEventCrossing *event) {
	KibaWin *win = KIBA_WIN(widget);
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	GtkWidget *dock;
	
	dock = KIBA_WIN_CHILD(win);
	
	if (GET_FLAG(win->probs, KIBA_WIN_BELOW)) {
		//FIXME there may is an leave event when hovering an object
		if (!priv->raise_id)
			priv->raise_id = g_timeout_add (500, __raise_window, win);
	}
	
	return FALSE;
}

/*
 * gtk_widget_show implementation.
 */
static void kiba_win_show (GtkWidget *widget) {
	GTK_WIDGET_SET_FLAGS (widget, GTK_VISIBLE);
	
	if (!GTK_WIDGET_REALIZED (widget))
		gtk_widget_realize (widget);
	
	if (!GTK_WIDGET_MAPPED (widget))
		gtk_widget_map (widget);
	
	/* clear the window to be transparent */
	__clear (KIBA_WIN(widget));
	/* shape the input events */
	__shape_input (widget);
}

/*
 * gtk_widget_map implementation.
 */
static void kiba_win_map (GtkWidget *widget) {
	GtkBin *bin = GTK_BIN (widget);
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
	
	gdk_window_show (widget->window);
	
	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/*
 * gtk_widget_unrealize implementation.
 */
static void kiba_win_unrealize (GtkWidget *widget) {
	KibaWinPrivate *priv = KIBA_WIN_GET_PRIVATE (widget);
	KibaWin *win = KIBA_WIN(widget);
	KibaScreen *screen;
	GList *l;
	
	screen = kiba_screen_get();
	g_signal_handlers_disconnect_by_func (G_OBJECT(screen->gdk_screen),
		G_CALLBACK(__screen_size_changed_cb), widget);
	
	for (l = priv->drag_reveive_data; l != NULL; l = priv->drag_reveive_data) {
		KibaWinDragData *data = l->data;
		priv->drag_reveive_data = g_list_remove (priv->drag_reveive_data, l->data);
		g_free (data);
	}
	g_list_free (priv->drag_reveive_data);
	
	for (l = priv->invalidate_areas; l != NULL; l =  priv->invalidate_areas) {
		GdkRectangle *damaged_area = l->data;
		priv->invalidate_areas = g_list_remove (priv->invalidate_areas, damaged_area);
		g_free (damaged_area);
	}
	g_list_free (priv->invalidate_areas);
	
	if (priv->raise_id)
		g_source_remove (priv->raise_id);
	if (priv->redraw_id)
		g_source_remove (priv->redraw_id);
	if (priv->check_hide_id)
		g_source_remove (priv->check_hide_id);
	if (priv->hide_id)
		g_source_remove (priv->hide_id);
	if (priv->move_dock_id)
		g_source_remove (priv->move_dock_id);
	
	if (KIBA_IS_TRIANGLE(win->triangle))
		kiba_triangle_destroy (win->triangle);
	
	if (KIBA_IS_PROGRESS(win->progress))
		g_object_unref (G_OBJECT(win->progress));
	
	(* GTK_WIDGET_CLASS (kiba_win_parent_class)->unrealize) (widget);
	
#ifdef USE_GLITZ
	kiba_glitz_fini (priv->pGlitzDrawable);
#endif
#ifdef USE_SDL
	kiba_sdl_fini ();
	if (priv->puc_surface)
		g_free (priv->puc_surface);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	if (GDK_IS_PIXMAP(priv->pixmap))
		g_object_unref (priv->pixmap);
#endif
}

/*
 * gtk_widget_realize implementation.
 */
static void kiba_win_realize (GtkWidget *widget) {
	GtkWindow *window = GTK_WINDOW (widget);
	KibaWinPrivate *priv;
	GdkWindowAttr attributes;
	gint attributes_mask;
	
	priv = KIBA_WIN_GET_PRIVATE (widget);
	
	/* make sure the size is properly allocated */
	if (widget->allocation.x == -1 &&
	    widget->allocation.y == -1 &&
	    widget->allocation.width == 1 &&
	    widget->allocation.height == 1) {
		GtkRequisition requisition;
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		gtk_widget_size_request (widget, &requisition);
		allocation.width = requisition.width;
		allocation.height = requisition.height;
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	attributes.window_type = GDK_WINDOW_TOPLEVEL;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.title = window->title;
	attributes.wmclass_name = window->wmclass_name;
	attributes.wmclass_class = window->wmclass_class;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.visual = gtk_widget_get_visual(widget);
	attributes.colormap = gtk_widget_get_colormap(widget);
	attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	if (attributes.visual != NULL && attributes.colormap != NULL)
		attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;
	attributes_mask |= (window->title ? GDK_WA_TITLE : 0);
	attributes_mask |= (window->wmclass_name ? GDK_WA_WMCLASS : 0);
	
	widget->window = gdk_window_new (gdk_get_default_root_window (),
				&attributes, attributes_mask);
	/* catch the window events on the widget */
	gdk_window_set_user_data (widget->window, window);
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	gdk_window_set_type_hint (widget->window, GDK_WINDOW_TYPE_HINT_DOCK);
	gdk_window_set_icon_name (widget->window, PACKAGE);
	gdk_window_stick (widget->window);
	
	if (GET_FLAG(KIBA_WIN(widget)->probs, KIBA_WIN_ABOVE))
		gdk_window_set_keep_above (widget->window, TRUE);
	if (GET_FLAG(KIBA_WIN(widget)->probs, KIBA_WIN_BELOW)) {
		gdk_window_set_keep_below (widget->window, TRUE);
		/* without that the window has border !?! */
		gdk_window_set_decorations (widget->window, 0);
	}
	
	/* initialling the offscreen drawable, the double buffering is disabled */
#ifdef USE_GLITZ
	kiba_glitz_create_drawable (GDK_WINDOW_XWINDOW(widget->window),
		widget->allocation.width, widget->allocation.height,
		priv->pGlitzDrawFormat,
		&priv->pGlitzFormat,
		&priv->pGlitzDrawable);
#endif
#ifdef USE_SDL
	priv->puc_surface = cairo_create_puc_data (widget->allocation.width,
					widget->allocation.height);
	priv->puc_width = widget->allocation.width;
	priv->puc_height = widget->allocation.height;
	kiba_sdl_init (GDK_WINDOW_XWINDOW(widget->window),
		widget->allocation.width,
		widget->allocation.height);
	kiba_sdl_resize_func (widget->allocation.width,
		widget->allocation.height);
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	priv->pixmap = kiba_cairo_utils_create_pixmap (widget->allocation.width, widget->allocation.height);
#endif
	
	KIBA_WIN (widget)->triangle = kiba_triangle_new (widget);
	KIBA_WIN (widget)->progress = kiba_progress_new (widget);
	
#if BENCHMARK_ENABLE == TRUE
	/* spitting out redraws */
	g_timeout_add (1000, __benchmark_cb, NULL);
#endif
}

/*** Class init ***/

static void kiba_win_class_init (KibaWinClass *class) {
	GtkWidgetClass    *widget_class;
	GtkContainerClass *container_class;
	
	widget_class    = GTK_WIDGET_CLASS (class);
	container_class = GTK_CONTAINER_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaWinPrivate));
	
	widget_class->realize = kiba_win_realize;
	widget_class->unrealize = kiba_win_unrealize;
	widget_class->show = kiba_win_show;
	widget_class->map = kiba_win_map;
	widget_class->enter_notify_event = kiba_win_enter_notify_event;
	widget_class->leave_notify_event = kiba_win_leave_notify_event;
	widget_class->size_allocate = kiba_win_size_allocate;
	widget_class->size_request = kiba_win_size_request;
	widget_class->expose_event = kiba_win_expose_event;
	widget_class->screen_changed = kiba_win_screen_changed;
	
	container_class->add = kiba_win_add;
	container_class->remove = kiba_win_remove;
	
	/* signals */
	kiba_win_signals[SIG_DOCK_HIDDEN] =
		g_signal_new ("dock-hidden",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWinClass, dock_hidden),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_DOCK);
	kiba_win_signals[SIG_DOCK_UNHIDDEN] =
		g_signal_new ("dock-unhidden",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWinClass, dock_unhidden),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			KIBA_TYPE_DOCK);
}

static void kiba_win_init (KibaWin *kiba_win) {
	KibaWinPrivate *priv;
	KibaScreen *screen;
	
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (kiba_win),
			GTK_NO_WINDOW |
			GTK_DOUBLE_BUFFERED);
	
	priv = KIBA_WIN_GET_PRIVATE (kiba_win);
	screen = kiba_screen_get();
	
	gtk_widget_add_events (GTK_WIDGET (kiba_win),
		GDK_EXPOSURE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	kiba_win_screen_changed (GTK_WIDGET (kiba_win), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (kiba_win), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (kiba_win), FALSE);
	gtk_window_set_resizable (GTK_WINDOW (kiba_win), FALSE);
	gtk_window_set_title (GTK_WINDOW (kiba_win), PACKAGE);
	gtk_window_set_type_hint (GTK_WINDOW(kiba_win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_stick (GTK_WINDOW(kiba_win));
	
	kiba_win->allocate_children = __resize_children;
	kiba_win->allocate_data = NULL;
	
	kiba_win->triangle = NULL;
	kiba_win->progress = NULL;
	
	priv->invalidate_areas = NULL;
	priv->check_hide_id = 0;
	priv->redraw_id = 0;
	priv->move_dock_id = 0;
	priv->hide_id = 0;
	priv->mouse_anchor = (GdkRectangle) { 0, 0, 0, 0 };
	priv->drag_reveive_data = NULL;
#ifdef USE_GLITZ
	priv->pGlitzDrawable = NULL;
	priv->pGlitzFormat = NULL;
#endif
#ifdef USE_SDL
	priv->puc_surface = NULL;
	priv->puc_width = 0;
	priv->puc_height = 0;
#endif
#ifdef USE_OFFSCREEN_PIXMAP
	priv->pixmap = NULL;
#endif
	
	g_signal_connect (G_OBJECT(screen->gdk_screen), "size-changed",
		G_CALLBACK(__screen_size_changed_cb), kiba_win);
}

/**
 * kiba_win_new:
 *
 * Creates and returns a KibaWin Instance.
 * 
 * Returns: the KibaWin Instance created.
 */
GtkWidget* kiba_win_new (void) {
	return g_object_new (KIBA_TYPE_WIN, NULL);
}
