/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author :	Daniel Bessler <danielb@kiba-dock.org>
 *
*/

/**
 * SECTION:kiba-window
 * @short_description: Object representing a window on #KibaScreen.
 * @see_also:#KibaScreen,#GdkWindow
 * @stability: Unstable
 *
 * Object representing a window on #KibaScreen.
 */

#include <gdk/gdkx.h>
#include <X11/Xatom.h>

#include "kiba-x11.h"
#include "kiba-window.h"
#include "kiba-debug.h"

#define KIBA_WINDOW_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_WINDOW, KibaWindowPrivate))

G_DEFINE_TYPE (KibaWindow, kiba_window, G_TYPE_OBJECT)

typedef struct _KibaWindowPrivate KibaWindowPrivate;

struct _KibaWindowPrivate {
	gboolean dispose_has_run;
	Pixmap   icon_pixmap;
};

enum {
	SIG_NAME_CHANGED = 0,
	SIG_STATE_CHANGED,
	SIG_ICON_CHANGED,
	SIG_WINDOW_LAST
};

static unsigned long WM_STATE;

static guint window_signals[SIG_WINDOW_LAST] = { 0 };

/*** Private methods ***/

/**
 * kiba_window_property_notify:
 * @kiba_window: The corresponding KibaWindow
 * @xproperty: The XPropertyEvent to filter
 *
 * PropertyEvent callback from X, the windowing system, when the corresponding window changes
 *
 * Returns: The value of what the KibaWindow state should be after the event.
 */
static int kiba_window_property_notify(KibaWindow     *kiba_window,
                                       XPropertyEvent *xproperty) {
	int new_state = kiba_window->state;
	
	// If the name changed...
	if (xproperty->atom == XA_WM_NAME) {
		if (kiba_window->name != NULL)
			g_free(kiba_window->name);
		
		if (xproperty->state == PropertyNewValue) {
			kiba_window->name = kiba_win_name (gdk_display,
			  kiba_window->xwindow);
		} else if (xproperty->state == PropertyDelete) { 
			kiba_window->name = NULL;
		} else {
			DEBUG_ERROR("Internal error:  xproperty->state is an unrecognized value.\n");
		}
		g_signal_emit (kiba_window, window_signals[SIG_NAME_CHANGED],
		  0, kiba_window->name);
	
	// If the icon changed...
	} else if (xproperty->atom == XA_WM_HINTS) {
		XWMHints *hints = XGetWMHints (gdk_display,
		  kiba_window->xwindow);
		
		if (hints != NULL) {
			if (hints->flags & IconMaskHint) {
				g_signal_emit (kiba_window,
			  	window_signals[SIG_ICON_CHANGED], 0);
			}
			if (hints->flags & XUrgencyHint) {
				new_state |= (1 << KIBA_WINDOW_URGENT);
			} else {
				new_state &= ~(1 << KIBA_WINDOW_URGENT);
			}
			
			XFree(hints);
		}
		
	// Else, if the state changed...
	} else if (xproperty->atom == WM_STATE) {
		switch (kiba_get_winstate(gdk_display, kiba_window->gdk_window)) {
			case NormalState:
				// The state is not iconified
				new_state &= ~(1 << KIBA_WINDOW_ICONIFIED);
				
				// The state is not withdrawn
				new_state &= ~(1 << KIBA_WINDOW_WITHDRAWN);
				
				// If maximized...
				if (kiba_window_state_hint (
				  kiba_window->gdk_window,
				  "_NET_WM_STATE_MAXIMIZED_VERT") &&
				  kiba_window_state_hint (
				  kiba_window->gdk_window,
				  "_NET_WM_STATE_MAXIMIZED_HORZ")) {
					new_state |= (1 << KIBA_WINDOW_MAXIMIZED);
				
				// Else, if not maximized...
				} else {
					new_state &= ~(1 << KIBA_WINDOW_MAXIMIZED);
				}
				break;
			case IconicState:
				// The state is not maximized
				new_state &= ~(1 << KIBA_WINDOW_MAXIMIZED);
				
				// The state is not withdrawn
				new_state &= ~(1 << KIBA_WINDOW_WITHDRAWN);
				
				// Iconified
				new_state |= (1 << KIBA_WINDOW_ICONIFIED);
				break;
			case WithdrawnState:
				// The state is not maximized
				new_state &= ~(1 << KIBA_WINDOW_MAXIMIZED);
				
				// The state is not iconified
				new_state &= ~(1 << KIBA_WINDOW_ICONIFIED);
				
				// The state is withdrawn
				new_state |= (1 << KIBA_WINDOW_WITHDRAWN);
				break;
		}
	}
	
	return new_state;
}

/**
 * x_callback:
 * @gdkxevent: The X event to filter
 * @event: The GDK event to which the X event will be translated
 * @data: The corresponding KibaWindow
 *
 * Callback from X, the windowing system, when the corresponding window changes
 *
 * Returns: GDK_FILTER_CONTINUE (event not handled, continue processing)
 */
static GdkFilterReturn x_callback (GdkXEvent *gdkxevent,
                                   GdkEvent  *event,
                                   gpointer   data) {
	XEvent *xevent = gdkxevent;
	KibaWindow *kiba_window = KIBA_WINDOW(data);
	int new_state = kiba_window->state;
	
	if (xevent->xany.window != kiba_window->xwindow)
		return;
	
	switch (xevent->type) {
		case UnmapNotify:
			new_state &= ~(1 << KIBA_WINDOW_MAPPED);
			break;

		case MapNotify:
			new_state |= (1 << KIBA_WINDOW_MAPPED);
			break;

		case DestroyNotify:
			new_state |= (1 << KIBA_WINDOW_CLOSED);
			break;

		case PropertyNotify:
			new_state = kiba_window_property_notify(kiba_window,
			                                   &xevent->xproperty);
			break;
		case FocusIn:
			new_state |= (1 << KIBA_WINDOW_FOCUSED);
			break;
		case FocusOut:
			new_state &= ~(1 << KIBA_WINDOW_FOCUSED);
			break;
	}

	if (new_state != kiba_window->state) {
		int d_state = kiba_window->state ^ new_state;
		kiba_window->state = new_state;
		g_signal_emit (kiba_window,
			  window_signals[SIG_STATE_CHANGED], 0,
			  d_state);
	}
	
	return GDK_FILTER_CONTINUE;
}

/*** Public Methods ***/

/**
 * kiba_window_set_minimize_restore_point:
 * @window: The KibaWindow instance.
 * @x: The x position of the minimize/restore area.
 * @y: The y position on the minimize/restore area.
 * @width: The width of the minimize/restore area.
 * @height: The height of the minimize/restore area.
 *
 * Sets the minimize/restore point of a KibaWindow.  In other words, when the window is minimized, it will shrink to the specified position on the screen. When the window is restored, it will expand from that position on the screen.
 */
void kiba_window_set_minimize_restore_point(KibaWindow *window,
                                            int x,	
                                            int y,
                                            int width,
                                            int height)
{
	kiba_set_icon_geometry (gdk_x11_drawable_get_xid(window->gdk_window),
	                        x,
	                        y,
	                        width,
	                        height);
}

/*** custom implementations ***/

static void kiba_window_dispose (GObject *obj) {
	KibaWindow *kiba_window;
	KibaWindowPrivate *priv;

	kiba_window = KIBA_WINDOW(obj);
	priv = KIBA_WINDOW_GET_PRIVATE (kiba_window);
	
	gdk_window_remove_filter (kiba_window->gdk_window, x_callback,
	  kiba_window);
	
	if (priv->dispose_has_run) {
		/* If dispose did already run, return. */
		return;
	}
	
	/* Make sure dispose does not run twice. */
	priv->dispose_has_run = TRUE;
	
	if (kiba_window->name) {
		g_free (kiba_window->name);
		kiba_window->name = NULL;
	}
	
	/* Chain up to the parent class */
	G_OBJECT_CLASS(kiba_window_parent_class)->dispose (obj);
}

/*** Class Functions ***/

static void kiba_window_class_init (KibaWindowClass *class) {
	GObjectClass *object_class = G_OBJECT_CLASS (class);
	
	object_class->dispose = kiba_window_dispose;
	
	g_type_class_add_private (object_class, sizeof (KibaWindowPrivate));
	
	/* Signals */
	window_signals[SIG_NAME_CHANGED] =
		g_signal_new ("name-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWindowClass, name_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__STRING,
			G_TYPE_NONE, 1,
			G_TYPE_CHAR);
	window_signals[SIG_STATE_CHANGED] =
		g_signal_new ("state-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWindowClass, state_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
	window_signals[SIG_ICON_CHANGED] =
		g_signal_new ("icon-changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaWindowClass, icon_changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0, G_TYPE_INVALID);

	WM_STATE = gdk_x11_get_xatom_by_name("WM_STATE");
}

static void kiba_window_init (KibaWindow *kiba_window) {
	KibaWindowPrivate *priv;
	
	priv = KIBA_WINDOW_GET_PRIVATE (kiba_window);
	
	kiba_window->state = (1 << KIBA_WINDOW_MAPPED);
	kiba_window->name = NULL;
}

/**
 * kiba_window_new:
 * @gdk_window: The GdkWindow.
 *
 * Creates and returns a new KibaWindow object from a GdkWindow.
 *
 * Returns: The KibaWindow object created.
 */
KibaWindow* kiba_window_new (GdkWindow *gdk_window) {
	KibaWindow *kiba_window;
	KibaWindowPrivate *priv;
	
	g_return_val_if_fail (GDK_IS_WINDOW(gdk_window), NULL);
	
	kiba_window = g_object_new (KIBA_TYPE_WINDOW, NULL);

	priv = KIBA_WINDOW_GET_PRIVATE (kiba_window);
	
	kiba_window->gdk_window = gdk_window;
	kiba_window->xwindow = gdk_x11_drawable_get_xid (
	  kiba_window->gdk_window);
	kiba_window->name = kiba_win_name(gdk_display, kiba_window->xwindow);
	
	kiba_show_interest_in_events(kiba_window->xwindow);
	priv->dispose_has_run = FALSE;
	
	gdk_window_add_filter (kiba_window->gdk_window, x_callback,
	  kiba_window);
	
	if (kiba_get_active_window (gdk_display) == kiba_window->xwindow)
		kiba_window->state |= (1 << KIBA_WINDOW_FOCUSED);
	
	switch (kiba_get_winstate (gdk_display, kiba_window->gdk_window)) {
		case IconicState:
			kiba_window->state |= (1 << KIBA_WINDOW_ICONIFIED);
			break;
		case WithdrawnState:
			kiba_window->state |= (1 << KIBA_WINDOW_WITHDRAWN);
			break;
	}
	
	if (kiba_window_state_hint (kiba_window->gdk_window,
	  "_NET_WM_STATE_MAXIMIZED_VERT") && kiba_window_state_hint (
	  kiba_window->gdk_window, "_NET_WM_STATE_MAXIMIZED_HORZ")) {
		kiba_window->state |= (1 << KIBA_WINDOW_MAXIMIZED);
	}
	
	XWMHints *hints = XGetWMHints (gdk_display, kiba_window->xwindow);

	if (hints != NULL) {
		if (hints->flags & XUrgencyHint) {
			kiba_window->state |= (1 << KIBA_WINDOW_URGENT);
		}
		XFree(hints);
	}

	return kiba_window;
}
