/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *
 *  Note : uses window functions from libwnck and the icon reading code is from metacity
*/

/**
 * SECTION:kiba-x11
 * @short_description: Xlib utilities.
 * @see_also:
 * @stability: Unstable
 *
 * Xlib utilities.
 */

#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gdk/gdkx.h>
#include <X11/Xatom.h>

#include <kiba-x11.h>
#include <kiba-utils.h>
#include <kiba-debug.h>

#define _(string) gettext (string)

static void kiba_handle_X_error_code (gint errorCode,
                                      int line)
{
	if (errorCode != Success) {
		if (errorCode == BadAlloc)
			kiba_debug_message(KIBA_DEBUG_ERROR, __FILE__, line, "Got X error code \"BadAlloc\"", errorCode);
		else if (errorCode == BadValue)
			kiba_debug_message(KIBA_DEBUG_ERROR, __FILE__, line, "Got X error code \"BadValue\"", errorCode);
		else if (errorCode == BadWindow)
			kiba_debug_message(KIBA_DEBUG_ERROR, __FILE__, line, "Got X error code \"BadWindow\"", errorCode);
		else
			kiba_debug_message(KIBA_DEBUG_ERROR, __FILE__, line, "Got X error code %d!", errorCode);
	}
}

#define KIBA_EXECUTE_IN_ERROR_TRAP(function) gdk_error_trap_push();\
                                               function\
                                               kiba_handle_X_error_code(gdk_error_trap_pop(), __LINE__);

static gboolean
kiba_get_cardinal (Window xwindow,
                   Atom   atom,
                   int   *val)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  gulong *num;
  int err, result;

  *val = 0;

  gdk_error_trap_push ();
  type = None;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result =
        XGetWindowProperty (gdk_display, xwindow, atom, 0, G_MAXLONG, False,
                            XA_CARDINAL, &type, &format, &nitems,
                            &bytes_after, (guchar **)&num);
  )
  err = gdk_error_trap_pop ();
  if (err != Success || result != Success)
      return FALSE;

  if (type != XA_CARDINAL)
  {
      XFree (num);
      return FALSE;
  }

  *val = *num;

  XFree (num);

  return TRUE;
}

static gboolean
kiba_get_cardinal_list (Window   xwindow,
                        Atom     atom,
                        gulong **cardinals,
                        int     *len)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  gulong *nums;
  int err, result;

  *cardinals = NULL;
  *len = 0;

  gdk_error_trap_push ();
  type = None;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (gdk_display, xwindow, atom, 0, G_MAXLONG,
        False, XA_CARDINAL, &type, &format, &nitems, &bytes_after, (void*)&nums);  
  )

  err = gdk_error_trap_pop ();
  if (err != Success ||
      result != Success)
    return FALSE;

  if (type != XA_CARDINAL)
    {
      XFree (nums);
      return FALSE;
    }

  *cardinals = g_new (gulong, nitems);
  memcpy (*cardinals, nums, sizeof (gulong) * nitems);
  *len = nitems;

  XFree (nums);

  return TRUE;
}

static char*
kiba_get_utf8_property (Display *display,
                        Window   xwindow,
                        Atom     atom)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  guchar *val;
  int result;
  char *retval = NULL;
  Atom utf8_string;

  utf8_string = kiba_get_atom (display, "UTF8_STRING");

  gdk_error_trap_push();
  type = None;
  val = NULL;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (display, xwindow, atom, 0, G_MAXLONG, False, utf8_string,
                                &type, &format, &nitems, &bytes_after, (guchar **)&val);  
  )
  gdk_error_trap_pop();

  if (result != Success)
  {
    if (retval != NULL)
      g_free (retval);
    return NULL;
  }

  if (type != utf8_string || format != 8 || nitems == 0)
  {
    if (val)
      XFree (val);
    if (retval != NULL)
      g_free (retval);
    return NULL;
  }

  if (!g_utf8_validate (val, nitems, NULL))
  {
    g_warning ("Property contained invalid UTF-8\n");
      XFree (val);
    if (retval != NULL)
      g_free (retval);
    return NULL;
  }
  retval = g_strndup (val, nitems);

  XFree (val);

  return retval;
}

static char**
kiba_get_utf8_list (Window   xwindow,
                    Atom     atom)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  char *val;
  int err, result;
  Atom utf8_string;
  char **retval;
  guint i;
  guint n_strings;
  char *p;
  
  utf8_string = kiba_get_atom (gdk_display, "UTF8_STRING");
  
  gdk_error_trap_push ();
  type = None;
  val = NULL;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (gdk_display,
			         xwindow,
			         atom,
			         0, G_MAXLONG,
			         False, utf8_string,
			         &type, &format, &nitems,
			         &bytes_after, (void*)&val);
  )
  err = gdk_error_trap_pop ();

  if (err != Success ||
      result != Success)
    return NULL;
  
  if (type != utf8_string ||
      format != 8 ||
      nitems == 0)
    {
      if (val)
        XFree (val);
      return NULL;
    }

  /* I'm not sure this is right, but I'm guessing the
   * property is nul-separated
   */
  i = 0;
  n_strings = 0;
  while (i < nitems)
    {
      if (val[i] == '\0')
        ++n_strings;
      ++i;
    }

  if (val[nitems - 1] != '\0')
    ++n_strings;

  /* we're guaranteed that val has a nul on the end
   * by XGetWindowProperty
   */
  
  retval = g_new0 (char*, n_strings + 1);

  p = val;
  i = 0;
  while (i < n_strings)
    {
      if (!g_utf8_validate (p, -1, NULL))
        {
          g_warning ("Property contained invalid UTF-8\n");
          XFree (val);
          g_strfreev (retval);
          return NULL;
        }

      retval[i] = g_strdup (p);
      
      p = p + strlen (p) + 1;
      ++i;
    }
  
  XFree (val);
  
  return retval;
}

static char*
text_property_to_utf8 (const XTextProperty *prop)
{
  char **list;
  int count;
  char *retval;

  list = NULL;

  count = gdk_text_property_to_utf8_list (gdk_x11_xatom_to_atom (prop->encoding), prop->format,
                                          prop->value, prop->nitems, &list);

  if (count == 0)
    return NULL;

  retval = list[0];
  list[0] = g_strdup (""); /* something to free */

  g_strfreev (list);

  return retval;
}

static char*
kiba_get_text_property (Display *display,
                        Window   xwindow,
                        Atom     atom)
{
  XTextProperty text;
  char *retval;
  int result;

  gdk_error_trap_push();
  text.nitems = 0;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetTextProperty (display, xwindow, &text, atom);
  )
  if (result)
    retval = text_property_to_utf8 (&text);
  else
    retval = NULL;
  gdk_error_trap_pop();

  if (text.value)
    XFree (text.value);

  return retval;
}

static char*
latin1_to_utf8 (const char *latin1)
{
  GString *str;
  const char *p;
  
  str = g_string_new (NULL);

  p = latin1;
  while (*p)
    {
      g_string_append_unichar (str, (gunichar) *p);
      ++p;
    }

  return g_string_free (str, FALSE);
}

static gboolean
kiba_get_atom_list (Window  xwindow,
                    Atom    atom,
                    Atom  **atoms,
                    int    *len)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  Atom *data;
  int err, result;

  *atoms = NULL;
  *len = 0;
  
  gdk_error_trap_push ();
  type = None;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (gdk_display,
			         xwindow,
			         atom,
			         0, G_MAXLONG,
			         False, XA_ATOM, &type, &format, &nitems,
			         &bytes_after, (void*)&data);
  )
  err = gdk_error_trap_pop ();
  if (err != Success ||
      result != Success)
    return FALSE;
  
  if (type != XA_ATOM)
    {
      XFree (data);
      return FALSE;
    }

  *atoms = g_new (Atom, nitems);
  memcpy (*atoms, data, sizeof (Atom) * nitems);
  *len = nitems;
  
  XFree (data);

  return TRUE;
}

Atom
kiba_get_atom (Display    *display,
               const char *atom_name)
{
  Atom retval;
  GHashTable *atom_hash = NULL, *reverse_atom_hash = NULL;
  g_return_val_if_fail (atom_name != NULL, None);

  atom_hash = g_hash_table_new (g_str_hash, g_str_equal);
  reverse_atom_hash = g_hash_table_new (NULL, NULL);

  retval = GPOINTER_TO_UINT (g_hash_table_lookup (atom_hash, atom_name));
  if (!retval)
  {
    KIBA_EXECUTE_IN_ERROR_TRAP (
      retval = XInternAtom (display, atom_name, FALSE);
    )

    if (retval != None)
    {
      char *name_copy;

      name_copy = g_strdup (atom_name);

      g_hash_table_insert (atom_hash, name_copy, GUINT_TO_POINTER (retval));
      g_hash_table_insert (reverse_atom_hash, GUINT_TO_POINTER (retval), name_copy);

      g_free (name_copy);
    }
  }

  g_hash_table_unref (atom_hash);
  g_hash_table_unref (reverse_atom_hash);

  return retval;
}

/*******
DESKTOP
*******/
/* returns TRUE if _NET_SHOWING_DESKTOP is set */
gboolean
kiba_get_show_desktop (void)
{
  Screen   *screen;
  int showing_desktop;

  showing_desktop = FALSE;
  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  kiba_get_cardinal (RootWindowOfScreen (screen),
                     kiba_get_atom (gdk_display, "_NET_SHOWING_DESKTOP"),
                    &showing_desktop);

  return showing_desktop != 0;
}

/* toggles _NET_SHOWING_DESKTOP */
void kiba_toggle_showing_desktop (GtkWidget *widget,
				  gpointer   data) {
	Screen *screen;
	XEvent xev;
	gboolean show = (gboolean)(GPOINTER_TO_INT (data));
	
	screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());
	
	xev.xclient.type = ClientMessage;
	xev.xclient.serial = 0;
	xev.xclient.send_event = True;
	xev.xclient.display = DisplayOfScreen (screen);
	xev.xclient.window = RootWindowOfScreen (screen);
	xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_SHOWING_DESKTOP");
	xev.xclient.format = 32;
	xev.xclient.data.l[0] = show != FALSE;
	xev.xclient.data.l[1] = 0;
	xev.xclient.data.l[2] = 0;
	xev.xclient.data.l[3] = 0;
	xev.xclient.data.l[4] = 0;
	
	KIBA_EXECUTE_IN_ERROR_TRAP (
		XSendEvent (DisplayOfScreen (screen),
		            RootWindowOfScreen (screen),
		            False,
		            SubstructureRedirectMask | SubstructureNotifyMask,
		            &xev);
	)
}

/* update _NET_WM_STRUT so that maximised windows doesnt overlap with the dock */
void kiba_set_wm_strut (Window     window,
			KibaStrut *strut) {
	long wm_strut[] = {
		strut->left, strut->right,
		strut->top,  strut->bottom
	};
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		XChangeProperty (gdk_display,
		                 window,
	 	                 kiba_get_atom (gdk_display, "_NET_WM_STRUT"),
		                 XA_CARDINAL,
		                 32,
		                 PropModeReplace,
		                 (guchar *) (&wm_strut),
		                 4);
	)
}

void kiba_set_wm_blur (Window xwindow) {
	long data[2];
	
	data[0] = 2; /* threshold */
	data[1] = 0; /* filter */
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		XChangeProperty (gdk_display,
		                 xwindow,
		                 kiba_get_atom (gdk_display, "_COMPIZ_WM_WINDOW_BLUR"),
		                 XA_INTEGER,
		                 32,
		                 PropModeReplace,
		                 (guchar *) data,
		                 2);
	)
}

void
kiba_focus_window (GtkWidget *widget,
                   gpointer   data)
{
  GdkWindow *window = data;
  Screen    *screen;
  XEvent     xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = gdk_x11_drawable_get_xid (window);
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_ACTIVE_WINDOW");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = 0;
  xev.xclient.data.l[1] = CurrentTime;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
                False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev); 
  )
}

void
kiba_iconify_window (GtkWidget *widget,
                     gpointer   data)
{
  GdkWindow *window = data;

  gdk_error_trap_push();
  KIBA_EXECUTE_IN_ERROR_TRAP (
    XIconifyWindow (gdk_display, gdk_x11_drawable_get_xid (window), DefaultScreen(gdk_display));
  )
  gdk_error_trap_pop();
}

void
kiba_deiconify_window (GtkWidget *widget,
                       gpointer   data)
{
  GdkWindow *window = data;

  gdk_error_trap_push();
  gdk_window_show (window);
  gdk_error_trap_pop();
}

void
kiba_unmaximize_window (GtkWidget *widget,
                        gpointer   data)
{
  GdkWindow *window = data;

  gdk_window_unmaximize (window);
}

void
kiba_maximize_window (GtkWidget *widget,
                      gpointer   data)
{
  GdkWindow *window = data;

  gdk_window_maximize (window);
}

void
kiba_set_window_above (GtkWidget *widget,
                       gpointer   data)
{
  GdkWindow *window = data;

  if (kiba_window_state_hint (window, "_NET_WM_STATE_ABOVE"))
  {
    gdk_window_set_keep_below (window, TRUE);
    gdk_window_set_keep_below (window, FALSE);
// this doesnt work, at least under metacity
//     gdk_window_set_keep_above (window, FALSE);
  }
  else
    gdk_window_set_keep_above (window, TRUE);
}

void
kiba_raise_window (GdkWindow *window) {
  gdk_window_raise(window);
}

void
kiba_set_window_below (GtkWidget *widget,
                       gpointer   data)
{
  GdkWindow *window = data;

  if (kiba_window_state_hint (window, "_NET_WM_STATE_BELOW"))
    gdk_window_set_keep_below (window, FALSE);
  else
    gdk_window_set_keep_below (window, TRUE);
}

static void kill_window_response (GtkDialog *dialog,
				  gint       response_id,
				  gpointer   user_data) {
	if (response_id == GTK_RESPONSE_ACCEPT) {
		Window window = (Window) user_data;
		
		gdk_error_trap_push ();
		KIBA_EXECUTE_IN_ERROR_TRAP (
			XKillClient (gdk_display, window);
		)
		gdk_flush ();
		gdk_error_trap_pop ();
	}

	gtk_widget_destroy (GTK_WIDGET (dialog));
}

/* From metacity */
void kill_window_question (gpointer window) {
	GtkWidget *dialog;
 
	dialog = gtk_message_dialog_new (NULL, 0,
					 GTK_MESSAGE_QUESTION,
					 GTK_BUTTONS_NONE,
					 _("Force this application to exit?\n"
					   "(Any open document will be lost.)"));
 
	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
				GTK_STOCK_CANCEL,
				GTK_RESPONSE_CANCEL,
				GTK_STOCK_OK,
				GTK_RESPONSE_ACCEPT,
				NULL);
 
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);

	g_signal_connect (dialog, "response",
			  G_CALLBACK (kill_window_response), window);

	gtk_widget_show (dialog);
}

static gint force_quit_timeout (gpointer data) {
	Window window;
	gint   pid = GPOINTER_TO_INT (data);
	
	window = kiba_get_win_for_pid (pid);
	if (window)
		kill_window_question ((gpointer) window);
	
	return FALSE;
}

void kiba_kill_window (GtkWidget *widget,
		       gpointer   data) {
	GdkWindow *window = data;
	Display   *display;
	Window     xwindow;
	XEvent     xev;
	gint       pid;
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		display = XOpenDisplay (NULL);
	)
	if (display == NULL){
		fprintf (stderr, "Unable to open display.\n");
		return;
	}
	
	xwindow = gdk_x11_drawable_get_xid (window);
	
	pid = kiba_get_win_pid (gdk_display, xwindow);
	
	xev.xclient.type = ClientMessage;
	xev.xclient.serial = 0;
	xev.xclient.send_event = True;
	xev.xclient.display = display;
	xev.xclient.window = xwindow;
	xev.xclient.message_type = kiba_get_atom (display, "_NET_CLOSE_WINDOW");
	xev.xclient.format = 32;
	xev.xclient.data.l[0] = 0;
	xev.xclient.data.l[1] = 0;
	xev.xclient.data.l[2] = 0;
	xev.xclient.data.l[3] = 0;
	xev.xclient.data.l[4] = 0;
	
	if (pid)
		g_timeout_add (5000, force_quit_timeout, GINT_TO_POINTER(pid));
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		XSendEvent (display, gdk_x11_get_default_root_xwindow (), False,
			SubstructureRedirectMask | SubstructureNotifyMask, &xev);
	)
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		XCloseDisplay (display);
	)
}

void
kiba_window_stick (GtkWidget *widget,
                   gpointer   data)
{
  GdkWindow *window = data;

  if (kiba_window_state_hint (window, "_NET_WM_STATE_STICKY"))
   gdk_window_unstick (window);
  else
   gdk_window_stick (window);
}

char* kiba_win_name (Display *display, Window window) {
	char *window_name,
             *ret;
	
	window_name = kiba_get_win_name (display, window);
	if (window_name == NULL)
		window_name = kiba_get_wmclass (display, window);
	
	if (window_name == NULL)
		return NULL;
	
	if (strlen (window_name) > 27) {
		char *cutted;
	
		cutted = g_strndup (window_name, 27);
		ret = g_strdup_printf ("%s...", cutted);
		g_free (cutted);
	}
	else
		ret = g_strdup (window_name);
	g_free (window_name);
	
	return ret;
}

char*
kiba_get_wintype (Display   *display,
                  GdkWindow *window)
{
  Window              xwindow;
  Atom                type_return, *type_atom;
  int                 format_return;
  unsigned long int   nitems_return;
  unsigned long int   bytes_after_return;
  unsigned char      *state = NULL;
  char               *state_string = NULL;
  int                 result;

  xwindow    = gdk_x11_drawable_get_xid (window);

  gdk_error_trap_push();
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (display, xwindow, kiba_get_atom (display,"_NET_WM_WINDOW_TYPE"),
        0, G_MAXLONG, False, AnyPropertyType, &type_return, &format_return, &nitems_return,
        &bytes_after_return, &state);
  )
  if (result == Success)
  {
    if (state != NULL && state != None)
    {
      type_atom = (Atom *)state;
      if (type_atom) {
        KIBA_EXECUTE_IN_ERROR_TRAP (
          state_string = XGetAtomName (display, *type_atom);
        )
      }
    }
  }

  if (state != NULL)
    g_free (state);

  if (state_string == NULL)
    state_string = g_strdup ("_NET_WM_WINDOW_TYPE_NORMAL");

  return state_string;
}

GList*
kiba_get_window_stack (Display *display)
{
  Screen *screen_x11;
  GList *ret = NULL;
  Atom type_return;
  gint format_return;
  gulong nitems_return;
  gulong bytes_after_return;
  guchar *data = NULL;
  GdkScreen *screen;
  int result;

  screen = gdk_screen_get_default ();

  g_return_val_if_fail (GDK_IS_SCREEN (screen), NULL);

  if (!gdk_x11_screen_supports_net_wm_hint (screen,
                                            gdk_atom_intern ("_NET_CLIENT_LIST_STACKING", FALSE)))
    return NULL;

  screen_x11 = gdk_x11_screen_get_xscreen (screen);

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (DisplayOfScreen(screen_x11), RootWindowOfScreen(screen_x11),
                       gdk_x11_get_xatom_by_name_for_display (gdk_screen_get_display (screen),
                                                    "_NET_CLIENT_LIST_STACKING"),
                  0, G_MAXLONG, False, XA_WINDOW, &type_return,
                  &format_return, &nitems_return,
                            &bytes_after_return, &data);
  )
  if (result == Success)
    {
      if ((type_return == XA_WINDOW) && (format_return == 32) &&
          (data) && (nitems_return > 0))
        {
          GdkNativeWindow *stack = (GdkNativeWindow *) data;
          GdkWindow *win;
          int i;

          for (i = 0; i < nitems_return; i++)
            {
              win = gdk_window_foreign_new_for_display (gdk_screen_get_display (screen),
                                                        stack[i]);

              if (win != NULL)
                ret = g_list_append (ret, win);
            }
        }
    }

  if (data)
    XFree (data);

  return ret;
}

char*
kiba_get_win_name (Display *display,
                   Window   xwindow)
{
  char *name = NULL;

  name = kiba_get_utf8_property (display, xwindow, kiba_get_atom (display, "_NET_WM_VISIBLE_NAME"));

  if (name == NULL)
    name = kiba_get_utf8_property (display, xwindow, kiba_get_atom (display, "_NET_WM_NAME"));

  if (name == NULL)
    name = kiba_get_text_property (display, xwindow, XA_WM_NAME);

  return name;
}

/* returns current windowstate:
 * -1=not running
 * WithdrawnState = systray/not mapped
 * NormalState    = shown
 * IconicState    = iconified
 */
int
kiba_get_winstate (Display   *display,
                   GdkWindow *window)
{
  Window            xwindow;
  Atom              type_return;
  int               format_return;
  unsigned long int nitems_return;
  unsigned long int bytes_after_return;
  unsigned char    *state = NULL;
  int               result;

  xwindow    = gdk_x11_drawable_get_xid (window);

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (display, xwindow, kiba_get_atom (display, "WM_STATE"),
         0, G_MAXLONG, False, AnyPropertyType, &type_return, &format_return,
         &nitems_return, &bytes_after_return, &state);
  )

  if (result == Success && state != NULL)
  {
    int winstate;

    winstate = (int)(((unsigned long *) state)[0]);
    g_free (state);
    return winstate;
  }

  if (state != NULL)
    g_free (state);

  return -1;
}

Window
kiba_get_active_window (Display *display)
{
  Window  window = 0;
  Atom    type_return;
  gint    format_return;
  gulong  nitems_return;
  gulong  bytes_after_return;
  guchar *data = NULL;
  Window rootWindow;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    Atom active_atom = XInternAtom (display, "_NET_ACTIVE_WINDOW", TRUE);
  )
  
  KIBA_EXECUTE_IN_ERROR_TRAP (
    rootWindow = XDefaultRootWindow (display);
  )
  
  KIBA_EXECUTE_IN_ERROR_TRAP (
    int result = XGetWindowProperty (display, rootWindow, active_atom,
          0, 1, False, XA_WINDOW, &type_return, &format_return, &nitems_return,
          &bytes_after_return, &data);
  )

  if (result == Success)
  {
    if ((type_return == XA_WINDOW) && (format_return == 32) && (data))
      window = *(Window *) data;
  }

  if (data)
    XFree (data);

  return window;
}

char*
kiba_get_wmclass (Display *display,
                  Window   window)
{
  XClassHint ch;
  char *retval;
  char *name = NULL, *class = NULL;

  gdk_error_trap_push ();

  ch.res_name = NULL;
  ch.res_class = NULL;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XGetClassHint (gdk_display, window, &ch);
  )

  gdk_error_trap_pop ();

  retval = NULL;


  if (ch.res_name)
  {
    // Why aren't we getting the name if we aren't doing anything with it
    name = latin1_to_utf8 (ch.res_name);

    XFree (ch.res_name);
  }

  if (ch.res_class)
  {
    class = latin1_to_utf8 (ch.res_class);

    XFree (ch.res_class);
  }

  g_free(name);
  return class;
}

gboolean kiba_window_urgency (Display  *display, GdkWindow *window) {
	XWMHints *wm_hints;
	int      urgency;
	
	KIBA_EXECUTE_IN_ERROR_TRAP (	
		wm_hints = XGetWMHints (display,  gdk_x11_drawable_get_xid (window));
	)
	
	if (wm_hints)
		urgency = wm_hints->flags & XUrgencyHint;
	else
		urgency = 0;
	
	XFree(wm_hints);
	
	return urgency;
}

gboolean
kiba_window_state_hint (GdkWindow  *window,
                        const char *state_hint)
{
  Window xwindow;
  Atom *atoms;
  int n_atoms;
  int i;

  atoms = NULL;
  n_atoms = 0;
  xwindow = gdk_x11_drawable_get_xid (window);
  kiba_get_atom_list (xwindow, kiba_get_atom (gdk_display, "_NET_WM_STATE"), &atoms, &n_atoms);

  i = 0;
  while (i < n_atoms)
  {
    if (atoms[i] == kiba_get_atom (gdk_display, state_hint))
    {
      g_free (atoms);
      return TRUE;
    }

    ++i;
  }

  g_free (atoms);

  return FALSE;
}

/* set icon geometry (for taskbar animation) */
void
kiba_set_icon_geometry  (Window    xwindow,
                         int       x,
                         int       y,
                         int       width,
                         int       height)
{
  gulong data[4];

  data[0] = x;
  data[1] = y;

  data[2] = width;
  data[3] = height;

  gdk_error_trap_push ();

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XChangeProperty (gdk_display, xwindow,
		     kiba_get_atom (gdk_display, "_NET_WM_ICON_GEOMETRY"),
		     XA_CARDINAL, 32, PropModeReplace,
		     (guchar *)&data, 4);
  )

  gdk_error_trap_pop ();
}

/* returns PID for given Window */
int kiba_get_win_pid (Display *display,
		      Window   window) {
	int val;
	
	if (!kiba_get_cardinal (window, kiba_get_atom (display, "_NET_WM_PID"), &val))
		return 0;
	
	return val;
}

/* returns Window for the given pid or 0 if there is no window with this pid */
Window kiba_get_win_for_pid (int pid) {
	GList *root_stack, *l;
	
	root_stack = (GList*) kiba_get_window_stack (gdk_display);
	
	for (l = root_stack; l != NULL; l = l->next) {
		GdkWindow *window = l->data;
		Window     xwindow;
		int        cpid;
		
		xwindow = gdk_x11_drawable_get_xid (window);
		
		cpid = kiba_get_win_pid (gdk_display, xwindow);
		
		if (cpid == pid)
			return xwindow;
	}
	g_list_free (root_stack);
	
	return 0;
}

/* The icon-reading code is copied
 * from metacity, please sync bugfixes
 */
static gboolean
find_largest_sizes (gulong *data,
                    gulong  nitems,
                    int    *width,
                    int    *height)
{
  *width = 0;
  *height = 0;

  while (nitems > 0)
    {
      int w, h;
      gboolean replace;

      replace = FALSE;

      if (nitems < 3)
        return FALSE; /* no space for w, h */

      w = data[0];
      h = data[1];

      if (nitems < ((w * h) + 2))
        return FALSE; /* not enough data */

      *width = MAX (w, *width);
      *height = MAX (h, *height);

      data += (w * h) + 2;
      nitems -= (w * h) + 2;
    }

  return TRUE;
}

static void
free_pixels (guchar  *pixels,
             gpointer data)
{
  g_free (pixels);
}

static gboolean
find_best_size (gulong  *data,
                gulong   nitems,
                int      ideal_width,
                int      ideal_height,
                int     *width,
                int     *height,
                gulong **start)
{
  int best_w;
  int best_h;
  gulong *best_start;
  int max_width, max_height;

  *width = 0;
  *height = 0;
  *start = NULL;

  if (!find_largest_sizes (data, nitems, &max_width, &max_height))
    return FALSE;

  if (ideal_width < 0)
    ideal_width = max_width;
  if (ideal_height < 0)
    ideal_height = max_height;

  best_w = 0;
  best_h = 0;
  best_start = NULL;

  while (nitems > 0)
  {
    int w, h;
    gboolean replace;

    replace = FALSE;

    if (nitems < 3)
      return FALSE; /* no space for w, h */

    w = data[0];
    h = data[1];

    if (nitems < ((w * h) + 2))
      break; /* not enough data */

    if (best_start == NULL)
    {
      replace = TRUE;
    }
    else
    {
      /* work with averages */
      const int ideal_size = (ideal_width + ideal_height) / 2;
      int best_size = (best_w + best_h) / 2;
      int this_size = (w + h) / 2;

      /* larger than desired is always better than smaller */
      if (best_size < ideal_size && this_size >= ideal_size)
        replace = TRUE;
      /* if we have too small, pick anything bigger */
      else if (best_size < ideal_size && this_size > best_size)
        replace = TRUE;
      /* if we have too large, pick anything smaller
       * but still >= the ideal
       */
      else if (best_size > ideal_size && this_size >= ideal_size && this_size < best_size)
        replace = TRUE;
    }

    if (replace)
    {
      best_start = data + 2;
      best_w = w;
      best_h = h;
    }

    data += (w * h) + 2;
    nitems -= (w * h) + 2;
  }

  if (best_start)
  {
    *start = best_start;
    *width = best_w;
    *height = best_h;
    return TRUE;
  }
  else
    return FALSE;
}

static void
argbdata_to_pixdata (gulong  *argb_data,
                     int      len,
                     guchar **pixdata)
{
  guchar *p;
  int i;

  *pixdata = g_new (guchar, len * 4);
  p = *pixdata;

  /* One could speed this up a lot. */
  i = 0;
  while (i < len)
  {
    guint argb;
    guint rgba;

    argb = argb_data[i];
    rgba = (argb << 8) | (argb >> 24);

    *p = rgba >> 24;
    ++p;
    *p = (rgba >> 16) & 0xff;
    ++p;
    *p = (rgba >> 8) & 0xff;
    ++p;
    *p = rgba & 0xff;
    ++p;

    ++i;
  }
}

static gboolean
read_rgb_icon (Window    xwindow,
               int       ideal_width,
               int       ideal_height,
               int       ideal_mini_width,
               int       ideal_mini_height,
               int      *width,
               int      *height,
               guchar  **pixdata,
               int      *mini_width,
               int      *mini_height,
               guchar  **mini_pixdata)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  int result, err;
  gulong *data;
  gulong *best;
  int w, h;
  gulong *best_mini;
  int mini_w, mini_h;

  gdk_error_trap_push ();
  type = None;
  data = NULL;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (gdk_display,
			         xwindow,
			         kiba_get_atom (gdk_display, "_NET_WM_ICON"),
			         0, G_MAXLONG,
			         False, XA_CARDINAL, &type, &format, &nitems,
			         &bytes_after, (void*)&data);
  )

  err = gdk_error_trap_pop ();

  if (err != Success ||
      result != Success)
    return FALSE;

  if (type != XA_CARDINAL)
  {
    XFree (data);
    return FALSE;
  }

  if (!find_best_size (data, nitems,
                       ideal_width, ideal_height,
                       &w, &h, &best))
  {
    XFree (data);
    return FALSE;
  }

  if (!find_best_size (data, nitems,
                       ideal_mini_width, ideal_mini_height,
                       &mini_w, &mini_h, &best_mini))
  {
    XFree (data);
    return FALSE;
  }

  *width = w;
  *height = h;

  *mini_width = mini_w;
  *mini_height = mini_h;

  argbdata_to_pixdata (best, w * h, pixdata);
  argbdata_to_pixdata (best_mini, mini_w * mini_h, mini_pixdata);

  XFree (data);

  return TRUE;
}

static void
kiba_get_pixmap_geometry (Pixmap       pixmap,
                          int         *w,
                          int         *h,
                          int         *d)
{
  Window root_ignored;
  int x_ignored, y_ignored;
  guint width, height;
  guint border_width_ignored;
  guint depth;

  if (w)
    *w = 1;
  if (h)
    *h = 1;
  if (d)
    *d = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XGetGeometry (gdk_display,
                  pixmap, &root_ignored, &x_ignored, &y_ignored,
                  &width, &height, &border_width_ignored, &depth);
  )

  if (w)
    *w = width;
  if (h)
    *h = height;
  if (d)
    *d = depth;
}

static GdkPixbuf*
apply_mask (GdkPixbuf *pixbuf,
            GdkPixbuf *mask)
{
  int w, h;
  int i, j;
  GdkPixbuf *with_alpha;
  guchar *src;
  guchar *dest;
  int src_stride;
  int dest_stride;

  w = MIN (gdk_pixbuf_get_width (mask), gdk_pixbuf_get_width (pixbuf));
  h = MIN (gdk_pixbuf_get_height (mask), gdk_pixbuf_get_height (pixbuf));

  with_alpha = gdk_pixbuf_add_alpha (pixbuf, FALSE, 0, 0, 0);

  dest = gdk_pixbuf_get_pixels (with_alpha);
  src = gdk_pixbuf_get_pixels (mask);

  dest_stride = gdk_pixbuf_get_rowstride (with_alpha);
  src_stride = gdk_pixbuf_get_rowstride (mask);

  i = 0;
  while (i < h)
  {
    j = 0;
    while (j < w)
    {
      guchar *s = src + i * src_stride + j * 3;
      guchar *d = dest + i * dest_stride + j * 4;

      if (s[0] == 0)
        d[3] = 0;   /* transparent */
      else
        d[3] = 255; /* opaque */

      ++j;
    }

    ++i;
  }

  return with_alpha;
}

static GdkColormap*
get_cmap (GdkPixmap *pixmap)
{
  GdkColormap *cmap;

  cmap = gdk_drawable_get_colormap (pixmap);
  if (cmap)
    g_object_ref (G_OBJECT (cmap));

  if (cmap == NULL)
  {
    if (gdk_drawable_get_depth (pixmap) == 1)
    {
      /* try null cmap */
      cmap = NULL;
    }
    else
    {
      /* Try system cmap */
      GdkScreen *screen = gdk_drawable_get_screen (GDK_DRAWABLE (pixmap));
      cmap = gdk_screen_get_system_colormap (screen);
      g_object_ref (G_OBJECT (cmap));
    }
  }

  /* Be sure we aren't going to blow up due to visual mismatch */
  if (cmap && (gdk_colormap_get_visual (cmap)->depth != gdk_drawable_get_depth (pixmap)))
    cmap = NULL;

  return cmap;
}

static GdkPixbuf*
kiba_gdk_pixbuf_get_from_pixmap (GdkPixbuf   *dest,
                                  Pixmap      xpixmap,
                                  int         src_x,
                                  int         src_y,
                                  int         dest_x,
                                  int         dest_y,
                                  int         width,
                                  int         height)
{
  GdkDrawable *drawable;
  GdkPixbuf *retval;
  GdkColormap *cmap;

  retval = NULL;
  cmap = NULL;

  drawable = gdk_xid_table_lookup (xpixmap);

  if (drawable)
    g_object_ref (G_OBJECT (drawable));
  else
    drawable = gdk_pixmap_foreign_new (xpixmap);

  if (drawable)
  {
    cmap = get_cmap (drawable);

    if (width < 0)
      gdk_drawable_get_size (drawable, &width, NULL);
    if (height < 0)
      gdk_drawable_get_size (drawable, NULL, &height);

    retval = gdk_pixbuf_get_from_drawable (dest, drawable, cmap, src_x, src_y, dest_x, dest_y, width, height);
  }

  if (cmap)
    g_object_unref (G_OBJECT (cmap));
  if (drawable)
    g_object_unref (G_OBJECT (drawable));

  return retval;
}

static GdkPixbuf*
scaled_from_pixdata (guchar *pixdata,
                     int     w,
                     int     h,
                     int     new_w,
                     int     new_h)
{
  GdkPixbuf *src;
  GdkPixbuf *dest;

  src = gdk_pixbuf_new_from_data (pixdata, GDK_COLORSPACE_RGB,
                                  TRUE, 8, w, h, w * 4, free_pixels, NULL);

  if (src == NULL)
    return NULL;

  if (w != h)
  {
    GdkPixbuf *tmp;
    int size;

    size = MAX (w, h);

    tmp = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, size, size);

    if (tmp != NULL)
    {
      gdk_pixbuf_fill (tmp, 0);
      gdk_pixbuf_copy_area (src, 0, 0, w, h, tmp, (size - w) / 2, (size - h) / 2);

      g_object_unref (src);
      src = tmp;
    }
  }

  if (w != new_w || h != new_h)
  {
    dest = gdk_pixbuf_scale_simple (src, new_w, new_h, GDK_INTERP_BILINEAR);

    g_object_unref (G_OBJECT (src));
  }
  else
    dest = src;

  return dest;
}

static gboolean
try_pixmap_and_mask (Pixmap      src_pixmap,
                     Pixmap      src_mask,
                     GdkPixbuf **iconp,
                     int         ideal_width,
                     int         ideal_height,
                     GdkPixbuf **mini_iconp,
                     int         ideal_mini_width,
                     int         ideal_mini_height)
{
  GdkPixbuf *unscaled = NULL;
  GdkPixbuf *mask = NULL;
  int w, h;

  if (src_pixmap == None)
    return FALSE;

  gdk_error_trap_push ();

  kiba_get_pixmap_geometry (src_pixmap, &w, &h, NULL);

  unscaled = kiba_gdk_pixbuf_get_from_pixmap (NULL, src_pixmap, 0, 0, 0, 0, w, h);

  if (unscaled && src_mask != None)
  {
    kiba_get_pixmap_geometry (src_mask, &w, &h, NULL);
    mask = kiba_gdk_pixbuf_get_from_pixmap (NULL, src_mask, 0, 0, 0, 0, w, h);
  }

  gdk_error_trap_pop ();

  if (mask)
  {
    GdkPixbuf *masked;

    masked = apply_mask (unscaled, mask);
    g_object_unref (G_OBJECT (unscaled));
    unscaled = masked;

    g_object_unref (G_OBJECT (mask));
    mask = NULL;
  }

  if (unscaled)
  {
    *iconp = gdk_pixbuf_scale_simple (unscaled,
                                 ideal_width > 0 ? ideal_width :
                                 gdk_pixbuf_get_width (unscaled),
                                 ideal_height > 0 ? ideal_height :
                                 gdk_pixbuf_get_height (unscaled),
                                 GDK_INTERP_BILINEAR);
    *mini_iconp = gdk_pixbuf_scale_simple (unscaled,
                                 ideal_mini_width > 0 ? ideal_mini_width :
                                 gdk_pixbuf_get_width (unscaled),
                                 ideal_mini_height > 0 ? ideal_mini_height :
                                 gdk_pixbuf_get_height (unscaled),
                                 GDK_INTERP_BILINEAR);      

    g_object_unref (G_OBJECT (unscaled));
    return TRUE;
  }
  else
    return FALSE;
}

static void
kiba_get_kwm_win_icon (Window  xwindow,
                       Pixmap *pixmap,
                       Pixmap *mask)
{
  Atom type;
  int format;
  gulong nitems;
  gulong bytes_after;
  Pixmap *icons;
  int err, result;

  *pixmap = None;
  *mask = None;

  gdk_error_trap_push ();
  icons = NULL;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty (gdk_display, xwindow,
			         kiba_get_atom (gdk_display, "KWM_WIN_ICON"),
			         0, G_MAXLONG,
			         False,
			         kiba_get_atom (gdk_display, "KWM_WIN_ICON"),
			         &type, &format, &nitems,
			         &bytes_after, (void*)&icons);
  )

  err = gdk_error_trap_pop ();
  if (err != Success ||
      result != Success)
    return;

  if (type != kiba_get_atom (gdk_display, "KWM_WIN_ICON"))
  {
    XFree (icons);
    return;
  }

  *pixmap = icons[0];
  *mask = icons[1];

  XFree (icons);

  return;
}

static gboolean
kiba_x11_load_icon (Window       xwindow,
                GtkWidget   *icon_cache,
                GdkPixbuf  **iconp,
                int          ideal_size,
                GdkPixbuf  **mini_iconp,
                int          ideal_mini_width,
                int          ideal_mini_height)
{
  guchar *pixdata;
  int w, h;
  guchar *mini_pixdata;
  int mini_w, mini_h;
  Pixmap pixmap;
  Pixmap mask;
  XWMHints *hints;

  *iconp = NULL;
  *mini_iconp = NULL;

  pixdata = NULL;

  if (read_rgb_icon (xwindow,
                     ideal_size, ideal_size,
                     ideal_mini_width, ideal_mini_height,
                    &w, &h, &pixdata,
                    &mini_w, &mini_h, &mini_pixdata))
  {
    *iconp = scaled_from_pixdata (pixdata, w, h, ideal_size, ideal_size);

    *mini_iconp = scaled_from_pixdata (mini_pixdata, mini_w, mini_h,
                                       ideal_mini_width, ideal_mini_height);

     return TRUE;
  }

  gdk_error_trap_push ();
  KIBA_EXECUTE_IN_ERROR_TRAP (
    hints = XGetWMHints (gdk_display, xwindow);
  )
  gdk_error_trap_pop ();
  pixmap = None;
  mask = None;
  if (hints)
  {
    if (hints->flags & IconPixmapHint)
      pixmap = hints->icon_pixmap;
    if (hints->flags & IconMaskHint)
      mask = hints->icon_mask;

    XFree (hints);
    hints = NULL;
  }


  if (try_pixmap_and_mask (pixmap, mask, iconp, ideal_size, ideal_size,
                           mini_iconp, ideal_mini_width, ideal_mini_height))
    return TRUE;

  kiba_get_kwm_win_icon (xwindow, &pixmap, &mask);

  if (try_pixmap_and_mask (pixmap, mask, iconp, ideal_size, ideal_size,
                           mini_iconp, ideal_mini_width, ideal_mini_height))
    return TRUE;

  return FALSE;
}

GdkPixbuf *
kiba_get_icon (Window    window,
               int       size)
{
  GdkPixbuf *icon = NULL;
  GdkPixbuf *mini_icon = NULL;
  char      *icon_path;

  if (kiba_x11_load_icon (window, NULL, &icon, size, &mini_icon, 64, 64) && icon)
    return icon;

  icon_path = kiba_get_wmclass (gdk_display, window);
  if (icon_path && kiba_load_icon (icon_path, size, &icon) && icon)
  {
    g_free (icon_path);

    return icon;
  }

  g_free (icon_path);

  return NULL;
}

/* checks if return data of _NET_WM_ICON is equal to icon_pixmap */
gboolean
kiba_icon_changed (Window  window,
                   Pixmap *icon_pixmap)
{
  Pixmap mask;
  XWMHints *hints;

  gdk_error_trap_push ();
  KIBA_EXECUTE_IN_ERROR_TRAP (
    hints = XGetWMHints (gdk_display, window);
  )
  gdk_error_trap_pop ();

  mask = None;
  if (hints)
  {
    if (hints->flags & IconMaskHint)
      mask = hints->icon_mask;

    XFree (hints);
    hints = NULL;
  }

  if (*icon_pixmap != mask)
  {
    *icon_pixmap = mask;
    return TRUE;
  }

  return FALSE;
}


/**************
WORKSPACES
***************/

int
kiba_num_workspaces ()
{
  int n_spaces;
  Window result;

  n_spaces = 0;
  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XDefaultRootWindow (gdk_display);
  )
  kiba_get_cardinal (result, kiba_get_atom (gdk_display, "_NET_NUMBER_OF_DESKTOPS"), &n_spaces);

  return n_spaces;
}

int
kiba_get_active_workspace_number (Display *display)
{
  static Atom _NET_CURRENT_DESKTOP = None;
  static Atom _WIN_WORKSPACE = None;
  Atom type_ret = None;
  int format_ret = 0;
  unsigned long nitems_ret = 0;
  unsigned long bytes_after_ret = 0;
  unsigned int *prop_ret = NULL;
  gint ws_num = 0;
  int result;
  Window rootWindow;

  if (G_UNLIKELY (!_NET_CURRENT_DESKTOP))
  {
    KIBA_EXECUTE_IN_ERROR_TRAP (
      _NET_CURRENT_DESKTOP = XInternAtom (display, "_NET_CURRENT_DESKTOP", False);
    )

    KIBA_EXECUTE_IN_ERROR_TRAP (
      _WIN_WORKSPACE = XInternAtom (display, "_WIN_WORKSPACE", False);
    )
  }

  gdk_error_trap_push();
  KIBA_EXECUTE_IN_ERROR_TRAP (
    rootWindow = XDefaultRootWindow (display);
  )

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty(display, rootWindow, _NET_CURRENT_DESKTOP,
                                  0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
                                  &bytes_after_ret, (unsigned char **)&prop_ret);
  )

  if(result != Success)
  {
    KIBA_EXECUTE_IN_ERROR_TRAP (
      rootWindow = XDefaultRootWindow (display);
    )

    KIBA_EXECUTE_IN_ERROR_TRAP (
      result = XGetWindowProperty(display, rootWindow, _WIN_WORKSPACE,
          0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
          &bytes_after_ret, (unsigned char **)&prop_ret);
    )

    if(result != Success)
      return 0;
  }

  gdk_error_trap_pop();

  if (type_ret == None || format_ret == 0 || !prop_ret)
  {
    if(prop_ret)
      XFree(prop_ret);
    return 0;
  }

  ws_num = *prop_ret;
  XFree(prop_ret);

  return ws_num;
}

char*
kiba_get_active_workspace_name ()
{
  Screen *screen;
  char  **names;
  char   *ret;
  int     active_workspace;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());
  names  = kiba_get_utf8_list (RootWindowOfScreen (screen),
                               kiba_get_atom (gdk_display, "_NET_DESKTOP_NAMES"));

  active_workspace = kiba_get_active_workspace_number (gdk_display);

  if (names && names[active_workspace] && strlen (names[active_workspace]) > 1)
  {
    ret = g_strdup (names[active_workspace]);
    g_strfreev (names);
  }
  else
    ret = NULL;

  g_strfreev (names);

  return ret;
}

int
kiba_get_win_workspace_number (Display *display,
                               Window   window)
{
  static Atom _NET_WM_DESKTOP = None;
  static Atom _WIN_WORKSPACE = None;
  Atom type_ret = None;
  int format_ret = 0;
  unsigned long nitems_ret = 0;
  unsigned long bytes_after_ret = 0;
  unsigned int *prop_ret = NULL;
  gint ws_num = 0;
  int result;

  if (G_UNLIKELY (!_NET_WM_DESKTOP))
  {
    KIBA_EXECUTE_IN_ERROR_TRAP (
      _NET_WM_DESKTOP = XInternAtom (display, "_NET_WM_DESKTOP", False);
    )

    KIBA_EXECUTE_IN_ERROR_TRAP (
      _WIN_WORKSPACE = XInternAtom (display, "_WIN_WORKSPACE", False);
    )
  }

  gdk_error_trap_push();

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty(display, window, _NET_WM_DESKTOP,
            0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
            &bytes_after_ret, (unsigned char **)&prop_ret);
  )

  if(result != Success)
  {
    KIBA_EXECUTE_IN_ERROR_TRAP (
      result = XGetWindowProperty(display, window, _WIN_WORKSPACE,
         0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
         &bytes_after_ret, (unsigned char **)&prop_ret);
    )

    if(result != Success)
      return 0;
  }
  gdk_error_trap_pop();

  if (type_ret == None || format_ret == 0 || !prop_ret)
  {
    if(prop_ret)
      XFree(prop_ret);
    return 0;
  }

  ws_num = *prop_ret;
  XFree(prop_ret);

  return ws_num;
}

void
kiba_activate_workspace (int new_active_space)
{
  Screen  *screen;
  XEvent   xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = RootWindowOfScreen (screen);
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_CURRENT_DESKTOP");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = new_active_space;
  xev.xclient.data.l[1] = CurrentTime;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
                False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev);
  )
}

void
kiba_change_workspace (Window xwindow,
                       int    new_space)
{
  Screen  *screen;
  XEvent   xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_DESKTOP");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = new_space;
  xev.xclient.data.l[1] = CurrentTime;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display, gdk_x11_get_default_root_xwindow (), False,
                SubstructureRedirectMask | SubstructureNotifyMask, &xev);
  )
}

void
kiba_window_to_right_workspace (GtkWidget *widget,
                                gpointer   data)
{
  GdkWindow *window = (GdkWindow*) data;
  int      workspace_number;

  workspace_number = kiba_get_win_workspace_number (gdk_display, gdk_x11_drawable_get_xid (window));

  kiba_change_workspace (gdk_x11_drawable_get_xid (window), workspace_number + 1);
}

void
kiba_window_to_left_workspace (GtkWidget *widget,
                               gpointer   data)
{
  GdkWindow *window = (GdkWindow*) data;
  int      workspace_number;

  workspace_number = kiba_get_win_workspace_number (gdk_display, gdk_x11_drawable_get_xid (window));

  kiba_change_workspace (gdk_x11_drawable_get_xid (window), workspace_number - 1);
}

void
kiba_window_change_workspace (GtkWidget *widget,
                              gpointer   data)
{
  GdkWindow  *window = (GdkWindow*) data;
  char   *widget_name;

  widget_name = (char*) gtk_widget_get_name (widget);

  kiba_change_workspace (gdk_x11_drawable_get_xid (window), atoi (widget_name) - 1);
  g_free (widget_name);
}

/**************
VIEWPORTS
***************/
void
kiba_get_num_viewports (int *x,
                        int *y)
{
  GdkScreen  *screen;
  int space_width, space_height;
  int n_spaces;
  gulong *p_coord;
  int     n_coord;

  screen   = gdk_screen_get_default ();
  n_spaces = kiba_num_workspaces ();

  /* If no property, use the screen's size */
  space_width  = gdk_screen_get_width (screen);
  space_height = gdk_screen_get_height (screen);

  if (kiba_get_cardinal_list (gdk_x11_get_default_root_xwindow (),
        kiba_get_atom (gdk_display, "_NET_DESKTOP_GEOMETRY"), &p_coord, &n_coord) && p_coord != NULL)
  {
    if (n_coord == 2*n_spaces)
    {
      space_width = p_coord[0];
      space_height = p_coord[1];

      if (space_width < gdk_screen_get_width (screen))
        space_width = gdk_screen_get_width (screen);

      if (space_height < gdk_screen_get_height (screen))
        space_height = gdk_screen_get_height (screen);
    }

    g_free (p_coord);
  }

  *x = space_width / gdk_screen_get_width (screen);
  *y = space_height / gdk_screen_get_height (screen);
}

void
kiba_change_win_viewport (Window xwindow,
                          int    x,
                          int    y)
{
  GdkScreen *screen;
  XWindowAttributes attr;
  int win_x, win_y, viewport_x, viewport_y, dest_x, dest_y;
  Status result;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowAttributes(gdk_display, xwindow, &attr);
  )

  if (!result)
    return;

  kiba_get_win_viewport_number (xwindow, &viewport_x, &viewport_y);

  screen = gdk_screen_get_default ();
  win_x  = attr.x;
  win_y  = attr.y;
  dest_x = (x - viewport_x + 1) * gdk_screen_get_width (screen);
  dest_y = (y - viewport_y + 1) * gdk_screen_get_height (screen);

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XMoveWindow (gdk_display, xwindow, win_x + dest_x, win_y + dest_y);
  )
}

void
kiba_change_viewport (int x,
                      int y)
{
  Screen *screen;
  XEvent xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = RootWindowOfScreen (screen);
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_DESKTOP_VIEWPORT");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = x;
  xev.xclient.data.l[1] = y;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
                False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev);
  )
}

void
kiba_get_avtive_viewport (int *x,
                          int *y)
{
  GdkScreen *screen;
  unsigned long data[2];
  Atom actual;
  int result, format, width, height;
  unsigned long n, left;
  unsigned char *propData;
  int hsize, vsize;

  screen = gdk_screen_get_default ();

  width  = gdk_screen_get_width (screen);
  height = gdk_screen_get_height (screen);

  kiba_get_num_viewports (&hsize, &vsize);
  hsize += 1;
  vsize += 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowProperty(gdk_display, gdk_x11_get_default_root_xwindow (),
                                kiba_get_atom (gdk_display, "_NET_DESKTOP_VIEWPORT"), 0L, 2L,
                                FALSE, XA_CARDINAL, &actual, &format, &n, &left, &propData);
  )

  if (result == Success && n && propData)
  {
    if (n == 2)
    {
      memcpy(data, propData, sizeof(unsigned long) * 2);

      if (data[0] / width < hsize - 1)
        *x = data[0] / width;

      if (data[1] / height < vsize - 1)
        *y = data[1] / height;
    }

    XFree(propData);
  }
}

void
kiba_get_win_viewport_number (Window window,
                              int   *x,
                              int   *y)
{
  GdkScreen *screen;
  XWindowAttributes attr;
  int width, height, serverX, serverY, active_x, active_y;
  int result;

  *x = 1;
  *y = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowAttributes(gdk_display, window, &attr);
  )

  if (!result)
    return;

  kiba_get_avtive_viewport (&active_x, &active_y);
  *x = *x + active_x;
  *y = *y + active_y;

  screen  = gdk_screen_get_default ();
  width   = gdk_screen_get_width (screen);
  height  = gdk_screen_get_height (screen);
  serverX = attr.x;
  serverY = attr.y;

  if (serverX > width)
  {
    while (serverX > width)
    {
      width += gdk_screen_get_width (screen);
      *x = *x + 1;
    }
  }
  else
  {
    while (serverX < width - gdk_screen_get_width (screen))
    {
      width -= gdk_screen_get_width (screen);
      *x = *x - 1;
    }
  }

  if ((width - attr.x) < attr.width / 2)
    *x = *x - 1;

/*
  if (serverY > height)
  {
    while (serverY > height)
    {
      height += gdk_screen_get_height (screen);
      *y = *y + 1;
    }
  }
  else
  {
    while (serverY < height - gdk_screen_get_height (screen))
    {
      height -= gdk_screen_get_width (screen);
      *y = *y - 1;
    }
  }

  if ((height - attr.y) < attr.height / 2)
    *y = *y - 1;
*/
// printf ("New Win VIEWPORT: %i %i\n", *x, *y);
}

gboolean
kiba_window_is_in_viewport (Window window,
                            int    viewport_x,
                            int    viewport_y)
{
  GdkScreen *screen;
  XWindowAttributes attr;
  Status result;

  screen = gdk_screen_get_default ();

  KIBA_EXECUTE_IN_ERROR_TRAP (
    result = XGetWindowAttributes(gdk_display, window, &attr);
  )

  if (!result)
    return FALSE;

  if (attr.x + attr.width  > 0 && attr.x < gdk_screen_get_width (screen) &&
      attr.y + attr.height > 0 && attr.y < gdk_screen_get_height (screen))
    return TRUE;

  return FALSE;
}

void
kiba_window_to_right_viewport (GtkWidget *widget,
                               gpointer   data)
{
  GdkScreen *screen;
  GdkWindow *window = (GdkWindow*) data;
  int viewport_x, viewport_y, num_x, num_y;

  screen = gdk_screen_get_default ();

  kiba_get_num_viewports (&num_x, &num_y);
  kiba_get_win_viewport_number (gdk_x11_drawable_get_xid (window), &viewport_x, &viewport_y);

  if (viewport_x > num_x - 1)
    viewport_x = 0;

  kiba_change_win_viewport (gdk_x11_drawable_get_xid (window), viewport_x, viewport_y - 1);
}

void
kiba_window_to_left_viewport (GtkWidget *widget,
                              gpointer   data)
{
  GdkScreen *screen;
  GdkWindow *window = (GdkWindow*) data;
  int viewport_x, viewport_y, num_x, num_y;

  screen = gdk_screen_get_default ();

  kiba_get_num_viewports (&num_x, &num_y);
  kiba_get_win_viewport_number (gdk_x11_drawable_get_xid (window), &viewport_x, &viewport_y);

  viewport_x -= 2;

  if (viewport_x < 0)
    viewport_x = num_x - 1;

  kiba_change_win_viewport (gdk_x11_drawable_get_xid (window), viewport_x, viewport_y - 1);
}

void
kiba_window_change_viewport (GtkWidget *widget,
                             gpointer   data)
{
  GdkScreen *screen;
  GdkWindow *window = (GdkWindow*) data;
  char   *widget_name;
  int     viewport_x, viewport_y;

  screen      = gdk_screen_get_default ();
  widget_name = (char*) gtk_widget_get_name (widget);

  kiba_get_win_viewport_number (gdk_x11_drawable_get_xid (window), &viewport_x, &viewport_y);
  kiba_change_win_viewport (gdk_x11_drawable_get_xid (window), atoi (widget_name) - 1, viewport_y - 1);

  g_free (widget_name);
}

/**********
OPACITY
**********/
void
kiba_change_opacity (Window  xwindow,
                     guint32 opacity)
{
  Screen *screen;
  XEvent xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_WINDOW_OPACITY");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = opacity;
  xev.xclient.data.l[1] = 0;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
	        False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev);
  )
}

int
kiba_get_opacity (Window xwindow)
{
  int val, ret;

  if (!kiba_get_cardinal (xwindow, kiba_get_atom (gdk_display, "_NET_WM_WINDOW_OPACITY"), &val))
    val = G_MAXUINT32;

  ret = ((guint64) val * 1005) / G_MAXUINT32;

  return ret / 10;
}

void
kiba_set_opacity (GtkWidget *widget,
                  gpointer   data)
{
  GdkWindow  *window = (GdkWindow*) data;
  char   *opacity;
  guint   value;

  opacity = (char*) gtk_widget_get_name (widget);
  value   = (guint) atoi (opacity);

  kiba_change_opacity (gdk_x11_drawable_get_xid (window), (((guint64) value * G_MAXUINT32) / 100));

  g_free (opacity);
}

/**********
SATURATION
**********/
void
kiba_change_saturation (Window  xwindow,
                        guint32 saturation)
{
  Screen *screen;
  XEvent xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_WINDOW_SATURATION");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = saturation;
  xev.xclient.data.l[1] = 0;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
	        False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev);
  )
}

int
kiba_get_saturation (Window xwindow)
{
  int val, ret;

  if (!kiba_get_cardinal (xwindow, kiba_get_atom (gdk_display, "_NET_WM_WINDOW_SATURATION"), &val))
    val = G_MAXUINT32;

  ret = ((guint64) val * 1005) / G_MAXUINT32;

  return ret / 10;
}

void
kiba_set_saturation (GtkWidget *widget,
                     gpointer   data)
{
  GdkWindow *window = (GdkWindow*) data;
  char   *saturation;
  guint   value;

  saturation = (char*) gtk_widget_get_name (widget);
  value      = (guint) atoi (saturation);
  g_free (saturation);

  kiba_change_saturation (gdk_x11_drawable_get_xid (window), (((guint64) value * G_MAXUINT32) / 100));
}

/**********
BRIGHTNESS
**********/
void
kiba_change_brightness (Window  xwindow,
                        guint32 brightness)
{
  Screen *screen;
  XEvent xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_WINDOW_BRIGHTNESS");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = brightness;
  xev.xclient.data.l[1] = 0;
  xev.xclient.data.l[2] = 0;
  xev.xclient.data.l[3] = 0;
  xev.xclient.data.l[4] = 0;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
	        RootWindowOfScreen (screen),
	        False,
	        SubstructureRedirectMask | SubstructureNotifyMask,
	        &xev);
  )
}

int
kiba_get_brightness (Window xwindow)
{
  int val, ret;

  if (!kiba_get_cardinal (xwindow, kiba_get_atom (gdk_display, "_NET_WM_WINDOW_BRIGHTNESS"), &val))
    val = G_MAXUINT32;

  ret = ((guint64) val * 1005) / G_MAXUINT32;

  return ret / 10;
}

void
kiba_set_brightness (GtkWidget *widget,
                     gpointer   data)
{
  GdkWindow *window = (GdkWindow*) data;
  char   *brightness;
  guint   value;

  brightness = (char*) gtk_widget_get_name (widget);
  value      = (guint) atoi (brightness);
  g_free (brightness);

  kiba_change_brightness (gdk_x11_drawable_get_xid (window), (((guint64) value * G_MAXUINT32) / 100));
}

#define _NET_WM_MOVERESIZE_SIZE_TOPLEFT      0
#define _NET_WM_MOVERESIZE_SIZE_TOP          1
#define _NET_WM_MOVERESIZE_SIZE_TOPRIGHT     2
#define _NET_WM_MOVERESIZE_SIZE_RIGHT        3
#define _NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT  4
#define _NET_WM_MOVERESIZE_SIZE_BOTTOM       5
#define _NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT   6
#define _NET_WM_MOVERESIZE_SIZE_LEFT         7
#define _NET_WM_MOVERESIZE_MOVE              8
#define _NET_WM_MOVERESIZE_SIZE_KEYBOARD     9
#define _NET_WM_MOVERESIZE_MOVE_KEYBOARD    10

void
kiba_keyboard_move (GtkWidget *widget,
                    gpointer   data)
{
  Screen  *screen;
  Window   xwindow = (Window) data;
  XEvent   xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_MOVERESIZE");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = 0; /* unused */
  xev.xclient.data.l[1] = 0; /* unused */
  xev.xclient.data.l[2] = _NET_WM_MOVERESIZE_MOVE_KEYBOARD;
  xev.xclient.data.l[3] = 0; /* unused */
  xev.xclient.data.l[4] = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
                RootWindowOfScreen (screen),
                False,
                SubstructureRedirectMask | SubstructureNotifyMask,
                &xev);
  )
}

void
kiba_keyboard_size (GtkWidget *widget,
                    gpointer   data)
{
  Screen  *screen;
  Window   xwindow = (Window) data;
  XEvent   xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_MOVERESIZE");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = 0; /* unused */
  xev.xclient.data.l[1] = 0; /* unused */
  xev.xclient.data.l[2] = _NET_WM_MOVERESIZE_SIZE_KEYBOARD;
  xev.xclient.data.l[3] = 0; /* unused */
  xev.xclient.data.l[4] = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
                RootWindowOfScreen (screen),
                False,
                SubstructureRedirectMask | SubstructureNotifyMask,
                &xev);
  )
}

void
kiba_move_resize (Window       xwindow,
                  gboolean     move,
                  gboolean     resize,
                  int          x,
                  int          y,
                  unsigned int width,
                  unsigned int height)
{
  Screen  *screen;
  XEvent  xev;

  screen = gdk_x11_screen_get_xscreen (gdk_screen_get_default ());

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_MOVERESIZE_WINDOW");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = ((move) ? 3 << 7 : 0) | ((resize) ? 3 << 9 : 0);
  xev.xclient.data.l[1] = x;
  xev.xclient.data.l[2] = y;
  xev.xclient.data.l[3] = width;
  xev.xclient.data.l[4] = height;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
                RootWindowOfScreen (screen),
                False,
                SubstructureRedirectMask | SubstructureNotifyMask,
                &xev);
  )
}

void
kiba_move_window (GtkWidget *widget,
                  gpointer   data)
{
  GdkWindow *gdk_window = data;
  XEvent     xev;
  Window     xwindow = gdk_x11_drawable_get_xid (gdk_window);

  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_MOVERESIZE");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = 0; /* unused */
  xev.xclient.data.l[1] = 0; /* unused */
  xev.xclient.data.l[2] = _NET_WM_MOVERESIZE_MOVE_KEYBOARD;
  xev.xclient.data.l[3] = 0; /* unused */
  xev.xclient.data.l[4] = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
                RootWindowOfScreen (gdk_x11_screen_get_xscreen (gdk_screen_get_default ())),
                False,
                SubstructureRedirectMask | SubstructureNotifyMask,
                &xev); 
  )
}

void
kiba_resize_window (GtkWidget *widget,
                    gpointer   data)
{
  GdkWindow *gdk_window = data;
  XEvent     xev;
  Window     xwindow = gdk_x11_drawable_get_xid (gdk_window);
  
  xev.xclient.type = ClientMessage;
  xev.xclient.serial = 0;
  xev.xclient.send_event = True;
  xev.xclient.display = gdk_display;
  xev.xclient.window = xwindow;
  xev.xclient.message_type = kiba_get_atom (gdk_display, "_NET_WM_MOVERESIZE");
  xev.xclient.format = 32;
  xev.xclient.data.l[0] = 0; /* unused */
  xev.xclient.data.l[1] = 0; /* unused */
  xev.xclient.data.l[2] = _NET_WM_MOVERESIZE_SIZE_KEYBOARD;
  xev.xclient.data.l[3] = 0; /* unused */
  xev.xclient.data.l[4] = 1;

  KIBA_EXECUTE_IN_ERROR_TRAP (
    XSendEvent (gdk_display,
                RootWindowOfScreen (gdk_x11_screen_get_xscreen (gdk_screen_get_default ())),
                False,
                SubstructureRedirectMask | SubstructureNotifyMask,
                &xev);
  )
}

/* loads a libwnck like right click menu in the given GtkMenu */
void
kiba_load_app_menu (GdkWindow *window,
                    GtkMenu   *menu)
{
  Window            xwindow;
  GtkMenu          *appearance_menu, *opacity_menu, *saturation_menu, *brightness_menu;
  GtkWidget        *separator, *change_state = NULL, *kill_item, *sticky, *appearance_label,
                   *maximize_item, *ontop_item, *below_item, *move_item, *resize_item,
                   *opacity_label, *saturation_label, *brightness_label;
  int               win_state, num_workspaces, viewports_x, viewports_y, i;

  xwindow   = gdk_x11_drawable_get_xid (window);
  win_state = kiba_get_winstate (gdk_display, window);
  appearance_menu  = (GtkMenu *) gtk_menu_new();
  opacity_menu     = (GtkMenu *) gtk_menu_new();
  saturation_menu  = (GtkMenu *) gtk_menu_new();
  brightness_menu  = (GtkMenu *) gtk_menu_new();

  if (win_state == IconicState)
  {
    change_state = gtk_menu_item_new_with_label (_("Restore"));
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_deiconify_window), window);
  }
  else if (win_state == NormalState && gdk_x11_drawable_get_xid (window) == kiba_get_active_window (gdk_display))
  {
    change_state = gtk_menu_item_new_with_label (_("Minimze"));
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_iconify_window), window);
  }
  else if (win_state == NormalState)
  {
    change_state = gtk_menu_item_new_with_label (_("Focus"));
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_focus_window), window);
  }
  if (change_state != NULL)
  {
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), change_state);
    gtk_widget_show (change_state);
  }

  if (kiba_window_state_hint (window, "_NET_WM_STATE_MAXIMIZED_VERT") ||
      kiba_window_state_hint (window, "_NET_WM_STATE_MAXIMIZED_HORZ"))
  {
    maximize_item = gtk_menu_item_new_with_label (_("Restore"));
    g_signal_connect (G_OBJECT (maximize_item), "activate", G_CALLBACK (kiba_unmaximize_window), window);
  }
  else
  {
    maximize_item = gtk_menu_item_new_with_label (_("Maximize"));
    g_signal_connect (G_OBJECT (maximize_item), "activate", G_CALLBACK (kiba_maximize_window), window);
  }
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), maximize_item);
  gtk_widget_show (maximize_item);

  ontop_item = gtk_check_menu_item_new_with_label (_("Above"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), ontop_item);
  if (kiba_window_state_hint (window, "_NET_WM_STATE_ABOVE"))
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (ontop_item), TRUE);
  else
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (ontop_item), FALSE);
  g_signal_connect (G_OBJECT (ontop_item), "activate", G_CALLBACK (kiba_set_window_above), window);
  gtk_widget_show (ontop_item);

  below_item = gtk_check_menu_item_new_with_label (_("Below"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), below_item);
  if (kiba_window_state_hint (window, "_NET_WM_STATE_BELOW"))
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (below_item), TRUE);
  else
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (below_item), FALSE);
  g_signal_connect (G_OBJECT (below_item), "activate", G_CALLBACK (kiba_set_window_below), window);
  gtk_widget_show (below_item);

  sticky = gtk_check_menu_item_new_with_label (_("Stick Window"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), sticky);
  if (kiba_window_state_hint (window, "_NET_WM_STATE_STICKY"))
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (sticky), TRUE);
  else
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (sticky), FALSE);
  g_signal_connect (G_OBJECT (sticky), "activate", G_CALLBACK (kiba_window_stick), window);
  gtk_widget_show (sticky);

  move_item = gtk_menu_item_new_with_label (_("Move Window"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), move_item);
  g_signal_connect (G_OBJECT (move_item), "activate", G_CALLBACK (kiba_move_window), window);
  gtk_widget_show (move_item);

  resize_item = gtk_menu_item_new_with_label (_("Resize Window"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), resize_item);
  g_signal_connect (G_OBJECT (resize_item), "activate", G_CALLBACK (kiba_resize_window), window);
  gtk_widget_show (resize_item);

  appearance_label = gtk_menu_item_new_with_label (_("Appearance"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), appearance_label);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (appearance_label), GTK_WIDGET(appearance_menu));
  gtk_widget_show (appearance_label);

  opacity_label = gtk_menu_item_new_with_label (_("Alpha"));
  gtk_menu_shell_append (GTK_MENU_SHELL (appearance_menu), opacity_label);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (opacity_label), GTK_WIDGET(opacity_menu));
  gtk_widget_show (opacity_label);

  int opacity_value;
  opacity_value = kiba_get_opacity (xwindow);
  if (opacity_value <= 0)
    opacity_value += 100;

  for (i = 25; i <= 100; i += 25)
  {
    if (opacity_value != i)
    {
      GtkWidget *opacity_widget;
      char *label, *opacity;;

      label   = g_strdup_printf ("%g", (float) i / 100.0f);
      opacity = g_strdup_printf ("%i", i);
      opacity_widget = gtk_menu_item_new_with_label (label);
      gtk_widget_set_name (opacity_widget, opacity);
      g_free (label);
      g_free (opacity);
      gtk_menu_shell_append (GTK_MENU_SHELL (opacity_menu), opacity_widget);
      g_signal_connect (G_OBJECT (opacity_widget), "activate", G_CALLBACK (kiba_set_opacity), window);
      gtk_widget_show (opacity_widget);
    }
  }

  saturation_label = gtk_menu_item_new_with_label (_("Saturation"));
  gtk_menu_shell_append (GTK_MENU_SHELL (appearance_menu), saturation_label);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (saturation_label), GTK_WIDGET(saturation_menu));
  gtk_widget_show (saturation_label);

  int saturation_value;
  saturation_value = kiba_get_saturation (xwindow);
  if (saturation_value <= 0)
    saturation_value += 100;

  for (i = 25; i <= 100; i += 25)
  {
    if (saturation_value != i)
    {
      GtkWidget *saturation_widget;
      char *label, *saturation;

      label      = g_strdup_printf ("%g", (float) i / 100.0f);
      saturation = g_strdup_printf ("%i", i);
      saturation_widget = gtk_menu_item_new_with_label (label);
      gtk_widget_set_name (saturation_widget, saturation);
      g_free (label);
      g_free (saturation);
      gtk_menu_shell_append (GTK_MENU_SHELL (saturation_menu), saturation_widget);
      g_signal_connect (G_OBJECT (saturation_widget), "activate", G_CALLBACK (kiba_set_saturation), window);
      gtk_widget_show (saturation_widget);
    }
  }

  brightness_label = gtk_menu_item_new_with_label (_("Brightness"));
  gtk_menu_shell_append (GTK_MENU_SHELL (appearance_menu), brightness_label);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (brightness_label), GTK_WIDGET(brightness_menu));
  gtk_widget_show (brightness_label);

  int brightness_value;
  brightness_value = kiba_get_brightness (xwindow);
  if (brightness_value <= 0)
    brightness_value += 100;

  for (i = 25; i <= 100; i += 25)
  {
    if (brightness_value != i)
    {
      GtkWidget *brightness_widget;
      char *label, *brightness;

      label      = g_strdup_printf ("%g", (float) i / 100.0f);
      brightness = g_strdup_printf ("%i", i);
      brightness_widget = gtk_menu_item_new_with_label (label);
      gtk_widget_set_name (brightness_widget, brightness);
      g_free (label);
      g_free (brightness);
      gtk_menu_shell_append (GTK_MENU_SHELL (brightness_menu), brightness_widget);
      g_signal_connect (G_OBJECT (brightness_widget), "activate", G_CALLBACK (kiba_set_brightness), window);
      gtk_widget_show (brightness_widget);
    }
  }

  separator = gtk_separator_menu_item_new();
  gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
  gtk_widget_show (separator);

  kill_item = gtk_menu_item_new_with_label (_("Close"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), kill_item);
  g_signal_connect (G_OBJECT (kill_item), "activate", G_CALLBACK (kiba_kill_window), window);
  gtk_widget_show (kill_item);

  //FIXME something here is broken
/*
  run_item = gtk_menu_item_new_with_label (_("Launch another instance"));
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), run_item);
  command = kiba_sys_get_command_from_pid (kiba_get_win_pid (gdk_display, xwindow));
  g_signal_connect (G_OBJECT (run_item), "activate", G_CALLBACK (launch_programm), command);
  g_free (command);
  gtk_widget_show (run_item);
*/

  num_workspaces = kiba_num_workspaces ();
  if (num_workspaces > 1)
  {
    GtkWidget *move_right, *move_left, *workspace_number;
    GtkMenu   *workspacemenu;
    int win_workspace;

    workspacemenu  = (GtkMenu *) gtk_menu_new();
    win_workspace  = kiba_get_win_workspace_number (gdk_display, xwindow);

    separator = gtk_separator_menu_item_new();
    gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
    gtk_widget_show (separator);

    if (win_workspace > 0)
    {
      move_left = gtk_menu_item_new_with_label (_("Move to Previous Workspace"));
      gtk_menu_shell_append (GTK_MENU_SHELL (menu), move_left);
      g_signal_connect (G_OBJECT (move_left), "activate", G_CALLBACK (kiba_window_to_left_workspace), window);
      gtk_widget_show (move_left);
    }
    if (win_workspace + 1 < num_workspaces)
    {
      move_right = gtk_menu_item_new_with_label (_("Move to Next Workspace"));
      gtk_menu_shell_append (GTK_MENU_SHELL (menu), move_right);
      g_signal_connect (G_OBJECT (move_right), "activate", G_CALLBACK (kiba_window_to_right_workspace), window);
      gtk_widget_show (move_right);
    }

    workspace_number = gtk_menu_item_new_with_label (_("Move to Another Workspace"));
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), workspace_number);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (workspace_number), GTK_WIDGET(workspacemenu));
    gtk_widget_show (workspace_number);

    for (i = 1; i != num_workspaces + 1; i++)
    {
      if (i != win_workspace + 1)
      {
        GtkWidget *workspace;
        char *label, *number;

        label  = g_strdup_printf ("Workspace %i", i);
        number = g_strdup_printf ("%i", i);
        workspace = gtk_menu_item_new_with_label (label);
        gtk_widget_set_name (workspace, number);
        g_free (label);
        g_free (number);
        gtk_menu_shell_append (GTK_MENU_SHELL (workspacemenu), workspace);
        g_signal_connect (G_OBJECT (workspace), "activate", G_CALLBACK (kiba_window_change_workspace), window);
        gtk_widget_show (workspace);
      }
    }
  }

  kiba_get_num_viewports (&viewports_x, &viewports_y);
  if (viewports_x > 1)
  {
    GtkWidget *move_right, *move_left, *viewport_number;
    GtkMenu   *viewportmenu;
    int win_viewport_x, win_viewport_y, i;

    viewportmenu  = (GtkMenu *) gtk_menu_new();

    kiba_get_win_viewport_number (xwindow, &win_viewport_x, &win_viewport_y);

    separator = gtk_separator_menu_item_new();
    gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
    gtk_widget_show (separator);

    if (win_viewport_x > 1)
    {
      move_left = gtk_menu_item_new_with_label (_("Move to Previous Viewport"));
      gtk_menu_shell_append (GTK_MENU_SHELL (menu), move_left);
      g_signal_connect (G_OBJECT (move_left), "activate", G_CALLBACK (kiba_window_to_left_viewport), window);
      gtk_widget_show (move_left);
    }
    if (win_viewport_x < viewports_x)
    {
      move_right = gtk_menu_item_new_with_label (_("Move to Next Viewport"));
      gtk_menu_shell_append (GTK_MENU_SHELL (menu), move_right);
      g_signal_connect (G_OBJECT (move_right), "activate", G_CALLBACK (kiba_window_to_right_viewport), window);
      gtk_widget_show (move_right);
    }

    viewport_number = gtk_menu_item_new_with_label (_("Move to Another Viewport"));
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), viewport_number);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (viewport_number), GTK_WIDGET(viewportmenu));
    gtk_widget_show (viewport_number);

    for (i = 1; i != viewports_x + 1; i++)
    {
      if (i != win_viewport_x)
      {
        GtkWidget *viewport;
        char *label, *number;

        label  = g_strdup_printf ("Viewport %i", i);
        number = g_strdup_printf ("%i", i);
        viewport = gtk_menu_item_new_with_label (label);
        gtk_widget_set_name (viewport, number);
        g_free (label);
        g_free (number);
        gtk_menu_shell_append (GTK_MENU_SHELL (viewportmenu), viewport);
        g_signal_connect (G_OBJECT (viewport), "activate", G_CALLBACK (kiba_window_change_viewport), window);
        gtk_widget_show (viewport);
      }
    }
  }
}

void kiba_show_interest_in_events(Window xwindow) {
	XWindowAttributes attrs;

	KIBA_EXECUTE_IN_ERROR_TRAP (
		XGetWindowAttributes (gdk_display, xwindow, &attrs);
	)

	int mask = attrs.your_event_mask;

	// Signal to X that we are interested in the PropertyNotify event
	mask |= PropertyChangeMask;

	// Signal to X that we are interested in the UnmapNotify, MapNotify,
	// and DestroyNotify events
	mask |= StructureNotifyMask;

	// Signal to X that we are interested in the FocusIn and FocusOut events
	mask |= FocusChangeMask;

	KIBA_EXECUTE_IN_ERROR_TRAP (
		XSelectInput (gdk_display, xwindow, mask);
	)
}
